/**
 * @author Magne Haveraaen
 * @since 2013
 */
// implementation: implementation BasicCxxRaw.BoundedArrayCxxRaw.boundedArrayRaw, file: "BasicCxxRaw/CxxBoundedArrayRaw.hxx" 
#ifndef _INCLUDE_BASICCXXRAW_CXXBOUNDEDARRAYRAW_HXX
#define _INCLUDE_BASICCXXRAW_CXXBOUNDEDARRAYRAW_HXX
namespace BasicCxxRaw {
template<typename E>
struct CxxBoundedArrayRaw {
	struct CxxBoundedArrayRaw_impl {

		/** The type Intex must match the generic index type size_t. */
		typedef size_t Intex;

		struct bar_A {
			  Intex data_size;
			  E * data;
		public:
			bar_A() : data_size(0) {
				data = new E[data_size];
				for(Intex i=0; i < data_size; ++i){
					data[i] = E();
				}
			}
			/** Constructor needed for the create function. */
			bar_A(const Intex& size) : data_size(size) {
				data = new E[data_size];
				for(Intex i=0; i < data_size; ++i){
					data[i] = E();
				}
			}
			/** Constructor needed for the create with element function. */
			bar_A(const Intex& size, const E& e) : data_size(size) {
				data = new E[data_size];
				for(Intex i=0; i < data_size; ++i){
					data[i] = e;
				}
			}
			bar_A(const bar_A& a) : data_size(a.data_size) {
				data = new E[data_size];
				for(Intex i=0; i < data_size; ++i){
					data[i] = a.data[i];
				}
			}
			bar_A(bar_A&& a) : data_size(a.data_size), data(a.data) {
				a.data_size = 0;
				a.data = new E[0];
			}
			~bar_A() {
				// std::cerr << "~bar_A Freeing bar_A data " << data_size << std::endl;
				delete [] data;
				// std::cerr << "~bar_A Freed   bar_A data " << data_size << std::endl;
			}

			void operator=(const bar_A& a) {
				Intex size = a.data_size;
				if(data_size != a.data_size) {
					// Magnolia::Assert::guardFailed("Source data size different from target data size","","BasicCxxRaw.CxxBoundedArrayRaw.CxxBoundedArrayRaw_impl.operator=");
					// std::cerr << "operator= Resizing bar_A data from " << data_size << " to " << size << std::endl;
					delete [] data;
					data_size = size;
					data = new E[data_size];
				}
				for(Intex i=0; i < size; ++i){
					data[i] = a.data[i];
				}
			}
			void operator=(bar_A&& a) {
				delete [] data;
				data_size = a.data_size;
				data = a.data;
				a.data_size = 0;
				a.data = new E[0];
				// std::cerr << "operator=&& " << data_size << std::endl;
			}
			bool operator==(const bar_A& a) const {
				bool ok=data_size==a.data_size;
				if(!ok) return ok;
				for(Intex i=0; i < data_size; ++i){
					if( !( data[i] == a.data[i] ) ) return false;
				}
				return ok;
			}
			template<typename _Stream> void operator>>=(_Stream& _stream) const {
				Magnolia::IO::Type _type_tag = Magnolia::IO::Type("Array", "");
				
				// std::cerr << "operator>>= bar_A pre  " << data_size << std::endl;
				Magnolia::IO::startData(_type_tag) >>= _stream;
				data_size >>= _stream;
				for(Intex i=0; i < data_size; ++i){
					// std::cerr << "operator>>= bar_A data[" << i << "]=" << data[i] << std::endl;
					data[i] >>= _stream;
				}
				Magnolia::IO::endData(_type_tag) >>= _stream;
				// std::cerr << "operator>>= bar_A post " << data_size << std::endl;
			}
			template<typename _Stream> void operator<<=(_Stream& _stream) {
				Magnolia::IO::Type _type_tag = Magnolia::IO::Type("Array", "");
				
				// std::cerr << "operator<<= bar_A pre  " << data_size << std::endl;
				Magnolia::IO::startData(_type_tag) <<= _stream;
				Intex new_data_size = 0;
				new_data_size <<= _stream;
				if(data_size < new_data_size) {
					// std::cerr << "operator<<= Resizing bar_A data from " << data_size << " to " << (new_data_size) << std::endl;
					delete [] data;
					data_size = new_data_size;
					data = new E[data_size];
				}
				for(Intex i=0; i < new_data_size; ++i){
					data[i] <<= _stream;
					// std::cerr << "operator<<= bar_A data[" << i << "]=" << data[i] << std::endl;
				}
				Magnolia::IO::endData(_type_tag) <<= _stream;
				// std::cerr << "operator<<= bar_A post " << data_size << std::endl;
			}
		};
		

		/** function bar_createArray(:Intex) : bar_A */
		bar_A bar_createArray(const Intex& size) const {
			return bar_A(size);
		}
		
		/** function bar_createArray_element(:Intex,:E) : bar_A */
		bar_A bar_createArray_element(const Intex& size, const E& e) const {
			return bar_A(size,e);
		}

		/** function bar_createArrayElement(:Intex) : E */
		E bar_createArrayElement(const Intex& i) const {
			return E();
		}

		/** predicate bar_equivalence(:bar_A, :bar_A, :Intex) */
		bool bar_equivalence(const bar_A& a1, const bar_A& a2, const Intex& length) const {
				for(Intex i=0; i < length; ++i){
					if( !( a1.data[i] == a2.data[i] ) ) return false;
				}
				return true;
		}

		/** function bar_getElement(:bar_A, :Intex) : E */
		E bar_getElement(const bar_A& a, const Intex& i) const {
			return a.data[i];
		}
		
		/** function bar_getSize(:bar_A) : Intex */
		Intex bar_getSize(const bar_A& a) const {
			return a.data_size;
		}
		
		/** procedure bar_concatenate(upd :bar_A, obs :Intex, obs :bar_A, obs :Intex) */
		void bar_concatenate(bar_A& a1, const Intex& low, const bar_A& a2, const Intex& high) const {
			// std::cout << " getSize(a1)=" << a1.data_size << " getSize(a2)=" << getSize(a2) << std::endl;
			for (Intex i=0; i < high; ++i) {
				// std::cout << " low=" << low << " i=" << i << " low+i=" << low+i << std::endl;
				// std::cout << " a1.data[low+i]=" << a1.data[low+i] << " a2.data[i]=" << a2.data[i] << std::endl;
				a1.data[low+i] = a2.data[i];
			}
		}
		
		/** procedure bar_replaceSegment(upd :bar_A, obs :Intex, obs :Intex, obs :E) */
		void bar_replaceSegment(bar_A& a, const Intex& low, const Intex& high, const E& e) const {
			for (Intex j=low; j < high; ++ j) {
				a.data[j] = e;
			}
		}
		
		/** procedure bar_replaceSegmentFromArray(upd :bar_A, obs :Intex, obs :Intex, obs :bar_A, obs :Intex) */
		void bar_replaceSegmentFromArray(bar_A& a1, const Intex& low, const Intex& high, const bar_A& a2, const Intex& i) const {
			for (Intex j=0; j < high-low; ++ j) {
				a1.data[low+j] = a2.data[i+j];
			}
		}
		
		/** procedure bar_reverseSegment(upd :bar_A, obs :Intex, obs :Intex) */
		void bar_reverseSegment(bar_A& a, const Intex& low, const Intex& high) const {
			for (Intex i = 0; i < (high-low+1)/2; ++i) {
				E e = a.data[high-i-1];
				a.data[high-i-1] = a.data[low+i];
				a.data[low+i] = e;
			}
		}
		
		/** procedure bar_shiftSegmentGroups(upd :bar_A, obs :Intex, obs :Intex, obs :Intex, obs :Intex, obs :Intex) */
		void bar_shiftSegmentGroups(bar_A& a, const Intex& segsize, const Intex& groupsize, const Intex& d, const Intex& low, const Intex& high) const {
			if ( high <= low || d == 0 ) return;

			E* tmp;
			tmp = new E[high-low];
			for (Intex j = low; j < high; j++) {
				tmp[j-low] = a.data[j];
			}
			//normalised shift to perfom
			Intex shift = (groupsize * d) % segsize;
			for (Intex j = low; j < high; j++) {
			  //local index within segment
			  Intex si = (j - low)% segsize;
			  // new index within segment after the local shift
			  Intex ni = (segsize + si - shift) % segsize;
			  // obtain the global position of ni within the whole array
			  Intex i =  low + ((j -low) / segsize)*segsize + ni;
			  //update a.data with the value from tmp corresponding to the shift operation
			  a.data[i] = tmp[j-low];

			}
			delete [] tmp;


/*
			Intex segs = (high - low) / segsize;
			* Groups Per Segment.
			Intex gps = segsize / groupsize;
			Intex posd = d % gps;
			for ( Intex seg = 0; seg < segs; seg++ ) {
//				std::cerr << "bar_shiftSegmentGroups segsize=" << segsize << " groupsize=" << groupsize << " d=" << d << " low=" << low << " high=" << high
//						  << " segs=" << segs << " seg=" << seg << " gps=" << gps << " posd=" << posd << std::endl;
				E e[posd][groupsize];
				// Save the groups being shifted into
				for(Intex i=0; i<posd; ++i) {
					Intex start1 = low + seg * segsize + i * groupsize;
//					std::cerr << "bar_shiftSegmentGroups l1 start1=" << start1 << " i=" << i << std::endl;
					for (Intex j = 0; j < groupsize; ++j) {
//						std::cerr << "bar_shiftSegmentGroups l1j start1=" << start1 << " i=" << i << " j=" << j << " e=" << a.data[start1+j] << std::endl;
						e[i][j] = a.data[start1+j];
					}
				}
				// Shift groups into vacated area
				{
					Intex start1 = low + seg * segsize;
					Intex start2 = low + seg * segsize + posd * groupsize;
					for(Intex i=0; i<(gps - posd) * groupsize; ++i) {
//						std::cerr << "bar_shiftSegmentGroups l2 start1=" << start1 << " start2=" << start2 << " i=" << i << " e=" << a.data[start2+i] << std::endl;
						a.data[start1+i] = a.data[start2+i];
					}
				}
				// Shift stored segments into open area
				for(Intex i=0; i<posd; ++i) {
					Intex start1 = low + seg * segsize + (gps - posd + i) * groupsize;
//					std::cerr << "bar_shiftSegmentGroups l3 start1=" << start1 << " i=" << i << std::endl;
					for (Intex j = 0; j < groupsize; ++j) {
//						std::cerr << "bar_shiftSegmentGroups l3j start1=" << start1 << " i=" << i << " j=" << j << " e=" << e[i][j] << std::endl;
						a.data[start1+j] = e[i][j];
					}
				}
			}
*/
		}

	};
	
	
};


}


#endif
