/**
 * A generic implementation of operations for a subset of raw (computer) integer types as built-in to C++.
 * The guards are to be implemented in the Magnolia context by code specific for the chosen number type.
 *
 * @author Magne Haveraaen
 * @since 2010-04-21
 * @revised 2010-07-05 Magne Haveraaen, defined 32bit int and cleaned up code.
 * @revised 2012-02-24 Magne Haveraaen, split file into a specific part and a generic part.
 * @revised 2013-09-06 Magne Haveraaen, freshly generated for the current basic library setup.
 */


		/** predicate gminus(:Integer, :Integer) */
		bool gminus(const Integer& a, const Integer& b) const {
			long double r = ((long double)a) - (long double) b;
			return std::numeric_limits < Integer > ::min() <= r && r <= std::numeric_limits < Integer > ::max();
		}

		/** predicate gmult(:Integer, :Integer) */
		bool gmult(const Integer& a, const Integer& b) const {
			long double r = ((long double)a) * (long double) b;
			return std::numeric_limits < Integer > ::min() <= r && r <= std::numeric_limits < Integer > ::max();
		}

		/** predicate gplus(:Integer, :Integer) */
		bool gplus(const Integer& a, const Integer& b) const {
			long double r = ((long double)a) + (long double) b;
			return std::numeric_limits < Integer > ::min() <= r && r <= std::numeric_limits < Integer > ::max();
		}

		/** function idiv(:Integer, :Integer) : Integer */
		Integer idiv(const Integer& a, const Integer& b) const {
			return a / b;
		}

		/** predicate le(:Integer, :Integer) */
		bool le(const Integer& a, const Integer& b) const {
			return a <= b;
		}

		/** function maximum() : Integer */
		Integer maximum() const {
			return std::numeric_limits < Integer > ::max();
		}

		/** function minimum() : Integer */
		Integer minimum() const {
			return std::numeric_limits < Integer > ::min();
		}

		/** function minus(:Integer, :Integer) : Integer */
		Integer minus(const Integer& a, const Integer& b) const {
			return a - b;
		}

		/** function mult(:Integer, :Integer) : Integer */
		Integer mult(const Integer& a, const Integer& b) const {
			return a * b;
		}

		/** function one() : Integer */
		Integer one() const {
			return 1;
		}

		/** function plus(:Integer, :Integer) : Integer */
		Integer plus(const Integer& a, const Integer& b) const {
			return a + b;
		}

		/** function rem(:Integer, :Integer) : Integer */
		Integer rem(const Integer& a, const Integer& b) const {
			return a % b;
		}

		/** function uminus(:Integer) : Integer */
		Integer uminus(const Integer& b) const {
			return -b;
		}

		/** function zero() : Integer */
		Integer zero() const {
			return 0;
		}

