/**
 * @author Jaakko Järvi, Magne Haveraaen
 * @since 2015-03-17
 * @see http://www.cplusplus.com/reference/string/string/
 */
// program: program BasicCxxRaw.StringCxxRaw.stringCxxRaw, file: "BasicCxxRaw/CxxStringUtf8Raw.hxx" 
#ifndef _INCLUDE_BASICCXXRAW_CXXSTRINGUTF8RAW_HXX
#define _INCLUDE_BASICCXXRAW_CXXSTRINGUTF8RAW_HXX
#include <string>
namespace BasicCxxRaw {
struct CxxStringUtf8Raw {
	struct CxxStringUtf8Raw_impl {

		/** type String */
		struct String {
			std::string utf8string;
		public:
			String() = default;
			String(const String&) = default;
			String(String&&) = default;
			~String() = default;
			String& operator=(const String&) = default;
			String& operator=(String&&) = default;
			bool operator==(const String& _other) const {
				return this->utf8string == _other.utf8string;
			}
			template<typename _Stream> void operator>>=(_Stream& _stream) const {
				Magnolia::IO::Type _type_tag = Magnolia::IO::Type("String", "");
				
				std::string out = u8"\"";
				for (char c : this->utf8string) { out.append(s_helper_escape_char(c)); }
				out.append(u8"\"");
				// Magnolia::IO::startData(_type_tag) >>= _stream;
				_stream.write(out);
				// Magnolia::IO::endData(_type_tag) >>= _stream;
			}
			template<typename _Stream> void operator<<=(_Stream& _stream) {
				Magnolia::IO::Type _type_tag = Magnolia::IO::Type("String", "");
				
				// Magnolia::IO::startData(_type_tag) <<= _stream;
				std::string str = s_helper_read_string(_stream.stream);
				// _stream.readId(this->utf8string);
				// Magnolia::IO::endData(_type_tag) <<= _stream;
				this->utf8string = str;
			}

			/** Private constructor to easily provide constants. */
			String(string str) : utf8string(move(str)) {}

			/** Escaping backslash and quotation marks with a backslash. */
			static std::string s_helper_escape_char(char c) {
			  switch (c) {
			  case '\"': return u8"\\\"";
			  case '\\': return u8"\\\\";
			  default: return std::string(1, c);
			  }
			}
			/** Reading a (raw) string without converting the escapes. Stops at any ASCII control code. */
			static std::string s_helper_read_rawstring(std::istream& in) {
			  char c;
			  std::string s;
			  in.get(c);
			  // Reads until an ASCII control code appears in the string.
			  while (c >= 32 || c < 0) {
				  s += c;
				  in.get(c);
			  }
			  // std::cerr << "s_helper_read_rawstring: " << s << " " << ((int)c) << " " << c << std::endl;
			  return s;
			}
			/** Reading a string with escapes (if it starts with a quotation mark), converting it to a raw string. */
			static std::string s_helper_read_string(std::istream& in) {
			  // Skip initial white space characters.
			  in >> std::ws;
			  char c = in.peek();
			  // Decide whether a raw or a formatted string.
			  if (c != '\"') return s_helper_read_rawstring(in);
			  in >> c; // this is the opening '\"' character
			  std::string s;
			  in.get(c);
			  while(in.good() && c != '\"') {
				// std::cerr << "s_helper_read_string: " << s << " " << ((int)c) << " " << c << std::endl;
			    if (c != '\\') s += c;
			    else {
			      if (!(in >> c)) return s; // leaves eofbit on
			      switch (c) {
			      case '\\': s += '\\'; break;
			      case '\"': s += '\"'; break;
			      default: in.setstate(std::ios::failbit); return s;
			      }
			    }
			    in.get(c);
			  }
			  // std::cerr << "s_helper_read_string: " << s << " " << ((int)c) << " " << c << std::endl;
			  return s;
			}
		};
		
		
		/** predicate isSubstring(:String, :String) */
		bool isSubstring(const String& a, const String& b) const {
			auto found = b.utf8string.find(a.utf8string);
			return found != std::string::npos;
		}

		/** predicate leEncodingLength(:String, :String) */
		bool leEncodingLength(const String& a, const String& b) const {
			return a.utf8string.length() <= b.utf8string.length();
		}

		/** predicate leEncodingPatterns(:String, :String) */
		bool leEncodingPatterns(const String& a, const String& b) const {
			return a.utf8string <= b.utf8string;
		}

		/** function s_concatenate(:String, :String) : String */
		String s_concatenate(const String& a, const String& b) const {
			return String(a.utf8string + b.utf8string);
		}
		
		/** function s_createEmptyString() : String */
		String s_createEmptyString() const {
			return String(u8"");
		}

		/** procedure splitString(obs :String, obs :String, out :String, out :String) */
		void splitString(const String& a, const String& b, String& c, String& d) const {
			auto found = b.utf8string.find(a.utf8string);
			if(found != std::string::npos) {
				c = String(b.utf8string.substr(0,found));
				d = String(b.utf8string.substr(found+a.utf8string.length()));
			} else {
				c = b;
				d = String(u8"");
			}
		}

	};
	
	
};


}


#endif
