#ifndef _MAGNOLIA_PRELUDE_ASSERT_HXX
#define _MAGNOLIA_PRELUDE_ASSERT_HXX

#ifdef __GLIBC__
#include <execinfo.h>
#endif

namespace Magnolia {
namespace Assert {
void print_trace() {
#ifdef __GLIBC__
	void *buffer[20];
	int n = backtrace(buffer, 20);

	std::cerr << "Backtrace:\n";
	std::cerr.flush();
	backtrace_symbols_fd(&buffer[2], n, 2);
#endif
}

void axiom_assert(bool cond, const char* msg, const char* src, const char* loc) {
	if(!cond) {
		if(*loc != '\0')
			std::cerr << loc << ": ";
		std::cerr << "Assertion failed";
		if(*src != '\0')
			std::cerr << ": " << src;
		std::cerr << std::endl;
		if(*msg != '\0')
			std::cerr << "    " << msg << std::endl;
		std::exit(1);
	}
}

template<typename T, typename U>
#ifdef __NVCC__
__host__ __device__
#endif
bool equal(const T& a, const U& b) { /*  std::cerr << a << " == " << b << " = " << (a == b) << std::endl; */ return a == b; }

template<typename T, typename U>
#ifdef __NVCC__
__host__ __device__
#endif
bool notEqual(const T& a, const U& b) { /* std::cerr << a << " != " << b << " = " << (a != b) << std::endl; */ return !(a == b); }


/** Counter for the number of times a guard has failed. */
static size_t guard_counter = 0;
void guardFailed(const char* msg, const char* src, const char* loc) {
	std::cerr << "Guard failed:"
	<< " loc=" << (*loc == '\0' ? "\"\"" : loc)
	<< " src=" << (*src == '\0' ? "\"\"" : src)
	<< " msg=" << (*msg == '\0' ? "\"\"" : msg)
	<< std::endl;

	assert ( guard_counter <= 20 );
	//print_trace();
	guard_counter++;
};

template<typename T, typename U>
const U& exprGuardFailed(const T& val, const char* msg, const char* src, const char* loc) {
	std::cerr << "Guard failed:"
	<< " loc=" << (*loc == '\0' ? "\"\"" : loc)
	<< " src=" << (*src == '\0' ? "\"\"" : src)
	<< " msg=" << (*msg == '\0' ? "\"\"" : msg);

//	std::ostringstream valprint;
//	Magnolia::IO::XmlOStream xmlstream(std::cerr);
//	Magnolia::IO::TextOStream textstream(valprint);
	std::cerr << " value=" << val << std::endl;
//	45.75 >>= xmlstream;
//	std::cerr << "X" << std::endl;
//	val >>= xmlstream;
//	std::cerr << "Y" << std::endl;
//	55.6 >>= textstream;
//	val >>= textstream;
//	std::ostringstream outprint = textstream.stream;
//	std::cerr << outprint.str() << "Z" << std::endl;
	// std::cerr << (std::ostringstream)(textstream.stream).str() << "Z" << std::endl;
	// std::cerr << "Value " << xmlstream.stream.str() << std::endl;
//	std::cerr << "W" << std::endl;
	if(*msg != '\0')
		std::cerr << "M " << msg << std::endl;

	assert ( guard_counter <= 20 );
	//print_trace();
	guard_counter++;
};

}

void compile_error(const char* msg) {
	throw msg;
}

bool implies(bool a, bool b) {
	return !a || b;
}
}

#endif

