#ifndef _MAGNOLIA_PRELUDE_XMLIO_HXX
#define _MAGNOLIA_PRELUDE_XMLIO_HXX

namespace Magnolia {
namespace IO {

struct XmlIStream : MgIStream {
	XmlIStream(std::istream& s) : MgIStream(s) {}
};

struct XmlOStream : MgOStream {
	XmlOStream(std::ostream& s) : MgOStream(s) {}
};


void operator>>=(const startData& tag, XmlOStream& s) {
	s.indent();
	s.write("<");
	s.write(tag.type.name);
	s.write("#");
	//TODO s.write(tag.type.fields);
	s.write(">");
	s.newline();
	s.nest();
}

void operator<<=(const startData& tag, XmlIStream& s) {
	std::string t;
	s.expect('<');
	s.readId(t);
	s.expect('>');
	s.nest();
}

void operator>>=(const sepData& tag, XmlOStream& s) {
}

void operator<<=(const sepData& tag, XmlIStream& s) {
}

void operator>>=(const endData& tag, XmlOStream& s) {
	s.unnest();
	s.indent();
	s.write("</");
	s.write(tag.type.name);
	s.write("#");
	//TODO s.write(tag.type.fields);
	s.write(">");
	s.newline();
}

void operator<<=(const endData& tag, XmlIStream& s) {
	s.unnest();

	std::string t;
	s.expect('<');
	s.expect('/');
	s.readId(t);
	s.expect('>');
}

void operator<<=(bool& a, XmlIStream& s) {
	// TODO
	int data;
	s.stream >> data;
	if( data < std::numeric_limits < bool >::min()
	 || data > std::numeric_limits < bool >::max())
	    throw "Integer out of range for signed bool";
	a = (int8_t) data;
}

void operator>>=(const bool& a, XmlOStream& s) {
	s.indent();
	// s.write((int)a);
	if (a) s.write("HOLDS");
	else s.write("DOES_NOT_HOLD");
	s.newline();
}

void operator<<=(signed char& a, XmlIStream& s) {
	int data;
	s.stream >> data;
	if( data < std::numeric_limits < signed char >::min()
	 || data > std::numeric_limits < signed char >::max())
	    throw "Integer out of range for signed char";
	a = (int8_t) data;
}

void operator>>=(const signed char& a, XmlOStream& s) {
	s.indent();
	s.write((int)a);
	s.newline();
}

void operator<<=(signed short& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed short& a, XmlOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(signed int& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed int& a, XmlOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(signed long int& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed long int& a, XmlOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(signed long long int& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed long long int& a, XmlOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(unsigned char& a, XmlIStream& s) {
	int data;
	s.stream >> data;
	if( data < std::numeric_limits < unsigned char >::min()
	 || data > std::numeric_limits < unsigned char >::max())
	    throw "Integer out of range for unsigned char";
	a = (int8_t) data;
}

void operator>>=(const unsigned char& a, XmlOStream& s) {
	s.indent();
	s.write((int)a);
	s.newline();
}

void operator<<=(unsigned short& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned short& a, XmlOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(unsigned int& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned int& a, XmlOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(unsigned long int& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned long int& a, XmlOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(unsigned long long int& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned long long int& a, XmlOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(float& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const float& a, XmlOStream& s) {
	s.indent();
	// s.stream.setf( std::ios::fixed, std:: ios::floatfield );
	s.stream.precision(10);
	s.write(a);
	s.newline();
}

void operator<<=(double& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const double& a, XmlOStream& s) {
	s.indent();
	// s.stream.setf( std::ios::fixed, std:: ios::floatfield );
	s.stream.precision(20);
	s.write(a);
	s.newline();
}
void operator<<=(long double& a, XmlIStream& s) {
	s.stream >> a;
}

void operator>>=(const long double& a, XmlOStream& s) {
	s.indent();
	// s.stream.setf( std::ios::fixed, std:: ios::floatfield );
	s.stream.precision(40);
	s.write(a);
	s.newline();
}

}
}

#endif
