import org.eclipse.jface.text.IInformationControl;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTError;
import org.eclipse.swt.browser.Browser;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.events.ShellListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;


public class BrowserTextHover implements IInformationControl {

	private Display display;
	private Shell shell;
	private Browser browser;
	private ShellListener shellListener;
	private String information;
	private int maxWidth;
	private int maxHeight;


	public BrowserTextHover(Shell shell) {
		browser = initializeBrowser(shell, display);
		if(browser == null) {
			displayFatalErrorMessage(shell, "Browser cannot be initialized.");
		}

		// TODO not sure if anything above this should be done in the constructor
		maxHeight = -1;
		maxWidth = -1;
	}


	public BrowserTextHover(String html, Rectangle subjectArea) {
		if(subjectArea == null || subjectArea.isEmpty()) {
			return;
		}

		if(html == null || html.equals("")) {
			return;
		}

		display = new Display();
		shell = initializeShell(display, subjectArea);
		browser = initializeBrowser(shell, display);
		if(browser == null) {
			displayFatalErrorMessage(shell, "Browser cannot be initialized.");
		}

		browser.setText(html);
		shell.open();
		while(!shell.isDisposed()) {
			if(!display.readAndDispatch()) {
				display.sleep();
			}
		}

		dispose();

		// TODO not sure if anything above this should be done in the constructor
		maxHeight = -1;
		maxWidth = -1;
	}


	public BrowserTextHover(Rectangle subjectArea, Object information) {

	}


	private static void displayFatalErrorMessage(final Shell shell, String message) {
		MessageBox messageBox = new MessageBox(shell, SWT.ICON_ERROR | SWT.OK);
		messageBox.setMessage(message);
		messageBox.setText("Exit");
		messageBox.open();
		System.exit(-1);
	}


	/**
	 * Initializes the shell, setting the correct size, location and closing
	 * behaviour
	 *
	 * @param display
	 * @return
	 */
	private final Shell initializeShell(Display display, Rectangle area) {
		final Shell shell = new Shell(display);
		shell.setLayout(new FillLayout());
		shell.setBounds(area);

		/* Listener that closes the shell/browser whenever it loses focus */
		shellListener = new ShellListener() {

			@Override
			public void shellActivated(ShellEvent e) {
			}


			@Override
			public void shellClosed(ShellEvent e) {
				System.out.println("Shell closed by X button, thus closing hover help");
				dispose();
			}


			@Override
			public void shellDeactivated(ShellEvent e) {
				System.out.println("Shell deactivated (e.g. lost focus), thus closing hover help");
				dispose();
			}


			@Override
			public void shellDeiconified(ShellEvent e) {
			}


			@Override
			public void shellIconified(ShellEvent e) {
			}

		};
		shell.addShellListener(shellListener);
		return shell;
	}


	/**
	 * Initializes a bare-bones browser
	 *
	 * @param shell
	 * @param display
	 * @return
	 */
	private Browser initializeBrowser(Shell shell, Display display) {

		Browser browser;
		try {
			browser = new Browser(shell, SWT.NONE);
		}
		catch(SWTError e) {
			e.printStackTrace();
			return null;
		}

		return browser;
	}


	@Override
	public void setInformation(String information) {
		this.information = information; // TODO maybe parse using html generator here, then feed into browser?
	}


	@Override
	public void setSizeConstraints(int maxWidth, int maxHeight) {
		this.maxWidth = maxWidth;
		this.maxHeight = maxHeight;
	}


	@Override
	public Point computeSizeHint() {
		return null; // TODO figure out what this should calculate
	}


	@Override
	public void setVisible(boolean visible) {
		shell.setVisible(visible);
	}


	public void setBounds(Rectangle area) {
		shell.setBounds(area);  // TODO make it honor the maxSize ?
	}


	@Override
	public void setSize(int width, int height) {
		shell.setSize(width, height); // TODO make it honor the maxSize ?
	}


	@Override
	public void setLocation(Point location) {
		shell.setLocation(location);
	}


	@Override
	public void dispose() {
		display.dispose(); // TODO not sure if display should be disposed. Ask Jurgen
		shell.dispose();
		browser.dispose();
	}


	@Override
	public void addDisposeListener(DisposeListener listener) {
		shell.addDisposeListener(listener);
	}


	@Override
	public void removeDisposeListener(DisposeListener listener) {
		shell.removeDisposeListener(listener);
	}


	@Override
	public void setForegroundColor(Color foreground) {
		// TODO implement
	}


	@Override
	public void setBackgroundColor(Color background) {
		// TODO implement
	}


	@Override
	public boolean isFocusControl() {
		return shell.isFocusControl() || browser.isFocusControl();
	}


	@Override
	public void setFocus() {
		shell.setFocus();
		browser.setFocus(); //TODO hmm
	}


	@Override
	public void addFocusListener(FocusListener listener) {
		shell.addFocusListener(listener);
	}


	@Override
	public void removeFocusListener(FocusListener listener) {
		shell.removeFocusListener(listener);
	}


}
