/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 Magne Haveraaen
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Magne Haveraaen
 * 
 *************************************************************************/
#ifndef _MAGNOLIA_PRELUDE_BINARYIO_HXX
#define _MAGNOLIA_PRELUDE_BINARYIO_HXX

namespace Magnolia {
namespace IO {
struct BinaryIStream : MgIStream {
	BinaryIStream(std::istream& s) : MgIStream(s) {}
};

struct BinaryOStream : MgOStream {
	BinaryOStream(std::ostream& s) : MgOStream(s) {}
};

void operator>>=(const startData& tag, BinaryOStream& s) {
	// s.indent();
	s.stream << tag.type.name << "(";
	// s.newline();
	s.nest();
}

void operator<<=(const startData& tag, BinaryIStream& s) {
	std::string t;
	s.readId(t);
	s.expect('(');
	s.nest();
}

void operator>>=(const sepData& tag, BinaryOStream& s) {
	s.stream << ", ";
}

void operator<<=(const sepData& tag, BinaryIStream& s) {
	s.expect(',');
}

void operator>>=(const endData& tag, BinaryOStream& s) {
	s.unnest();
	// s.indent();
	s.stream << ")";
	// s.newline();
}

void operator<<=(const endData& tag, BinaryIStream& s) {
	s.unnest();
	s.expect(')');
}

void operator<<=(signed char& a, BinaryIStream& s) {
	s.stream >> a;
	std::cerr << "read: " << a << std::endl;
}

void operator>>=(const signed char& a, BinaryOStream& s) {
	s.indent();
	std::cerr << "write: " << a << std::endl;
	s.write(a);
	s.newline();
}

void operator<<=(signed short& a, BinaryIStream& s) {
	s.stream >> a;
	std::cerr << "read: " << a << std::endl;
}

void operator>>=(const signed short& a, BinaryOStream& s) {
	s.indent();
	std::cerr << "write: " << a << std::endl;
	s.write(a);
	s.newline();
}

void operator<<=(signed int& a, BinaryIStream& s) {
	s.stream >> a;
	std::cerr << "read: " << a << std::endl;
}

void operator>>=(const signed int& a, BinaryOStream& s) {
	s.indent();
	std::cerr << "write: " << a << std::endl;
	s.write(a);
	s.newline();
}

void operator<<=(signed long int& a, BinaryIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed long int& a, BinaryOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(signed long long int& a, BinaryIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed long long int& a, BinaryOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(unsigned char& a, BinaryIStream& s) {
	s.stream >> a;
	std::cerr << "read: " << a << std::endl;
}

void operator>>=(const unsigned char& a, BinaryOStream& s) {
	s.indent();
	std::cerr << "write: " << a << std::endl;
	s.write(a);
	s.newline();
}

void operator<<=(unsigned short& a, BinaryIStream& s) {
	s.stream >> a;
	std::cerr << "read: " << a << std::endl;
}

void operator>>=(const unsigned short& a, BinaryOStream& s) {
	s.indent();
	std::cerr << "write: " << a << std::endl;
	s.write(a);
	s.newline();
}

void operator<<=(unsigned int& a, BinaryIStream& s) {
	s.stream >> a;
	std::cerr << "read: " << a << std::endl;
}

void operator>>=(const unsigned int& a, BinaryOStream& s) {
	s.indent();
	std::cerr << "write: " << a << std::endl;
	s.write(a);
	s.newline();
}

void operator<<=(unsigned long int& a, BinaryIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned long int& a, BinaryOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(unsigned long long int& a, BinaryIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned long long int& a, BinaryOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(float& a, BinaryIStream& s) {
	s.stream >> a;
	std::cerr << "read: " << a << std::endl;
}

void operator>>=(const float& a, BinaryOStream& s) {
	s.indent();
	std::cerr << "write: " << a << std::endl;
	s.write(a);
	s.newline();
}

void operator<<=(double& a, BinaryIStream& s) {
	s.stream >> a;
}

void operator>>=(const double& a, BinaryOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

void operator<<=(long double& a, BinaryIStream& s) {
	s.stream >> a;
}

void operator>>=(const long double& a, BinaryOStream& s) {
	s.indent();
	s.write(a);
	s.newline();
}

}
}
#endif
