/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
#ifndef _MAGNOLIA_PRELUDE_IO_HXX
#define _MAGNOLIA_PRELUDE_IO_HXX
#include <magnolia/prelude/BinaryIO.hxx>
#include <magnolia/prelude/TextIO.hxx>
#include <magnolia/prelude/XmlIO.hxx>
#include <string>
#include <iostream>
#include <cctype>
namespace Magnolia {
namespace IO {

struct Type {
	const char* name;
	const char* fields;

	constexpr Type(const char* n, const char* fs) : name(n), fields(fs) {}
};

template<typename Stream>
struct MgIOStream {
	Stream& stream;
	int nesting;
	bool bol;

	MgIOStream(Stream& s) : stream(s), nesting(0), bol(true) {}

	void nest() { ++nesting; }
	void unnest() { --nesting; }
};

struct MgIStream : MgIOStream<std::istream> {
	MgIStream(std::istream& s) : MgIOStream(s) {
		s.setf(std::ios_base::skipws);
	}

	void expect(char expected) {
		char c;
		stream >> c;
		if(expected != c)
			stream.setstate(std::ios_base::failbit);
	}

	void readId(std::string& outString) {
		char c;
		while(stream.good()) {
			stream.get(c);
			if(c == '/' || c == '<' || c == '>' || c == '[' || c == ']' || c == '(' || c == ')' || isspace(c)) {
				stream.unget();
				break;
			}
			outString += c;
		}
	}
};

struct MgOStream : MgIOStream<std::ostream> {
	MgOStream(std::ostream& s) : MgIOStream(s) {}

	void newline() {
		if(!bol) {
			stream << std::endl;
			bol = true;
		}
	}
	void indent() {
		if(bol) {
			for(int i = 0; i < nesting; i++)
				stream << "  ";
			bol = false;
		}
	}

	template<typename T>
	void write(const T& t) {
		stream << t;
		bol = false;
	}

	template<typename T>
	void writeBinary(const T& t) {
		stream << t;
		bol = false;
	}
};


struct startData {
	const Type& type;
	startData(const Type& t) : type(t) {}
};

struct sepData {
	sepData() {}
};

struct endData {
	const Type& type;
	endData(const Type& t) : type(t) {}
};

struct startSequence {
	const Type& type;
	const int length;
	startSequence(const Type& t, int l) : type(t), length(l) {}
};

struct sepSequence {
	sepSequence() {}
};

struct endSequence{
	const Type& type;
	endSequence(const Type& t) : type(t) {}
};

}
}

typedef int int_t;
#endif
