/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 Magne Haveraaen
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Magne Haveraaen
 * 
 *************************************************************************/
#ifndef _MAGNOLIA_PRELUDE_TEXTIO_HXX
#define _MAGNOLIA_PRELUDE_TEXTIO_HXX

namespace Magnolia {
namespace IO {

struct TextIStream : MgIStream {
	TextIStream(std::istream& s) : MgIStream(s) {}
};

struct TextOStream : MgOStream {
	TextOStream(std::ostream& s) : MgOStream(s) {}
};

void operator>>=(const startData& tag, TextOStream& s) {
	// s.indent();
	s.write(tag.type.name);
	s.write("(");
	// s.newline();
	s.nest();
}

void operator<<=(const startData& tag, TextIStream& s) {
	std::string t;
	s.readId(t);
	s.expect('(');
	s.nest();
}

void operator>>=(const sepData& tag, TextOStream& s) {
	s.stream << ", ";
}

void operator<<=(const sepData& tag, TextIStream& s) {
	s.expect(',');
}

void operator>>=(const endData& tag, TextOStream& s) {
	s.unnest();
	// s.indent();
	s.write(")");
	// s.newline();
}

void operator<<=(const endData& tag, TextIStream& s) {
	s.unnest();
	s.expect(')');
}

void operator<<=(signed char& a, TextIStream& s) {
	int data;
	s.stream >> data;
	if( data < std::numeric_limits < signed char >::min()
	 || data > std::numeric_limits < signed char >::max())
	    throw "Integer out of range for signed char";
	a = (int8_t) data;
}

void operator>>=(const signed char& a, TextOStream& s) {
	// s.indent();
	s.write((int)a);
	// s.newline();
}

void operator<<=(signed short& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed short& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(signed int& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed int& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(signed long int& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed long int& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(signed long long int& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const signed long long int& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(unsigned char& a, TextIStream& s) {
	int data;
	s.stream >> data;
	if( data < std::numeric_limits < unsigned char >::min()
	 || data > std::numeric_limits < unsigned char >::max())
	    throw "Integer out of range for unsigned char";
	a = (int8_t) data;
}

void operator>>=(const unsigned char& a, TextOStream& s) {
	// s.indent();
	s.write((int)a);
	// s.newline();
}

void operator<<=(unsigned short& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned short& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(unsigned int& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned int& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(unsigned long int& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned long int& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(unsigned long long int& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const unsigned long long int& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(float& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const float& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(double& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const double& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

void operator<<=(long double& a, TextIStream& s) {
	s.stream >> a;
}

void operator>>=(const long double& a, TextOStream& s) {
	// s.indent();
	s.write(a);
	// s.newline();
}

}
}
#endif
