/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.compiler;

import org.rascalmpl.value.IList;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IValueFactory;

public class Util {

	private final IValueFactory vf;


	public Util(IValueFactory vf) {
		this.vf = vf;
	}


	public IString strJoin(IList l) {
		if(l.isEmpty()) {
			return vf.string("");
		}
		else {
			/*
			StringBuilder buf = new StringBuilder();
			for(IValue v : l) {
				buf.append(((IString) v).getValue());
			}
			 */
			int length = 0;
			for(IValue v : l) {
				length += ((IString) v).length();
			}

			StringBuilder buf = new StringBuilder(length);
			for(IValue v : l) {
				buf.append(((IString) v).getValue());
			}
			if(length != buf.length() || buf.capacity() != buf.length()) {
				System.err.println("cap: " + buf.capacity() + ", len: " + buf.length());
			}

			return vf.string(buf.toString());
		}
	}


	public IString strJoin(IList l, IString sep) {
		if(l.isEmpty()) {
			return vf.string("");
		}
		else {
			String sepStr = sep.getValue();

			int length = sepStr.length() * (l.length() - 1);
			for(IValue v : l) {
				length += ((IString) v).length();
			}

			StringBuilder buf = new StringBuilder(length);
			boolean first = true;
			for(IValue v : l) {
				if(!first) {
					buf.append(sepStr);
				}
				else {
					first = false;
				}
				buf.append(((IString) v).getValue());
			}
			if(length != buf.length() || buf.capacity() != buf.length()) {
				System.err.println("cap: " + buf.capacity() + ", len: " + buf.length());
			}

			return vf.string(buf.toString());
		}
	}

}
