/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::compiler::Util
import List;
import String;
import IO;
import org::nuthatchery::pica::terms::Terms;
@doc{Join a list of strings, with a separator}
@javaClass{org.magnolialang.compiler.Util}
public java str strJoin(list[str] l, str sep);/* {
	if(size(l) == 0)
		return "";
	else
		return (head(l)|it+sep+x|x <- tail(l));
}
*/

@doc{Join a list of strings, without a separator}
@javaClass{org.magnolialang.compiler.Util}
public java str strJoin(list[str] l);

public list[str] strSplit(str s, str splitAt) {
	if(contains(splitAt, "\\"))
		throw "strSplit: splitAt should no longer be a RE: <splitAt>";
	if(/^<pre:.*?><splitAt><post:.*>$/ := s) {	
	//		println("(<pre>)(<splitAt>)(<post>)");
		return [pre, *strSplit(post, splitAt)];
	}
	else {
		return [s];
	}
}

/** Insert a separator in a list */
public list[&T] sepBy(list[&T] l, &T sep) {
	if(size(l) == 0)
		return [];
	else
		return ([head(l)]|it+[sep,x]|x <- tail(l));
}

public str hello() {
	return "hello";
}

public str unquote(str s) {
	return visit(s) {
		case /\\\\/ => "\\"
		case /\\n/ => "\n"
		case /\\r/ => "\r"
		case /\\"/ => "\""
		case /\\<c:.>/ => "<c>"
	}
}

public str quote(str s) {
	return visit(s) {
		case /\\/ => "\\\\"
		case /\n/ => "\\n"
		case /\r/ => "\\r"
		case /"/ => "\\\""
	}
}

public str spaces(int n) {
	return ("" | "<it>  " | x <- [0..n], x > 0);
}

public str tabs(int n) {
	return ("" | "<it>\t" | x <- [0..n], x > 0);
}
