/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.annotation.Nullable;
import org.osgi.framework.Bundle;

public final class MagnoliaFileLocator {

	private MagnoliaFileLocator() {

	}


	/**
	 * Locates a file in the Magnolia plugin, and returns a URL.
	 * 
	 * @see org.eclipse.core.runtime.FileLocator#find(Bundle, IPath, Map)
	 */
	@Nullable
	public static URL find(IPath file) {
		URL url = FileLocator.find(MagnoliaPlugin.MAGNOLIA_BUNDLE, file, null);
		if(url == null) {
			url = FileLocator.find(MagnoliaPlugin.MAGNOLIA_BUNDLE, new Path("/src").append(file), null);
		}
		return url;
	}


	@Nullable
	public static IFile findInProject(String project, IPath file) {
		IProject proj = ResourcesPlugin.getWorkspace().getRoot().getProject(project);
		if(proj == null) {
			return null;
		}
		else {
			return proj.getFile(file);
		}
	}


	@Nullable
	public static IFile findInProject(String project, String file) {
		return findInProject(project, new Path(file));
	}


	/**
	 * Locates a file in the Magnolia plugin, and returns an open stream to it,
	 * or null if the file is not found.
	 * 
	 * 
	 * @see org.eclipse.core.runtime.FileLocator#openStream(Bundle, IPath,
	 *      boolean)
	 */
	@Nullable
	public static InputStream openStream(IPath file) throws IOException {
		URL url = find(file);
		if(url == null) {
			return null;
		}
		else {
			return url.openStream();
		}
	}
}
