/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.ICommand;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IProjectNature;
import org.eclipse.core.runtime.CoreException;
import org.magnolialang.eclipse.builders.MagnoliaBuilder;
import org.magnolialang.infra.Infra;
import org.nuthatchery.pica.resources.eclipse.EclipseWorkspaceManager;

public class MagnoliaNature implements IProjectNature {
	public static final String NATURE_ID = MagnoliaPlugin.PLUGIN_ID + ".magnolia.nature";
	private IProject project = null;


	@Override
	public void configure() throws CoreException {
		IProjectDescription desc = project.getDescription();
		for(ICommand cmd : desc.getBuildSpec()) {
			if(cmd.getBuilderName().equals(MagnoliaBuilder.BUILDER_ID)) {
				return;
			}
		}
		ICommand[] cmds = desc.getBuildSpec();
		cmds = Arrays.copyOf(cmds, cmds.length + 1);

		ICommand cmd = desc.newCommand();
		cmd.setBuilderName(MagnoliaBuilder.BUILDER_ID);
		cmds[cmds.length - 1] = cmd;
		desc.setBuildSpec(cmds);
		project.setDescription(desc, null);
		((EclipseWorkspaceManager) Infra.getWorkspaceManager()).openProject(project);

		System.err.println("Magnolia nature configured for " + project.getName());
	}


	@Override
	public void deconfigure() throws CoreException {
		IProjectDescription desc = project.getDescription();
		ICommand[] cmds = desc.getBuildSpec();
		List<ICommand> list = new ArrayList<ICommand>(Arrays.asList(cmds));
		Iterator<ICommand> iter = list.iterator();
		while(iter.hasNext()) {
			ICommand cmd = iter.next();
			if(cmd.getBuilderName().equals(MagnoliaBuilder.BUILDER_ID)) {
				iter.remove();
			}
		}
		desc.setBuildSpec(list.toArray(new ICommand[list.size()]));
		project.setDescription(desc, null);
		((EclipseWorkspaceManager) Infra.getWorkspaceManager()).closeProject(project);
	}


	@Override
	public IProject getProject() {
		return project;
	}


	@Override
	public void setProject(IProject project) {
		this.project = project;
	}
}
