/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;

import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.builders.MagnoliaBuilder;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;

public class BuildPackage extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
		if(editorPart instanceof UniversalEditor) {
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final IProjectManager mng = CommandUtil.getModuleManager(parseCtrl);
			final IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);
			final String CMD_NAME = "Build Package";
			WorkspaceJob job = new WorkspaceJob(CMD_NAME) {
				@Override
				public IStatus runInWorkspace(IProgressMonitor monitor) {
					ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
					tm.begin(new TaskId(getClass().getCanonicalName(), CMD_NAME, pkg.getResource()));
					//parserBenchmark(mng, pkg);

					long t = System.currentTimeMillis();
					MagnoliaBuilder builder = new MagnoliaBuilder();
					try {
						List<IResource> toClean = new ArrayList<IResource>();
						//builder.copyFiles(monitor, toClean);
						builder.buildOutputFiles(mng, tm, pkg, toClean);
					}
					catch(CoreException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					finally {
						CommandUtil.commitMarks(parseCtrl);
						tm.done();
						monitor.done();
					}
					System.err.println("Build package '" + pkg.getName() + "': done in " + (System.currentTimeMillis() - t) + "ms");
					return Status.OK_STATUS;
				}
			};
			job.schedule();
		}
		return null;
	}
}
