/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import java.util.Collection;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.ITuple;
import org.rascalmpl.value.IValue;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.editor.Editor;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.resources.IMagnoliaCodeUnit;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.nuthatchery.pica.terms.TermAdapter;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.control_exceptions.Throw;

public class CollectAxioms extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
		if(editorPart instanceof UniversalEditor) {
			System.out.println("----------------------------------------------------------------------------------------------------------------------------------------");
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final ICompiler compiler = CommandUtil.getCompiler(parseCtrl);
			final IConstructor symbol = CommandUtil.getModuleId(parseCtrl);
			final IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);

			try {

				Job job = new Job("Collect Axioms") {
					@Override
					public boolean belongsTo(Object obj) {
						return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
					}


					@Override
					protected IStatus run(IProgressMonitor monitor) {
						ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
						tm.begin(new TaskId(getClass().getCanonicalName(), "Collect Axioms", pkg.getResource()));
						RefreshHandler.autoRefresh(event);

						try {
							Collection<? extends IMagnoliaCodeUnit> children = pkg.getChildren(tm);
							StringBuilder b = new StringBuilder();

							for(IMagnoliaCodeUnit c : children) {
								System.out.println("Collect Axioms: " + c.getKind(tm) + " " + c.getName());
								if(c.getKind(tm).equals("program")) {
									IConstructor defInfo = c.getDefInfo(tm);
									defInfo = defInfo.set("body", c.getTypechecked(tm));
									IConstructor result = (IConstructor) compiler.call("collectAxioms", defInfo.get("body"));
									if(result.has("val")) {
										IList defs = (IList) ((ITuple) result.get("val")).get(0);
										IList axioms = (IList) ((ITuple) result.get("val")).get(1);
										b.append("ToTransform(\n");
										b.append("[");
										boolean first = true;
										for(IValue v : defs) {
											if(!first) {
												b.append(",");
											}
											b.append(TermAdapter.yieldTerm(v, false));
											b.append("\n");
											first = false;
										}
										b.append("],\n");
										b.append("[");
										first = true;
										for(IValue v : axioms) {
											if(!first) {
												b.append(",");
											}
											b.append(TermAdapter.yieldTerm(v, false));
											b.append("\n");
											first = false;
										}
										b.append("])\n");
									}
								}

							}
							System.out.println("Displaying axioms for " + TermAdapter.yieldTerm(symbol, false));
							// pretty printed AST
							Editor.edit(b.toString(), parseCtrl.getPath().removeFileExtension().toString() + "#" + Magnolia.yieldName(symbol) + ".term");
						}
						catch(Throw t) {
							System.err.println(t);
						}
						catch(Throwable t) { // NOPMD by anya on 1/5/12 3:52 AM
							t.printStackTrace();
						}
						CommandUtil.commitMarks(parseCtrl);

						// compiler.visualize();
						System.out.println("Done!!");
						tm.done();
						monitor.done();
						return Status.OK_STATUS;
					}

				};
				job.schedule();
			}
			catch(final Throwable e1) { // NOPMD by anya on 1/5/12 3:52 AM
				e1.printStackTrace();
			}
		}
		return null;
	}
}
