/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import java.util.ArrayList;
import java.util.Collection;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;

import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.eclipse.ui.texteditor.AbstractTextEditor;
import org.magnolialang.eclipse.editor.Editor;
import org.magnolialang.errors.CompilationErrorException;
import org.magnolialang.magnolia.MagnoliaCompiler;
import org.magnolialang.magnolia.resources.IMagnoliaCodeUnit;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.NullTaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.NullRascalMonitor;

public class GenerateCxxCode extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
		if(editorPart instanceof UniversalEditor) {
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final MagnoliaCompiler compiler = CommandUtil.getCompiler(parseCtrl);
			final IConstructor symbol = CommandUtil.getModuleId(parseCtrl);
			final IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);
			final ITextSelection selection = (ITextSelection) ((AbstractTextEditor) editorPart).getSelectionProvider().getSelection();

			WorkspaceJob job = new WorkspaceJob("Generate C++ Code") {
				@Override
				public IStatus runInWorkspace(IProgressMonitor monitor) {
					ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
					tm.begin(new TaskId(getClass().getCanonicalName(), "Generate C++ Code", pkg.getResource()));
					//parserBenchmark(mng, pkg);

					long t0 = System.currentTimeMillis();
					RefreshHandler.autoRefresh(event);
					long t1 = System.currentTimeMillis();
					long t2 = System.currentTimeMillis();
					;
					try {
						String fileName = parseCtrl.getPath().removeFileExtension().toString() + ".cxx";
						IConstructor module = pkg.getTypechecked(tm);
						Collection<? extends IMagnoliaCodeUnit> children = pkg.getChildren(tm);
						ArrayList<IValue> r = new ArrayList<IValue>();
						StringBuilder result = new StringBuilder();
						int offset = selection.getOffset();

						for(IMagnoliaCodeUnit c : children) {
							if(offset == -1 || c.containsOffset(offset)) {
								IConstructor defInfo = c.getDefInfo(tm);
								defInfo = defInfo.set("body", c.getTypechecked(tm));
								IConstructor mutified = (IConstructor) compiler.call(tm, "getMutified", defInfo);
								defInfo = defInfo.set("body", mutified.get("val"));
								IConstructor cxxResult = (IConstructor) compiler.call(tm, "cxxCode", defInfo, TermFactory.vf.string(fileName), TermFactory.vf.bool(true));
								if(cxxResult.has("val")) {
									result.append(((IString) cxxResult.get("val")).getValue());
								}
								r.add(mutified.get("val"));
							}
						}

						t2 = System.currentTimeMillis();
/*						for(IValue v : r) {
							result += ((IString) compiler.call("prettyMagnolia", v)).getValue();
							result += "\n\n";
						}
 */						//String result = ((IString)value).getValue();

/*						IValue mutKey = MagnoliaFacts
								.keyedName(compiler.getEvaluator(), Arrays.asList(MagnoliaFacts.Type_Alerted, MagnoliaFacts.Type_Mutified, MagnoliaFacts.Type_TypeChecked), symbol);
						IValue foo = CommandUtil.getVal(compiler.getFact(tm, MagnoliaFacts.Type_CxxCode, mutKey));
 */ if(result.length() > 0) {
							System.out.println("Displaying C++ for " + TermAdapter.yieldTerm(symbol, false));
							// String yield = TermAdapter.yieldTerm(info, false);
							// String yield = TermAdapter.yield(value, new MagnoliaSkin(), false);
							//IList contents = (IList) value.get("val");
							//or(IValue c : contents)
							//	yield += TermAdapter.yieldTerm(c, false) + "\n";
							Editor.edit(result.toString(), fileName);
						}
// compiler.visualize();
					}
					catch(final CompilationErrorException ce) {
						ce.printStackTrace();
						// manager.reportError(ce);
					}
					catch(final Throwable e1) { // NOPMD by anya on 1/5/12 3:52 AM
						if(e1.getStackTrace() != null) {
							e1.printStackTrace();
						}
						else {
							System.err.println(e1);
							// throw new ImplementationError("Flattening failed:",
							// e1);
						}
					}
					finally {
						CommandUtil.commitMarks(parseCtrl);

						tm.done();
						monitor.done();
					}
					long t3 = System.currentTimeMillis();
					System.err.printf("Generate C++ Code: %dms refresh + %dms work + %dms prettyprint = %dms total%n", t1 - t0, t2 - t1, t3 - t2, t3 - t0);
					return Status.OK_STATUS;
				}


				@SuppressWarnings("unused")
				private void parserBenchmark(final IProjectManager mng, final IMagnoliaPackage pkg) {
					int N = 20;
					long[] timing = new long[N];
					for(int i = 0; i < N; i++) {
						System.out.println("Iteration " + i);
						long t0 = System.currentTimeMillis();
						for(IManagedCodeUnit p : mng.allPackages(pkg.getLanguage())) {
							p.onResourceChanged();
							((IMagnoliaCodeUnit) p).getAST(new NullTaskMonitor());
						}
						timing[i] = System.currentTimeMillis() - t0;
					}
					System.err.flush();
					System.out.flush();
					System.out.println("REPORT: ");
					System.out.print("  Timings: ");
					long smallest = Long.MAX_VALUE;
					long sum = 0;
					for(long l : timing) {
						sum += l;
						smallest = Math.min(smallest, l);
						System.out.printf("%5d ", l);
					}
					System.out.println();
					System.out.printf("  Smallest: %6d%n", smallest);
					System.out.printf("  Average:  %6d%n", sum / N);
					System.out.printf("  TOTAL:    %6d%n", sum);
				}
			};
			job.schedule();
		}
		return null;
	}
}
