/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.ITuple;
import org.rascalmpl.value.IValue;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.errors.CompilationErrorException;
import org.magnolialang.magnolia.MagnoliaCompiler;
import org.magnolialang.magnolia.resources.IMagnoliaModule;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;

public class GenerateCxxSkeleton extends AbstractHandler {
	GeneratedFileSaver skeletonUI = new GeneratedFileSaver("C++ Library Skeleton");


	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);

		if(editorPart instanceof UniversalEditor) {
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final MagnoliaCompiler compiler = CommandUtil.getCompiler(parseCtrl);
			final IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);

			Job job = new Job("Generate C++ Library Skeleton") {
				@Override
				public boolean belongsTo(Object obj) {
					return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
				}


				@Override
				protected IStatus run(IProgressMonitor monitor) {
					ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
					tm.begin(new TaskId(getClass().getCanonicalName(), "Generate C++ Library Skeleton", pkg.getResource()));
					RefreshHandler.autoRefresh(event);
					try {
						for(IMagnoliaModule member : pkg.getChildren(tm)) {
							tm.subTask("Generating skeleton for " + member.getName());
							IConstructor defInfo = member.getDefInfo(tm);
							IConstructor body = member.getTypechecked(tm);
							ISet externals = member.getExternals(tm);
							// defInfo = defInfo.set("body", body);
							// TODO
							IConstructor result = (IConstructor) compiler.call(tm, "generateCxxSkeleton", defInfo, body, externals);
							if(result.has("val")) {
								IList results = (IList) result.get("val");
								for(IValue value : results) {
									ITuple tuple = (ITuple) value;
									final String code = ((IString) tuple.get(0)).getValue();
									final String fileName = ((IString) tuple.get(1)).getValue();
									if(!fileName.equals("")) {
										skeletonUI.addFile(fileName, code, CommandUtil.getModuleManager(parseCtrl), parseCtrl.getProject().getRawProject(), GeneratedFileSaver.Root.SOURCEFOLDER, IResource.NONE);
										Display.getDefault().asyncExec(skeletonUI);
									}
								}
							}
							else {
								System.err.println(result.get("message"));
							}
							tm.done();
						}
					}
					catch(final CompilationErrorException ce) {
						ce.printStackTrace();
						// manager.reportError(ce);
					}
					catch(final Throwable e1) { // NOPMD by anya on 1/5/12 3:52 AM

						e1.printStackTrace();
						// throw new ImplementationError("Flattening failed:",
						// e1);
					}
					finally {
						CommandUtil.commitMarks(parseCtrl);

						tm.done();
						monitor.done();
					}
					return Status.OK_STATUS;
				}
			};
			job.schedule();
		}
		return null;
	}


}
