/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.ltk.core.refactoring.DocumentChange;
import org.eclipse.ltk.core.refactoring.PerformChangeOperation;
import org.eclipse.ltk.core.refactoring.TextChange;
import org.eclipse.swt.widgets.Display;
import org.eclipse.text.edits.DeleteEdit;
import org.eclipse.text.edits.InsertEdit;
import org.eclipse.text.edits.MultiTextEdit;
import org.eclipse.text.edits.ReplaceEdit;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.editor.MagnoliaEditor;
import org.magnolialang.eclipse.editor.MagnoliaParseController;
import org.magnolialang.magnolia.MagnoliaCompiler;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.control_exceptions.Throw;

public class Inline extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);

		if(editorPart instanceof MagnoliaEditor) {
			final MagnoliaEditor editor = (MagnoliaEditor) editorPart;
			InlineJob job = new InlineJob(event, (MagnoliaParseController) editor.getParseController(), (ITextSelection) editor.getSelectionProvider().getSelection());
			job.schedule();
		}
		return null;
	}


	@Override
	public boolean isEnabled() {
		return false;
	}


	static class InlineJob extends Job {
		public ISourceLocation loc = null;
		protected final ExecutionEvent event;
		protected final ITextSelection selection;
		protected final MagnoliaParseController parseCtrl;


		InlineJob(ExecutionEvent event, MagnoliaParseController parseCtrl, ITextSelection iSelection) {
			super("Inline");
			this.event = event;
			this.selection = iSelection;
			this.parseCtrl = parseCtrl;
		}


		@Override
		public boolean belongsTo(Object obj) {
			return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
		}


		@Override
		public IStatus run(IProgressMonitor monitor) {

			RefreshHandler.autoRefresh(event);

			try {
				IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);
				MagnoliaCompiler compiler = CommandUtil.getCompiler(parseCtrl);
				ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
				tm.begin(new TaskId(getClass().getCanonicalName(), "Inline", pkg.getResource()));
				IConstructor typechecked = pkg.getTypechecked(tm);
				IValue result = compiler.call(tm, "inlineByPoint", typechecked, TermFactory.vf.integer(selection.getOffset()));
				IList diff = (IList) compiler.call(tm, "treeEditDiff", typechecked, result);
				IDocument doc = parseCtrl.getEditor().getDocumentProvider().getDocument(parseCtrl.getEditor().getEditorInput());
				TextChange docChange = new DocumentChange(pkg.getName(), doc); // EclipseWorkspaceManager.getEclipseFile(pkg));
				docChange.initializeValidationData(monitor);
				MultiTextEdit rootEdit = new MultiTextEdit();
				docChange.setEdit(rootEdit);
				for(IValue d : diff) {
					IConstructor change = (IConstructor) d;
					ISourceLocation loc = (ISourceLocation) change.get("location");
					String newCode = null;

					System.out.println("Change: " + change.getName() + " at " + change.get(0));
					if(change.has("old")) {
						System.out.println("    " + TermAdapter.yieldTerm(change.get("old"), false));
					}
					if(change.has("new")) {
						System.out.println(" -> " + TermAdapter.yieldTerm(change.get("new"), false));
						newCode = ((IString) compiler.call(tm, "prettyMagnolia", compiler.call(tm, "ensugarTree", change.get("new")))).getValue();
					}

					if(change.getName().equals("Replace")) {
						rootEdit.addChild(new ReplaceEdit(loc.getOffset(), loc.getLength(), newCode));
					}
					else if(change.getName().equals("Delete")) {
						rootEdit.addChild(new DeleteEdit(loc.getOffset(), loc.getLength()));
					}
					else if(change.getName().equals("Insert")) {
						rootEdit.addChild(new InsertEdit(loc.getOffset(), newCode));
					}
					else if(change.getName().equals("Copy")) {
						// rootEdit.addChild(new CopySourceEdit(loc.getOffset(), loc.getLength(), newCode));
					}
				}
				if(rootEdit.hasChildren()) {
					PerformChangeOperation operation = new PerformChangeOperation(docChange);
					ResourcesPlugin.getWorkspace().run(operation, monitor);
				}
				//IValue value = compiler.call(tm, "prettyMagnolia", result);
				//Editor.edit(((IString) value).getValue(), parseCtrl.getPath().removeFileExtension().toString() + ".mg");
			}
			catch(Throw t) {
				IValue exception = t.getException();
				if(exception instanceof IConstructor) {
					final IConstructor ex = (IConstructor) exception;
					if(ex.getName().equals("NoDefFound")) {
						final ISourceLocation loc = (ISourceLocation) ex.get("location");
						Display.getDefault().syncExec(new Runnable() {
							@Override
							public void run() {
								MagnoliaEditor editor = parseCtrl.getEditor();
								if(editor != null) {
									editor.selectAndReveal(loc.getOffset(), loc.getLength());
								}
								MessageDialog.openError(null, "Inlining Error", "No definition found for '" + ((IString) ex.get(0)).getValue() + "'");
							}
						});
						//Status status = new Status(IStatus.ERROR, MagnoliaPlugin.PLUGIN_ID, "No definition found for '" + ((IString) ex.get(0)).getValue() + "'\nmore info");
						//StatusManager.getManager().handle(status, StatusManager.BLOCK);
					}
					else if(ex.getName().equals("NoInlineableFound")) {
						// Status status = new OperationStatus(IStatus.ERROR, MagnoliaPlugin.PLUGIN_ID, Status.OK, , null);
						Display.getDefault().syncExec(new Runnable() {
							@Override
							public void run() {
								MessageDialog.openError(null, "Inlining Error", "Nothing to inline at this position");
							}
						});
						// ErrorDialog.openError(null, null, null, status);
						//StatusManager.getManager().handle(status, StatusManager.BLOCK);
					}
				}
				else {
					throw t;
				}
			}
			catch(ExecutionException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch(CoreException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			try {
				CommandUtil.getModuleManager(parseCtrl).commitMarks();
			}
			catch(ExecutionException e) {
				e.printStackTrace();
			}

			monitor.done();

			return Status.OK_STATUS;
		}
	};
}
