/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.runtime.IPath;
import io.usethesource.impulse.editor.EditorUtility;
import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.eclipse.jface.text.IRegion;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.editor.MagnoliaParseController;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.magnolialang.magnolia.services.IMgXRefInfo;
import org.nuthatchery.pica.rascal.RascalUtil;
import org.nuthatchery.pica.resources.IXRefInfo;
import org.nuthatchery.pica.resources.eclipse.EclipseWorkspaceManager;
import org.nuthatchery.pica.tasks.NullTaskMonitor;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.interpreter.NullRascalMonitor;

public class OpenDeclaration extends AbstractHandler {

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
		if(editorPart instanceof UniversalEditor) {
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);

			IRegion region = ((UniversalEditor) editorPart).getSelectedRegion();

			IMgXRefInfo xRefInfo = (IMgXRefInfo) pkg.getXRefs(RascalUtil.toCodeRegion(TermFactory.vf.sourceLocation(((MagnoliaParseController) parseCtrl).getSourceLocation(), region.getOffset(), region.getLength())), new NullTaskMonitor());

			if(xRefInfo != null) {
				IConstructor info = xRefInfo.getDefInfo();

				if(info != null) {
					IConstructor locInfo = (IConstructor) info.get("locInfo");
					if(locInfo.has("defLoc")) {
						ISourceLocation loc = (ISourceLocation) locInfo.get("defLoc");

						IPath path = EclipseWorkspaceManager.getPath(loc.getURI());
						IEditorPart editor = EditorUtility.isOpenInEditor(path);
						if(editor == null) {
							try {
								System.err.println("Opening new editor at " + loc);
								editor = EditorUtility.openInEditor(path);
							}
							catch(PartInitException e) {
								MagnoliaPlugin.getInstance().logException("Failed to open editor for hyperlink", e);
							}
						}
						System.err.println("Opening location in editor at " + loc);
						EditorUtility.revealInEditor(editor, loc.getOffset(), loc.getLength());
						IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
						page.activate(editor);
					}
				}
			}
			else {
				// TODO: pop-up list of targets
			}
			CommandUtil.commitMarks(parseCtrl);

		}
		return null;
	}

}
