/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import nuthatch.rascal.adapter.ValuesCursor;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.type.Type;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.editor.Editor;
import org.magnolialang.magnolia.backend.Instantiator;
import org.magnolialang.magnolia.lib.Pretty;
import org.magnolialang.magnolia.resources.IMagnoliaCodeUnit;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pgf.trees.PrettyPrinter;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.control_exceptions.Throw;

public class RunOptimiser extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
		try {
			System.out.println("---------------------------------------------------------------------------------------------------------------------------------------");
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final ICompiler compiler = CommandUtil.getCompiler(parseCtrl);
			final IConstructor symbol = CommandUtil.getModuleId(parseCtrl);
			final IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);
			final String CMD_NAME = "Magnolia Optimser";
			Job job = new Job("Magnolia optimiser") {
				@Override
				public boolean belongsTo(Object obj) {
					return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
				}


				@Override
				protected IStatus run(IProgressMonitor monitor) {
					ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
					tm.begin(new TaskId(getClass().getCanonicalName(), CMD_NAME, pkg.getResource()));

					RefreshHandler.autoRefresh(event);

					System.out.println("Magnolia optimiser: " + symbol);

					// IList contents = (IList) ((IConstructor) compiler.getFact(tm, MagnoliaFacts.Type_Contents, symbol)).get("val");
					StringBuilder s = new StringBuilder();
					s.append("package ");
					s.append(pkg.getName());
					s.append(";\n\n");
					try {
						for(IMagnoliaCodeUnit c : pkg.getChildren(tm)) {
							IValue value = c.getTypechecked(tm);

							IConstructor tree = Instantiator.instantiate((IConstructor) value);
							IConstructor info = c.getDefInfo(tm);
							info = info.set("body", tree);

							IConstructor result = (IConstructor) compiler.call(tm, "getOptimised", info);
							info = info.set("body", result.get("val"));
							PrettyPrinter<IValue, Type> printer = Pretty.makePrinter();
							s.append(c.getKind(tm));
							s.append(" ");
							s.append(c.getName().replaceAll("^.*\\.", ""));
							s.append(" = ");
							s.append(printer.toString(new ValuesCursor(result.get("val"))) + ";\n\n\n");
							// value = compiler.call(tm, "prettyMagnolia", result.get("val"));
						}
						Editor.edit(s.toString(), parseCtrl.getPath().removeFileExtension().toString()
						// + "#"
						// + TermAdapter.yield(sym,
						// new MagnoliaSkin(), false)
						+ "_opt.mg");

					}
					catch(Throw t) {
						System.err.println(t);
						System.err.println(t.getTrace());
					}
					catch(Throwable t) { // NOPMD by anya on 1/5/12 3:51 AM
						t.printStackTrace();
					}
					CommandUtil.commitMarks(parseCtrl);

					// compiler.visualize();
					System.out.println("Done!!");
					tm.done();
					monitor.done();
					return Status.OK_STATUS;
				}

			};
			job.schedule();

			/*
			 * for(IValue sym : db.symbols()) { System.out.println(sym);
			 * IValue markers = db.getAttr(sym, IProgramDB.ATTR_MARKERS);
			 * if(markers != null && markers instanceof ISet) { for(IValue
			 * markerVal : (ISet) markers) { if(markerVal instanceof ITuple)
			 * { ITuple marker = (ITuple) markerVal; ISourceLocation loc =
			 * (ISourceLocation) marker .get(0); IInteger severity =
			 * (IInteger) marker.get(1); IString message = (IString)
			 * marker.get(2); manager.reportError(message.getValue(), loc,
			 * ErrorMarkers.TYPE_COMPILATION_ERROR, ErrorMarkers.SEVERITY_INT_MAP
			 * .get(severity.intValue())); } } } }
			 *
			 * Editor.edit(result.getValue(), parseCtrl.getPath()
			 * .removeFileExtension().toString() + ".rascal");
			 *
			 * Editor.edit(db.dumpDB(), parseCtrl.getPath()
			 * .removeFileExtension().toString() + ".db_dump");
			 */
		}
		catch(final Throwable e1) { // NOPMD by anya on 1/5/12 3:51 AM
			e1.printStackTrace();
		}
		return null;
	}
}
