/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Andreas Hjortland
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.eclipse.ui.texteditor.AbstractTextEditor;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.errors.CompilationErrorException;
import org.magnolialang.magnolia.MagnoliaCompiler;
import org.magnolialang.magnolia.backend.Instantiator;
import org.magnolialang.magnolia.evaluator.AlgoEvaluator;
import org.magnolialang.magnolia.evaluator.values.Value;
import org.magnolialang.magnolia.repr.ConvertAST;
import org.magnolialang.magnolia.resources.IMagnoliaModule;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.nuthatchery.pica.util.depgraph.IDepGraph;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.uri.URIUtil;

public class RunProgram extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);

		if(editorPart instanceof UniversalEditor) {
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final MagnoliaCompiler compiler = CommandUtil.getCompiler(parseCtrl);
			final IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);
			final ITextSelection selection = (ITextSelection) ((AbstractTextEditor) editorPart).getSelectionProvider().getSelection();
			final PrintWriter errorWriter = CommandUtil.getConsoleErrorWriter();
			final PrintWriter outputWriter = CommandUtil.getConsoleOutputWriter();
			final Reader inputReader = CommandUtil.getConsoleInputReader();
			final String CMD_NAME = "Run Program";
			CommandUtil.showConsole();
			IProject rawProject = parseCtrl.getProject().getRawProject();
			URLClassLoader loader = null;
			IJavaProject p = JavaCore.create(rawProject);
			IClasspathEntry[] classpath;
			try {
				IPath fullPath = rawProject.getFullPath();
				URI locationURI = rawProject.getLocationURI();

				IPath path = p.getOutputLocation().makeRelativeTo(rawProject.getFullPath());
				URL projOutURL = URIUtil.createFile(rawProject.getLocationURI().getPath() + "/" + path.toString() + "/").toURL();
				//URI rawLocationURI = rawProject.getRawLocationURI();

				//URI uri = rawLocationURI..resolve(p.getOutputLocation().toPortableString());
				//IPath outputLocation = rawProject.getWorkspace().getRoot().getLocation().append(p.getOutputLocation()).addTrailingSeparator();
				//URL projOutURL = URIUtil.createFile(outputLocation.toString()).toURL();
//				URL projOutURL = URIUtil.assumeCorrect(locationURI.toString() + p.getOutputLocation().toPortableString()).toURL();
				loader = new URLClassLoader(new URL[] { projOutURL }, getClass().getClassLoader());
				//classpath = p.getResolvedClasspath(true);
				//System.out.println(Arrays.toString(classpath));
			}
			catch(JavaModelException | URISyntaxException | MalformedURLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}


			final URLClassLoader classLoader = loader;
			Job job = new Job(CMD_NAME) {
				@Override
				public boolean belongsTo(Object obj) {
					return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
				}


				@Override
				protected IStatus run(IProgressMonitor monitor) {
					ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
					tm.begin(new TaskId(getClass().getCanonicalName(), CMD_NAME, pkg.getResource()));
					RefreshHandler.autoRefresh(event);
					try {
						int offset = selection.getOffset();
						for(IMagnoliaModule member : pkg.getChildren(tm)) {
							if(offset == -1 || member.containsOffset(offset)) {
								tm.subTask("Running " + member.getName() + ".main()");
								errorWriter.println("Executing " + member.getName() + ".main()...");
								IConstructor body = member.getTypechecked(tm);

								IConstructor tree = Instantiator.instantiate(body);

								/*try {
									ConvertAST converted = new ConvertAST();
									IDepGraph<IConstructor> graph = converted.getDFGraph(body);
								}
								catch(Throwable t) {
									t.printStackTrace();
									outputWriter.flush();
									t.printStackTrace(errorWriter);
								}*/
								AlgoEvaluator evaluator = new AlgoEvaluator(member.getName(), tree, classLoader, outputWriter, errorWriter, inputReader);

								errorWriter.flush();

								Object value = evaluator.apply("main");
								if(value != null && value != Value.UNDEFINED) {
									outputWriter.println(value.toString());
									System.err.println(value);
								}
								outputWriter.flush();
								errorWriter.println("OK");
								tm.done();
							}
						}
					}
					catch(final CompilationErrorException ce) {
						ce.printStackTrace();
						outputWriter.flush();
						ce.printStackTrace(errorWriter);
					}
					catch(final Throwable e1) { // NOPMD by anya on 1/5/12 3:52 AM

						e1.printStackTrace();
						outputWriter.flush();
						e1.printStackTrace(errorWriter);
						// throw new ImplementationError("Flattening failed:",
						// e1);
					}
					finally {
						try {
							if(classLoader != null) {
								classLoader.close();
							}
						}
						catch(IOException e) {
							e.printStackTrace();
						}
						CommandUtil.commitMarks(parseCtrl);
						errorWriter.close();
						outputWriter.close();
						try {
							inputReader.close();
						}
						catch(IOException e) {
							// ignore
						}

						tm.done();
						monitor.done();
					}
					return Status.OK_STATUS;
				}
			};
			job.schedule();
		}
		return null;
	}
}
