/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.editor.MagnoliaEditor;
import org.magnolialang.eclipse.editor.MagnoliaParseController;
import org.magnolialang.eclipse.editor.NodeLocator;
import org.nuthatchery.pica.terms.SortNameMatcher;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.values.uptr.ITree;
import org.rascalmpl.values.uptr.TreeAdapter;

public class SelectConstruct extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);

		if(editorPart instanceof MagnoliaEditor) {
			final MagnoliaEditor editor = (MagnoliaEditor) editorPart;
			SelectJob job = new SelectJob(event, (MagnoliaParseController) editor.getParseController(), editor.getSelection());
			job.schedule();

		}
		return null;
	}


	static class SelectJob extends Job {
		public ISourceLocation loc = null;
		final ExecutionEvent event;
		final MagnoliaParseController parseCtrl;
		final Point point;


		SelectJob(ExecutionEvent event, MagnoliaParseController parseCtrl, Point p) {
			super("Select Construct");
			this.event = event;
			this.point = p;
			this.parseCtrl = parseCtrl;
		}


		@Override
		public boolean belongsTo(Object obj) {
			return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
		}


		@Override
		public IStatus run(IProgressMonitor monitor) {
			IRascalMonitor rm = new RascalMonitor(monitor, new WarningsToMarkers());
			rm.startJob("Select construct");
			RefreshHandler.autoRefresh(event);
			final NodeLocator locator = parseCtrl.getSourcePositionLocator();
			ITree tree = parseCtrl.getCurrentAst();
			if(tree != null) {
				tree = locator.findNode(tree, point.x, new SortNameMatcher("Expr", "TopExpr", "Type", "Decl", "TopDecl", "RealProgram"), NodeLocator.INNERMOST);
				if(tree != null) {
					loc = TreeAdapter.getLocation(tree);
				}
			}
			if(loc != null) {
				Display.getDefault().syncExec(new Runnable() {
					@Override
					public void run() {
						MagnoliaEditor editor = parseCtrl.getEditor();
						if(editor != null) {
							editor.selectAndReveal(loc.getOffset(), loc.getLength());
						}
					}
				});
			}
			monitor.done();
			return Status.OK_STATUS;
		}
	};

}
