/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import io.usethesource.impulse.services.IDocumentationProvider;
import org.eclipse.jdt.annotation.Nullable;
import org.rascalmpl.values.uptr.RascalValueFactory;

/*
 * Assuming the innermost lexical node in the parse tree is given, we simply
 * return the annotation labeled
 * "doc" which is a string.
 */
public class DocumentationProvider implements IDocumentationProvider {

	@Override
	@Nullable
	public String getDocumentation(final Object target, final IParseController parseController) {
		if(target instanceof IConstructor) {
			IConstructor tgt = (IConstructor) target;
			if(tgt.getType().equivalent(RascalValueFactory.Tree)) {
				return getDocString(tgt);
			}
		}

		return null;
	}


	@Nullable
	private String getDocString(final IConstructor arg) {
		final IValue val = arg.asAnnotatable().getAnnotation("doc");

		if(val != null && val.getType().isString()) {
			return ((IString) val).getValue();
		}

		return null;
	}
}
