/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.net.URI;
import java.net.URISyntaxException;

import org.eclipse.core.runtime.IPath;
import io.usethesource.impulse.editor.EditorUtility;
import org.rascalmpl.value.IValue;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.ide.IDE;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.nuthatchery.pica.errors.ImplementationError;

public final class Editor {

	public static void edit(final IPath path) {
		IEditorPart editor = EditorUtility.isOpenInEditor(path);
		if(editor == null) {
			try {
				editor = EditorUtility.openInEditor(path);
			}
			catch(PartInitException e) {
				MagnoliaPlugin.getInstance().logException("Failed to open editor for path " + path, e);
				return;
			}
		}
		editor.getEditorSite().getPage().activate(editor);
	}


	public static void edit(final IPath path, int offset, int length) {
		IEditorPart editor = EditorUtility.isOpenInEditor(path);
		if(editor == null) {
			try {
				editor = EditorUtility.openInEditor(path);
			}
			catch(PartInitException e) {
				MagnoliaPlugin.getInstance().logException("Failed to open editor for path " + path, e);
				return;
			}
		}
		EditorUtility.revealInEditor(editor, offset, length);
		editor.getEditorSite().getPage().activate(editor);
	}


	public static void edit(final IValue v) {
		openEditor(new ValueEditorInput(v));
	}


	public static void edit(final String value, final String uri) {
		try {
			openEditor(new StringEditorInput(uri, value));
		}
		catch(URISyntaxException e) {
			throw new ImplementationError("Illegal path '" + uri + "' in editor URI", e);
		}
	}


	public static void edit(final String value, final URI uri) {
		openEditor(new StringEditorInput(uri, value));
	}


	private static void openEditor(final IEditorInput input) {
		IWorkbench wb = PlatformUI.getWorkbench();
		IWorkbenchWindow win = wb.getActiveWorkbenchWindow();

		if(win == null && wb.getWorkbenchWindowCount() != 0) {
			win = wb.getWorkbenchWindows()[0];
		}

		if(win != null) {
			final IWorkbenchPage page = win.getActivePage();
			if(page != null) {
				Display.getDefault().asyncExec(new Runnable() {
					@Override
					public void run() {
						String editorId = "org.eclipse.ui.DefaultTextEditor";
						try {
							editorId = IDE.getEditorDescriptor(input.getName()).getId();
						}
						catch(PartInitException e1) { // NOPMD by anya on 1/5/12 5:02 AM
							// never mind, use default
						}
						try {
							IDE.openEditor(page, input, editorId, true);
						}
						catch(PartInitException e) {
							MagnoliaPlugin.getInstance().logException("Failed to open editor", e);
						}
					}
				});
			}
		}
	}


	private Editor() {

	}
}
