/**************************************************************************
 * Copyright (c) 2013 Anya Helene Bagge
 * Copyright (c) 2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import io.usethesource.impulse.services.ISourceFormatter;
import org.nuthatchery.pgf.config.TokenizerConfigBase;
import org.nuthatchery.pgf.examples.Indenter;
import org.nuthatchery.pgf.examples.SensibleSpacing;
import org.nuthatchery.pgf.plumbing.impl.BufferedSyncPipeComponent;
import org.nuthatchery.pgf.processors.Printer;
import org.nuthatchery.pgf.rascal.uptr.UPTRTokenizer;
import org.nuthatchery.pgf.tokens.CategoryStore;
import org.nuthatchery.pgf.tokens.Token;
import org.nuthatchery.pgf.tokens.TokensToString;

public class Formatter implements ISourceFormatter {

	@Override
	public String format(IParseController parseController, String content, boolean isLineStart, String indentation, int[] positions) {
		MagnoliaParseController ctrl = (MagnoliaParseController) parseController;

		IConstructor pt = ctrl.getCurrentAst();
		if(pt != null) {
			Config config = new Config();
			UPTRTokenizer tokenizer = new UPTRTokenizer(config);
			BufferedSyncPipeComponent<Token, Token> output = new BufferedSyncPipeComponent<>(new Printer<Token>(new PrintWriter(System.out), " "));

			SensibleSpacing respacer = new SensibleSpacing(config.cfgCategories());
			Indenter indenter = new Indenter(config.cfgCategories());
			TokensToString tokensToString = new TokensToString();
			output.connect(new BufferedSyncPipeComponent<>(respacer)).connect(new BufferedSyncPipeComponent<>(indenter)).connect(new BufferedSyncPipeComponent<>(new Printer<Token>(new PrintWriter(System.err), " "))).connect(new BufferedSyncPipeComponent<>(tokensToString));
			tokenizer.tokenize(pt, output);
			output.end();
			System.out.println();
			System.out.println("RESULT: ");
			System.out.println(tokensToString.toString());
			System.out.println("***");
			return tokensToString.toString();
		}
		return null;
	}


	@Override
	public void formatterStarts(String initialIndentation) {
		// TODO Auto-generated method stub

	}


	@Override
	public void formatterStops() {
		// TODO Auto-generated method stub

	}


	static class Config extends TokenizerConfigBase {
		Set<String> nestSorts = Collections.unmodifiableSet(new HashSet<String>(Arrays.asList("Expr", "Stat", "Decl*")));


		@Override
		public Set<String> cfgNestSorts() {
			return nestSorts;
		}


		@Override
		protected String cfgIdentifierRegex() {
			return DEFAULT_ID_REGEX;
		}


		@Override
		protected CatSet cfgInitialCategorySet() {
			return CatSet.CATSET_DEFAULT;
		}


		@Override
		protected String cfgKeywordRegex() {
			return DEFAULT_KW_REGEX;
		}


		@Override
		protected Collection<String> cfgKeywords() {
			return Arrays.asList("implementation", "require", "type", "function", "satisfaction", "if", "then", "else", "end");
		}


		@Override
		protected boolean cfgUseDefaultLiteralClassification() {
			return true;
		}


		@Override
		protected void moreCategories(CategoryStore store) {
			store.declare("EQ", "BINOP", "PUNCT");
			addLitString("=", "EQ");
		}

	}
}
