/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import org.eclipse.jdt.annotation.Nullable;

import java.net.URI;

import org.eclipse.core.runtime.IPath;
import org.rascalmpl.value.ISourceLocation;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.Region;
import org.eclipse.jface.text.hyperlink.IHyperlink;
import org.eclipse.ui.texteditor.ITextEditor;
import org.nuthatchery.pica.resources.eclipse.EclipseWorkspaceManager;
import org.nuthatchery.pica.resources.regions.ICodeRegion;
import org.nuthatchery.pica.resources.regions.IOffsetLength;

public class Hyperlink implements IHyperlink {
	private final IRegion region;
	private final String text;
	private final String type;
	private final ISourceLocation dest;
	private final IOffsetLength loc;
	private final ITextEditor editor;


	public Hyperlink(IOffsetLength iOffsetLength, ISourceLocation dest, String text, String type, ITextEditor editor) {
		this.region = new Region((int) iOffsetLength.getOffset(), (int) iOffsetLength.getLength());
		this.loc = iOffsetLength;
		this.dest = dest;
		this.text = text;
		this.type = type;
		this.editor = editor;
	}


	@Override
	public IRegion getHyperlinkRegion() {
		return region;
	}


	@Override
	@Nullable
	public String getHyperlinkText() {
		return text;
	}


	@Override
	@Nullable
	public String getTypeLabel() {
		return type;
	}


	@Override
	public void open() {
		if(((ICodeRegion<URI>) loc).getBase().equals(dest.getURI())) {
			editor.selectAndReveal(dest.getOffset(), dest.getLength());
			editor.setFocus();
		}
		else {
			IPath path = EclipseWorkspaceManager.getPath(dest.getURI());
			Editor.edit(path, dest.getOffset(), dest.getLength());
		}
	}

}
