/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import io.usethesource.impulse.services.ISourceHyperlinkDetector;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.ITextViewer;
import org.eclipse.jface.text.hyperlink.IHyperlink;
import org.eclipse.ui.texteditor.ITextEditor;
import org.magnolialang.eclipse.util.XRefHelper;
import org.magnolialang.infra.Infra;
import org.magnolialang.magnolia.MagnoliaCompiler;
import org.magnolialang.magnolia.MagnoliaFacts;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.resources.regions.IOffsetLength;
import org.nuthatchery.pica.tasks.NullTaskMonitor;
import org.nuthatchery.pica.util.Pair;

public class HyperlinkDetector implements ISourceHyperlinkDetector {

	@Override
	public IHyperlink[] detectHyperlinks(IRegion region, ITextEditor editor, ITextViewer textViewer, IParseController ctrl) {
		IProjectManager manager = Infra.getResourceManager(ctrl.getProject().getName());
		IManagedCodeUnit pkg = manager.findCodeUnit(((MagnoliaParseController) ctrl).getURI());

		Collection<Pair<IOffsetLength, IConstructor>> xRefs = XRefHelper.getXRefs((IMagnoliaPackage) pkg, region.getOffset(), new NullTaskMonitor());
		manager.commitMarks();
		return makeHyperlink(null, xRefs, editor);
	}


	private IHyperlink[] makeHyperlink(MagnoliaCompiler compiler, Collection<Pair<IOffsetLength, IConstructor>> xRefs, ITextEditor editor) {
		List<IHyperlink> links = new ArrayList<IHyperlink>();
		for(Pair<IOffsetLength, IConstructor> xRef : xRefs) {

			IConstructor info = null;

			if(xRef.getSecond().getConstructorType().equivalent(MagnoliaFacts.Cons_Ref)) {
				info = null; // TODO: (IConstructor) compiler.getFactOrNull(new NullRascalMonitor(), MagnoliaFacts.Type_DefInfo, xRef.getSecond().get("qName"));
			}
			else if(xRef.getSecond().getConstructorType().equivalent(MagnoliaFacts.Cons_LocalRef)) {
				info = (IConstructor) xRef.getSecond().get("info");
			}

			if(info != null) {

				IConstructor locInfo = (IConstructor) info.get("locInfo");
				if(locInfo.getConstructorType().equivalent(MagnoliaFacts.Cons_LocInfo1)) {
					ISourceLocation loc = (ISourceLocation) locInfo.get("defLoc");
					if(!xRef.getFirst().equals(loc)) {
						links.add(new Hyperlink(xRef.getFirst(), loc, "Open Declaration", "Declaration", editor));
					}
				}
				else if(locInfo.getConstructorType().equivalent(MagnoliaFacts.Cons_LocInfo2)) {
					ISourceLocation loc = (ISourceLocation) locInfo.get("defLoc");
					if(!xRef.getFirst().equals(loc)) {
						links.add(new Hyperlink(xRef.getFirst(), loc, "Open Declaration", "Declaration", editor));
					}
					loc = (ISourceLocation) locInfo.get("refLoc");
					if(!xRef.getFirst().equals(loc)) {
						links.add(new Hyperlink(xRef.getFirst(), loc, "Open Reference", "Reference", editor));
					}
				}
			}
		}
		if(links.isEmpty()) {
			return null;
		}
		else {
			return links.toArray(new IHyperlink[links.size()]);
		}
	}
}
