/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.net.URI;
import java.net.URISyntaxException;

import org.eclipse.core.filesystem.EFS;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.ui.IPersistableElement;
import org.eclipse.ui.IURIEditorInput;
import org.eclipse.ui.editors.text.ILocationProvider;
import org.eclipse.ui.editors.text.ILocationProviderExtension;
import org.magnolialang.eclipse.filesystem.MgGenFileSystem;
import org.eclipse.jdt.annotation.Nullable;

public class StringEditorInput implements IURIEditorInput {
	private final String value;
	private final String label;
	private final URI uri;


	public StringEditorInput(String label, String value) throws URISyntaxException {
		this(new URI("mggen", null, label.startsWith("/") ? label : "/" + label, null), value);

	}


	public StringEditorInput(URI uri, String value) {
		this.value = value;
		this.uri = uri;
		this.label = getPath().lastSegment();
		try {
			MgGenFileSystem fileSystem = (MgGenFileSystem) EFS.getFileSystem("mggen");
			fileSystem.addFile(uri, value);
		}
		catch(CoreException e) {
			e.printStackTrace();
		}
	}


	@Override
	public boolean equals(@Nullable Object obj) {
		if(this == obj) {
			return true;
		}
		if(obj == null) {
			return false;
		}
		if(getClass() != obj.getClass()) {
			return false;
		}
		StringEditorInput other = (StringEditorInput) obj;
		if(value == null) {
			if(other.value != null) {
				return false;
			}
		}
		else if(!value.equals(other.value)) {
			return false;
		}
		return true;
	}


	@Override
	public boolean exists() {
		return value != null;
	}


	@Override
	public Object getAdapter(Class adapter) {
		if(adapter.equals(IFile.class)) {
			return null;
		}
		else if(adapter.equals(IResource.class)) {
			return null;
		}
		else if(adapter.equals(ILocationProvider.class)) {
			return new LocationProvider();
		}
		else if(adapter.equals(ILocationProviderExtension.class)) {
			return new LocationProvider();
		}
		else {
			return null;
		}
	}


	@Override
	public ImageDescriptor getImageDescriptor() {
		return null;
	}


	@Override
	public String getName() {
		return label;
	}


	public IPath getPath() {
		return new Path(uri.getPath());
	}


	@Override
	public IPersistableElement getPersistable() {
		return null;
	}


	@Override
	public String getToolTipText() {
		if(uri == null) {
			return label;
		}
		else {
			return uri.getPath();
		}
	}


	@Override
	public URI getURI() {
		return uri;
	}


	public String getValue() {
		return value;
	}


	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (value == null ? 0 : value.hashCode());
		return result;
	}


	static final class LocationProvider implements ILocationProvider, ILocationProviderExtension {
		@Override
		public IPath getPath(Object element) {
			if(element instanceof StringEditorInput) {
				return ((StringEditorInput) element).getPath();
			}
			else {
				return null;
			}
		}


		@Override
		public URI getURI(Object element) {
			if(element instanceof StringEditorInput) {
				return ((StringEditorInput) element).getURI();
			}
			else {
				return null;
			}
		}
	}
}
