/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IValue;
import org.rascalmpl.values.uptr.ITree;
import org.rascalmpl.values.uptr.ProductionAdapter;
import org.rascalmpl.values.uptr.TreeAdapter;
import org.rascalmpl.values.uptr.visitors.TreeVisitor;

public class TokenIterator implements Iterator<Object> {
	private class LexicalCollector extends TreeVisitor<RuntimeException> {
		private int location;


		public LexicalCollector() {
			super();

			location = 0;
		}


		@Override
		public ITree visitTreeAmb(final ITree arg) {
			// we just go into the first, it's arbitrary but at least we'll get
			// some nice highlighting
			return (ITree) TreeAdapter.getAlternatives(arg).iterator().next().accept(this);
			// TreeAdapter.getAlternatives(arg).iterator().next().accept(this);
			// ISourceLocation loc =
			// TreeAdapter.getLocation((ITree)TreeAdapter.getAlternatives(arg).iterator().next());
			// tokenList.add(new Token(TokenColorer., loc));

			/*
			 * ISet alts = TreeAdapter.getAlternatives(arg); String message =
			 * "Ambiguity:\n"; for(IValue alt : alts) { IConstructor prod =
			 * TreeAdapter.getProduction((ITree)alt); message += "    " +
			 * ProductionAdapter.getSortName(prod) + " ::= " +
			 * ProductionAdapter.getLhs(prod) + "\n"; }
			 *
			 * if(msgHandler != null) { msgHandler.handleSimpleMessage(message,
			 * loc.getOffset(), loc.getOffset()+loc.getLength(), 0, 0, 0, 0); }
			 * System.out.println(message);
			 */
			// return arg;
		}


		@Override
		public ITree visitTreeAppl(final ITree arg) {
			String category = null;
			int offset = location;

			if(TreeAdapter.isAppl(arg)) {
				final IConstructor prod = TreeAdapter.getProduction(arg);
				final String sort = ProductionAdapter.getSortName(prod);
				category = ProductionAdapter.getCategory(prod);
				// String code = TreeAdapter.yield(arg);

				if(category == null && TokenColorer.CATEGORIES.contains(sort)) {
					category = sort;
				}

				if(TokenColorer.ERROR.equals(ProductionAdapter.getConstructorName(prod))) {
					category = TokenColorer.ERROR;
				}
			}

			if(category == null && TreeAdapter.isLiteral(arg)) {
				final String yield = TreeAdapter.yield(arg);
				for(final byte c : yield.getBytes()) {
					if(c != '-' && !Character.isJavaIdentifierPart(c)) {
						location += TreeAdapter.yield(arg).length();
						return arg;
					}
				}
				category = TokenColorer.KEYWORD;
			}

			if(category == null) {
				for(final IValue child : TreeAdapter.getArgs(arg)) {
					child.accept(this);
				}
			}
			else {
				location += TreeAdapter.yield(arg).length();
				tokList.add(new Token(category, offset, location - offset));
			}
			return arg;
		}


		@Override
		public ITree visitTreeChar(final ITree arg) {
			++location;
			return arg;
		}


		@Override
		public ITree visitTreeCycle(final ITree arg) {
			return arg;
		}
	}

	protected final List<Token> tokList;


	protected final Iterator<Token> tokIterator;


	public TokenIterator(final IConstructor parseTree) {
		tokList = new ArrayList<Token>(1000);

		if(parseTree != null) {
			parseTree.accept(new LexicalCollector());
		}
		tokIterator = tokList.iterator();

	}


	@Override
	public boolean hasNext() {
		return tokIterator.hasNext();
	}


	@Override
	public Token next() {
		return tokIterator.next();
	}


	@Override
	public void remove() {
		throw new UnsupportedOperationException();
	}
}
