/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.io.ByteArrayInputStream;
import java.io.InputStream;

import org.eclipse.core.resources.IStorage;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.rascalmpl.value.IValue;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.ui.IPersistableElement;
import org.eclipse.ui.IStorageEditorInput;
import org.eclipse.jdt.annotation.Nullable;

public class ValueEditorInput implements IStorageEditorInput {
	protected final IValue value;
	protected final String label;
	protected final String fileName;


	public ValueEditorInput(IValue value) {
		this.value = value;
		this.label = value.getType().toString();
		this.fileName = null;
	}


	public ValueEditorInput(String label, IValue value) {
		this.value = value;
		this.label = label;
		this.fileName = label;
	}


	@Override
	public boolean equals(@Nullable Object obj) {
		if(this == obj) {
			return true;
		}
		if(obj == null) {
			return false;
		}
		if(getClass() != obj.getClass()) {
			return false;
		}
		ValueEditorInput other = (ValueEditorInput) obj;
		if(value == null) {
			if(other.value != null) {
				return false;
			}
		}
		else if(!value.equals(other.value)) {
			return false;
		}
		return true;
	}


	@Override
	public boolean exists() {
		return value != null;
	}


	@Override
	public Object getAdapter(Class adapter) {
		if(IValue.class.equals(adapter)) {
			return value;
		}
		else if(IValue.class.isAssignableFrom(adapter)) {
			return value;
		}
		return null;
	}


	@Override
	public ImageDescriptor getImageDescriptor() {
		return ImageDescriptor.getMissingImageDescriptor();
	}


	@Override
	public String getName() {
		return label;
	}


	@Override
	public IPersistableElement getPersistable() {
		return null;
	}


	@Override
	public IStorage getStorage() throws CoreException {
		return new IStorage() {

			@Override
			public Object getAdapter(Class adapter) {
				return null;
			}


			@Override
			public InputStream getContents() throws CoreException {
				return new ByteArrayInputStream(value.toString().getBytes());
			}


			@Override
			public IPath getFullPath() {
				if(fileName == null) {
					return null;
				}
				else {
					return new Path(fileName);
				}
			}


			@Override
			public String getName() {
				return value.getType().toString();
			}


			@Override
			public boolean isReadOnly() {
				return false;
			}
		};
	}


	@Override
	public String getToolTipText() {
		return getName();
	}


	public IValue getValue() {
		return value;
	}


	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (value == null ? 0 : value.hashCode());
		return result;
	}
}
