/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.preferences;

import org.eclipse.jface.preference.BooleanFieldEditor;
import org.eclipse.jface.preference.FieldEditorPreferencePage;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;
import org.magnolialang.eclipse.MagnoliaPlugin;

/**
 * This class represents a preference page that
 * is contributed to the Preferences dialog. By
 * subclassing <samp>FieldEditorPreferencePage</samp>, we
 * can use the field support built into JFace that allows
 * us to create a page that is small and knows how to
 * save, restore and apply itself.
 * <p>
 * This page is used to modify preferences only. They are stored in the
 * preference store that belongs to the main plug-in class. That way,
 * preferences can be accessed directly via the preference store.
 */

public class BuilderPreferencePage extends FieldEditorPreferencePage implements IWorkbenchPreferencePage {

	public BuilderPreferencePage() {
		super(GRID);
		setPreferenceStore(MagnoliaPlugin.getInstance().getPreferenceStore());
		setDescription("Magnolia Builder Preferences:");
	}


	/**
	 * Creates the field editors. Field editors are abstractions of
	 * the common GUI blocks needed to manipulate various types
	 * of preferences. Each field editor knows how to save and
	 * restore itself.
	 */
	@Override
	public void createFieldEditors() {
		for(Pref p : Pref.values()) {
			if(p.getPage().equals("BUILD") && p.isBoolean()) {
				BooleanFieldEditor editor = new BooleanFieldEditor(p.name(), p.getLabel(), getFieldEditorParent());
				if(!p.getTooltip().equals("")) {
					editor.getDescriptionControl(getFieldEditorParent()).setToolTipText(p.getTooltip());
				}
				addField(editor);
			}
		}
	}


	@Override
	public void init(IWorkbench workbench) {
	}

}
