/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.preferences;

import java.io.File;

import org.eclipse.core.runtime.preferences.AbstractPreferenceInitializer;
import org.eclipse.jface.preference.IPreferenceStore;
import org.magnolialang.eclipse.MagnoliaPlugin;

/**
 * Class used to initialize default preference values.
 */
public class PreferenceInitializer extends AbstractPreferenceInitializer {

	@Override
	public void initializeDefaultPreferences() {
		IPreferenceStore store = MagnoliaPlugin.getInstance().getPreferenceStore();

		for(Pref p : Pref.values()) {
			if(p.isString()) {
				store.setDefault(p.name(), p.getStringValue());
			}
			else if(p.isBoolean()) {
				store.setDefault(p.name(), p.getBooleanValue());
			}
		}

		String optPath = ".cabal" + File.separator + "bin" + File.separator + "Rewriter";
		try {
			String home = System.getenv("HOME");
			if(home != null) {
				optPath = home + File.separator + optPath;
			}
		}
		catch(SecurityException e) {

		}
		store.setDefault(Pref.MAIN_OPTIMIZER_PATH.name(), optPath);
		String z3path = "";
		try {
			String path = System.getenv("PATH");
			if(path != null) {
				for(String s : path.split(File.pathSeparator)) {
					File file = new File(s + File.separator + "z3");
					if(file.canExecute()) {
						z3path = file.toString();
						break;
					}
				}
			}
		}
		catch(SecurityException e) {

		}
		store.setDefault(Pref.MAIN_Z3_PATH.name(), z3path);

	}
}
