/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.tests;

import static org.junit.Assert.assertEquals;

import java.net.URI;
import java.util.ArrayList;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.magnolialang.eclipse.MagnoliaPlugin;

public class UtilityTest {
	String[] names = { "FOO", "foo", "Foo Bar", "Foo.bar", "foo&bar", " foo ", "0123", "" };
	IPath[] paths = { new Path("/foo"), new Path("/foo/"), new Path("/ bar /"), new Path(" bar"), new Path("foo+bar"), new Path("/") };
	private ArrayList<IProject> projects;


	public void constructProjectURIAxiom(IProject project, IPath path) {
		URI uri = MagnoliaPlugin.constructProjectURI(project, path);
		System.err.println("constructProjectURI: " + uri);
		assertEquals(project.getName(), uri.getAuthority());
		if(path.isAbsolute()) {
			assertEquals(path.toString(), uri.getPath());
		}
		else {
			assertEquals("/" + path.toString(), uri.getPath());
		}
	}


	@Before
	public void setUp() {
		projects = new ArrayList<IProject>();
		for(String s : names) {
			if(Path.ROOT.isValidSegment(s)) {
				try {
					projects.add(ProjectUtil.createProject(s, true));
				}
				catch(Exception e) {
					e.printStackTrace();
				}
			}
			else {
				System.err.println("Skipping invalid path segment '" + s + "'");
			}
		}
	}


	@After
	public void tearDown() {
		for(IProject p : projects) {
			try {
				p.delete(true, null);
			}
			catch(CoreException e) {
				e.printStackTrace();
			}
		}
	}


	@Test
	public void uriTest() {
		for(IProject project : projects) {
			for(IPath path : paths) {
				constructProjectURIAxiom(project, path);
			}
		}
	}
}
