/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.errors;

import java.net.URI;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IValue;
import org.nuthatchery.pica.errors.ErrorMarkers;
import org.rascalmpl.values.uptr.ITree;
import org.rascalmpl.values.uptr.TreeAdapter;

/**
 * @author anya
 *
 */
public class CompilationErrorException extends RuntimeException {

	private static final long serialVersionUID = 2660588367962898009L;
	private ISourceLocation loc = null;
	private IConstructor tree = null;


	/**
	 * @param message
	 *            Message describing the error
	 */
	public CompilationErrorException(final String message) {
		super(message);
	}


	/**
	 * @param message
	 *            Message describing the error
	 * @param locOrTree
	 *            An ISourceLocation describing the error location or an
	 *            IConstructor tree with location information
	 */
	public CompilationErrorException(final String message, final IValue locOrTree) {
		super(message);
		if(locOrTree != null) {
			if(locOrTree instanceof ISourceLocation) {
				loc = (ISourceLocation) locOrTree;
			}
			else if(locOrTree instanceof ITree) {
				loc = TreeAdapter.getLocation((ITree) locOrTree);
				tree = (IConstructor) locOrTree;
			}
		}
	}


	/**
	 * @param message
	 *            Message describing the error
	 * @param cause
	 *            The exception which caused this exception
	 */
	public CompilationErrorException(final String message, final Throwable cause) {
		super(message, cause);
	}


	/**
	 * @param message
	 *            Message describing the error
	 * @param cause
	 *            The exception which caused this exception
	 * @param locOrTree
	 *            An ISourceLocation describing the error location or an
	 *            IConstructor tree with location information
	 */
	public CompilationErrorException(final String message, final Throwable cause, final IValue locOrTree) {
		super(message, cause);
		if(locOrTree instanceof ISourceLocation) {
			loc = (ISourceLocation) locOrTree;
		}
		else if(locOrTree instanceof ITree) {
			loc = TreeAdapter.getLocation((ITree) locOrTree);
			tree = (IConstructor) locOrTree;
		}
	}


	public String getErrorType() {
		return "compilation error";
	}


	public ISourceLocation getLocation() {
		return loc;
	}


	public String getMarkerMessage() {
		final Throwable cause = getCause();
		if(cause instanceof CompilationErrorException) {
			return super.getMessage() + "\n(" + ((CompilationErrorException) cause).getErrorType() + ")";
		}
		else {
			return super.getMessage();
		}
	}


	public String getMarkerType() {
		return ErrorMarkers.TYPE_COMPILATION_ERROR;
	}


	@Override
	public String getMessage() {
		if(loc != null) {
			final URI url = loc.getURI();

			return (url.getScheme().equals("file") || url.getScheme().equals("project") ? url.getPath() : url) + ":" + loc.getBeginLine() + "," + loc.getBeginColumn() + ": " + super.getMessage();
		}

		// TODO remove once all errors have locations
		return super.getMessage();
	}


	public IConstructor getTree() {
		return tree;
	}


	public void setLocation(final ISourceLocation loc) {
		this.loc = loc;
	}


	public void setTree(final IConstructor tree) {
		this.tree = tree;
	}

}
