/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.errors;

import org.rascalmpl.value.IValue;
import org.nuthatchery.pica.errors.ErrorMarkers;

public class ModuleNotFoundException extends ModuleLoadException {
	private static final String ERROR_TYPE = "module not found";
	private static final String ERROR_MESSAGE = "Module not found: ";
	private static final long serialVersionUID = -3355800366920728861L;
	private final String moduleName;


	public ModuleNotFoundException(final String moduleName) {
		super(ERROR_MESSAGE + moduleName);
		this.moduleName = moduleName;
	}


	public ModuleNotFoundException(final String moduleName, final IValue locOrTree) {
		super(ERROR_MESSAGE + moduleName, locOrTree);
		this.moduleName = moduleName;
	}


	public ModuleNotFoundException(final String moduleName, final Throwable cause) {
		super(ERROR_MESSAGE + moduleName, cause);
		this.moduleName = moduleName;
	}


	public ModuleNotFoundException(final String moduleName, final Throwable cause, final IValue locOrTree) {
		super(ERROR_MESSAGE + moduleName, cause, locOrTree);
		this.moduleName = moduleName;
	}


	@Override
	public String getErrorType() {
		final Throwable cause = getCause();
		if(cause == null) {
			return ERROR_TYPE;
		}
		else {
			return cause.getMessage();
		}
	}


	@Override
	public String getMarkerType() {
		return ErrorMarkers.TYPE_LOADER;
	}


	public String getModuleName() {
		return moduleName;
	}
}
