/**************************************************************************
 * Copyright (c) 2011-2013 Anya Helene Bagge
 * Copyright (c) 2011-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.errors;

import org.eclipse.core.runtime.Path;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.rascalmpl.interpreter.control_exceptions.Throw;

public final class RascalErrors {
	/*
	public static RuntimeException decodeRascalError(StaticError t) {

	}*/
	public static RuntimeException decodeRascalError(Throw t) {
		IValue val = t.getException();
		if(val instanceof IConstructor) {
			IConstructor ex = (IConstructor) val;
			String name = ex.getConstructorType().getName();
			if(name.equals("ParseError")) {
				IString msg = (IString) ex.get("message");
				ISourceLocation loc = (ISourceLocation) ex.get("location");
				CompilationErrorException error = new SyntaxErrorException(msg.getValue(), loc);
				setLocation(error, t);
				return error;
			}
			else if(name.equals("FactProductionError")) {
				IString msg = (IString) ex.get("message");
				ISourceLocation loc = (ISourceLocation) ex.get("location");
				CompilationErrorException error = new CompilationErrorException(msg.getValue(), loc);
				setLocation(error, t);
				return error;
			}
			else if(ex.has("message")) {
				IString msg = (IString) ex.get("message");
				CompilationErrorException error;
				if(ex.has("location")) {
					ISourceLocation loc = (ISourceLocation) ex.get("location");
					error = new CompilationErrorException(msg.getValue(), loc);
					setLocation(error, t);
				}
				else {
					error = new CompilationErrorException(msg.getValue(), t.getLocation());
				}
				return error;
			}
			else {
				String msg = ex.toString();
				CompilationErrorException error;
				if(ex.has("location")) {
					ISourceLocation loc = (ISourceLocation) ex.get("location");
					error = new CompilationErrorException(msg, loc);
					setLocation(error, t);
				}
				else {
					error = new CompilationErrorException(msg + t.getTrace(), t.getLocation());
				}
				return error;
			}

		}
		return new CompilationErrorException("Unknown Rascal error: " + val.toString() + "\n" + t.getTrace());
	}


	private static void setLocation(CompilationErrorException error, Throw t) {
		ISourceLocation location = t.getLocation();
		if(location != null) {
			StackTraceElement[] stackTrace = error.getStackTrace();

			String mod = new Path(location.getURI().getPath()).removeFileExtension().lastSegment();
			String fun = "";
			String file = location.getURI().getPath();
			int line = location.getBeginLine();
			stackTrace[0] = new StackTraceElement(mod, fun, file, line);
			error.setStackTrace(stackTrace);
		}
	}


	private RascalErrors() {

	}
}
