/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
package org.magnolialang.infra;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.eclipse.MagnoliaFileLocator;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.nuthatchery.pica.errors.Severity;
import org.nuthatchery.pica.rascal.EclipseEvaluatorFactory;
import org.nuthatchery.pica.rascal.IEvaluatorFactory;

public final class EclipseInfra extends AbstractInfra {
	private final static MagnoliaWorkspaceConfig config = new MagnoliaWorkspaceConfig();
	private final static IEvaluatorFactory evaluatorFactory = new EclipseEvaluatorFactory(config);


	private EclipseInfra() {
		super(new org.nuthatchery.pica.eclipse.EclipsePicaInfra(config));
	}


	@Override
	public boolean areModuleFactsPreloaded() {
		return true;
	}


	@Override
	public File getDataFile(String fileName) {
		IPath path = MagnoliaPlugin.getInstance().getStateLocation();
		path = path.append(fileName);
		return new File(path.toOSString());
	}


	@Override
	public IEvaluatorFactory getEvaluatorFactory() {
		return evaluatorFactory;
	}


	@Override
	public void logException(@Nullable String msg, @Nullable Throwable t) {
		MagnoliaPlugin.getInstance().logException(msg, t);
	}


	@Override
	public void logMessage(String msg) {
		logMessage(msg, Severity.DEFAULT);
	}


	@Override
	public void logMessage(String msg, Severity severity) {
		MagnoliaPlugin.getInstance().logException(severity.toString() + ": " + msg, null);
	}


	@Override
	public InputStream openStream(String path) throws IOException {
		InputStream stream = MagnoliaFileLocator.openStream(new Path(path));
		if(stream == null) {
			throw new FileNotFoundException(path);
		}
		else {
			return stream;
		}
	}


	public static void setInfra() {
		Infra.set(new EclipseInfra());
	}
}
