/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.infra;

import java.net.URI;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;

import org.rascalmpl.value.IConstructor;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.eclipse.MagnoliaNature;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.magnolialang.magnolia.resources.MagnoliaPackage;
import org.nuthatchery.pgf.rascal.PGFRascal;
import org.nuthatchery.pica.rascal.EvaluatorFactory;
import org.nuthatchery.pica.rascal.ISearchPathProvider;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.IWorkspaceConfig;
import org.nuthatchery.pica.resources.handles.IFileHandle;
import org.nuthatchery.pica.resources.storage.IStorage;
import org.rascalmpl.interpreter.Evaluator;

public class MagnoliaWorkspaceConfig implements IWorkspaceConfig, ISearchPathProvider {
	private Collection<String> natures = Collections.unmodifiableCollection(Collections.singletonList(MagnoliaNature.NATURE_ID));


	@Override
	public Collection<ClassLoader> additionalClassLoaders() {
		return Arrays.asList(getClass().getClassLoader());
	}


	@Override
	public void addRascalSearchPaths(Evaluator evaluator) {
		EvaluatorFactory.addClassInputSearchPath(evaluator, "magnolia", getClass(), "/");
		EvaluatorFactory.addClassInputSearchPath(evaluator, "pgf", PGFRascal.class, "/");
	}


	@Override
	public Collection<String> getActiveNatures() {
		return natures;
	}


	@Override
	public synchronized void initCompiler() {
		Magnolia.getInstance().getCompiler().ensureInit();
	}


	@Override
	public IMagnoliaPackage makePackage(IProjectManager manager, IFileHandle res, @Nullable IStorage storage, Object id, ILanguage lang) {
		if(storage == null) {
			throw new IllegalArgumentException();
		}
		return new MagnoliaPackage(manager, res, storage, (IConstructor) id, lang);
	}
}
