/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.load;

import java.net.URI;
import java.util.Collection;

import io.usethesource.impulse.utils.Pair;

/**
 * @author anya
 * 
 */
public interface IBuildPath {
	/**
	 * Add a new path to the end of the build path. The main binary path is
	 * used.
	 * 
	 * @param src
	 *            The source path
	 */
	void addPath(URI src);


	/**
	 * Add a new path to the end of the build path.
	 * 
	 * @param src
	 *            The source path
	 * @param bin
	 *            The binary output path for this source path
	 */
	void addPath(URI src, URI bin);


	/**
	 * Find the given module in the path. Prefers binary/compiled modules unless
	 * source is newer.
	 * 
	 * @param moduleName
	 *            Name of the module
	 * @return A pair of the URI of the module and string of the unused
	 *         components of the module name (null if none), or null if not
	 *         found.
	 */
	URI find(String moduleName, Collection<String> extensions);


	/**
	 * Get the binary path associated with a source path.
	 * 
	 * @param src
	 *            The source path
	 * @return The binary path associated with src
	 */
	URI getBinPath(URI src);


	/**
	 * @return The main/default binary path
	 */
	URI getMainBinPath();


	/**
	 * Construct a dotted module name based on a URI.
	 * 
	 * @param uri
	 */
	Pair<URI, String> getModuleName(URI uri);


	/**
	 * @return A new empty build path object, with the same main binary path as
	 *         this.
	 */
	IBuildPath newPath();


	/**
	 * @param mainBinPath
	 * @return A new empty build path object, with the given main binary path.
	 */
	IBuildPath newPath(URI mainBinPath);


	/**
	 * Set the main/default binary path.
	 * 
	 * @param bin
	 *            New main binary path
	 */
	void setMainBinPath(URI bin);
}
