/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 Tero Hasu
 * Copyright (c) 2010-2013 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 *
 *************************************************************************/
package org.magnolialang.load;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.magnolialang.infra.Infra;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.parser.gtd.IGTD;
import org.rascalmpl.parser.gtd.io.InputConverter;
import org.rascalmpl.parser.gtd.result.action.VoidActionExecutor;
import org.rascalmpl.parser.gtd.result.out.DefaultNodeFlattener;
import org.rascalmpl.parser.uptr.UPTRNodeFactory;
import org.rascalmpl.values.uptr.ITree;

public final class ModuleParser {
	private static final UPTRNodeFactory NODE_CONSTRUCTOR_FACTORY = new UPTRNodeFactory();

	private static final Map<ILanguage, ModuleParser> parsers = new HashMap<ILanguage, ModuleParser>();


	public static ModuleParser getParser(ILanguage lang, String modName) {
		ModuleParser parser = parsers.get(lang);
		if(parser == null) {
			parser = new ModuleParser(modName);
			parsers.put(lang, parser);
		}
		return parser;
	}


	private final String module;


	private ModuleParser(String modName) {
		super();
		this.module = modName;
	}


	public String getModuleName() {
		return module;
	}


	public void initialize() {
		Infra.get().getParser(module);
	}


	public ITree parseModule(final URI location) throws IOException {
		ISourceLocation loc = TermFactory.vf.sourceLocation(location);
		InputStream inputStream = Infra.getResolverRegistry().getInputStream(loc);
		try {
			char[] data = InputConverter.toChar(inputStream, Infra.getResolverRegistry().getCharset(loc));
			return parseModule(location, data);
		}
		finally {
			inputStream.close();
		}
	}


	public ITree parseModule(final URI location, final char[] moduleString) {
		IGTD<IConstructor, ITree, ISourceLocation> parser = Infra.get().getParser(module);
		ITree pt = (ITree) parser.parse("start__Program", location, moduleString,
				//new VoidActionExecutor<IConstructor>(),
				new DefaultNodeFlattener<ITree, ITree, ISourceLocation>(), NODE_CONSTRUCTOR_FACTORY);
		return pt;
	}


	public ITree parseModule(final URI location, final String moduleString) {
		return parseModule(location, InputConverter.toChar(moduleString));
	}
}
