/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia;

import nuthatch.pattern.Pattern;

import org.rascalmpl.value.IValue;
import org.rascalmpl.value.type.Type;
import org.magnolialang.magnolia.grammar.MagnoliaPatterns;
import org.nuthatchery.pica.terms.TermFactory;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;

@SuppressFBWarnings("NM_METHOD_NAMING_CONVENTION")
public class AbstractSyntax extends MagnoliaPatterns {

	public static Pattern<IValue, Type> Api() {
		return pf.cons(TermFactory.consType("Api", 0));
	}


	public static Pattern<IValue, Type> Binding(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("Binding", 2), arg0, arg1);
	}


	public static Pattern<IValue, Type> BlockExpr(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("BlockExpr", 2), arg0, arg1);
	}


	public static Pattern<IValue, Type> Builtin(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("Builtin", 2), arg0, arg1);
	}


	public static Pattern<IValue, Type> Close(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.consType("Close", 1), arg0);
	}


	public static Pattern<IValue, Type> Concrete(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.consType("Concrete", 1), arg0);
	}


	public static Pattern<IValue, Type> DeclBody(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.consType("DeclBody", 1), arg0);
	}


	public static Pattern<IValue, Type> Define(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2) {
		return pf.cons(TermFactory.consType("Define", 3), arg0, arg1, arg2);
	}


	public static Pattern<IValue, Type> DefinesOf(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.consType("DefinesOf", 1), arg0);
	}


	public static Pattern<IValue, Type> DotOp(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2) {
		return pf.cons(TermFactory.consType("DotOp", 3), arg0, arg1, arg2);
	}


	public static Pattern<IValue, Type> Equal(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2) {
		return pf.cons(TermFactory.consType("Equal", 3), arg0, arg1, arg2);
	}


	public static Pattern<IValue, Type> External(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("External", 2), arg0, arg1);
	}


	public static Pattern<IValue, Type> ExternalFun(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2, Pattern<IValue, Type> arg3, Pattern<IValue, Type> arg4, Pattern<IValue, Type> arg5) {
		return pf.cons(TermFactory.consType("ExternalFun", 6), arg0, arg1, arg2, arg3, arg4, arg5);
	}


	public static Pattern<IValue, Type> ExternalParams(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("ExternalParams", 2), arg0, arg1);
	}


	public static Pattern<IValue, Type> ExternalProc(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2, Pattern<IValue, Type> arg3, Pattern<IValue, Type> arg4) {
		return pf.cons(TermFactory.consType("ExternalProc", 5), arg0, arg1, arg2, arg3, arg4);
	}


	public static Pattern<IValue, Type> ExternalType(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2, Pattern<IValue, Type> arg3) {
		return pf.cons(TermFactory.consType("ExternalType", 4), arg0, arg1, arg2, arg3);
	}


	public static Pattern<IValue, Type> Impl() {
		return pf.cons(TermFactory.consType("Impl", 0));
	}


	public static Pattern<IValue, Type> Instance(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("Instance", 2), arg0, arg1);
	}


	public static Pattern<IValue, Type> IntegerLiteralAttr(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2, Pattern<IValue, Type> arg3, Pattern<IValue, Type> arg4, Pattern<IValue, Type> arg5) {
		return pf.cons(TermFactory.consType("IntegerLiteralAttr", 6), arg0, arg1, arg2, arg3, arg4, arg5);
	}


	public static Pattern<IValue, Type> LiftStatExpr(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2) {
		return pf.cons(TermFactory.consType("LiftStatExpr", 3), arg0, arg1, arg2);
	}


	public static Pattern<IValue, Type> LiftVar() {
		return pf.cons(TermFactory.consType("LiftVar", 0));
	}


	public static Pattern<IValue, Type> Modifier(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.consType("Modifier", 1), arg0);
	}


	public static Pattern<IValue, Type> Origin(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.tf.constructor(TermFactory.ts, TermFactory.tf.abstractDataType(TermFactory.ts, "Origin"), "Origin", //
				TermFactory.tf.stringType(), "name", //
				TermFactory.tf.sourceLocationType(), "location"), //
				arg0, arg1);
	}


	public static Pattern<IValue, Type> Origins(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.tf.constructor(TermFactory.ts, TermFactory.Type_AST, "Origins", TermFactory.tf.listType(TermFactory.tf.abstractDataType(TermFactory.ts, "Origin")), "locs"), arg0);
	}


	public static Pattern<IValue, Type> RenameSeq(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("RenameSeq", 2), arg0, arg1);
	}


	public static Pattern<IValue, Type> RequiresOf(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.consType("RequiresOf", 1), arg0);
	}


	public static Pattern<IValue, Type> seqElement(Pattern<IValue, Type> prefix, Pattern<IValue, Type> element, Pattern<IValue, Type> suffix) {
		return pf.cons(TermFactory.Cons_Seq, pf.listElement(TermFactory.tf.listType(TermFactory.Type_AST), prefix, element, suffix));
	}


	public static Pattern<IValue, Type> SignatureOf(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.consType("SignatureOf", 1), arg0);
	}


	public static Pattern<IValue, Type> TopDefine(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1, Pattern<IValue, Type> arg2) {
		return pf.cons(TermFactory.consType("TopDefine", 3), arg0, arg1, arg2);
	}


	public static Pattern<IValue, Type> TopName(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("TopName", 2), arg0, arg1);
	}


	public static Pattern<IValue, Type> Undefined() {
		return pf.cons(TermFactory.consType("Undefined", 0));
	}


	public static Pattern<IValue, Type> UnlessReturned(Pattern<IValue, Type> arg0) {
		return pf.cons(TermFactory.consType("UnlessReturned", 1), arg0);
	}
}
