/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::AbstractSyntax

extend org::magnolialang::magnolia::grammar::MagnoliaAST;

data AST
	= RenameList(AST arg0)  // a list of renaming operations
	| RenameSeq(AST arg0, AST arg1) // a sequence of renamings
	| TopName(AST arg0, AST arg1) // name, kind
	| Equal(AST arg0, AST arg1, AST arg2) // lhs, op, rhs
	| Builtin(AST arg0, AST arg1) // qname, idstring
	| Define(AST arg0, AST arg1, AST arg2) // dcl, attrs, body
	| TopDefine(AST arg0, AST arg1, AST arg2) // dcl, attrs, body
	| LiftStatExpr(AST varType, AST stat, str name) // sort("Expr")
	| LiftVar()
	| UnlessReturned(AST arg0)
	| ExternalType(AST arg0, AST arg1, AST arg2, AST arg3) // language, extname, extparams, name
	| ExternalFun(AST arg0, AST arg1, AST arg2, AST arg3, AST arg4, AST arg5) // language, extname, extparams, name, params, return
	| ExternalProc(AST arg0, AST arg1, AST arg2, AST arg3, AST arg4) // language, extname, extparams, name, params 
	| ExternalParams(AST arg0, AST arg1) // types, operations (both are sorted lists of full names) -- used in extparams above
	| Instance(AST arg0, AST arg1) // op/type, list of bindings (instantiation of a Fun, Proc, or a type)
	| Binding(AST arg0, AST arg1) // from-name, to-name (for use in Instance) 
	| Concrete(AST arg0) // paramlist (list of names). (potentially parameterised concrete implementation of something) 
	| Api() // modifier for api definitions
	| Impl() // modifier for implementation (non-api) definitions
	| DotOp(AST arg0, AST arg1, AST arg2)
	| DeclBody(AST arg0)
	| Origins(list[Origin] locs) // attribute: a list of locations tracing the origin of a declaration
	| External(AST arg0, AST arg1) // language, name
	| Undefined()
	| IntegerLiteralAttr(AST arg0, AST arg1, AST arg2, AST arg3, AST arg4, AST arg5) // type, 0, 1, +, *, -
// Core top level
	| SignatureOf(AST arg0)
	| RequiresOf(AST arg0)
	| DefinesOf(AST arg0)
	| Modifier(AST arg0)
	| Close(AST arg0)
	| Id(AST arg0) // wraps a unique identifier, such as FunName, ProcName, TypeName, etc.
	| BlockExpr(AST arg0, AST arg1) // statements, type
	;

data Origin = Origin(str name, loc location);
