/**************************************************************************
 * Copyright (c) 2011-2013 Anya Helene Bagge
 * Copyright (c) 2011-2013 Tero Hasu
 * Copyright (c) 2011-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
package org.magnolialang.magnolia;

import static org.nuthatchery.pica.terms.TermFactory.*;

import java.util.List;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.IListWriter;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.ISetWriter;
import org.rascalmpl.value.ITuple;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IValueFactory;
import org.rascalmpl.value.type.Type;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.interpreter.IEvaluatorContext;
import org.rascalmpl.interpreter.TypeReifier;

public final class MagnoliaFacts {
	public static final Type Type_Severity = tf.abstractDataType(ts, "Severity");
	public static final Type Cons_Error = tf.constructor(ts, Type_Severity, "Error");
	public static final Type Cons_Warning = tf.constructor(ts, Type_Severity, "Warning");
	public static final Type Cons_Internal = tf.constructor(ts, Type_Severity, "Internal");
	public static final Type Cons_Info = tf.constructor(ts, Type_Severity, "Info");

	public static final Type Type_ErrorMark = tf.abstractDataType(ts, "ErrorMark");
	public static final Type Cons_ErrorMark = tf.constructor(ts, Type_ErrorMark, "ErrorMark", tf.stringType(), "message", Type_Severity, "severity", tf.sourceLocationType(), "location");

	public static final Type Type_LocInfo = tf.abstractDataType(ts, "LocInfo");
	public static final Type Cons_LocInfo0 = tf.constructor(ts, Type_LocInfo, "LocInfo");
	public static final Type Cons_LocInfo1 = tf.constructor(ts, Type_LocInfo, "LocInfo", tf.sourceLocationType(), "defLoc");
	public static final Type Cons_LocInfo2 = tf.constructor(ts, Type_LocInfo, "LocInfo", tf.sourceLocationType(), "defLoc", tf.sourceLocationType(), "refLoc");

	public static final Type Type_DefInfo = tf.abstractDataType(ts, "DefInfo");
	public static final Type Cons_DefInfo = tf.constructor(ts, Type_DefInfo, "DefInfo", tf.stringType(), "kind", TermFactory.Type_AST, "qName", TermFactory.Type_AST, "defName", TermFactory.Type_AST, "parent", tf.listType(TermFactory.Type_AST), "attrs", tf.setType(TermFactory.Type_AST), "depends", TermFactory.Type_AST, "decl", TermFactory.Type_AST, "body", tf.stringType(), "doc", Type_LocInfo, "locInfo", tf.sourceLocationType(), "srcLoc", tf.valueType(), "resource");

	public static final Type Type_Ref = tf.abstractDataType(ts, "Ref");
	public static final Type Cons_Ref = tf.constructor(ts, Type_Ref, "Ref", TermFactory.Type_AST, "qName");
	public static final Type Cons_LocalRef = tf.constructor(ts, Type_Ref, "LocalRef", Type_DefInfo, "info");
	public static final Type Type_XRefs = tf.aliasType(ts, "XRefs", tf.relType(tf.sourceLocationType(), "token", Type_Ref, "ref"));

	public static final String KIND_PACKAGE = "package";
	public static final String KIND_MODULE = "module";
	public static final String KIND_PROGRAM = "program";
	public static final String KIND_IMPLEMENTATION = "implementation";
	public static final String KIND_RENAMING = "renaming";
	public static final String KIND_SATISFACTION = "satisfaction";
	public static final String KIND_CONCEPT = "concept";
	public static final String KIND_TEST = "test";
	public static final String KIND_AXIOM = "axiom";
	public static final String KIND_FUNCTION = "function";
	public static final String KIND_PROCEDURE = "procedure";
	public static final String KIND_TYPE = "type";


	private MagnoliaFacts() {

	}


	public static IConstructor defInfo(IValueFactory vf) {
		IConstructor nop = TermFactory.cons("Nop");
		IList nilList = vf.list(TermFactory.Type_AST);
		ISet nilSet = vf.set(TermFactory.Type_AST);
		return vf.constructor(Cons_DefInfo, vf.string(""), nop, nop, nop, nilList, nilSet, nop, nop, vf.string(""), vf.constructor(Cons_LocInfo0), vf.sourceLocation("unknown:///"), nop);
// 	return DefInfo("",Nop(),Nop(),Nop(),[],{},Nop(),Nop(), "", LocInfo(), Nop());
	}


	public static ITuple keyedName(IEvaluatorContext ctx, List<Type> types, IValue value) {
		TypeReifier tr = new TypeReifier(ctx.getValueFactory());
		IListWriter lw = vf.listWriter();
		for(Type t : types) {
			lw.append(tr.typeToValue(t, ctx).getValue());
		}
		return ctx.getValueFactory().tuple(lw.done(), value);
	}


	public static ISetWriter xrefsWriter() {
		return TermFactory.vf.setWriter(TermFactory.tf.tupleType(TermFactory.tf.sourceLocationType(), MagnoliaFacts.Type_Ref));
	}
}
