/**************************************************************************
 * Copyright (c) 2011-2013 Anya Helene Bagge
 * Copyright (c) 2011-2013 Tero Hasu
 * Copyright (c) 2011-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
// Everything you need to know about Magnolia programs!
module org::magnolialang::magnolia::MagnoliaFacts

import org::nuthatchery::pica::terms::Terms;
import ParseTree;
import Node;

data AST = QName(AST) | Nop();
data AST = Cycle(AST cycle_name);
alias Name = AST;

data Exception =
	  ParseError(str message, loc location)
	| FactProductionError(str message, loc location)
	;
	
alias Color = int;

@doc{AST[ModuleName] -> Outline. A tree outline.}
data Outline = Outline(list[Outline] nodes) | OutlineLeaf();
anno str Outline@label; 
anno loc Outline@\loc;
anno str Outline@icon;
@doc{fgcolor can be either a color value from the Vis library, or a string with the name of a
colour defined in the Magnolia style file.}
anno value Outline@fgcolor;
anno value Outline@bgcolor;

data DefInfo = DefInfo(str kind, AST qName, AST defName, AST parent, list[AST] attrs, set[AST] depends, AST decl, AST body, str doc, LocInfo locInfo, loc srcLoc, value resource);
anno DefInfo AST@hoverInfo;

data LocInfo
	= LocInfo()
	| LocInfo(loc defLoc)
	| LocInfo(loc defLoc, loc refLoc);

public LocInfo LocInfo(AST tree) {
	if((tree@\loc)? && (tree@from)?)
		return LocInfo(tree@\loc, tree@from);
	else if((tree@\loc)?)
		return LocInfo(tree@\loc);
	else
		return LocInfo();
}

public AST getVal(value v) {
	if(node n := v) {
		return astValue(n[0]);
	}
}

public set[ErrorMark] getErrors(value v) {
	if(node n := v) {
		if(set[ErrorMark] errs := n[arity(n)-1])
			return errs;
	}
}

alias XRefs
	= rel[loc token, Ref ref];
	
data Ref
	= Ref(AST qName)
	| LocalRef(DefInfo info)
	;
	
public DefInfo definfo() {
	return DefInfo("",Nop(),Nop(),Nop(),[],{},Nop(),Nop(), "", LocInfo(), |unknown:///|, Nop());
}

public bool hasParent(DefInfo info) {
	return info.parent != Nop();
}

@doc{Relation between names and fully qualified names in a scope}
data Scope = Scope(rel[AST,AST] val);

public Scope Scope() {
	return Scope({});
}

public Scope add(Scope sc, AST name, AST qName) {
	return Scope(sc.val + {<name,qName>});
}

@doc{DefInfo -> TopDecl.}
data TopDecl = TopDecl(AST val);

// data Library = Library(AST qName, AST body);


data Severity = Warning() | Error() | Internal() | Info();

data ErrorMark = ErrorMark(str message, Severity severity, loc location);

data Result[&T] = Result(&T val, XRefs xrefs, set[ErrorMark] errors, set[ErrorMark] warnings) | NoResult(str message, set[ErrorMark] errors, set[ErrorMark] warnings);

public str KIND_PACKAGE = "package";
public str KIND_FILE = "file";
public str KIND_PROGRAM = "program";

public int SEVERITY_DEBUG = 0;
public int SEVERITY_INFO = 1;
public int SEVERITY_WARNING = 2;
public int SEVERITY_ERROR = 3;
public int SEVERITY_FATAL = 4;
