/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::analysis::Optimiser
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::TypeChecker;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::typechecker::algo::Primitives;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import String;
import List;
import Exception;
import util::Benchmark;

public Result[AST] getOptimised(DefInfo info) {
	env = newEnv(info.qName, info.parent, locOf(info.body));
	switch(info.body) {
		case DeclBody(seq(ds)): {
			env = loadDefs(info.body, true, env);
			ds = for(def <- ds) {
			
				<optDef, env> = optimiseOperation(def, env);
			
				
				append optDef;
			}
			return makeResult(preserveAnnos(DeclBody(seq(ds)), info.body), env);
			}
		default: {
			println("Can\'t optimise <consOf(info.body)>");
			return makeResult(info.body, env);
		}
	}
}
public tuple[AST, Env] optimiseOperation(def:Define(dcl,attrs,body), Env env) {
	tmpCounter = 0;
	if(!isOperation(def))
		return <def, env>;
	
	// termPrintln(def);
	
	switch(body) {
		case Body(b:Block(ss)): {
			<ss, env> = <ss, env>; // TODO
			return <def[arg2 = Body(b[arg0=ss])], env>;
		}
		case Body(b:BlockExpr(ss, rt)): {
			<ss, env> = <ss, env>;
			return <def[arg2 = Body(b[arg0=ss])], env>;
		}
		case EmptyBody():
			return <def[arg2 = body][arg1 = attrs], env>;
		case Body(b:Apply(FunOf(ProcName(p, seq(ps)), leaf(_argn)), _)):
			return <def, env>;
		case Instance(_,_):
			return <def, env>;
		case FunName(_,_,_):
			return <def, env>;
		case ProcName(_,_):
			return <def, env>;
		case Body(e): {
			<b, env> = <seq([Return(e)]), env>; // TODO
			return <def[arg2 = Body(Block(b))][arg1 = attrs], env>;
		}
		default: {
			errorMark(env, "INTERNAL COMPILER ERROR: optimiseOperation(): Unknown operation body cons <consOf(body)>", Internal(), locOf(dcl));
			return <def, env>;
		}
	};
}
