/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::analysis::Permissions

import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::typechecker::algo::Primitives;
import org::magnolialang::magnolia::typechecker::TypeChecker;
import org::nuthatchery::pica::terms::Terms;
import Set;
import Node;
import List;
import IO;

data AST
	= Alt(set[AST] alt0, set[AST] alt1)
	| Not(AST arg0);

public tuple[AST, Env] analysePermissions(progBody:DeclBody(seq(ds)), Env env) {
	env = enterScope(env);
	env = loadDefs(progBody, false, env);
		
	set[AST] perms = {};
	
	for(Define(dcl, attrs, _) <- ds) {
		for(Alert(AlertName(Name(leaf("RequiresPermission"))), AlertPre(expr)) <- attrs.args) {
			<ps, env> = splitPredicateExpr(expr, env);
			perms += {negate(p) | p <- ps};
		}
	}
	
	// make decisions
	set[AST] neededPerms = decidePermissionSet(perms);
	
	termPrintln("Permissions needed: ", neededPerms);
	
	ds = [updatePredicate(d, neededPerms) | d <- ds];
	
	progBody.arg0.args = ds;
	env = exitDefScope(env);
	return <progBody, env>;
}


tuple[set[AST], Env] splitPredicateExpr(AST expr, Env env) {
	and = predicateName("_&&_", 2);
	or = predicateName("_||_", 2);
	not = predicateName("!_", 1);
	TRUE = predicateName("TRUE", 0);
	
	set[AST] result = {};
	switch(expr) {
		case Apply(Builtin(and,_), seq([l, r])): {
			<ls, env> = splitPredicateExpr(l, env);
			<rs, env> = splitPredicateExpr(r, env);
			result = ls + rs;
		}
		case Apply(Builtin(or,_), seq([l, r])): {
			<ls, env> = splitPredicateExpr(l, env);
			<rs, env> = splitPredicateExpr(r, env);
			result = {Alt(ls, rs)};
		}
		case Apply(Builtin(not,_), seq([x])): {
			<xs, env> = splitPredicateExpr(x, env);
			result = {negate(p) | p <- xs};
		}
		case Apply(Builtin(TRUE,_), seq([])):
			;
		case Apply(Fun(fn), seq(as)): {
			if(hasDef(fn, env)) {
				info = lookupDef(fn, env);
				if(Body(Apply(Fun(FunName(Name(leaf("Permission")), seq([]), SpecialType(Predicate()))), seq([]))) := info.body) {
					result = {fn};
				}
				else {
					termPrintln("Permissions::splitPredicateExpr: illegal predicate expression: ", expr);
					env = errorMark(env, "Illegal predicate call in permissions declaration", locOf(expr));
				}
			}
			else {
				env = errorMark(env, "No definition found for <dispNameOf(fn)>", locOf(fn));
			}
		}
		default: {
			termPrintln("Permissions::splitPredicateExpr: illegal predicate expression: ", expr);
			env = errorMark(env, "Illegal predicate expression in permissions declaration", locOf(expr));
		}
	};
	
	return <result, env>;
}

AST negate(AST p) {
	not = predicateName("!_", 1);
	if(Apply(Builtin(not, _), seq([a])) := p) {
		return a;
	}
	else {
		return Apply(Builtin(not, leaf("!_(\\predicate)")), seq([p]));
	}
}

AST updatePredicate(d:Define(dcl:FunClause(qName,_,SpecialType(Predicate())), attrs, body), set[AST] neededPerms) {
	permName = FunName(Name(leaf("Permission")), seq([]), SpecialType(Predicate()));

	if(Body(app:Apply(Fun(permName), seq([]))) := body) {
		if(qName in neededPerms) {
			d.arg2 = preserveAnnos(Body(preserveAnnos(Apply(Builtin(predicateName("TRUE", 0), leaf("TRUE()")), seq([])), app)), body);
		}
		else {
			d.arg2 = preserveAnnos(Body(preserveAnnos(Apply(Builtin(predicateName("FALSE", 0), leaf("FALSE()")), seq([])), app)), body);
		}
	}
	else if(qName == permName) {
		d.arg2 = preserveAnnos(Body(preserveAnnos(Apply(Builtin(predicateName("FALSE", 0), leaf("FALSE()")), seq([])), body)), body);
	}
	return d;
}


default AST updatePredicate(AST d, set[AST] neededPerms) {
	return d;
}

set[AST] decidePermissionSet(set[AST] perms) {
	set[AST] neededPerms = {};
	while(size(perms) > 0) {
		<p, perms> = takeOneFrom(perms);
		if(Alt(l, r) := p) {
			perms += l;
		}
		else {
			neededPerms += p;
		}
	}
	return neededPerms;
}