package org.magnolialang.magnolia.backend;

import static nuthatch.rascal.pattern.StaticValuesPatternFactory.*;
import static org.magnolialang.magnolia.AbstractSyntax.*;
import static org.magnolialang.magnolia.grammar.MagnoliaPatterns.*;
import static org.nuthatchery.pgf.trees.PPUtil.*;

import java.util.Collection;

import nuthatch.library.ActionFactory;
import nuthatch.library.BaseMatchAction;
import nuthatch.library.FactoryFactory;
import nuthatch.library.MatchBuilder;
import nuthatch.pattern.Environment;
import nuthatch.pattern.Pattern;
import nuthatch.rascal.adapter.ValuesCursor;
import nuthatch.rascal.adapter.ValuesWalker;
import nuthatch.rascal.pattern.ValuesBuildContext;
import nuthatch.rascal.pattern.impl.ValuesPatternFactory;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.type.Type;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.Magnolia;
import org.nuthatchery.pgf.config.TokenizerConfig;
import org.nuthatchery.pgf.config.TokenizerConfigBase;
import org.nuthatchery.pgf.plumbing.ForwardStream;
import org.nuthatchery.pgf.tokens.CategoryStore;
import org.nuthatchery.pgf.tokens.DataToken;
import org.nuthatchery.pgf.tokens.Token;
import org.nuthatchery.pgf.trees.PPBuilder;
import org.nuthatchery.pgf.trees.PrettyPrinter;
import org.nuthatchery.pgf.trees.TreeToStream;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;

@NonNullByDefault(false)
public class CaslBackend {
	protected static final ActionFactory<IValue, Type, ValuesCursor, ValuesWalker> af = FactoryFactory.getActionFactory(IValue.class, Type.class, ValuesCursor.class, ValuesWalker.class);
	protected static final ValuesPatternFactory pf = ValuesPatternFactory.getInstance();
	protected static final ValuesBuildContext context = new ValuesBuildContext(TermFactory.vf, TermFactory.ts);

	static final TokenizerConfig config = new TokenizerConfigBase() {

		@Override
		protected @Nullable String cfgIdentifierRegex() {
			// TODO Auto-generated method stub
			return null;
		}


		@Override
		protected CatSet cfgInitialCategorySet() {
			return CatSet.CATSET_DEFAULT;
		}


		@Override
		protected @Nullable String cfgKeywordRegex() {
			// TODO Auto-generated method stub
			return null;
		}


		@Override
		protected @Nullable Collection<String> cfgKeywords() {
			// TODO Auto-generated method stub
			return null;
		}


		@Override
		protected boolean cfgUseDefaultLiteralClassification() {
			return true;
		}


		@Override
		protected void moreCategories(@Nullable CategoryStore store) {
			// TODO Auto-generated method stub

		}
	};


	public static @Nullable Pattern<IValue, Type> AxiomClause(Pattern<IValue, Type> arg0, Pattern<IValue, Type> arg1) {
		return pf.cons(TermFactory.consType("AxiomClause", 2), arg0, arg1);
	};


	public static PrettyPrinter<IValue, Type> makePrinter() {
		return makePrinter(PrinterMode.STANDARD);
	}


	public static PrettyPrinter<IValue, Type> makePrinter(PrinterMode mode) {
		PPBuilder<IValue, Type> axiomBuilder = new PPBuilder<>(config, context);

		axiomBuilder.addList(leaf(_), PPBuilder.custom(new TreeToStream<IValue>() {
			@Override
			public void printTree(IValue tree, ForwardStream<Token> output) {
				if(tree instanceof IString) {
					String s = ((IString) tree).getValue();
					output.put(new DataToken(s, config.getCatForLexical(s)));
				}
			}
		}, 1));


		axiomBuilder.addTypedTmpl(Apply(descendant(leaf("_=>_")), seq(_, _)), sort("Expr"),//
				"<2.1.1>  =>  <2.1.2>");
		axiomBuilder.addTypedTmpl(Apply(descendant(leaf("_<=>_")), seq(_, _)), sort("Expr"),//
				"<2.1.1>  <=>  <2.1.2>");
		axiomBuilder.addTypedTmpl(Apply(descendant(leaf("_<->_")), seq(_, _)), sort("Expr"),//
				"<2.1.1>  =  <2.1.2>");
		axiomBuilder.addTypedTmpl(Apply(descendant(leaf("_&&_")), seq(_, _)), sort("Expr"),//
				"<2.1.1>  /\\  <2.1.2>");
		axiomBuilder.addTypedTmpl(Apply(descendant(leaf("_||_")), seq(_, _)), sort("Expr"),//
				"<2.1.1>  \\/  <2.1.2>");
		axiomBuilder.addTypedTmpl(Apply(descendant(leaf("!_")), seq(_)), sort("Expr"),//
				"not ( <2.1.1> )");
		axiomBuilder.addTypedTmpl(Apply(_, seq()), sort("Expr"),//
				"<1>");
		axiomBuilder.addTypedTmpl(Apply(_, _), sort("Expr"),//
				"<1> ( <commaSep(2.1)> )");

		axiomBuilder.addTypedTmpl(Assert(_, _), sort("Stat"),//
				".  <1> \n");

		axiomBuilder.addTypedTmpl(Assign(_, _), sort("Stat"),//
				"<1>  =  <2>  ;");

		axiomBuilder.addTypedTmpl(Assume(_, _), sort("Stat"),//
				".  <1>");
		axiomBuilder.addTypedTmpl(Fun(_), sort("Fun"),//
				"<1>");
		axiomBuilder.addTypedTmpl(IfThenElseExpr(_, _, _), sort("Expr"),//
				"if  <1>  then  <2>  else  <3>  end");

		axiomBuilder.addTypedTmpl(Literal(_), sort("Expr"),//
				"<1>");

		axiomBuilder.addTypedTmpl(Literal(_, _), sort("Expr"),//
				"<1>  :  <2>");
		axiomBuilder.addTypedTmpl(OctalIntegerLiteral(_), lex("IntegerLiteral"),//
				"<1>");
		axiomBuilder.addTypedTmpl(StringLiteral(_), sort("StringLiteral"),//
				"<1>");
		axiomBuilder.addTypedTmpl(Struct(_, _), sort("Expr"),//
				"<1> ( <commaSep(2.1)> )");
		axiomBuilder.addTypedTmpl(Undefined(_), sort("Expr"),//
				"_");
		axiomBuilder.addTypedTmpl(Var(_), sort("Expr"),//
				"<1>");
		axiomBuilder.addTypedTmpl(Param(_, _, _), sort("ProcedureParam"),//
				"<2> : <3>");

		axiomBuilder.addTypedTmpl(ParamList(seq()), sort("ParamList"),//
				"");
		axiomBuilder.addTypedTmpl(ParamList(_), sort("ParamList"),//
				"forall  <sepBy(1.1,',')>");
		axiomBuilder.addTypedTmpl(FunName(_, _, _), sort("QualifiedName"),//
				"<1>");

		axiomBuilder.addTypedTmpl(IdName(_, _), sort("NameIdentifier"),//
				"<1> <2>");

		axiomBuilder.addTmpl(Define(AxiomClause(_, _), descendant(ImplyMod()), _), "<1.2> \n <3>   %( <1.1> )%  %implied \n\n");
		axiomBuilder.addTmpl(Define(AxiomClause(_, _), _, _), "<1.2> \n <3>   %( <1.1> )% \n\n");
		axiomBuilder.addTypedTmpl(QName(_), sort("QualifiedNameIdentifier"),//
				"<sepBy(1.1,'.')>");
		axiomBuilder.addTypedTmpl(Name(_), sort("NameIdentifier"),//
				"<1>");
		axiomBuilder.addTypedTmpl(TypeName(_), sort("QualifiedName"),//
				"<1>");
		axiomBuilder.addTypedTmpl(Body(_), sort("ExprDefBody"),//
				"<1>");

		axiomBuilder.addTypedTmpl(Return(_), sort(""),//
				"");
		axiomBuilder.addTypedTmpl(EmptyBody(), sort("ExprDefBody"),//
				"");

		axiomBuilder.addTypedTmpl(Equal(Fun(Name(leaf("_==_"))), _, _), sort(""), // lhs, op, rhs
				"<2>  =  <3>");
		axiomBuilder.addTypedTmpl(Equal(Fun(Name(leaf("_!=_"))), _, _), sort(""), // lhs, op, rhs
				"<2>  !=  <3>");
		axiomBuilder.addTypedTmpl(Equal(_, _, _), sort(""), // lhs, op, rhs
				"<2>  <1>  <3>");
		axiomBuilder.addTypedTmpl(Builtin(_, _), sort(""), // qname, idstring
				"<1>");
		axiomBuilder.addTypedTmpl(VarName(_, _), sort("QualifiedName"),//
				"<1>");

		//axiomBuilder.addList(_, "");


		axiomBuilder.addList(Unknown(), PPBuilder.custom(new TreeToStream<IValue>() {
			@Override
			public void printTree(IValue tree, ForwardStream<Token> output) {
				if(tree instanceof IConstructor) {
					output.put(new DataToken(((IConstructor) tree).getName(), config.cfgCategories().category("TXT")));
				}
			}
		}, 0), "(", PPBuilder.sepBy(0, ","), ")");

		final PrettyPrinter<IValue, Type> axiomPrinter = axiomBuilder.compile();
		axiomBuilder = null;

		PPBuilder<IValue, Type> specBuilder = new PPBuilder<>(config, context);

		specBuilder.addList(leaf(_), PPBuilder.custom(new TreeToStream<IValue>() {
			@Override
			public void printTree(IValue tree, ForwardStream<Token> output) {
				if(tree instanceof IString) {
					String s = ((IString) tree).getValue();
					output.put(new DataToken(s, config.getCatForLexical(s)));
				}
			}
		}, 1));


		specBuilder.addTypedTmpl(RenameList(_), sort(""), // a list of renaming operations
				"[ <commaSep(1.1)> ]");
		specBuilder.addTypedTmpl(RenameSeq(_, _), sort(""), // a sequence of renamings
				"<1> <2>");
		specBuilder.addTypedTmpl(Equal(_, _, _), sort(""), // lhs, op, rhs
				"<1> ( <2> , <3> )");
		specBuilder.addTypedTmpl(Builtin(_, _), sort(""), // qname, idstring
				"<1>");
		specBuilder.addTypedTmpl(Define(FunClause(_, _, _), _, _), sort(""), // dcl, attrs, body
				"<1> ; \n");
		specBuilder.addTypedTmpl(Define(TypeClause(_), _, _), sort(""), // dcl, attrs, body
				"<1> ; \n");
		specBuilder.addList(Define(AxiomClause(_, _), _, Body(_)), PPBuilder.custom(new TreeToStream<IValue>() {
			@Override
			public void printTree(IValue tree, ForwardStream<Token> output) {
				axiomPrinter.print(new ValuesCursor(tree), output);
			}
		}));
		specBuilder.addTypedTmpl(Instance(_, _), sort(""), // op/type, list of bindings (instantiation of a Fun, Proc, or a type)
				"<1> [ <commaSep(2.1)> ]");
		specBuilder.addTypedTmpl(Binding(_, _), sort(""), // from-name, to-name (for use in Instance), sort("Modifier"),
				"<1> => <2>");
		specBuilder.addTypedTmpl(DotOp(_, _, _), sort(""), // expr, name, type
				"_._ ( <1> ,  <2> )");
		specBuilder.addTypedTmpl(DeclBody(_), sort(""), //
				"<sepBy(1.1,'')>");
		specBuilder.addTypedTmpl(Undefined(), sort(""), //
				"_");
		specBuilder.addTypedTmpl(AnonParam(_, _), sort("AnonParam"),//
				"<1>  <2>");


		specBuilder.addTypedTmpl(Body(_), sort("ExprDefBody"),//
				"=  <1>");

		specBuilder.addTypedTmpl(EmptyBody(), sort("ExprDefBody"),//
				"");


		specBuilder.addTypedTmpl(FunClause(_, ParamList(_), SpecialType(Predicate())), sort("ExprDeclarative"),//
				"pred  <1>  :  <2>");

		specBuilder.addTypedTmpl(FunClause(_, ParamList(seq()), _), sort("ExprDeclarative"),//
				"ops  <1>  :  <3>");

		specBuilder.addTypedTmpl(FunClause(_, ParamList(_), _), sort("ExprDeclarative"),//
				"ops  <1>  :  <2>  ->  <3>");
		specBuilder.addTypedTmpl(FunName(_, _, _), sort("QualifiedName"),//
				"<1>");

		specBuilder.addTypedTmpl(IdName(_, _), sort("NameIdentifier"),//
				"<1> <2>");


		specBuilder.addTypedTmpl(Models(_, _), sort("SatisfactionExpr"),//
				"<1>  approximates  <2>");

		specBuilder.addTypedTmpl(Models(_, _), sort("SatisfactionExpr"),//
				"<1>  models  <2>");

		specBuilder.addTypedTmpl(Name(_), sort("NameIdentifier"),//
				"<1>");

		specBuilder.addTypedTmpl(Nop(), sort("Decl"),//
				";");


		specBuilder.addTypedTmpl(Param(_, _, _), sort("ProcedureParam"),//
				"<3>");

		specBuilder.addTypedTmpl(ParamList(_), sort("ParamList"),//
				"<sepBy(1.1,'*')>");

		specBuilder.addTypedTmpl(SpecialType(Predicate()), sort("SpecialType"),//
				"\\predicate");

		specBuilder.addTypedTmpl(QName(_), sort("QualifiedNameIdentifier"),//
				"<sepBy(1.1,'.')>");


		specBuilder.addTypedTmpl(Struct(_), sort("Type"),//
				"( <sepBy(1.1,'\n')> )");


		specBuilder.addTypedTmpl(Type(_), sort("Type"),//
				"<1>");

		specBuilder.addTypedTmpl(TypeClause(_), sort("TypeDeclarative"),//
				"sort  <1>");

		specBuilder.addTypedTmpl(TypeField(VarDecl(_, _, _, _)), sort("TypeField"),//
				"<1.3>");

		specBuilder.addTypedTmpl(TypeName(_), sort("QualifiedName"),//
				"<1>");


		specBuilder.addTypedTmpl(Unknown(), sort("Type"),//
				"?");

		specBuilder.addTypedTmpl(VarClause(_, _), sort("ExprDeclarative"),//
				"var  <1>  :  <2>");

		specBuilder.addTypedTmpl(VarName(_, _), sort("QualifiedName"),//
				"<1>");

		specBuilder.addTypedTmpl(WithModels(_, _, _), sort("SatisfactionExpr"),//
				"<1>  with  <2>  approximates  <3>");

		specBuilder.addTypedTmpl(WithModels(_, _, _), sort("SatisfactionExpr"),//
				"<1>  with  <2>  models  <3>");

		if(mode == PrinterMode.FULL) {
			specBuilder.addList(_, PPBuilder.custom(new TreeToStream<IValue>() {
				@Override
				public void printTree(IValue tree, ForwardStream<Token> output) {
					if(tree instanceof IConstructor) {
						output.put(new DataToken(((IConstructor) tree).getName(), config.cfgCategories().category("TXT")));
					}
				}
			}, 0), "(", PPBuilder.sepBy(0, ","), ")");
		}
		else {
			specBuilder.addList(_, "");
		}


		return specBuilder.compile();
	}


	public static String toCasl(IValue tree) {
		PrettyPrinter<IValue, Type> printer = CaslBackend.makePrinter();
		if(TermAdapter.isCons(tree, "DeclBody")) {
			tree = TermFactory.cons("DeclBody", TermFactory.seq((IList) Magnolia.getInstance().getCompiler().call("sortDefs", TermAdapter.getArg((IConstructor) tree, 0).get("args"))));
		}
		ValuesCursor cursor = new ValuesCursor(tree);
		cursor = caslPreprocess(cursor);
		return printer.toString(cursor).replace("_", "__");
	}


	@SuppressWarnings("null")
	private @Nullable static ValuesCursor caslPreprocess(ValuesCursor tree) {
		MatchBuilder<IValue, Type, ValuesCursor, ValuesWalker> mb = af.matchBuilder(context);

		mb.add(Define(ProcClause(var("n"), var("ps")), var("attrs", descendant(Axiom())), var("body")), //
				Define(AxiomClause(var("n"), var("ps")), var("attrs"), var("body")));
		mb.add(IdName(var("n"), var("id")), new BaseMatchAction<IValue, Type, ValuesCursor, ValuesWalker>() {


			@Override
			public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
				String n = TermAdapter.getString((IConstructor) env.get("n").getData());
				String id = TermAdapter.getString((IConstructor) env.get("id").getData());
				String m = opToName(n);
				if(n != m) {
					walker.replace(IdName(leaf(m), leaf(id)).build(context, env));
				}
				return PARENT;
			}
		});
		ValuesWalker walker = new ValuesWalker(tree, af.walk(af.down(mb.one())));
		walker.start();
		return walker.copySubtree();
	}


	private static String opToName(String n) {
		switch(n) {
		case "+_":
			return "pos";
		case "-_":
			return "neg";
		case "!_":
			return "not";
		case "~_":
			return "inv";
			// binary operators
			// MULOP: level 1
		case "_*_":
			return "mul";
		case "_/_":
			return "div";
		case "_%_":
			return "mod";
		case "_**_":
			return "mulmul";
			// ADDOP: level 2
		case "_-_":
			return "sub";
		case "_+_":
			return "add";
		case "_++_":
			return "addadd";
			// SHFOP: level 3
		case "_<<_":
			return "lshift";
		case "_>>_":
			return "rshift";
			// RNGOP: level 4
		case "_.._":
			return "range";
			// CMPOP, EQUOP, INOP, NOTINOP: level 5
		case "_<_":
			return "lt";
		case "_>_":
			return "gt";
		case "_<=_":
			return "le";
		case "_>=_":
			return "ge";
		case "_===_":
			return "eq";
		case "_!==_":
			return "ne";
			// BITAND: level 6
		case "_&_":
			return "and";
			// BITXOR: level 7
		case "_^_":
			return "xor";
			// BITOR: level 8
		case "_|_":
			return "or";
			// SUBST: level 9
			// LGAND: level 10
			//case "_&&_": return "andand";
			// LGOR: level 11
			//case "_||_": return "oror";
			//case "_^^_": return "xorxor";
			// LGIMP: level 12
			// Others
		case "_[_]":
			return "getelem";
		case "_[_]=_":
			return "setelem";

		}
		return n;
	}


	enum PrinterMode {
		STANDARD, FULL
	}
}
