/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::backend::CxxBackend
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::CxxAPI;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import org::magnolialang::magnolia::backend::CxxPP;
import org::magnolialang::magnolia::backend::OperatorInfo;
import org::magnolialang::magnolia::backend::ProgramBuilder;
import Set;
import Node;
import List;
import Map;
import Relation;
import String;
import util::Benchmark;
import analysis::graphs::Graph;

public Result[str] cxxCode(DefInfo info, str outname, bool acceptErroneous) {
	env = minimalEnv(info.body@\loc);

	str body = "";
	bool hasCode = false;
	env = startWork("cxxCode", nameToStr(info.qName), env,false);
	// library
	if(DeclBody(_) := info.body) {
		if(info.kind == "program" || info.kind == "satisfaction") {
			if(DeclBody(seq(ds)) := info.body) {
				try {
					<incl, exts> = findCxxIncludes(info.body);
					<body, _, env> = generateImpl(ds, makeCxxNonKeywordName(info.qName, "_mod"), exts, false, env);
					if(body != "") {
						markWork("generateImpl", env);
						body = makeNameSpace(info.parent, body);
						markWork("findCxxIncludes", env);
						for(i <- incl) {
							body = "#include \"<i>\"\n" + body;
						}
						// if(info.kind == "program")
						body = generateCommandInterface(env, ds, info.qName, body); 
						hasCode = true;
						markWork("cmdInterface", env);
					}
				}
				catch "not c++": {;
					body = "  /* no code generated */";
					env = errorMark(env, "Java code ignored by C++ code generator", Info(), locOf(info.qName));
				}
			}
		}
		else {
			body = "  /* no code generated */";
		}
	}
	else {
		println("Don\'t know how to deal with <consOf(info.body)>");
		body = "  /* Don\'t know how to deal with <consOf(info.body)> */";
	}
	env = endWork(env);
	if(hasCode) {
		//env = errorMark(env, "C++ code output in \'<outname>\'", Info(), locOf(info.qName));
		return makeResult("// <info.kind>: <dispNameOf(info.qName)>\n<body>\n", env);
	}
	else {
		//env = errorMark(env, "No code generated for \'<shortDispNameOf(info.qName)>\' by CxxBackend", Info(), locOf(info.qName));
		return makeNoResult("no code generated", env);
	}
}


tuple[set[str], rel[AST,AST]] findCxxIncludes(&T tree) {
	set[str] modules = {};
	rel[AST,AST] externals = {};
	tree = visit(tree) {
		case ExternalType(l, m, onPart, t): {
			termPrintln("*** External type: ", t);
			if(!isCxxLanguage(l))
				throw "not c++";
			modules += "<replaceAll(nameToStr(m), ".", "/")>.hxx";
			externals[m] = onPart;
//			insert qualify(qualify(m, Name(leaf("defines"))), t);
		}
		case ExternalProc(l, m, onPart, t, ps): {
			termPrintln("*** External proc: ", t);
			if(!isCxxLanguage(l))
				throw "not c++";
			modules += "<replaceAll(nameToStr(m), ".", "/")>.hxx";
			externals[m] = onPart;
//			insert qualify(qualify(m, Name(leaf("defines"))), t);
		}
		case ExternalFun(l, m, onPart, t, ps, rt): {
			termPrintln("*** External fun:  ", t);
			if(!isCxxLanguage(l))
				throw "not c++";
			modules += "<replaceAll(nameToStr(m), ".", "/")>.hxx";
			externals[m] = onPart;
//			insert qualify(qualify(m, Name(leaf("defines"))), t);
		}
	}

	for(<m, e> <- externals) {
		termPrintln("*** Externals: module: ", m); 
		termPrintln("********************** ", e); 
	}
	return <modules, externals>;
}

// 
//
//
//
//

tuple[str, list[AST], Env] generateImpl(list[AST] defs, str structName, rel[AST,AST] exts, bool allowRequire, Env env) {
	env = startWork("generateImpl", structName, env, true);
	dcls = sortDefs(defs);
	list[str] typeParams = [];
	list[str] opParams = [];
	list[str] ftorFwds = [];
	set[str] fields = {};
	Graph[AST] types = {};
	list[str] consParams = [];
	list[str] consArgs = [];
	set[str] cons = {};
	set[str] usedNames = {};
	map[AST,AST] typeDefs = ();
	map[AST,AST] typeNames = ();
	map[AST,AST] renaming = ();
	lrel[AST,str] runtimeDefs = [];
	int i = 0;
	markWork("0", env);
	for(d:Define(dcl,seq(mods),body) <- dcls) {
		markWork("d1", env);
		dn = defNameOf(dcl);
		mn = basename(unqualify(dn));
		n = makeCxxName(dcl);
		termPrintln("defNameOf: ", mn, " dn: ", dn, " cxxName: ", n);
		if(Assume() in mods) { // we can't generate code for this
			termPrintln("Found Assume() in declaration list for ", structName, " -- bailing out");
			return <"", [], env>;
		}
		else if(allowRequire && Require() in mods) { // it's a parameter
			if(Export() notin mods) {
				n = "_<n>";
			}	
			markWork("makeUnique", env);
			<N,usedNames> = makeUnique(toUpperCase(n), usedNames);
			markWork("makeUnique done", env);
			if(isOperation(dcl)) {
				opParams += "<N>_OP"; // template parameter
				fields += "const <N>_OP& <n>;\n"; // local field
				consParams += "const <N>_OP& arg<i>"; // constructor param
				cons += "<n>(arg<i>)"; // constructor initialiser
				consArgs += "arg<i>"; // constructor argurment
				i += 1;
				typeNames[dn] = leaf("<N>_OP");
			}
			else if(isTypeDecl(dcl)) {
				typeParams += "<N>_T"; // template parameter
				types[Nop()] = dn;
				typeDefs[dn] = Typedef(leaf("<N>_T"), leaf(n)); // local field
				typeNames[dn] = leaf("<N>_T");
			}
			else if(isAlertDecl(dcl)) {
				defs = defs - d;
			}
			markWork("d1a", env);
		}
		else if(isAlertDecl(dcl)) {
			defs = defs - d;
		}
//		else if(body == EmptyBody()) { // it's derived
//			defs = defs - d;
//		}
		else { // it's defined here
			if(Require() in mods) {
				defs = defs - d;
				d.arg2 = body = Missing(dn);
				d.arg1 = seq(mods - Require());
				defs = defs + d;
			}
			
			if(isOperation(dcl)) {
				if(isName(body)) {
					fields += "const <makeCxxName(body)>_ftor& <n> = <makeCxxName(body)>;\n"; // local field
					//cons += "<n>{*this}"; // constructor initialiser
					defs = defs - d;
				}
				else {
					ftorFwds += "struct <n>_ftor;\n"; // forward declaration
					fields += "const <n>_ftor <n>;\n"; // local field
					cons += "<n>{*this}"; // constructor initialiser
					typeNames[dn] = leaf("<n>_ftor");
					runtimeDefs += <d, "<n>_ftor">;
				}
			}
			else if(isTypeDecl(dcl)) {
				if(isName(body)) {
					types[Nop()] = dn;
					types[body] = dn;
					typeDefs[dn] = Typedef(leaf("<makeCxxName(body)>"), leaf(n));
					typeNames[dn] = leaf(n);
				}
				else {
					if(Body(Struct(_)) := body) {
						types[Nop()] = dn;
						types += {<dep, dn> | /dep:TypeName(_) <- body};
						typeDefs[dn] = d;
						typeNames[dn] = leaf(n);
					}
					else if(Missing(_) := body) {
						types[Nop()] = dn;
						typeDefs[dn] = d;
						typeNames[dn] = leaf(n);
					}
					else if(Body(ExternalType(lang, qname, onPart, extName)) := body) {
						fullExtName = QName(seq([externalImplTypeName(qname,onPart), leaf(ppName(extName))]));
						types[Nop()] = dn;
						types += {<dep, dn> | /dep:TypeName(_) <- fullExtName};
						typeDefs[dn] = Typedef(fullExtName, leaf(n));
						typeNames[dn] = fullExtName;
					}
				}
				runtimeDefs += <d, n>;
				defs = defs - d;
			}
			
			markWork("d1b", env);
		}
		
		//termPrintln("Rename: ", defNameOf(dcl), " : ", Name(leaf(n)));
		markWork("d2a", env);
		if(isOperation(dcl)) {
			renaming[Fun(dn)] = Fun(Name(leaf("base.<n>")));
			renaming[Proc(dn)] = Proc(Name(leaf("base.<n>")));
			//renaming += (Fun(dn) : Fun(Name(leaf("base.<n>"))));
			//renaming += (Proc(dn) : Proc(Name(leaf("base.<n>"))));
		}
		renaming[defNameOf(dcl)] = Name(leaf(n));
		//renaming += (defNameOf(dcl) : Name(leaf(n)));
		markWork("d2b", env);
	}
	
	// inline references to other types in typedefs, so we don't have to sort them
	// note: names in the list of types that shouldn't be rewritten should be
	// wrapped in just leaf() not Name().
	// types = [rename(t, typeNames) | t <- types];
	// sorting should place locally defined Structs in front of TypeDefs that may reference
	// them in template arguments
	typeOrder = order(types);
	typeDefsInOrder = [trToCxx(rename(typeDefs[tn] ? Nop(), typeNames), renaming) | tn <- typeOrder, tn != Nop()];
	
	// we need both the renamed external parameter names, and the
	// original ones
	list[tuple[AST,AST,AST]] externals;
	externals = for(<e,p> <- exts) {
		AST p2, pTmp;
		p2 = p;
		do {
			pTmp = p2;
	 		p2 = rename(p2, typeNames);
	 	}
	 	while(p2 != pTmp);
	 	
	 	//termPrintln("Externals: ", p, " =\> ", p2);
		append <e,p,p2>;
	}
	markWork("1", env);
	
	str ctor = "";
	consInits = toList(cons) + ["<externalImplName(e, p)>(<strJoin([makeCxxName(x) | x <- ops], ", ")>)" | <e,p:ExternalParams(_,seq(ops)),_> <- externals];


	opParamList = isEmpty(opParams) ? "" : "\<<strJoin(["typename <t>" | t <- sort(opParams)], ", ")>\>";
	opArgList = isEmpty(opParams) ? "" : "\<<strJoin(sort(opParams), ", ")>\>";
	markWork("2", env);
	ds = [trToCxx(addModifier(x, Member()), renaming) | x <- definesOf(DeclBody(seq(defs)))];
	markWork("3", env);
	defs = groupByName(ds, Name(leaf("<structName>_impl")));
	body = "";
	markWork("4", env);

	runtimeInfo = ""; // makeRuntimeInfo(runtimeDefs, renaming);

	ctor = "<structName>_impl(<strJoin(consParams, ", ")>)\n"
	     + (isEmpty(consInits) ? "" : "\t: <strJoin(consInits, ", ")>\n")
	     + "\t{<runtimeInfo>}\n";
	
	body = (isEmpty(typeParams) ? "" : "template\<<strJoin(["typename <t>" | t <- typeParams], ", ")>\>\n")
	     + "struct <structName> {\n"
	     +   (isEmpty(opParams) ? "" : "template<opParamList>\n")
		 +   "struct <structName>_impl {\n"
		 +     strJoin(ftorFwds, "")
		 +     "<strJoin([ppDecl(t) | t <- typeDefsInOrder], "\n")>\n"
		 +     strJoin(["const <ppName(externalImplTypeName(e, p2))> <externalImplName(e, p)>;\n" | <e,p,p2> <- externals], "")
		 +     "<ctor>\n"
		 +     "<strJoin([ppDecl(d) | d <- defs], "\n")>\n"
		 +     "<strJoin(toList(fields), "")>\n"
		 +   "};\n\n"
	     +   (isEmpty(opParams) ? "" : "template<opParamList>\n")
	     +   "<structName>_impl<opArgList>\n"
		 +   "operator()(<strJoin(consParams, ", ")>) {\n"
		 +     "return <structName>_impl<opArgList>(<strJoin(consArgs, ", ")>);\n"
		 +   "}\n"
		 + "};\n";

	env = endWork(env);
	return <body, ds, env>;
}



tuple[str, set[str]] makeUnique(str name, set[str] usedNames) {
	n = name;
	j = 0;
	while(n in usedNames) {
		if(/^<prefix:.*>_<suffix:[^_]*>$/ := n)
			n = "<prefix><j>_<suffix>";
		else
			n = "<name><j>";
		j = j + 1;
	}
	return <n, usedNames + n>;
}

list[AST] groupByName(list[AST] defs, AST base) {
	set[AST] result = {};
	rel[AST,AST] grouping = {};
	
	for(d <- defs) {
		if(isOperation(d))
			grouping[defNameOf(d)] = d;
		else
			result += d;
	}

	for(n <- domain(grouping)) {
		g = toList(grouping[n]);
		result += FunObjClass(transformName(n, str(str x) {return "<x>_ftor";}), base, seq(g));
	}
	
	return toList(result);
}

str makeRuntimeInfo(lrel[AST,str] defs, map[AST,AST] renaming) {
	for(def:Define(dcl,seq(attrs),body) <- sortDefs(defs<0>)) {
		dn = defNameOf(def);
		cxxName = nameToStr(renaming[dn] ? dn);
		entry = "";
			
		termPrintln("makeRuntimeInfo: ");
		termPrintln("     def=", def);
		termPrintln(" cxxName=", cxxName);
		termPrintln("dispName=", dispNameOf(dn));
		
		if(isOperation(def)) {
			ps = paramsOf(def);
			inPs = inParams(ps);
			outPs = outParams(ps);
			entry = "Magnolia::Runtime::opEntry()";
			if(isFunctionDecl(def)) {
				entry += ".functionKind()";
				entry += ".returnType(\"<nameToStr(returnOf(def))>\")";
				entry += ".outputs({Magnolia::Runtime::paramList().param(\"return\", \"<nameToStr(returnOf(def))>\")})";
			}
			else if(isProcedureDecl(def)) {
				entry += ".procedureKind()";
			}
			
			for(<lst, kind> <- [<ps, "params">, <inPs, "inputs">, <outPs, "outputs">]) {
				if(lst != []) {
					entry += ".params(Magnolia::Runtime::paramList()";
					for(Param(m,n,t) <- ps) {
						entry += ".param(\"<modeToStr(m)>\", \"<nameToStr(n)>\", \"<nameToStr(t)>\")";
					}
					entry += ")";
				}
			}
			
			entry += ".operation(&<cxxName>.operator(<strJoin(["<nameToStr(renaming[t] ? t)>" | Param(m,n,t) <- ps], ", ")>))";
			
		}
		else {
			entry = "Magnolia::Runtime::typeEntry()";
			entry += "";
		}
		
		entry += ".fullName(\"<dispNameOf(dn)>\")";
		list[str] mods = [];
		// termPrintln("XRefs: def: ", def);
		//entry = quoteXML(shortDispNameOf(def));
		//if(body != EmptyBody())
		//	entry = "<entry> = ...";
		//if(Body(Apply(Builtin(_,leaf(b)), seq([]))) := body) {
		//	entry = "<entry> = <b>";
		//}
	
		//if(Guard(g) <- attrs) {
		//	entry += " \<b\>guard\</b\> <prettyMagnolia(g)>";
		//}
		//if(Alert(_,_) <- attrs) {
		//	entry += " \<i\>(alerted)\</i\>";
		//}
		
		//if(IntegerLiteralAttr(typeName, zeroName, oneName, plusName, mulName, negName) <- attrs) {
		//	entry += "\<br\>\n";
		//	entry += "            literal integer(";
		//	entry += "0 =\> <quoteXML(shortDispNameOf(zeroName))>";
		//	entry += ", 1 =\> <quoteXML(shortDispNameOf(oneName))>";
		//	entry += ", _+_ =\> <quoteXML(shortDispNameOf(plusName))>";
		//	entry += ", _*_ =\> <quoteXML(shortDispNameOf(mulName))>";
		//	if(negName != Nop())
		//		entry += ", -_ =\> <quoteXML(shortDispNameOf(negName))>";
		//	entry += ")";
		//}
		if(hasModifier(def, Derived()))
			mods += "DERIVED";
		if(hasModifier(def, External()))
			mods += "EXTERNAL";
		if(hasModifier(def, Default()))
			mods += "DEFAULT";
		if(hasModifier(def, Predicate()))
			mods += "PREDICATE";
		if(hasModifier(def, Require()))
			mods += "REQUIRE";
		if(hasModifier(def, Extend()))
			mods += "EXTEND";
		if(hasModifier(def, Axiom()))
			mods += "AXIOM";
		if(hasModifier(def, Use()))
			mods += "USE";
		if(hasModifier(def, Protect()))
			mods += "PROTECT";
		if(hasModifier(def, Broken()))
			mods += "BROKEN";
		if(hasModifier(def, Assume()))
			mods += "ASSUME";
		if(hasModifier(def, NoBuild()))
			mods += "NOBUILD";
	
		if(mods != [])
			entry += ".attrs(<strJoin(["Magnolia::Runtime::<m> " | m <- mods], "|")>)";
			
		termPrintln("  output=", entry);

	}

	return "";
}

private str bindingsToParams(list[AST] bindings) {
	return strJoin(["<makeCxxName(y)>" | Binding(x, y) <- bindings], ", ");
}
