/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 Tero Hasu
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
module org::magnolialang::magnolia::backend::CxxPP
import org::nuthatchery::pica::memo::Memo;

import IO;
import Node;
import List;
import String;

// Cannot pattern match arguments without the ADT definitions.
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::AbstractSyntax;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::CxxAPI;
import org::magnolialang::magnolia::backend::OperatorInfo;

// spaces, sepBy, etc.
import org::magnolialang::compiler::Util;

data AST 
	= FunObjClass(AST arg0, AST arg1, AST arg2)
	| BinOp(AST arg0, AST arg1, AST arg2, int pri)
	| PreOp(AST arg0, AST arg1, int pri)
	| Member()
	| Comment(AST arg0)
	| Typedef(AST arg0, AST arg1)
	| Struct(AST arg0)
	| TName(AST arg0, AST arg1)
	| Missing(AST arg0)
	;
	
void printlnNoAnnos(value x) {
	println(delAnnotationsRec(x));
}

AST qnameAppend(QName(seq(lst)), AST n) {
	return QName(seq(lst + [n]));
}

default AST qnameAppend(AST qname, AST n) {
	return QName(seq([qname, n]));
}
// --------------------------------------------------
// indentation
// --------------------------------------------------

// Splits into separate lines, discarding line break characters.
list[str] splitLines(str s) {
	list[str] result = [];
	while(/^<pre:.*?>\r?\n<post:.*>$/s := s) {	
		result += pre;
		s = post;
	}
	
	return result + s;
}

int nonneg(int x) {
	if (x < 0) return 0;
	return x;
}

public str indentStr(str expr) {
	lines = splitLines(expr);
	lv = 0;
	r = "";
	for (line <- lines) {
		if (/^\s*\}/ := line) lv = nonneg(lv - 1);
		thislv = lv;
		switch (line) {
		case /^#/: thislv = 0;
		case /:$/: thislv = nonneg(lv - 1);
		};
		r += (tabs(thislv) + line + "\n");
		if (/\{\s*$/ := line) lv += 1;
	}

	return r;
}

// --------------------------------------------------
// translation
// --------------------------------------------------

// An AST node for valid C++ operators only.
// Can use when renaming Magnolia operators with C++ equivalents. xxx
data AST = CxxOperator(str op);


public str externalImplName(AST name, AST params) {
	if(ExternalParams(seq(onTypes), seq(onOps)) := params) {
		<onType, onOps> = bottom-up visit(<onTypes, onOps>) {
			case TName(n, seq(p)) => Name(leaf("<ppName(n)><strJoin([ppName(x) | x <- p], "_")>"))
		}
		typeParam = size(onTypes) > 0 ? "__<strJoin([ppName(n) | n <- onTypes], "_")>" : "";
		opParam = size(onOps) > 0 ? "__<strJoin([ppName(n) | n <- onOps], "_")>" : "";
				
		r = replaceChars("<replaceAll(nameToStr(name), ".", "_")><typeParam><opParam>__xpkg");
//		termPrintln("externalImplName: <r>");
		return r;
	}
}

public AST externalImplTypeName(AST name, AST params) {
	if(ExternalParams(seq(onTypes), seq(onOps)) := params) {
		r = QName(seq([TName(leaf(replaceAll(nameToStr(name), ".", "::")), seq(onTypes)),
		              TName(leaf("<nameToStr(unqualify(name))>_impl"), seq(onOps))]));
//		termPrintln("externalImplTypeName: ", name);
//		termPrintln("                   =\> ", r);
		return r;
	}
}

/* 
The compound assignment operators are the following, and we could use
them to ensugar, but this might change semantics as they are
separately overloadable in C++. In Magnolia they are not defined, so
no such danger there, but there can be foreign definitions.

a += b
a -= b
a *= b
a /= b
a %= b
a &= b
a |= b
a ^= b
a <<= b
a >>= b
*/

// translate to C++
public AST trToCxx(AST ast, map[AST,AST] renaming) {
	return top-down visit (ast) {
		case AnonParam(m, t) => Param(m, Name(leaf("")), t)
		case Define(PredClause(n, pl), p, body) =>
			Define(FunClause(n, pl, TypeName(Name(leaf("bool")))), p,
				makeReturn(body))
		case Define(FunClause(n, pl, rt), p, body) =>
			Define(FunClause(n, pl, rt), p, makeReturn(body))
		case Define(ProcClause(n, pl), p, body) =>
			Define(FunClause(n, pl, TypeName(Name(leaf("void")))), p, body)
		case d:Define(TypeClause(t), p, body) => d
		case Define(dcl, p, body) =>
			leaf("Unknown decl clause <unparse(dcl)>")
		case Apply(ExternalFun(lang, ns, onPart, name, ps, t), seq(args)): {
			//if(!isCxxLanguage(lang))
			//	throw "Call to non-C++ function <nameToStr(name)> in <nameToStr(ns)> (language <nameToStr(lang)>)";
			n = nameToStr(name);

			insert Apply(Fun(Name(leaf("base.<externalImplName(ns,onPart)>.<n>"))), seq(args));
		}
		case Apply(Builtin(name, _), seq(args)): {
			n = nameToStr(name);
			if(n in opMapCxx) {
				info = opMapCxx[n];
				if(info.typ == Infix() && [x, y] := args)
					insert BinOp(x, leaf(info.opSym), y, info.pri);
				else if(info.typ == Prefix() && [x] := args)
					insert PreOp(leaf(info.opSym), x, info.pri);
				else if(info.typ == Funfix())
					insert Apply(Fun(Name(leaf(info.opName))), seq(args));
			}
			else if(n == "TRUE")
				insert Var(Name(leaf("true")));
			else if(n == "FALSE")
				insert Var(Name(leaf("false")));
		}

		case Call(ExternalProc(lang, ns, onPart, name, ps), seq(args)): {
			//if(!isCxxLanguage(lang))
			//	throw "Call to non-C++ procedure <nameToStr(name)> in <nameToStr(ns)> (language <nameToStr(lang)>)";
			n = nameToStr(name);

			insert Call(Proc(Name(leaf("base.<externalImplName(ns,onPart)>.<n>"))), seq(args));
			}
		case n:Name(_):        if(n in renaming) insert renaming[n]; else fail;
		case n:QName(_):       if(n in renaming) insert renaming[n]; else fail;
		case n:FunName(_,_,_): if(n in renaming) insert renaming[n]; else fail;
		case n:VarName(_,_):   if(n in renaming) insert renaming[n]; else fail;
		case n:TypeName(_):    if(n in renaming) insert renaming[n]; else fail;
		case n:ProcName(_,_):  if(n in renaming) insert renaming[n]; else fail;
		case n:Fun(f):         if(n in renaming) insert renaming[n]; else if(!(Name(_) := f)) { termPrintln("*** CxxPP: Missing in renaming: ", f); insert Missing(f); }
		case n:Proc(p):        if(n in renaming) insert renaming[n]; else if(!(Name(_) := p)) { termPrintln("*** CxxPP: Missing in renaming: ", p);insert Missing(p); }
	};
}

AST makeReturn(Body(BlockExpr(b, _))) { return Body(Block(b)); }
AST makeReturn(Body(Block(b))) { return Body(Block(b)); }
AST makeReturn(Body(e)) { return Body(Block(seq([Return(e)]))); }
AST makeReturn(b:EmptyBody()) {	return b; }
AST makeReturn(b:Missing(_)) {	return b; }
AST makeReturn(b:Instance(_,_)) {	return b; }
default AST makeReturn(AST tree) { return Block(seq([Return(tree)])); }



// --------------------------------------------------
// pretty printing
// --------------------------------------------------
public str ppBody(Body(arg0)) {
	return ppStat(arg0);
}

public str ppBody(EmptyBody()) {
	return ";";
}

public str ppBody(Missing(n)) {
		msg = "missing implementation of operation \'<ppName(n)>\'";
		return "{<cxxError(msg, "Magnolia::compile_error(\"<quote(msg)>\");", ";")>}";
}

public str ppTree(MagnoliaTree(arg0, arg1)) {
	return "<ppTree(arg0)> strJoin([ppDecl(d) | d <- arg1.args], "\n\n")>";
}

public str ppTree(PackageHead(arg0, arg1)) {
	return "// package <ppName(arg0)> ;\n";
}

public str ppExpr(Apply(fn:ExternalFun(lang, ns, _, name, ps, rt), args)) {
	return "<ppExpr(fn)>(<strJoin([ppExpr(x) | x <- args.args], ", ")>)";
}

public str ppExpr(Apply(Fun(arg0), args)) {
	return "<ppExpr(Fun(arg0))>(<strJoin([ppExpr(x) | x <- args.args], ", ")>)";
}

public str ppExpr(Apply(Instance(f, bs), args)) {
	return "<ppExpr(Fun(f))>\<<strJoin([ppName(y) | Binding(x, y) <- bs.args], ", ")>\>(<strJoin([ppExpr(x) | x <- args.args], ", ")>)";
}

public str ppExpr(Apply(Missing(f), args)) {
	msg = "missing implementation of \'<ppExpr(Fun(f))>\'"; //  in expression \'<ppExpr(Fun(f))>(<strJoin([ppExpr(x) | x <- args.args], ", ")>)\'";
	return cxxError(msg, "Magnolia::compile_error(\"<quote(msg)>\")", "0");
}

public str ppExpr(BinOp(x, leaf(op), y, pri)) {
	return "<ppPri(x, pri)> <op> <ppPri(y, pri-1)>";	
}

public str ppExpr(PreOp(leaf(op), x, pri)) {
	return "<op><ppPri(x, pri)>";
}

public str ppPri(op:BinOp(_, _, _, pri1), pri2) {
	if(pri1 > pri2)
		return "(<ppExpr(op)>)";
	else
		return ppExpr(op);
}

public str ppPri(op:PreOp(_, _, pri1), pri2) {
	if(pri1 > pri2)
		return "(<ppExpr(op)>)";
	else
		return ppExpr(op);
}

public default str ppPri(AST expr, pri2) {
	return ppExpr(expr);
}

public str ppExpr(BlockExpr(seq(lst), rt)) {
	// TODO: translate return to goto
	return "{ <strJoin([ppStat(l) | l <- lst], "\n")> }";
}

public str ppStat(a:Assert(arg0, seq(clauses))) {
	loc l = locOf(a);
	str comment = "";
	str src = "";
	str doc = "";
	for(c <- clauses) {
		switch(c) {
		case Src(leaf(s)):	src = s;
		case Doc(leaf(s)):  doc = s;
		default:	comment += "<unparse(c)>";
		}
	}
	if(l != unknownLoc) {
		return "Magnolia::Assert::axiom_assert(<ppExpr(arg0)>, \"<quote(doc)>\", \"<quote(src)>\", \"<locInfo(l, unknownLoc)>\"); /* <comment> */";
	}
	else {
		return "Magnolia::Assert::axiom_assert(<ppExpr(arg0)>, \"<quote(doc)>\", \"<quote(src)>\", \"\"); /* <comment> */";
	}
}

public str ppStat(a:Assume(arg0, seq(clauses))) {
	loc l = locOf(a);
	str comment = "";
	str src = "";
	str doc = "";
	for(c <- clauses) {
		switch(c) {
		case Src(leaf(s)):	src = s;
		case Doc(leaf(s)):  doc = s;
		default:	comment += "<unparse(c)>";
		}
	}
	if(l != unknownLoc) {
		return "Magnolia::Assert::axiom_assert(<ppExpr(arg0)>, \"<quote(doc)>\", \"<quote(src)>\", \"<locInfo(l, unknownLoc)>\"); /* <comment> */";
	}
	else {
		return "Magnolia::Assert::axiom_assert(<ppExpr(arg0)>, \"<quote(doc)>\", \"<quote(src)>\", \"\"); /* <comment> */";
	}
}

public str ppStat(Assign(arg0, arg1)) {
	return "<ppExpr(arg0)> = <ppExpr(arg1)> ;";
}

public str ppStat(Block(seq(lst))) {
	s = "{";
	for (stmt <- lst) {
		s += ("\n" + ppStat(stmt));
	}
	s += "\n}";
	return s;
}

public str ppStat(Break()) {
	return "break;";
}

public str ppStat(Call(arg0, arg1)) {
	//printlnNoAnnos(Call(arg0, arg1));
	if(Missing(p) := arg0) {
		msg = "missing implementation of \'<ppExpr(Proc(p))>\'"; //  in call \'<ppExpr(Proc(p))>(<strJoin([ppExpr(x) | x <- arg1.args], ", ")>)\'";
		return cxxError(msg, "Magnolia::Assert::missing_implementation(\"<quote(msg)>\");", ";");
	}
	else if(Instance(p, bs) := arg0) {
		return "<ppExpr(Proc(p))>\<<strJoin([ppName(y) | Binding(x, y) <- bs.args], ", ")>\>(<strJoin([ppExpr(x) | x <- arg1.args], ", ")>)";
	}
	else {
		return "<ppExpr(arg0)>(<strJoin([ppExpr(a) | a <- arg1.args], ", ")>);";
	}
}

public str ppStat(Comment(leaf(s))) {
	return "// <replaceAll(s, "\n", "\n// ")>";
}

public str ppDecl(Comment(leaf(s))) {
	return "// <replaceAll(s, "\n", "\n// ")>";
}

public str ppDecl(leaf(d)) {
	return d;
}

public str ppDecl(Struct(n)) {
	return "struct <ppName(n)>;";
}

public str ppDecl(Typedef(d, n)) {
	return "typedef <ppName(d)> <ppName(n)>;";
}

public str ppDecl(DeclBody(seq(lst))) {
	if ([] := lst)
		return "{\n}";
	return "{\n<for (d <- lst) {><ppDecl(d)>\n<}>}";
}

public str ppDecl(def:Define(FunClause(name, paramList, retType), seq(attrs), body)) {
	str comment = "";
	for(Comment(leaf(s)) <- attrs)
		comment += "/** <s> */\n";
	str modComment = "";
	for(m <- attrs, !(Member() == m || Comment(_) := m)) {
		modComment += "<yieldTerm(m, false)> ";
	}
	if(modComment != "")
		modComment = "// <modComment>\n";
	if(hasModifier(def, Member()))
		return "<comment><modComment><ppName(retType)> <ppName(unqualify(name))><ppParams(paramList)> const <ppBody(body)>";
	else
		return "<comment><modComment><ppName(retType)> <ppName(unqualify(name))><ppParams(paramList)> <ppBody(body)>";
}

public str ppDecl(Define(TypeClause(name), seq(attrs), body)) {
	str comment = "";
	for(Comment(leaf(s)) <- attrs)
		comment += "/** <s> */\n";

	if(EmptyBody() := body){
		if (Require() in attrs)
			return "";
		else
			return "<comment>struct <ppName(unqualify(name))>;";
	}

	str origName(str name) {
		if(/^<n:.*>_t/ := name)
			return n;
		else
			return name;
	}
	list[str] ppFields = [];
	list[str] consArgs = [];
	list[str] consInits = [];
	list[str] fieldNames = [];
	list[str] fieldTypes = [];
	list[str] outStats = [];
	list[str] inStats = [];
	list[str] eqStats = [];
	consBody = "";
	clsName = ppName(unqualify(name)) ? ppName(name); // might be a leaf rather than a proper name

	if(Body(Struct(seq(fields))) := body) {
		for(f <- fields) {
			if(TypeField(VarDecl(m,n,t,_)) := f) {
				ppFields += ppStat(VarDecl(m,n,t,Undefined()));
				consArgs += "const <ppName(t)>& _<ppName(n)>";
				consInits  += "<ppName(n)>(_<ppName(n)>)";
				fieldNames += ppName(n);
				fieldTypes += origName(ppName(t));
				outStats += "<ppName(n)> \>\>= _stream;";
				inStats += "<ppName(n)> \<\<= _stream;";
				eqStats += "if(!(<ppName(n)> == _other.<ppName(n)>)) return false;";
			}
			else {
				ppFields += ppStat(f);
			}
		}
	}
	else if(Missing(_) := body) {
		msg = "missing implementation of type \'<clsName>\'";
		consBody = cxxError(msg, "Magnolia::compile_error(\"<quote(msg)>\");", ";");
	}
	
	return "<comment>struct <clsName> {\n"
	     + "<strJoin(ppFields, "\n")>\n"
	     + "public:\n"
	     + "<clsName>() = default;\n"
	     + (consArgs != [] ? "<clsName>(<strJoin(consArgs, ", ")>) : <strJoin(consInits, ", ")> {<consBody>}\n" : "")
	     + "<clsName>(const <clsName>&) <consBody != "" ? "{<consBody>}" : "= default">;\n"
	     + "~<clsName>() = default;\n"
	     + "<clsName>& operator=(const <clsName>&) = default;\n"
	     + "bool operator==(const <clsName>& _other) const {\n"
	     + "<strJoin(eqStats, "\n")>\n"
	     + "return true;\n"
	     + "}\n"
	     + "template\<typename _Stream\> void operator\>\>=(_Stream& _stream) const {\n"
	     + "Magnolia::IO::Type _type_tag = Magnolia::IO::Type(\"<origName(clsName)>\", \"<strJoin(fieldTypes, ".")>\");\n\n"
	     + "Magnolia::IO::startData(_type_tag) \>\>= _stream;\n"
	     + "<strJoin(outStats, "\nMagnolia::IO::sepData() \>\>= _stream;\n")>\n"
	     + "Magnolia::IO::endData(_type_tag) \>\>= _stream;\n"
	     + "}\n"
	     + "template\<typename _Stream\> void operator\<\<=(_Stream& _stream) {\n"
	     + "Magnolia::IO::Type _type_tag = Magnolia::IO::Type(\"<origName(clsName)>\", \"<strJoin(fieldTypes, ".")>\");\n\n"
	     + "Magnolia::IO::startData(_type_tag) \<\<= _stream;\n"
	     + "<strJoin(inStats, "\nMagnolia::IO::sepData() \<\<= _stream;\n")>\n"
	     + "Magnolia::IO::endData(_type_tag) \<\<= _stream;\n"
	     + "}\n"
	     + "};\n\n"
	     ;
}

public str ppDecl(FunObjClass(n, base, defs)) {
	defs = visit(defs) {
		case FunClause(_, ps, t) => FunClause(Name(leaf("operator()")), ps, t)
		case ProcClause(_, ps) => ProcClause(Name(leaf("operator()")), ps)
	}
	return "struct <ppName(unqualify(n))> {\nconst <ppName(base)> &base;\n<strJoin([ppDecl(d) | d <- defs.args], "\n")>\n};\n";
}

public str ppDecl(Nop()) {
	return ";";
}

public str ppExpr(DotOp(arg0, arg1, _)) {
	return "<ppExpr(arg0)> . <ppName(arg1)>";
}

public str ppExpr(ExternalFun(lang, ns, _, name, ps, rt)) {
	qname = qnameAppend(ns, name);
	return ppName(qname);
}

public str ppExpr(ExternalProc(lang, ns, _, name, ps)) {
	qname = qnameAppend(ns, name);
	return ppName(qname);
}

public str ppExpr(Fun(arg0)) {
	return ppName(arg0);
}

public str ppName_impl(leaf(n)) {
	return n;
}

public str ppName_impl(FunName(arg0, _, _)) {
	return ppName(arg0);
}

public str ppExpr(IfThenElseExpr(arg0, arg1, arg2)) {
	return "((<ppExpr(arg0)>) ? (<ppExpr(arg1)>) : (<ppExpr(arg2)>))";
}

public str ppExpr(Guard(arg0, arg1)) {
	return "((<ppExpr(arg1)>) ? (<ppExpr(arg0)>) : Magnolia::Assert::exprGuardFailed\<decltype(<ppExpr(arg1)>),decltype(<ppExpr(arg0)>)\>(<ppExpr(arg1)>, \"guard failed\", \"\", \"<locOf(arg0)>\"))";
}

public str ppStat(If(c, b1, b2)) {
	return "if (<ppExpr(c)>) <ppStat(Block(b1))> else <ppStat(Block(b2))>";
}

public str ppStat(IfSugar(c, b)) {
	return "if (<ppExpr(c)>) <ppStat(Block(b))>";
}

public str ppStat(Let(arg0, arg1)) {
if(seq(seq(_)) := arg1)
	throw InternalError("XXX: wrong arg in ppStat Let: ", arg1);
	return "<ppStat(arg0)>\n<strJoin([ppStat(s) | s <- arg1.args], "\n")>";
}

public str ppStat(Nop()) {
	return ";";
}

public str ppStat(Return()) {
	return "return;";
}

public str ppStat(Return(arg0)) {
	return "return <ppExpr(arg0)>;";
}

public str ppStat(VarDecl(arg0, arg1, arg2, Undefined())) {
	return "<ppName(arg2)> <ppName(arg1)>;";
}

public str ppStat(VarDecl(arg0, arg1, arg2, arg3)) {
	if(arg0 == Obs())
		return "const <ppName(arg2)> <ppName(arg1)> = <ppExpr(arg3)>;";
	else
		return "<ppName(arg2)> <ppName(arg1)> = <ppExpr(arg3)>;";
}

public str ppStat(While(arg0, arg1)) {
	return "while (<ppExpr(arg0)>) <ppStat(Block(arg1))>";
}

public str ppStat(Guard(arg0, arg1)) {
	return "if(<ppExpr(arg1)>) <ppStat(Block(arg0))>\nelse {\nMagnolia::Assert::guardFailed(\"guard failed\", \"\", \"<locOf(arg0)>\");\n}";
}

public str ppExpr(Literal(arg0)) {
	return ppExpr(arg0);
}

public str ppExpr(DecimalIntegerLiteral(leaf(arg0))) {
	return arg0;
}

public str ppExpr(HexIntegerLiteral(leaf(arg0))) {
	return arg0;
}

public str ppExpr(OctalIntegerLiteral(leaf(arg0))) {
	return arg0;
}

public str ppExpr(Proc(arg0)) {
	return ppName(arg0);
}

public str ppExpr(StringLiteral(leaf(arg0))) {
	return unquote(arg0);
}

public str ppExpr(Struct(arg0, arg1)) {
	return "<ppName(arg0)> { <strJoin(["<ppName(n)> : <ppExpr(e)>" | Field(n, e) <- arg1.args], ", ")> }";
}

public str ppExpr(Equal(Fun(arg0), arg1, arg2)) {
	if(nameToStr(arg0) == "_==_")
		return "Magnolia::Assert::equal(<ppExpr(arg1)>, <ppExpr(arg2)>)";
	else
		return "Magnolia::Assert::notEqual(<ppExpr(arg1)>, <ppExpr(arg2)>)";
}

public str ppExpr(Var(arg0)) {
	return ppName(arg0);
}

public str ppParams(ParamList(arg0)) {
	return "(<strJoin([ppParam(p) | p <- arg0.args], ", ")>)";
}

str ppParam(Param(arg0, arg1, arg2)) {
	if (Obs() := arg0) 
		return "const <ppName(arg2)>& <ppName(arg1)>";
	else
		return "<ppName(arg2)>& <ppName(arg1)>";
}
default str ppParam(AST tree) {
	println("Missing PP Param case: <consOf(tree)>");
	return cxxError("INTERNAL MAGNOLIA COMPILER ERROR: not a C++ Param {<yieldTerm(tree)>}\n");
}

public str ppName_impl(Name(leaf(n))) {
	return n;
}

public str ppName_impl(IdName(leaf(n), leaf(i))) {
	return "<n>_<i>";
}

public str ppName_impl(TName(arg0, seq(args))) {
	if(args == [])
		return ppName(arg0);
	else
		return "<ppName(arg0)>\<<strJoin([ppName(a) | a <- args], ", ")>\>";
}

public str ppName_impl(ProcName(arg0, arg1)) {
	return ppName(arg0);
}

public str ppName_impl(TopName(arg0, arg1)) {
	return ppName(arg0);
}

public str ppName_impl(QName(seq(lst))) {
	if(size(lst) > 0)
		return strJoin([makeCxxNonKeywordName(ppName(n), "_pkg") | n <- lst[..-1]] + ppName(lst[-1]), "::");
	else
		return "";
}

public str ppName_impl(SpecialType(t)) {
	switch(t) {
		case Predicate():
			return "bool";
		case Literal():
			return "int";
		default:
			throw "ppName: Invalid SpecialType in C++ backend: <consOf(t)>"; 
	}
}

public str ppName_impl(Instance(n, bs)) {
	return "<ppName(n)>\<<strJoin([ppName(y) | Binding(x, y) <- bs.args], ", ")>\>";
}

public str ppName_impl(TypeName(arg0)) {
	return ppName(arg0);
}

public str ppName_impl(Unknown()) {
	return "auto";
}

public str ppName_impl(VarName(arg0, arg1)) {
	return ppName(arg0);
}

public default str pp(AST tree) {
	println("Missing PP case: <consOf(tree)>");
	return cxxError("INTERNAL MAGNOLIA COMPILER ERROR: not a C++ AST: {<yieldTerm(tree)>}\n");
}

public default str ppExpr(AST tree) {
	println("Missing PP Expr case: <consOf(tree)>");
	msg = "INTERNAL MAGNOLIA COMPILER ERROR: not a C++ Expr: {<yieldTerm(tree)>}";
	return cxxError(msg, "Magnolia::compile_error(\"<quote(msg)>\")", "0");
}

public default str ppStat(AST tree) {
	println("Missing PP Stat case: <consOf(tree)>");
	msg = "INTERNAL MAGNOLIA COMPILER ERROR: not a C++ Stat: {<yieldTerm(tree)>}";
	return cxxError(msg, "Magnolia::compile_error(\"<quote(msg)>\");", ";");
} 

public default str ppBody(AST tree) {
	println("Missing PP Body case: <consOf(tree)>");
	return cxxError("INTERNAL MAGNOLIA COMPILER ERROR: not a C++ Body: {<yieldTerm(tree)>}\n");
}

public default str ppDecl(AST tree) {
	println("Missing PP Decl case: <consOf(tree)>");
	return cxxError("INTERNAL MAGNOLIA COMPILER ERROR: not a C++ Decl: {<yieldTerm(tree)>}");
}

public default str ppName_impl(AST tree) {
	println("Missing PP Name case: <consOf(tree)>");
	return cxxError("INTERNAL MAGNOLIA COMPILER ERROR: not a C++ Name: {<yieldTerm(tree)>}");
}

public str cxxError(str msg) {
  return "\n#error \"<quote(msg)>\"\n";
}

public str cxxError(str msg, str code, str alt) {
  return "\n#ifdef MAGNOLIA_COMPILE_WITH_ERRORS\n<code>\n#else\n#error <msg>\n<alt>\n#endif\n";
}

public str(AST) ppName = memo(ppName_impl);

public str makeNameSpace(n:Name(_), str body) {
	return "namespace <makeCxxNonKeywordName(n, "_pkg")> {\n<body>\n}\n";
}

public str makeNameSpace(QName(seq([n, *ns])), str body) {
	return "namespace <makeCxxNonKeywordName(n, "_pkg")> {\n<makeNameSpace(QName(seq(ns)), body)>\n}\n";
}

public str makeNameSpace(QName(seq([])), str body) {
	return body;
}
