/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::backend::CxxSkeleton
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::CxxAPI;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import org::magnolialang::magnolia::backend::CxxPP;
import org::magnolialang::magnolia::backend::CxxBackend;
import org::magnolialang::magnolia::backend::OperatorInfo;
import org::magnolialang::magnolia::backend::ProgramBuilder;
import Set;
import Node;
import List;
import Relation;
import String;
import util::Benchmark;

public Result[list[tuple[str,str]]] generateCxxSkeleton(DefInfo info, AST tree, set[AST] externals) {
	Result[list[tuple[str,str]]] result;
	
	env = minimalEnv(info.body@\loc);

	env = startWork("cxxSkeleton", nameToStr(info.qName), env);

	list[tuple[str, str]] bodies = [];
	
	for(External(language, name) <- externals) {
		str body = "";
		fileName = replaceAll(nameToStr(name), ".", "/") + ".hxx";
		bool generated = false;
		// library
		if(DeclBody(seq(ds)) := tree) {
			<_, exts> = findCxxIncludes(tree);
			
			body = generateSkeleton(ds, name, info.qName);
			generated = true;
		}
		else {
			println("Don\'t know how to deal with <consOf(tree)>");
			body = "  /* Don\'t know how to deal with <consOf(tree)> */";
		}
		symbol = toUpperCase("_INCLUDE_<replaceAll("<nameToStr(name)>.hxx", ".", "_")>");
		body = indentStr(makeNameSpace(qualifierOf(name), body));
		body = "// <info.kind>: <dispNameOf(info.qName)>, file: \"<fileName>\" \n"
			 + "#ifndef <symbol>\n"
			 + "#define <symbol>\n"
			 + "<body>\n"
			 + "#endif";
		bodies += <body, fileName>;
	}
	
	if(bodies != []) {
		result = makeResult(bodies, env);
	}
	else {
		result = makeNoResult("not external; no code generated", env);
	}
		
	env = endWork(env);
	
	return result;
}

str generateSkeleton(list[AST] defs, AST fileName, AST parent) {
	list[AST] entries = [];
	map[AST,AST] renaming = ();
	set[AST] onParts = {};
	for(d:Define(dcl,seq(attrs),body) <- defs) {
		suf = declSuffix(dcl);

		if(Require() in attrs) { // it's a parameter
			;
		}
		else { // it's defined here
			switch(body) {
				case Body(ExternalType(lang, fileName, onPart, name)): {
					extName = makeExtCxxName(nameToStr(name), suf);
					entries += Define(TypeClause(Name(leaf(extName))),
						seq([Comment(leaf(dispNameOf(unprefix(defNameOf(dcl), parent)))), Member()]), 
						Body(Struct(seq([Comment(leaf("TODO: auto-generated type stub"))]))));
					renaming += (defNameOf(dcl) : Name(leaf(extName)));
					onParts += onPart;
				}
				case Body(Apply(ExternalFun(lang, fileName, onPart, name, ps, rt), args)): {
					extName = makeExtCxxName(nameToStr(name), suf);
					// return false if predicate, otherwise call default constructor
					ret = (SpecialType(Predicate()) := rt) ? Var(VarName(leaf("false"), rt)) : Apply(Fun(rt), seq([])); 
					entries += Define(FunClause(Name(leaf(extName)), ps, rt),
						seq([Comment(leaf(dispNameOf(unprefix(defNameOf(dcl), parent)))), Member()]), 
						Body(BlockExpr(seq([Comment(leaf("TODO: auto-generated function stub")),
						                    Return(ret)]), rt)));
					renaming += (defNameOf(dcl) : Name(leaf(extName)));
					onParts += onPart;
				}
				case Body(Block(seq([Call(ExternalProc(lang, fileName, onPart, name, ps), args)]))): {
					extName = makeExtCxxName(nameToStr(name), suf);
					entries += Define(ProcClause(Name(leaf(extName)), ps),
						seq([Comment(leaf(dispNameOf(unprefix(defNameOf(dcl), parent)))), Member()]), 
						Body(Block(seq([Comment(leaf("TODO: auto-generated function stub"))]))));
					renaming += (defNameOf(dcl) : Name(leaf(extName)));
					onParts += onPart;
				}
			}
		}
	}
	entries = sortDefs(entries);

	str outerTemplate = "";
	str innerTemplate = "";
	str body = "";
	if(ExternalParams(seq(onTypes), seq(onOps)) := getOneFrom(onParts)) {
		if(onTypes != []) {
			list[str] tmplArgs = [];
			for(t <- onTypes) {
				n = makeExtCxxName(ppName(t), "_t");
				renaming += (t : Name(leaf(n)));
				tmplArgs +=  "typename <n>";
			}
			outerTemplate = "template\<<strJoin([a | a <- tmplArgs], ", ")>\>\n";
		}
		if(onOps != []) {
			set[str] usedNames = {};
			list[str] opTypeNames = [];
			list[str] opFieldDecls = [];
			list[str] opConsArgs = [];
			list[str] opConsInits = [];
			
			for(op <- onOps) {
				<N,usedNames> = makeUnique(toUpperCase(ppName(op) + "_OP"), usedNames);
				<n,usedNames> = makeUnique(makeExtCxxName(ppName(op), "_op"), usedNames);
				renaming += (op : Name(leaf(n)));
				opTypeNames += N;
				opConsArgs += "const <N>& _<n>";
				opConsInits += "<n>(_<n>)";
				opFieldDecls += "/** require <dispNameOf(unprefix(op, parent))> */\nconst <N>& <n>;\n";
			}
			
			innerTemplate = "template\<<strJoin(["typename <n>" | n <- opTypeNames], ", ")>\>\n";

			body = strJoin(opFieldDecls, "\n");
			body += "\n";
			body += "<ppName(unqualify(fileName))>_impl(<strJoin(opConsArgs, ", ")>)\n\t: <strJoin(opConsInits, ", ")>\n\t{};\n\n\n";
		}
	}
	else {	
		termPrintln("INTERNAL COMPILER ERROR: CxxSkeleton::generateSkeleton: onParts should be a singleton");
	}
	
	body += strJoin([ppDecl(trToCxx(e, renaming)) | e <- entries], "\n\n");
	body = "<innerTemplate>struct <ppName(unqualify(fileName))>_impl {\n<body>\n};\n\n";
	body = "<outerTemplate>struct <ppName(unqualify(fileName))> {\n<body>\n};\n\n";
	return body;
}

private str makeExtCxxName(str name, str suf) {
	if(name in opMapCxx) {
		info = opMapCxx[name];
		if(info.typ == Infix() || info.typ == Prefix())
			return "operator<info.opSym>";
		else if(info.typ == Funfix())
			return info.opName;
		else
			return "operator_<info.opName>";
	}
		
	if(name in reservedKeywords)
		name = "<name><suf>";
	return name;
}
