package org.magnolialang.magnolia.backend;

import static nuthatch.rascal.pattern.StaticValuesPatternFactory.*;
import static org.magnolialang.magnolia.AbstractSyntax.*;
import static org.magnolialang.magnolia.grammar.MagnoliaPatterns.*;
import static org.nuthatchery.pica.terms.TermFactory.vf;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;

import nuthatch.library.ActionBuilder;
import nuthatch.library.ActionFactory;
import nuthatch.library.BaseMatchAction;
import nuthatch.library.MatchAction;
import nuthatch.library.Walk;
import nuthatch.pattern.Environment;
import nuthatch.rascal.adapter.ValuesCursor;
import nuthatch.rascal.adapter.ValuesUtil;
import nuthatch.rascal.adapter.ValuesWalker;
import nuthatch.rascal.pattern.ValuesBuildContext;

import org.rascalmpl.value.IBool;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.IListWriter;
import org.rascalmpl.value.IMap;
import org.rascalmpl.value.IMapWriter;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.type.Type;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.frontend.MagnoliaLoader;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.nuthatchery.pica.util.NullnessHelper;
import org.nuthatchery.pica.util.Pair;

public class Instantiator {
	protected final static String MARK_SOURCE = MagnoliaLoader.class.getName();
	protected final static String MARK_SOURCE_GET_PKG_INFO = MARK_SOURCE + "#instantiate";
	protected final static String MARK_SOURCE_GET_PKG_CONTENTS = MARK_SOURCE + "#instantiate";
	private static final ValuesBuildContext context = new ValuesBuildContext(vf, TermFactory.ts);
	private static final ValuesUtil valuesUtil = new ValuesUtil(vf, TermFactory.ts);
	private static final ActionFactory<IValue, Type, ValuesCursor, ValuesWalker> af = ValuesUtil.getActionFactory();
	protected static final IValue UNKNOWN = TermFactory.cons("Unknown");
	private final boolean DEBUG_PRINT = MagnoliaPlugin.getInstance().isInSecondLevel();

	private final ICompiler compiler = Magnolia.getInstance().getCompiler();

	protected IMap instanceMap = NullnessHelper.assertNonNull(vf.mapWriter().done());
	protected IMap workingMap = NullnessHelper.assertNonNull(vf.mapWriter().done());
	protected IMap placeholderMap = NullnessHelper.assertNonNull(vf.mapWriter().done());
	protected IMapWriter finalNames = NullnessHelper.assertNonNull(vf.mapWriter());
	protected ISet instancesTodo = NullnessHelper.assertNonNull(vf.setWriter().done());
	protected ISet usedPlaceholders = NullnessHelper.assertNonNull(vf.setWriter().done());
	protected IListWriter resultWriter = NullnessHelper.assertNonNull(vf.listWriter());
	protected final Walk<ValuesWalker> collectWalk;
	protected final IConstructor moduleTree;
	protected IMap defEnv = NullnessHelper.assertNonNull(vf.mapWriter().done());
	protected int counter = 0;
	protected int nesting = 0;
	protected PrintStream log = System.out;


	final MatchAction<IValue, Type, ValuesCursor, ValuesWalker> skipExternalChildren = af.action(af.redirect(1, 3, 2, 3, 4, 5));


	private Instantiator(IConstructor moduleTree) {
		this.collectWalk = collectWalk();
		this.moduleTree = moduleTree;
	}


	public IConstructor instantiate() {
		long t = System.currentTimeMillis();
		if(TermAdapter.isCons(moduleTree, "DeclBody", 1)) {
			try {
				log = new PrintStream(new FileOutputStream("/tmp/instantiate.log"));
			}
			catch(FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			// pre-declare everything
			IMapWriter envWriter = vf.mapWriter();
			for(IConstructor def : TermAdapter.getChildren(TermAdapter.getArg(moduleTree, 0))) {
				assert def != null && TermAdapter.isCons(def, "Define", 3);

				IConstructor defName = (IConstructor) compiler.call("defNameOf", def);

				if(DEBUG_PRINT) {
					log.println(prefix() + "Defined: " + TermAdapter.yieldTerm(defName) + " => " + TermAdapter.yieldTerm(TermAdapter.getArg(def, 2)));
				}
				envWriter.put(defName, def);
			}
			defEnv = NullnessHelper.assertNonNull(envWriter.done());
			log.println("\n" + prefix() + "************** PREDECLARE DONE\n");

			// collect from and redeclare API declarations
			for(IConstructor def : TermAdapter.getChildren(TermAdapter.getArg(moduleTree, 0))) {
				assert def != null && TermAdapter.isCons(def, "Define", 3);

				if(!((IBool) compiler.call("isConcrete", def)).getValue()) {
					def = collectInstances(def);
					resultWriter.append(def);
					IConstructor defName = (IConstructor) compiler.call("defNameOf", def);

					if(DEBUG_PRINT) {
						log.println(prefix() + "Defined: " + TermAdapter.yieldTerm(defName) + " => " + TermAdapter.yieldTerm(TermAdapter.getArg(def, 2)));
					}
					defEnv = NullnessHelper.assertNonNull(defEnv.put(defName, def));
				}
			}

			log.println("\n" + prefix() + "************** COLLECT/REDECLARE DONE\n");

			int c = 0, d = 0;
			ISet done = vf.setWriter().done();
			while(!instancesTodo.isEmpty()) {
				IConstructor i = (IConstructor) instancesTodo.iterator().next();
				instancesTodo = NullnessHelper.assertNonNull(instancesTodo.delete(i));
				assert i != null;

				IConstructor inst = placeholderMap.isEmpty() ? i : (IConstructor) valuesUtil.transform(i, af.walk(af.down(valuesUtil.renameAction(placeholderMap))));

				if(!done.contains(inst)) {
					done = done.insert(inst);

					IConstructor name = (IConstructor) instanceMap.get(inst);
					assert name != null;
					log.println(prefix() + "Instantiate: " + TermAdapter.yieldTerm(name));
					if(DEBUG_PRINT) {
						log.println(prefix() + "\nMake Instance: " + TermAdapter.yieldTerm(inst));
						if(!i.isEqual(inst)) {
							log.println(prefix() + "\nremapped from: " + TermAdapter.yieldTerm(i));

						}
						log.println(prefix() + "         id: " + TermAdapter.yieldTerm(name));
					}
					IConstructor instance = makeInstance(inst, name);
					if(DEBUG_PRINT) {
						log.println(prefix() + "       => " + TermAdapter.yieldTerm(instance));
					}
					resultWriter.append(instance);
					c++;
				}
				else {
					d++;
				}
			}

			IConstructor result = TermFactory.seq(resultWriter.done());
			IMap names = finalNames.done();
			System.out.println("Map of final names: ");
			log.println("Map of final names: ");
			for(IValue k : names) {
				System.out.println("  " + TermAdapter.yieldTerm(k) + " => " + TermAdapter.yieldTerm(names.get(k)));
				log.println("  " + TermAdapter.yieldTerm(k) + " => " + TermAdapter.yieldTerm(names.get(k)));
			}
			result = (IConstructor) valuesUtil.transform(result, af.walk(af.down(valuesUtil.renameAction(names))));

			t = System.currentTimeMillis() - t;
			System.out.println("Instantiation complete, with " + c + " instances, " + d + " duplicates ignored, in " + t + "ms");
			log.println("Instantiation complete, with " + c + " instances, " + d + " duplicates ignored, in " + t + "ms");

			log.close();
			log = System.out;

			return moduleTree.set(0, result);
		}
		else {
			return moduleTree;
		}
	}


	private IConstructor collectInstances(IConstructor defTree) {
		ValuesWalker walker = new ValuesWalker(defTree, collectWalk);
		walker.start();
		return NullnessHelper.assertNonNull((IConstructor) walker.getData());
	}


	private Walk<ValuesWalker> collectWalk() {
		ActionBuilder<ValuesWalker> actions = af.combineBuilder();

		actions.add(af.down(af.match(Binding(_, _), af.action(af.go(2)))));
		actions.add(af.down(af.match(Instance(_, _), af.action(af.go(2)))));
		actions.add(af.match(ExternalFun(_, _, _, _, _, _), skipExternalChildren));
		actions.add(af.match(ExternalProc(_, _, _, _, _), skipExternalChildren));
		actions.add(af.match(ExternalType(_, _, _, _), skipExternalChildren));
		actions.add(af.up(af.match(Instance(var("name"), var("bindings")), new BaseMatchAction<IValue, Type, ValuesCursor, ValuesWalker>() {

			@Override
			@NonNullByDefault(false)
			public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
				nesting++;

				IConstructor origInstance = (IConstructor) walker.getData();


				if(DEBUG_PRINT) {
					log.println(prefix() + "Looking at instance: " + TermAdapter.yieldTerm(origInstance));
				}


				if(instanceMap.containsKey(origInstance)) {
					if(DEBUG_PRINT) {
						log.println(prefix() + "        found  => " + TermAdapter.yieldTerm(instanceMap.get(origInstance)));
					}
					walker.replace(instanceMap.get(origInstance));
				}
				else if(workingMap.containsKey(origInstance)) {
					IValue ph = workingMap.get(origInstance);
					walker.replace(ph);
					usedPlaceholders = usedPlaceholders.insert(ph);
					if(DEBUG_PRINT) {
						log.println(prefix() + "  working map  => " + TermAdapter.yieldTerm(origInstance) + " => " + TermAdapter.yieldTerm(ph));
					}
				}
				else {
					IConstructor origName = (IConstructor) env.get("name").getData();
					IConstructor bindings = (IConstructor) env.get("bindings").getData();
					IMap workingMapTmp = workingMap;
					IConstructor ph = TermFactory.cons("InstancePlaceholder", TermFactory.leaf("" + counter++));
					workingMap = workingMap.put(origInstance, ph); // TODO: we should check whether it turns out we've already seen this instance, in which case we should patch the original name back over any occurences of the newly generated name

					IConstructor instance = substituteBindings(origInstance, TermAdapter.getChildrenList(bindings));
					IConstructor newBindings = TermFactory.seq(expandBindings(TermAdapter.getChildrenList(bindings)).getSecond());
					if(DEBUG_PRINT) {
						log.println(prefix() + "Substitute bindings: " + TermAdapter.yieldTerm(instance));
						log.println(prefix() + " origBindings==bindings? " + bindings.isEqual(instance.get(1)));
						log.println(prefix() + " origBindings==newBindings? " + bindings.isEqual(newBindings));
						log.println(prefix() + " newBindings==bindings? " + newBindings.isEqual(instance.get(1)));
						log.println(prefix() + " origBindings=" + TermAdapter.yieldTerm(bindings));
						log.println(prefix() + " substBindings=" + TermAdapter.yieldTerm(instance.get(1)));
						log.println(prefix() + " newBindings=" + TermAdapter.yieldTerm(newBindings));
					}
					instance = origInstance.set(1, newBindings);

					workingMap = workingMapTmp;
					assert !valuesUtil.contains(instance, ph);

					if(instanceMap.containsKey(instance)) { // this instance is already known
						IConstructor knownName = (IConstructor) instanceMap.get(instance);
						if(DEBUG_PRINT) {
							log.println(prefix() + "..Found instance: " + TermAdapter.yieldTerm(knownName));
							log.println(prefix() + "               => " + TermAdapter.yieldTerm(knownName));
						}
						if(usedPlaceholders.contains(ph)) {
							// has been used, but was already known under another name
							placeholderMap = placeholderMap.put(ph, knownName);
							log.println(prefix() + "....replacing placeholder: " + TermAdapter.yieldTerm(ph) + " => " + TermAdapter.yieldTerm(knownName));
						}
						walker.replace(instanceMap.get(instance));
					}
					else if(TermAdapter.getChildren(bindings).iterator().hasNext()) { // new instance, with bindings
						if(DEBUG_PRINT) {
							log.println(prefix() + "..New instance:   " + TermAdapter.yieldTerm(instance));
							log.println(prefix() + "               => " + TermAdapter.yieldTerm(ph));
						}
						instanceMap = instanceMap.put(instance, ph);
						instancesTodo = instancesTodo.insert(instance);
						walker.replace(ph);
					}
					else { // new instance, without bindings
						if(DEBUG_PRINT) {
							log.println(prefix() + "..Direct instance:" + TermAdapter.yieldTerm(instance));
							log.println(prefix() + "               => " + TermAdapter.yieldTerm(ph));
						}
						instanceMap = instanceMap.put(instance, ph);
						instancesTodo = instancesTodo.insert(instance);
						walker.replace(ph);
					}

//					throw new RuntimeException();
				}

				nesting--;
				return PROCEED;
			}
		})));

		return af.walk(actions.done());
	}


	private Pair<IMap, IList> expandBindings(IList bindings) {
		IMapWriter renaming = vf.mapWriter();
		IListWriter newBindings = vf.listWriter();
		for(IValue v : bindings) {
			IConstructor b = (IConstructor) v;
			assert b != null && TermAdapter.isCons(b, "Binding", 2);
			IConstructor from = TermAdapter.getArg(b, 0);
			IConstructor to = TermAdapter.getArg(b, 1);
			IConstructor orig = to;
			while(defEnv.containsKey(to)) {
				IConstructor body = TermAdapter.getArg((IConstructor) defEnv.get(to), 2);
//			if(DEBUG_PRINT) {
//				log.println(prefix() + "  BINDING: " + TermAdapter.yieldTerm(from) + " => " + TermAdapter.yieldTerm(body));
//			}
				if(((IBool) compiler.call("isName", body)).getValue() || TermAdapter.isCons(body, "InstancePlaceholder")) {
					if(DEBUG_PRINT) {
						log.println(prefix() + "  Binding: [following " + TermAdapter.yieldTerm(to));
						log.println(prefix() + "  Binding:         =] " + TermAdapter.yieldTerm(body));
					}
					if(!to.isEqual(body)) {
						renaming.put(to, body);
					}
					to = body;
				}
				else if(((IBool) compiler.call("isInstance", body)).getValue()) {
					if(DEBUG_PRINT) {
						log.println(prefix() + "  Binding: [following " + TermAdapter.yieldTerm(to));
						log.println(prefix() + "  Binding:         =] " + TermAdapter.yieldTerm(body));
					}
					body = collectInstances(body);
					if(DEBUG_PRINT) {
						log.println(prefix() + "  Binding: [instantiated " + TermAdapter.yieldTerm(to));
						log.println(prefix() + "  Binding:            =] " + TermAdapter.yieldTerm(body));
					}
					if(!to.isEqual(body)) {
						renaming.put(to, body);
					}
					to = body;
				}
				else {
					break;
				}
			}


			if(from.isEqual(to)) {
				if(DEBUG_PRINT) {
					log.println(prefix() + "  Binding: " + TermAdapter.yieldTerm(from) + " => " + TermAdapter.yieldTerm(orig));
					log.println(prefix() + "        == " + TermAdapter.yieldTerm(to));
				}
			}
			else {
				if(DEBUG_PRINT) {
					log.println(prefix() + "  Binding: " + TermAdapter.yieldTerm(from) + " => " + TermAdapter.yieldTerm(orig));
					log.println(prefix() + "        => " + TermAdapter.yieldTerm(to));
				}
				renaming.put(from, to);
				newBindings.append(b.set(0, from).set(1, to));
			}
		}
		return new Pair<>(renaming.done(), newBindings.done());
	}


	@Nullable
	private IConstructor lookup(IConstructor name, IMap env) {
		return (IConstructor) env.get(name);
	}


	private IConstructor makeInstance(IConstructor in, IConstructor id) {
		assert TermAdapter.isCons(in, "Instance", 2);
		IConstructor implName = TermAdapter.getArg(in, 0); // (IConstructor) compiler.call("basename", TermAdapter.getArg(in, 0));
		assert TermAdapter.isCons(implName);

		IConstructor bindings = TermAdapter.getArg(in, 1);
		assert TermAdapter.isSeq(bindings);


		IConstructor impl = lookup(implName, defEnv);
		if(impl == null) {
			log.println(prefix() + "Can't find " + TermAdapter.yieldTerm(implName));
		}

		if(DEBUG_PRINT) {
			log.println(prefix() + "  Original name: " + TermAdapter.yieldTerm(implName));
		}

		IConstructor name = makeInstanceName(collectInstances(substituteBindings(implName, TermAdapter.getChildrenList(bindings))));
		if(DEBUG_PRINT) {
			log.println(prefix() + "  Instance name: " + TermAdapter.yieldTerm(name));
		}
		finalNames.put(id, name);

		assert impl != null;

		assert TermAdapter.isCons(impl, "Define", 3);

		IConstructor renamedImpl = (IConstructor) compiler.call("renameDef", impl, name);

		if(DEBUG_PRINT) {
			log.println(prefix() + "  Original: " + TermAdapter.yieldTerm(impl));
		}


		IConstructor result = substituteBindings(renamedImpl, TermAdapter.getChildrenList(bindings));
		result = collectInstances(result);
		// TODO: remove Concrete(_) attribute
		assert result != null;
		return result;
	}


	private IConstructor makeInstanceName(IConstructor name) {
		name = (IConstructor) compiler.call("makeId", name);
		return name;
	}


	private String prefix() {
		StringBuilder b = new StringBuilder();
		for(int i = 0; i < nesting; i++) {
			b.append("  ");
		}
		return b.toString();
	}


	private IConstructor substituteBindings(IConstructor tree, IList bindings) {
		IMap ren = expandBindings(bindings).getFirst();
		if(!ren.isEmpty()) {
			ActionBuilder<ValuesWalker> actions = af.combineBuilder();
			ActionBuilder<ValuesWalker> skipActions = af.combineBuilder();

			if(DEBUG_PRINT) {
				log.println(prefix() + "  Renaming:");
				for(IValue n : ren) {
					log.println(prefix() + "    " + TermAdapter.yieldTerm(n) + " => " + TermAdapter.yieldTerm(ren.get(n)));
				}

				log.println(prefix() + "  In tree: " + TermAdapter.yieldTerm(tree));
			}
			actions.add(af.down(valuesUtil.renameAction(ren))); // always do this when first entering node
			actions.add(af.down(af.match(Binding(_, _), af.action(af.go(2)))));
			actions.add(af.down(af.match(Instance(_, _), af.action(af.go(2)))));
			actions.add(af.match(ExternalFun(_, _, _, _, _, _), skipExternalChildren));
			actions.add(af.match(ExternalProc(_, _, _, _, _), skipExternalChildren));
			actions.add(af.match(ExternalType(_, _, _, _), skipExternalChildren));
			skipActions.add(af.down(af.match(FunName(_, _, _), af.action(af.go(0)))));
			skipActions.add(af.down(af.match(TypeName(_), af.action(af.go(0)))));
			skipActions.add(af.down(af.match(ProcName(_, _), af.action(af.go(0)))));
			Walk<ValuesWalker> walk = af.walk(af.seq(actions.done(), skipActions.done()));
			ValuesWalker walker = new ValuesWalker(tree, walk);
			walker.start(1);
			IConstructor result = (IConstructor) walker.getData();
			if(DEBUG_PRINT) {
				log.println(prefix() + "  To: " + TermAdapter.yieldTerm(result));
			}
			assert result != null;
			return result;
		}
		else {
			return tree;
		}

	}


	public static IConstructor instantiate(IConstructor tree) {
		return new Instantiator(tree).instantiate();
	}
}
