/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::backend::Mutifier
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::TypeChecker;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import String;
import List;
import Exception;
import util::Benchmark;
import IO;

public Result[AST] getMutified(DefInfo info) {
	env = newEnv(info.qName, info.parent, info.body@\loc);
	switch(info.body) {
		case DeclBody(seq(ds)): {
			env = loadDefs(info.body, true, env);
			ds = for(def <- ds) {
				<def, env> = mutifyOperation(def, env);
				append def;
			}
			return makeResult(preserveAnnos(DeclBody(seq(ds)), info.body), env);
			}
		default: {
			println("Can\'t mutify <consOf(info.body)>");
			return makeResult(info.body, env);
		}
	}
}
public default tuple[AST,Env] mutifyOperation(AST def, Env env) {
	env = errorMark(env, "INTERNAL COMPILER ERROR: mutifyOperation: Unknown definition kind <consOf(def)>", Internal(), locOf(def));
	return <def, env>;
}

public tuple[AST,Env] mutifyOperation(def:Define(dcl,attrs,body), Env env) {
	tmpCounter = 0;
	if(!isOperation(def))
		return <def, env>;
	try {
	switch(body) {
		case Body(b:Block(ss)): {
			<ss, env> = mutify(ss, env);
			return <def[arg2 = Body(b[arg0=ss])], env>;
		}
		case Body(b:BlockExpr(ss, rt)): {
			<ss, env> = mutify(ss, env);
			return <def[arg2 = Body(b[arg0=ss])], env>;
		}
		case EmptyBody():
			return <def[arg2 = body], env>;
		case Body(b:Apply(FunOf(ProcName(p, seq(ps)), leaf(_argn)), _)):
			return makeStub(def, env);
		case Instance(_,_):
			return <def[arg1 = attrs], env>;
		case FunName(_,_,_):
			return <def[arg1 = attrs], env>;
		case ProcName(_,_):
			return <def[arg1 = attrs], env>;
		case Body(e): {
			<b, env> = mutify(seq([Return(e)]), env);
			return <def[arg2 = Body(Block(b))], env>;
		}
		default: {
			env = errorMark(env, "INTERNAL COMPILER ERROR: mutifyOperation(): Unknown operation body cons <consOf(body)>", locOf(dcl));
			return <def, env>;
		}
	};
	}
	catch e:InternalError(msg, l): {
		env = errorMark(env, "While processing <consOf(body)>: <msg>", locOf(l));
	}
	catch e: {
	println(e);
		env = errorMark(env, "<e>", locOf(body));
	}
	return <def, env>;
	
}

private tuple[AST, Env] makeStub(def:Define(FunClause(n,ps,t),_,_), Env env) {
	//if(hasModifier(def, Derived()))
	//	return <Nop(), env>;
	//else {
	// def = changeModifiers(def, {Derived()}, {});
	<result, env> = mutify(Return(Apply(Fun(n), seq(paramsToArgs(ps)))), env);
	return <def[arg2=Body(BlockExpr(seq([result]), t))], env>;
	//}
}

public tuple[AST, Env] mutify(seq(todo), Env env) {
	todo = for(t <- todo) { <t, env> = mutify(t, env); append t; }
	return <seq(todo), env>;
}

public tuple[AST, Env] mutify(stat:If(_,_,_), Env env) {
	return mutifySubExpr(stat, 0, [1,2], env);
}

public tuple[AST, Env] mutify(stat:Guard(_,_), Env env) {
	return mutifySubExpr(stat, 1, [0], env);
}

public tuple[AST, Env] mutify(stat:While(_,_), Env env) {
	return mutifySubExpr(stat, 0, [1], env);
}

public tuple[AST, Env] mutify(stat:For(_,_,_), Env env) {
	return mutifySubExpr(stat, 1, [2], env);
}

public tuple[AST, Env] mutify(stat:Assert(_,_), Env env) {
	return mutifySubExpr(stat, 0, [], env);
}

public tuple[AST, Env] mutify(stat:Assume(_,_), Env env) {
	return mutifySubExpr(stat, 0, [], env);
}

public tuple[AST, Env] mutify(stat:Yield(_), Env env) {
	return mutifySubExpr(stat, 0, [], env);
}

public tuple[AST, Env] mutify(stat:Return(_), Env env) {
	return mutifySubExpr(stat, 0, [], env);
}

public tuple[AST, Env] mutify(stat:Block(ss), Env env) {
	<ss, env> = mutify(ss, env);
	return <stat[arg0 = ss], env>;
}

public tuple[AST, Env] mutify(stat:Call(p,seq(as)), Env env) {
	list[AST] dcls = [];
	
	as = for(a <- as) {
		if(Apply(_,_) := a) {
			n = mkTmpName("callTmp");
			t = getExprType(a);
			dcls += VarDecl(Obs(), n, t, a);
			append Var(VarName(n, t));
		}
		else {
			append a;
		}
	}

	if(dcls == [])	
		return <preserveAnnos(Call(p, seq(as)), stat), env>;
	else
		return mutify(LetMulti(dcls, seq([preserveAnnos(Call(p, seq(as)), stat)])), env);
}

public tuple[AST, Env] mutifySubExpr(AST tree, int childIndex, list[int] recurse, Env env) {
	<e, dcls, env> = varOnlyExpr(astValue(tree[childIndex]), env);
	tree[childIndex] = e;
	for(i <- recurse) {
		<tree[i], env> = mutify(astValue(tree[i]), env);
	}
	if(dcls == [])
		return <tree, env>;
	else
		return mutify(LetMulti(dcls, seq([tree])), env);
}

public tuple[AST,list[AST], Env] varOnlyExpr(AST expr, Env env) {
	<expr, dcls, env> = asgnExpr(expr, env);
	
	if(Var(_) := expr) {
		return <expr, dcls, env>;
	}
	else {
		n = mkTmpName("asgn");
		t = getExprType(expr);
		return <Var(VarName(n, t)), [*dcls, VarDecl(Obs(), n, t, expr)], env>;
	}
}

// asgnExpr -- split a nested expression up into a sequence of assignments, keeping
// an outer Apply, if any
public tuple[AST,list[AST],Env] asgnExpr(AST expr, Env env) {
	if(Apply(fun, as) := expr) {
		<as, dcls, env> = asgnExpr2(as, env);
		return <preserveAnnos(Apply(fun, as), expr), dcls, env>;	
	}
	else {
		return asgnExpr2(expr, env);
	}
}

// asgnExpr2 -- lift Apply()s out of an expression and into a sequence of assignments
private tuple[AST,list[AST],Env] asgnExpr2(AST expr, Env env) {
	list[AST] dcls = [];
	expr = top-down-break visit(expr) {
		case x:Apply(_,_): {
			n = mkTmpName("applyTmp");
			t = getExprType(x);
			dcls += VarDecl(Obs(), n, t, x);
			insert preserveAnnos(Var(VarName(n, t)), x);
		}
		case x:IfThenElseExpr(_, _, _): {
			n = mkTmpName("applyTmp");
			t = getExprType(x);
			dcls += VarDecl(Obs(), n, t, x);
			insert preserveAnnos(Var(VarName(n, t)), x);
		}
		case x:BlockExpr(_,_): {
			n = mkTmpName("applyTmp");
			t = getExprType(x);
			dcls += VarDecl(Obs(), n, t, x);
			insert preserveAnnos(Var(VarName(n, t)), x);
		}
/*		case x:Guard(_,_): {
			n = mkTmpName("applyTmp");
			t = getExprType(x);
			dcls += VarDecl(Obs(), n, t, x);
			insert preserveAnnos(Var(VarName(n, t)), x);
		}*/
	}
	return <expr, dcls, env>;
}



int tmpCounter = 0;

public AST mkTmpName(str prefix) {
	name = Name(leaf("<prefix><tmpCounter>"));
	tmpCounter += 1;
	return name;
}

public AST mkTmpName() {
	return mkTmpName("tmp");
}

tuple[AST,Env] mutify(stat:Assign(v, e), Env env) {
	<m, env> = doMutify(e, v, [], false, env);

	if(m == Nop()) {
		if(Var(_) := e) {
			return <Assign(v, e), env>;
		}
		else if(IfThenElseExpr(cnd, thn, els) := e) {
			return mutify(If(cnd, seq([Assign(v, thn)]), seq([Assign(v, els)])), env);
		}
		else if(BlockExpr(ss, rt) := e) {
			ss = returnToAssign(ss, v);
			return mutify(Block(ss), env);
		}
		else {
			<e, dcls, env> = asgnExpr(e, env);
			if(dcls == [])
				return <preserveAnnos(Assign(v, e), stat), env>;
			else
				return mutify(LetMulti(dcls, seq([preserveAnnos(Assign(v, e), stat)])), env);
		}
	}
	else
		return <m, env>;
}

tuple[AST,Env] mutify(tree:Let(decl:VarDecl(mode,n,t,e), seq(stats)), Env env) {
	AST m;
	v = VarName(n, t);
	<m, env> = doMutify(e, Var(v), stats, true, env);
 	
 	if(m == Nop()) {
 		if(IfThenElseExpr(cnd, thn, els) := e) {
 			return mutify(Let(VarDecl(mode,n,t,Undefined()), 
 			                  seq([If(cnd, seq([Assign(Var(v), thn)]), seq([Assign(Var(v), els)])), *stats])), env); 			
 		}
		else if(BlockExpr(seq(ss), rt) := e) {
			seq(ss) = returnToAssign(seq(ss), v);
			return mutify(Let(VarDecl(mode,n,t,Undefined()),
							 seq([*ss, *stats])), env);
		}
		else {
			<e, preDcls, env> = asgnExpr(e, env);
			if(preDcls != [])
				return mutify(LetMulti([*preDcls, VarDecl(mode,n,t,e)], seq(stats)), env);
			else {
				<x, env> = mutify(seq(stats), env);
				return <Let(VarDecl(mode,n,t,e), x), env>;
			}
		}
	}
	else {
		return <m, env>;
	}
}

private tuple[AST,Env] doMutify(AST expr, AST v, list[AST] cont, bool isLet, Env env) {
	if(Apply(Fun(f:FunName(_,_,_)), seq(args)) := expr) {
		DefInfo info;
		try {
			info = lookupDef(f, env);
			if(isName(info.body)) {
				termPrintln("Followed link from: ", info.body);
				info = lookupDef(info.body, env);
				termPrintln("                to: ", info.qName);
			}
		}
		catch NoSuchKey(_):
			return <Nop(), env>;

		if(Body(Apply(fo:FunOf(proc:ProcName(_, seq(ps)), leaf(_argn)), _)) := info.body) {
			expr = preserveAnnos(Apply(fo, seq(args)), expr);
		}
	}
	
	if(Apply(FunOf(proc:ProcName(_, seq(ps)), leaf(_argn)), seq(args)) := expr) {
		argn = toInt(_argn);
		list[AST] inits = [];
		list[AST] asgns = [];
		list[AST] pargs = [];

		for(<argNum, param:AnonParam(m,t)> <- zip(index(ps),ps)) {
			if(m == Obs()) {
				if(argNum == argn)
					env = errorMark(env, "Illegal functionalisation of procedure <nameToStr(p)> for obs param <argNum>", locOf(param));
				if(a:Apply(_,_) := head(args)) {
					tmpName = mkTmpName();
					inits += VarDecl(Obs(), tmpName, t, a);
					pargs += Var(VarName(tmpName, t));
				}
				else {
					pargs += head(args);
				}
				args = tail(args);
			}
			else if(m == Upd()) {
				if(argNum == argn) {
					if(isLet) // declare the variable we're declaring
						inits += VarDecl(Upd(), basename(defNameOf(v)), getExprType(v), head(args));
					else if(v != head(args)) // load arg value if the var is different from the update arg
						asgns += Assign(v, head(args));
					pargs += v;
				}
				else { // make temp to hold (and discard) the update result
					tmpName = mkTmpName();
					inits += VarDecl(Upd(), tmpName, t, head(args));
					pargs += Var(VarName(tmpName, t));
				}
				args = tail(args);
			}
			else if(m == Out()) {
				if(argNum == argn) {
					if(isLet)  // declare the variable we're declaring
						inits += VarDecl(Upd(), basename(defNameOf(v)), getExprType(v), Undefined());
					pargs += v;
				}
				else { // make temp to hold (and discard) the output result
					tmpName = mkTmpName();
					inits += VarDecl(Upd(), tmpName, t, Undefined());
					pargs += Var(VarName(tmpName, t));
				}
			}
			else {
				inits += leaf("Unimplemented parameter mode <m>");
			}
		}
		termPrintln("asgns: ", asgns);
		termPrintln("inits: ", inits);
		return mutify(LetMulti(inits, seq([*asgns, preserveAnnos(Call(Proc(proc), seq(pargs)), expr), *cont])), env);
	}
	return <Nop(), env>;
}

default tuple[AST,Env] mutify(AST tree, Env env) {
	termPrintln("mutify default: ", tree);
	return <tree, env>;
}

AST returnToAssign(AST tree, AST var) {
	tree = top-down-break visit(tree) {
		case Return(e) => Assign(var, e)
		case BlockExpr(_,_): ; 
	}
	
	return tree;
}