/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::backend::OperatorInfo

import IO;
import Node;

data OpType = Prefix() | Infix() | Funfix();
data OpInfo = OpInfo(str opSym, str opName, OpType typ, int arity, int pri);

public map[str,OpInfo] opMapCxx = (
	// unary operators
	// NEGOP, BTNOT, LGNOT: level 0
	"+_":	OpInfo("+", "pos", Prefix(), 1, 0),
	"-_":	OpInfo("-", "neg", Prefix(), 1, 0),
	"!_":	OpInfo("!", "not", Prefix(), 1, 0),
	"~_":	OpInfo("~", "inv", Prefix(), 1, 0),
	// binary operators
	// MULOP: level 1
	"_*_":		OpInfo("*", "mul", Infix(), 2, 1),
	"_/_":		OpInfo("/", "div", Infix(), 2, 1),
	"_%_":		OpInfo("%", "mod", Infix(), 2, 1),
	"_**_":		OpInfo("", "mulmul", Funfix(), 2, 1),
	// ADDOP: level 2
	"_-_":		OpInfo("-", "sub", Infix(), 2, 2),
	"_+_":		OpInfo("+", "add", Infix(), 2, 2),
	"_++_":		OpInfo("", "addadd", Funfix(), 2, 2),
	// SHFOP: level 3
	"_\<\<_":	OpInfo("\<\<", "lshift", Infix(), 2, 3),
	"_\>\>_":	OpInfo("\>\>", "rshift", Infix(), 2, 3),
	// RNGOP: level 4
	"_.._":		OpInfo("", "range", Funfix(), 2, 4),
	// CMPOP, EQUOP, INOP, NOTINOP: level 5
	"_\<_":		OpInfo("\<", "lt", Infix(), 2, 5),
	"_\>_":		OpInfo("\>", "gt", Infix(), 2, 5),
	"_\<=_":	OpInfo("\<=", "le", Infix(), 2, 5),
	"_\>=_":	OpInfo("\>=", "ge", Infix(), 2, 5),
	"_===_":		OpInfo("==", "eq", Infix(), 2, 5),
	"_!==_":		OpInfo("!=", "ne", Infix(), 2, 5),
	// BITAND: level 6
	"_&_":		OpInfo("&", "and", Infix(), 2, 6),
	// BITXOR: level 7
	"_^_":		OpInfo("^", "xor", Infix(), 2, 7),
	// BITOR: level 8
	"_|_":		OpInfo("|", "or", Infix(), 2, 8),
	// SUBST: level 9
	"_\<-\>_":	OpInfo("", "Magnolia::Assert::equal", Funfix(), 2, 9),
	// LGAND: level 10
	"_&&_":		OpInfo("&&", "andand", Infix(), 2, 10),
	// LGOR: level 11
	"_||_":		OpInfo("||", "oror", Infix(), 2, 11),
	"_^^_":		OpInfo("^^", "xorxor", Infix(), 2, 11),
	// LGIMP: level 12
	"_=\>_":	OpInfo("", "Magnolia::implies", Funfix(), 2, 12),
	"_\<=\>_":	OpInfo("==", "equiv", Infix(), 2, 12),
	// Others
	"_[_]":     OpInfo("", "getelem", Funfix(), 2, 0),
	"_[_]=_":   OpInfo("", "setelem", Funfix(), 3, 0)
	
);

public map[str,OpInfo] opMapJava = opMapCxx + (
	"_\<-\>_":	OpInfo("", "Magnolia.Assert.equal", Funfix(), 2, 9),
	"_=\>_":	OpInfo("", "Magnolia.implies", Funfix(), 2, 12)
);


public set[str] reservedKeywords = {
	"and", "and_eq", "alignas", "alignof", "asm", "auto",
	"bitand", "bitor", "bool", "break",
	"case", "catch", "char", "char16_t", "char32_t",
	"class", "compl", "const", "constexpr", "const_cast", "continue",
	"decltype", "default", "delete", "double", "dynamic_cast",
	"else", "enum", "explicit", "export", "extern",
	"false", "float", "for", "friend",
	"goto",
	"if", "inline", "int",
	"long",
	"mutable", "namespace", "new", "noexcept", "not", "not_eq", "nullptr",
	"operator", "or", "or_eq",
	"private", "protected", "public",
	"register", "reinterpret_cast", "return",
	"short", "signed", "sizeof", "static", "static_assert", "static_cast", "struct", "switch",
	"template", "this", "thread_local", "throw", "true", "try", "typedef", "typeid", "typename",
	"union", "unsigned", "using",
	"virtual", "void", "volatile",
	"wchar_t", "while",
	"xor", "xor_eq"
};

public set[str] reservedJavaKeywords = {
	"abstract", "continue", "for", "new", "switch",
	"assert", "default", "goto", "package", "synchronized",
	"boolean", "do", "if", "private", "this",
	"break", "double", "implements", "protected", "throw",
	"byte", "else", "import", "public", "throws",
	"case", "enum", "instanceof", "return", "transient",
	"catch", "extends", "int", "short", "try",
	"char", "final", "interface", "static", "void",
	"class", "finally", "long", "strictfp", "volatile",
	"const", "float", "native", "super", "while"
};