/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::backend::Plainifier
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::TypeChecker;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::typechecker::algo::Primitives;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import String;
import List;
import Exception;
import util::Benchmark;

public Result[AST] getPlainified(DefInfo info) {
	env = newEnv(info.qName, info.parent, locOf(info.body));
	switch(info.body) {
		case DeclBody(seq(ds)): {
			env = loadDefs(info.body, true, env);
			ds = for(def <- ds) {
				if(/Apply(Fun(Name(_)),_) := def) {
					termPrintln("Body contains wrong Apply before plainify: ", def);
				}
				<plainDef, env> = plainifyOperation(def, env);
				if(/Apply(Fun(Name(_)),_) := plainDef) {
					termPrintln("Body contains wrong Apply after plainify:  ", plainDef);
				}
				
				append def; // TODO: plainifier disabled
			}
			return makeResult(preserveAnnos(DeclBody(seq(ds)), info.body), env);
			}
		default: {
			println("Can\'t plainify <consOf(info.body)>");
			return makeResult(info.body, env);
		}
	}
}
public tuple[AST, Env] plainifyOperation(def:Define(dcl,attrs,body), Env env) {
	tmpCounter = 0;
	if(!isOperation(def))
		return <def, env>;
	
	attrs.args = [a | a <- attrs.args, !(Guard(_) := a)];
	
	switch(body) {
		case Body(b:Block(ss)): {
			<ss, env> = plainify(ss, env);
			return <def[arg2 = Body(b[arg0=ss])][arg1 = attrs], env>;
		}
		case Body(b:BlockExpr(ss, rt)): {
			<ss, env> = plainify(ss, env);
			return <def[arg2 = Body(b[arg0=ss])][arg1 = attrs], env>;
		}
		case EmptyBody():
			return <def[arg2 = body][arg1 = attrs], env>;
		case Body(b:Apply(FunOf(ProcName(p, seq(ps)), leaf(_argn)), _)):
			return <def[arg1 = attrs], env>;
		case Instance(_,_):
			return <def[arg1 = attrs], env>;
		case FunName(_,_,_):
			return <def[arg1 = attrs], env>;
		case ProcName(_,_):
			return <def[arg1 = attrs], env>;
		case Body(e): {
			<b, env> = plainify(seq([Return(e)]), env);
			return <def[arg2 = Body(Block(b))][arg1 = attrs], env>;
		}
		default: {
			errorMark(env, "INTERNAL COMPILER ERROR: plainifyOperation(): Unknown operation body cons <consOf(body)>", Internal(), locOf(dcl));
			return <def, env>;
		}
	};
}

public tuple[AST, Env] plainifyOperation(def:Nop(), Env env) {
	return <def, env>;
}

default tuple[AST, Env] plainifyOperation(AST tree, Env env) {
	termPrintln("plainifyOperation default: ", tree);
	errorMark(env, "plainifyOperation(): not implemented for <consOf(tree)>", Warning(), locOf(tree));
	return <tree, env>;
}

public tuple[AST, Env] plainify(seq(todo), Env env) {
	todo = for(t <- todo) {
		<t, env> = plainify(t, env);
		append t;
	}
	return <seq(todo), env>;
}

public tuple[AST, Env] plainify(stat:If(_,_,_), Env env) {
	return plainifySubExpr(stat, 0, [1,2], env);
}

public tuple[AST, Env] plainify(stat:While(_,_), Env env) {
	return plainifySubExpr(stat, 0, [1], env);
}

public tuple[AST, Env] plainify(stat:For(_,_,_), Env env) {
	return plainifySubExpr(stat, 1, [2], env);
}

public tuple[AST, Env] plainify(stat:Assert(_,_), Env env) {
	return plainifySubExpr(stat, 0, [], env);
}

public tuple[AST, Env] plainify(stat:Assume(_,_), Env env) {
	return plainifySubExpr(stat, 0, [], env);
}

public tuple[AST, Env] plainify(stat:Yield(_), Env env) {
	return plainifySubExpr(stat, 0, [], env);
}

public tuple[AST, Env] plainify(stat:Return(_), Env env) {
	return plainifySubExpr(stat, 0, [], env);
}

public tuple[AST, Env] plainify(stat:Block(ss), Env env) {
	<ss, env> = plainify(ss, env);
	return <stat[arg0 = ss], env>;
}

public tuple[AST, Env] plainify(stat:Assign(v, e), Env env) {
	return plainifySubExpr(stat, 1, [], env);
}

public tuple[AST, Env] plainify(stat:Call(Proc(p),seq(as)), Env env) {
	DefInfo info;
	try {
		info = lookupDef(p, env);
		if(isName(info.body)) {
			termPrintln("Followed link from: ", info.qName);
			info = lookupDef(info.body, env);
			termPrintln("                to: ", info.qName);
		}
	}
	catch NoSuchKey(_):
		return <stat, env>;
		
	g = getGuard(info);
	
	if(g == Nop())
		return <stat, env>;
	else {
		if(ProcClause(_, ParamList(seq(ps))) := info.decl) {
			map[AST,AST] renaming = ();
			for(<Param(_,n,t), a> <- zip(ps, as)) {
				renaming[VarName(n,t)] = a;
			}
			g = top-down-break visit(g) {
				case Var(v): if(v in renaming) insert renaming[v];
			}
			return <Guard(preserveAnnos(seq([stat]), stat), g), env>;
		}
		else {
			env = errorMark(env, "INTERNAL COMPILER ERROR: plainifyStat: ProcClause not found", Internal(), locOf(stat));
			return <stat, env>;
		}
	}
}


tuple[AST, Env] plainify(tree:Let(decl:VarDecl(m,v,t,e), seq(stats)), Env env) {
	<e, g, env> = plainifyExpr(e, env);
	if(g == Nop()) {
		<seqStats, env> = plainify(seq(stats), env);
		return <Let(decl, seqStats), env>;
	}
	else {
		<seqStats, env> = plainify(seq(stats), env);
		return <Guard(preserveAnnos(seq([Let(VarDecl(m,v,t,e), seqStats)]), tree), g), env>;
	}
}

default tuple[AST, Env] plainify(AST tree, Env env) {
	termPrintln("plainify default: ", tree);
	errorMark(env, "plainify(): not implemented for <consOf(tree)>", Warning(), locOf(tree));
	return <tree, env>;
}

public tuple[AST, Env] plainifySubExpr(AST tree, int childIndex, list[int] recurse, Env env) {
	<e, g, env> = plainifyExpr(astValue(tree[childIndex]), env);
	tree[childIndex] = e;
	for(i <- recurse) {
		<tree[i], env> = plainify(astValue(tree[i]), env);
	}
	if(g == Nop())
		return <tree, env>;
	else
		return <Guard(seq([tree]), g), env>;
}

public tuple[AST,AST, Env] plainifyExpr(expr:Apply(fexpr,seq(as)), Env env) {
	AST g = Nop();
	
	as = for(a <- as) {
		<a, g2, env> = plainifyExpr(a, env);
		if(g2 == Nop())
			append a;
		else
			append Guard(a, g2);
	}
	
	DefInfo info;
	if(Fun(f) := fexpr) {
		try 
			info = lookupDef(f, env);
		catch NoSuchKey(_): {
			env = errorMark(env, "plainifyExpr: definition not found: <yieldTerm(f)>", Internal(), locOf(expr));
			return <expr, Nop(), env>;
		}
		g = getGuard(info);
	}
	
	if(g == Nop())
		return <expr[arg1=seq(as)], g, env>;
	else {
		if(FunClause(_, ParamList(seq(ps)), _) := info.decl) {
			//termPrintln("plainfyExpr expr: ", expr);
			//termPrintln("plainfyExpr: ps: ", ps);
			//termPrintln("plainfyExpr: as: ", as);
			map[AST,AST] renaming = ();
			for(<Param(_,n,t), a> <- zip(ps, as)) {
				renaming[VarName(n,t)] = a;
			}
			//termPrintln("plainifyExpr:renaming: ", renaming);
			//termPrintln("plainifyExpr:g: ", g);
			g = top-down-break visit(g) {
				case Var(v): if(v in renaming) insert renaming[v];
			}
			//termPrintln("plainifyExpr:g\': ", g);
			return <expr[arg1=seq(as)], g, env>;
		}
		else {
			env = errorMark(env, "INTERNAL COMPILER ERROR: plainifyExpr: FunClause not found", Internal(), locOf(expr));
			return <expr[arg1=seq(as)], g, env>;
		}
	}
}

public tuple[AST,AST, Env] plainifyExpr(expr:Guard(e, g), Env env) {
	<e, g2, env> = plainifyExpr(e, env);
	if(g2 == Nop())
		return <e, g, env>;
	else
		return <a, Apply(binaryPredicate("_&&_"), seq([g, g2])), env>;
}

public tuple[AST,AST, Env] plainifyExpr(expr:Var(_), Env env) {
	return <expr, Nop(), env>;
}

public tuple[AST,AST, Env] plainifyExpr(expr:Undefined(), Env env) {
	return <expr, Nop(), env>;
}
public tuple[AST,AST, Env] plainifyExpr(expr:Equal(op,a,b), Env env) {
	AST g = Nop();
	
	<a, ga, env> = plainifyExpr(a, env);
	<b, gb, env> = plainifyExpr(b, env);
	g = combine(ga, gb);
	
	return <expr, g, env>;
}

default tuple[AST,AST, Env] plainifyExpr(AST tree, Env env) {
	termPrintln("plainifyExpr default: ", tree);
	errorMark(env, "plainifyExpr(): not implemented for <consOf(tree)>", Warning(), locOf(tree));
	return <tree, Nop(), env>;
}

AST getGuard(DefInfo info) {
	for(Guard(g) <- info.attrs) {
		return g;
	}
	return Nop();
}


public AST combine(AST g1, AST g2) {
	if(g1 == Nop())
		return g2;
	else if(g2 == Nop())
		return g1;
	else
		return Apply(binaryPredicate("_&&_"), seq([g1, g2]));
}
