/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::backend::ProgramBuilder
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import org::magnolialang::magnolia::backend::CxxPP;
import org::magnolialang::magnolia::lib::CxxAPI;
import org::magnolialang::magnolia::backend::OperatorInfo;
import Set;
import Node;
import List;
import String;
import util::Benchmark;


public str generateCommandInterface(Env env, list[AST] ds, AST qName, str body) {
	str structName = nameToStr(unqualify(qName));
	list[str] usageString = [];
	list[str] docString = [];
	list[str] callCode = [];
	str template = getBundledResource("org/magnolialang/magnolia/templates/MagicMain.cxx");
	//str ioLib = getBundledResource("org/magnolialang/magnolia/templates/IO.cxx");
	//ioLib += getBundledResource("org/magnolialang/magnolia/templates/XmlIO.cxx");
	//ioLib += getBundledResource("org/magnolialang/magnolia/templates/TextIO.cxx");
	//ioLib += getBundledResource("org/magnolialang/magnolia/templates/BinaryIO.cxx");
	modName = makeCxxNonKeywordName(structName, "_mod");

	str libSetup = "\t\t\ttypedef <ppName(qualifierOf(qName))>::<modName>::<modName>_impl impl;\n\t\t\timpl lib = <ppName(qualifierOf(qName))>::<modName>()();\n";
	time0 = getMilliTime();	
	
	ds = sort([d | d:Define(_,_,_) <- ds, !hasModifier(d, Derived()), !hasModifier(d, Impl())], bool(AST a, AST b) {
		return nameToStr(unqualify(defNameOf(a))) < nameToStr(unqualify(defNameOf(b)));
	});
	
	list[AST] realDefs = [d | d <- ds, !hasModifier(d, Axiom())];
	list[AST] axDefs = [d | d <- ds, hasModifier(d, Axiom())];
	
	set[str] seenOpNames = {};
	str makeUniqueOpName(str n) {
		if(n in seenOpNames) {
			i = 1;
			while("<n><i>" in seenOpNames)
				i = i + 1;
			seenOpNames += {"<n><i>"};
			return "<n><i>";
		}
		else {
			seenOpNames += {n};
			return n;
		}
	}
	void processDefs(list[AST] ds) {
		for(d:Define(def,_,_) <- ds) {
			suf = declSuffix(def);
			defName = defNameOf(def);
			opName = makeUniqueOpName(nameToStr(unqualify(defName)));
			//if(opName in opMapCxx)
			//	opName = opMapCxx[opName].opName;
			cxxName = makeCxxName(def);
			
			switch(def) {
				case TypeClause(n):
					;
				case AlertClause(_,_):
					;
				case FunClause(n,ParamList(seq(ps)),t): {
					nIn = size(inParams(ps));
					usageString += "\t\<\<\t\"       \" \<\< programName \<\< \" [opts] <quote(opName)>";
					usageString += usageParams(ps, qName, nIn == 1, true);
					if(t != SpecialType(Predicate()))
						usageString += " [-o result.<quote(shortNameToStr(t))>]";
					usageString += "\\n\"\n";
					docString += "\t\t\telse if(cmd == \"<quote(opName)>\") {\n\t\t\t\tstd::cerr \<\< \"<quote(shortDispNameOf(def))>\\n\";\n\t\t}\n";
					callCode += "\t\telse if(cmd == \"<quote(opName)>\") {\n";
					callCode += libSetup;
					
					callCode += "\t\t\tbool ok = true;\n";
					callCode += expectParams(ps, opName, nIn == 1, true, true);
					callCode += "\t\t\tok &= Magnolia::MagicMain::expectArg(\"<quote(opName)>\", \"result\", <nIn+1>, args, PARAM_MODE_OUT, false, true);\n";
					callCode += "\t\t\tMagnolia::MagicMain::expectNoMoreArgs(\"<quote(opName)>\", <nIn+1>, args, ok);\n";

					i = 0;
					for(p <- ps) {
						callCode += inParam(p, i, qName);
						i = i + 1;
					} 
					callCode += "\t\t\t<makeCxxName("impl", t)> result = lib.<cxxName>(<strJoin(["arg<j>" | j <- index(ps)], ", ")>);\n";
					myT = makeCxxName("impl", t);
					callCode += "\t\t\tMagnolia::MagicMain::writeValue\<<myT>\>(\"<quote(shortNameToStr(t))>\", args[<nIn+1>], result);\n";
					if(t == SpecialType(Predicate())) {
						callCode += "\t\t\tif(result) return 0; else return 1;\n";
					}
					callCode += "\t\t}\n";
				}
				case ProcClause(n,ParamList(seq(ps))): {
					nIn = size(inParams(ps));
					nOut = size(outParams(ps));
					usageString += "\t\<\<\t\"       \" \<\< programName \<\< \" [opts] <quote(opName)>";
					usageString += usageParams(ps, qName, nIn == 1, nOut == 1);
					usageString += "\\n\"\n";
					docString += "\t\t\telse if(cmd == \"<quote(opName)>\") {\n\t\t\t\tstd::cerr \<\< \"<quote(shortDispNameOf(def))>\\n\";\n\t\t}\n";
					callCode += "\t\telse if(cmd == \"<quote(opName)>\") {\n";
					callCode += libSetup;
				
					callCode += "\t\t\tbool ok = true;\n";
					callCode += expectParams(ps, opName, nIn == 1, nOut == 1, false);
					callCode += "\t\t\tMagnolia::MagicMain::expectNoMoreArgs(\"<quote(opName)>\", <size(ps)>, args, ok);\n";

					i = 0;
					for(p <- ps) {
						callCode += inParam(p, i, qName);
						i = i + 1;
					} 
	
					callCode += "\t\t\tlib.<cxxName>(<strJoin(["arg<j>" | j <- index(ps)], ", ")>);\n";
					i = 0;
				
					for(p <- ps) {
						callCode += outParam(p, i, qName);
						i = i + 1;
					} 
					callCode += "\t}\n";
				}
				default:
					// TODO: make throw
					println(InternalError("Unknown declaration clause <consOf(def)>", def));
			}
		}
	}

	processDefs(realDefs);
	if(axDefs != []) {
		usageString += "\t\<\<\t\"\\n\"\n";
		usageString += "\t\<\<\t\"\\n\"\n";
		usageString += "\t\<\<\t\"     Axioms:\\n\"\n";
		processDefs(axDefs);
	}
	template = replaceAll(template, "/* INSERT COMMAND HELP HERE */", strJoin(docString));
	template = replaceAll(template, "/* INSERT COMMAND CALLS HERE */", strJoin(callCode));
	template = replaceAll(template, "/* INSERT COMMAND USAGE HERE */", strJoin(usageString));
	template = replaceAll(template, "/* INSERT BODY HERE */", body);
	// template = replaceAll(template, "/* INSERT IO LIBRARY HERE */", ioLib);
	println("ProgramBuilder: <getMilliTime()-time0> ms");
	
	return template;
}


str expectParams(list[AST] ps, str opName, bool inFallback, bool outFallback, bool hasResult) {
	i = 0;
	firstIn = 0;
	firstOut = 0;
	result = "";
	for(p <- ps) {
		switch(p) {
			case Param(m, n, t): {
				if(m == Obs()) {
					i = i + 1;
					result += "\t\t\tok &= Magnolia::MagicMain::expectArg(\"<quote(opName)>\", \"<ppName(n)>\", <i>, args, PARAM_MODE_OBS, <inFallback>, <outFallback>);\n";
					if(firstIn == 0) {
						firstIn = i;
					}
				}
				else if(m == Upd()) {
					i = i + 1;
					result += "\t\t\tok &= Magnolia::MagicMain::expectArg(\"<quote(opName)>\", \"<ppName(n)>\", <i>, args, PARAM_MODE_UPD, <inFallback>, <outFallback>);\n";
					if(firstIn == 0) {
						firstIn = i;
					}
					if(firstOut == 0) {
						firstOut = i;
					}
				}
				else if(m == Out()) {
					i = i + 1;
					result += "\t\t\tok &= Magnolia::MagicMain::expectArg(\"<quote(opName)>\", \"<ppName(n)>\", <i>, args, PARAM_MODE_OUT, <inFallback>, <outFallback>);\n";
					if(firstOut == 0) {
						firstOut = i;
					}
				}
			}
			default:
				println("Unknown parameter <p>");
		}
	}
	if(hasResult && firstOut == 0)
		firstOut = i + 1; // result will be last argument
	result = "\t\t\tsetIOArgs(args, argIn, argOut, <firstIn>, <firstOut>);\n" + result;
	return result;
}

str inParam(AST p, int i, AST parent) {
	switch(p) {
		case Param(m, n, t): {
			myT = makeCxxName("impl", t);
			if(m == Obs())
				return "\t\t\tconst <myT> arg<i> = Magnolia::MagicMain::readValue\<<myT>\>(\"<quote(shortNameToStr(t))>\", args[<i+1>]);\n";
			else if(m == Upd())
				return "\t\t\t<myT> arg<i> = Magnolia::MagicMain::readValue\<<myT>\>(\"<quote(shortNameToStr(t))>\", args[<i+1>]);\n";
			else if(m == Out())
				return "\t\t\t<myT> arg<i>;\n";
		}
		default:
			println("Unknown parameter <p>");
	}
	return "";
}

AST param(AST p, int i, AST parent) {
	switch(p) {
		case ParamSugar(n, t): {
			return Var(VarName(Name(leaf("arg<i>")), t));
		}
		case Param(m, n, t): {
			return Var(VarName(Name(leaf("arg<i>")), t));
		}
		default:
			println("Unknown parameter <p>");
	}
}

list[str] usageParams(list[AST] ps, AST parent, bool inIsOptional, bool outIsOptional) {
	i = 1;
	list[str] result = [];

	for(p <- ps) {
		switch(p) {
			case Param(m, n, t): {
				switch(m) {
					case Obs():
						result += " <inIsOptional ? "[" : "">[-i<i>] <quote(nameToStr(n))>.<quote(shortNameToStr(t))><inIsOptional ? "]" : "">";
					case Upd():
						result += " <inIsOptional && outIsOptional ? "[" : "">[-io<i>] <quote(nameToStr(n))>.<quote(shortNameToStr(t))><inIsOptional && outIsOptional ? "]" : "">";
					case Out():
						result += " <outIsOptional ? "[" : "">[-o<i>] <quote(nameToStr(n))>.<quote(shortNameToStr(t))><outIsOptional ? "]" : "">";
					case Nrm():
						result += " <inIsOptional ? "[" : "">[-i<i>] <quote(nameToStr(n))>.<quote(shortNameToStr(t))><inIsOptional ? "]" : "">";
					default:
						println("Unknown parameter mode <m>");
				}
			}
			default:
				println("Unknown parameter <p>");
		}
		i = i + 1;
	}
	return result;
}

str outParam(AST p, int i, AST parent) {
	switch(p) {
		case Param(m, n, t): {
			if(m == Out() || m == Upd()) {
				myT = makeCxxName("impl", t);
				return "\t\t\tMagnolia::MagicMain::writeValue\<<myT>\>(\"<quote(shortNameToStr(t))>\", args[<i+1>], arg<i>);\n";
			}
		}
		default:
			println("Unknown parameter <p>");
	}
	return "";
}
/*
AST unwrapExternal(AST tree) {
	// TODO: check that external kind is C++
	if(ExternalType(_, m, _, t) := tree)
		return t;
	else if(ExternalProc(_, m, _, t, ps) := tree)
		return t;
	else if(ExternalFun(_, m, _, t, ps, rt) := tree)
		return t;
//	else if(always)
//		throw InternalError("Expected external definition, but got <consOf(tree)>", tree);
	else
		return tree;
}
*/

str makeHelp(str s) {
	return "void help(const std::string& programName) {
	         'std::cerr
	         '\t\<\<\t\"usage: \" \<\< programName \<\< \" [opts] command args...\\n\"
	         '<s>;
	         'std::exit(0);
	         '}\n";
}

str makeCmdHelp(str s) {
	return "void cmdHelp(const std::string& programName, const std::string& cmdName) {
	         'std::cerr
	         '\t\<\<\t\"usage: \" \<\< programName \<\< \" [opts] command args...\\n\"
	         '<s>;
	         'std::exit(0);
	         '}\n";
}
