/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator;

import java.io.PrintWriter;
import java.io.Reader;

import org.rascalmpl.value.ISourceLocation;
import org.magnolialang.magnolia.evaluator.callables.Callable;

public interface IEvaluationContext {
	/**
	 * Record that we are starting evaluation of the given source location.
	 * 
	 * @param loc
	 *            The source location.
	 * @return A new, updated context.
	 */
	IEvaluationContext enterLocation(ISourceLocation loc);


	/**
	 * Record that we are starting evaluation of the given operation.
	 * 
	 * @param operation
	 *            The operation.
	 * @return A new, updated context.
	 */
	IEvaluationContext enterOperation(Callable operation);


	/**
	 * @return Current source location.
	 */
	ISourceLocation getCurrentLocation();


	/**
	 * @return Current location, as a string.
	 */
	String getCurrentLocationString();


	/**
	 * @return The operation we're currently evaluating.
	 */
	Callable getCurrentOperation();


	/**
	 * @return The current evaluator.
	 */
	AlgoEvaluator getEvaluator();


	/**
	 * @return Module name of the current evaluator.
	 */
	String getModuleName();


	/**
	 * @return Standard error stream of the evaluator.
	 */
	PrintWriter getStderr();


	/**
	 * @return Standard input stream of the evaluator.
	 */
	Reader getStdin();


	/**
	 * @return Standard output stream of the evaluator.
	 */
	PrintWriter getStdout();
}
