/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::evaluator::Values
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import List;
import String;


data Value = IntValue(int i) | RealValue(real r) | StrValue(str s) | BoolValue(bool b) | AstValue(AST a) | VoidValue();

public AST voidType = TypeName(QName(seq([Name(leaf("void"))])));
public AST intType = TypeName(QName(seq([Name(leaf("int"))])));
public AST realType = TypeName(QName(seq([Name(leaf("real"))])));
public AST strType = TypeName(QName(seq([Name(leaf("str"))])));
public AST boolType = SpecialType(Predicate());
public AST astType = TypeName(QName(seq([Name(leaf("ast"))])));
public AST declType = SpecialType(Decl());
public AST decllistType = SpecialType(DeclList());
public AST renamingType = SpecialType(Renaming());

public AST typeOf(VoidValue()) {
	return voidType;
}

public AST typeOf(IntValue(_)) {
	return intType;
}

public AST typeOf(RealValue(_)) {
	return realType;
}

public AST typeOf(StrValue(_)) {
	return strType;
}

public AST typeOf(BoolValue(_)) {
	return boolType;
}

public AST typeOf(AstValue(Define(_,_,_))) {
	return declType;
}

public AST typeOf(AstValue(DeclBody(_))) {
	return decllistType;
}

public AST typeOf(AstValue(RenameList(_))) {
	return renamingType;
}
public AST typeOf(AstValue(_)) {
	return astType;
}

public AST typeOf(Value v) {
	throw InternalError("Internal Compiler Error: Values::typeOf: unable to find type of <v>", Nop());
}
