/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.callables;

import java.lang.reflect.InvocationTargetException;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IMap;
import org.magnolialang.magnolia.evaluator.AlgoEvaluator;
import org.magnolialang.magnolia.evaluator.IEvaluationContext;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.magnolialang.magnolia.evaluator.types.ExternalModule;
import org.magnolialang.magnolia.evaluator.types.ExternalModule.MethodHandle;
import org.magnolialang.magnolia.evaluator.values.Value;
import org.nuthatchery.pica.terms.TermAdapter;

public abstract class ExternalOperation extends Operation {

	protected final MethodHandle method;
	private final ExternalModule module;


	public ExternalOperation(ExternalModule module, IConstructor opName, IConstructor paramList, IMap env) {
		super(opName, paramList, env);
		this.module = module;
		this.method = module.getMethod(name, paramTypes, paramModes);
	}


	@Override
	public Object call(IEvaluationContext ctx, IMap env, IConstructor argList) {
		Object[] args = new Object[paramModes.size()];
		int i = 0;
		Iterable<IConstructor> arguments = TermAdapter.getChildren(argList);
		for(IConstructor arg : arguments) {
			if(AlgoEvaluator.isOut(paramModes.get(i))) {
				args[i] = ctx.getEvaluator().evalVar(arg, env, ctx);
			}
			else {
				args[i] = ctx.getEvaluator().evalExpr(arg, env, ctx);
			}
			i++;
		}

		try {
			Object result = method.invoke(ctx, args);
			if(result != null) {
				return result;
			}
			else {
				return Value.UNDEFINED;
			}
		}
		catch(IllegalAccessException | IllegalArgumentException e) {
			throw new EvaluationError("Error invoking method '" + module.getName() + "." + name + "': " + e.getMessage(), e);
		}
		catch(InvocationTargetException e) {
			Throwable cause = e.getCause();
			if(cause instanceof RuntimeException) {
				throw (RuntimeException) cause;
			}
			else {
				throw new EvaluationError("Caught exception while invoking method '" + module.getName() + "." + name + "': " + e.getMessage(), e);
			}
		}
	}
}
