/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.callables;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IMap;
import org.rascalmpl.value.IValue;
import org.magnolialang.magnolia.evaluator.IEvaluationContext;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.nuthatchery.pica.terms.TermAdapter;

public class Function extends InternalOperation {

	public Function(IConstructor funDef, IMap env) {
		super(funDef, env);

		if(!TermAdapter.getName(clause).equals("FunClause")) {
			throw new EvaluationError("Not a function");
		}
	}


	@Override
	public Object call(IEvaluationContext ctx, IMap env, IConstructor argList) {
		env = bindArguments(ctx, env, argList);
		IConstructor body = getBody();

		if(TermAdapter.isCons(body, "FunName")) {
			IValue value = env.get(body);

			if(value instanceof Function) {
				return ((Function) value).call(ctx, env, argList);
			}
		}
		return ctx.getEvaluator().evalExprDefBody(body, env, ctx);

	}


	@Override
	public String getKind() {
		return "function";
	}


	@Override
	public boolean isFunction() {
		return true;
	}

}
