/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.callables;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IMap;
import org.magnolialang.magnolia.evaluator.AlgoEvaluator;
import org.magnolialang.magnolia.evaluator.IEvaluationContext;
import org.magnolialang.magnolia.evaluator.errors.WrongNumberOfArguments;
import org.magnolialang.magnolia.evaluator.values.VarValue;
import org.nuthatchery.pica.terms.TermAdapter;

public abstract class InternalOperation extends Operation {

	private final IConstructor body;
	protected final IConstructor clause;


	public InternalOperation(IConstructor opDef, IMap env) { // Define(clause, attrs, body)
		super(opDef, env);
		this.body = TermAdapter.getArg(opDef, 2);
		this.clause = TermAdapter.getArg(opDef, 0);
	}


	public IConstructor getBody() {
		return body;
	}


	@Override
	public String toString() {
		return getKind() + " " + name;
	}


	protected IMap bindArguments(IEvaluationContext ctx, IMap callerEnv, IConstructor argList) {
		IMap calleeEnv = ctx.getEvaluator().getTopEnv();
		if(numParams() != TermAdapter.arity(argList)) {
			throw new WrongNumberOfArguments(this, TermAdapter.arity(argList));
		}

		int i = 0;
		Iterable<IConstructor> arguments = TermAdapter.getChildren(argList);
		for(IConstructor arg : arguments) {
			IConstructor mode = getParamMode(i);
			IConstructor name = getParamName(i);
			if(AlgoEvaluator.isOut(mode)) {
				calleeEnv = calleeEnv.put(name, ctx.getEvaluator().evalVar(arg, callerEnv, ctx));
			}
			else {
				calleeEnv = calleeEnv.put(name, new VarValue(ctx.getEvaluator().evalExpr(arg, callerEnv, ctx), mode));
			}
			System.err.println("Defining: " + TermAdapter.yieldTerm(name, false) + " = " + calleeEnv.get(name));
			i++;
		}

		return calleeEnv;
	}
}
