/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.callables;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IMap;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.evaluator.AlgoEvaluator;
import org.magnolialang.magnolia.evaluator.types.Type;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;

public abstract class Operation extends Callable {

	protected final List<IConstructor> paramNames;
	protected final List<IConstructor> paramModes;
	protected final List<Type> paramTypes;
	protected final String name;
	protected final IConstructor id;


	public Operation(IConstructor opName, IConstructor paramList, IMap env) {
		this.id = opName;
		this.name = Magnolia.yieldName(Magnolia.basename(id));
		paramList = TermAdapter.getArg(paramList, 0);
		List<IConstructor> paramNames = new ArrayList<>(TermAdapter.arity(paramList));
		List<IConstructor> paramModes = new ArrayList<>(TermAdapter.arity(paramList));
		List<Type> paramTypes = new ArrayList<>(TermAdapter.arity(paramList));
		for(IConstructor c : TermAdapter.getChildren(paramList)) {
			paramNames.add(TermFactory.cons("VarName", TermAdapter.getArg(c, 1), TermAdapter.getArg(c, 2)));
			paramModes.add(TermAdapter.getArg(c, 0));
			paramTypes.add(AlgoEvaluator.lookupType(TermAdapter.getArg(c, 2), env));
		}
		this.paramNames = Collections.unmodifiableList(paramNames);
		this.paramModes = Collections.unmodifiableList(paramModes);
		this.paramTypes = Collections.unmodifiableList(paramTypes);
	}


	public Operation(IConstructor opDef, IMap env) { // Define(clause, attrs, body)
		this(TermAdapter.getArg(opDef, 0, 0), TermAdapter.getArg(opDef, 0, 1), env);
	}


	@Override
	public IConstructor getId() {
		return id;
	}


	@Override
	public String getName() {
		return name;
	}


	@Override
	public IConstructor getParamMode(int i) {
		return paramModes.get(i);
	}


	@Override
	public IConstructor getParamName(int i) {
		return paramNames.get(i);
	}


	@Override
	public Type getParamType(int i) {
		return paramTypes.get(i);
	}


	@Override
	public boolean isOperation() {
		return true;
	}


	@Override
	public int numParams() {
		return paramModes.size();
	}

}
