/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.types;

import org.magnolialang.magnolia.evaluator.values.BooleanValue;
import org.nuthatchery.pica.terms.TermFactory;

public abstract class BuiltinType extends Type {

	public static final Type BOOLEAN_TYPE = new BooleanType();
	public static final Type UNKNOWN_TYPE = new UnknownType();


	public BuiltinType(String name) {
		super(TermFactory.cons("Name", TermFactory.leaf(name)));
	}


	static class BooleanType extends BuiltinType {

		public BooleanType() {
			super("\\predicate");
		}


		@Override
		public Class<?> getValueClass() {
			return BooleanValue.class;
		}

	}


	static class UnknownType extends BuiltinType {

		public UnknownType() {
			super("?");
		}


		@SuppressWarnings("null")
		@Override
		public Class<?> getValueClass() {
			return Object.class;
		}

	}
}
