/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.types;

import org.rascalmpl.value.IConstructor;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.nuthatchery.pica.terms.TermAdapter;

public class ExternalType extends Type {

	private final String packageName;
	private final String className;
	private final Class<?> clazz;


	public ExternalType(ExternalModule module, IConstructor typeDef) {
		super(TermAdapter.getArg(typeDef, 0, 0));
		IConstructor body = TermAdapter.getArg(typeDef, 2);

		IConstructor langId = TermAdapter.getArg(body, 0, 0);
		String language = TermAdapter.getString(TermAdapter.getArg(langId, 0));
		if(language.equalsIgnoreCase("Java")) {
			packageName = Magnolia.yieldName(TermAdapter.getArg(body, 0, 1));
			className = Magnolia.yieldName(TermAdapter.getArg(body, 0, 3));
			clazz = module.getValueClass(className);
		}
		else {
			throw new EvaluationError("Unsupported language " + language);
		}
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(@Nullable Object obj) {
		if(this == obj) {
			return true;
		}
		if(!super.equals(obj) || obj == null) {
			return false;
		}
		if(getClass() != obj.getClass()) {
			return false;
		}
		ExternalType other = (ExternalType) obj;
		if(!className.equals(other.className)) {
			return false;
		}
		if(!clazz.equals(other.clazz)) {
			return false;
		}
		if(!packageName.equals(other.packageName)) {
			return false;
		}
		return true;
	}


	@Override
	public Class<?> getValueClass() {
		return clazz;
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + className.hashCode();
		result = prime * result + clazz.hashCode();
		result = prime * result + packageName.hashCode();
		return result;
	}


	@Override
	public String toString() {
		return "class " + packageName + "." + className;
	}

}
