/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.types;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IMap;
import org.rascalmpl.value.IMapWriter;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.magnolialang.magnolia.evaluator.values.IValueBuilder;
import org.magnolialang.magnolia.evaluator.values.StructValue;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;

public class StructType extends Type {
	private final IMap fields;


	@SuppressWarnings("null")
	public StructType(IConstructor typeDef) {
		super(TermAdapter.getArg(typeDef, 0, 0));
		IConstructor body = TermAdapter.getArg(typeDef, 2);
		// body is Body(Struct(Field(VarDecl(...))...))
		IMapWriter mapWriter = TermFactory.vf.mapWriter();
		for(IConstructor field : TermAdapter.getChildren(TermAdapter.getArg(TermAdapter.getArg(body, 0), 0))) {
			IConstructor varDecl = TermAdapter.getArg(field, 0); // VarDecl(mod, name, type, init)

			mapWriter.put(TermAdapter.getArg(varDecl, 1), TermAdapter.getArg(varDecl, 3));
		}
		fields = mapWriter.done();
	}


	@Override
	public IValueBuilder getBuilder() {
		return new StructValue(this);
	}


	public IConstructor getFieldInit(IConstructor name) {
		IConstructor v = (IConstructor) getFields().get(name);
		if(v == null) {
			throw new EvaluationError("No such field: " + Magnolia.yieldName(name));
		}
		return v;
	}


	public IMap getFields() {
		return fields;
	}


	@Override
	public Class<?> getValueClass() {
		return StructValue.class;
	}


	public boolean hasField(IConstructor name) {
		return getFields().containsKey(name);
	}

}
