/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.types;

import org.rascalmpl.value.IConstructor;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.magnolialang.magnolia.evaluator.values.IValueBuilder;
import org.magnolialang.magnolia.evaluator.values.Value;

public abstract class Type extends Value {

	private final IConstructor id;
	private final String name;


	public Type(IConstructor id) {
		this.id = id;
		this.name = Magnolia.yieldName(id);
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(@Nullable Object obj) {
		if(this == obj) {
			return true;
		}
		if(obj == null) {
			return false;
		}
		if(getClass() != obj.getClass()) {
			return false;
		}
		Type other = (Type) obj;
		if(!id.isEqual(other.id)) {
			return false;
		}
		return true;
	}


	public IValueBuilder getBuilder() {
		throw new EvaluationError("Can't build values of this type");
	}


	@Override
	public IConstructor getId() {
		return id;
	}


	@Override
	public String getKind() {
		return "type";
	}


	@Override
	public String getName() {
		return name;
	}


	public abstract Class<?> getValueClass();


	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + id.hashCode();
		return result;
	}


	@Override
	public boolean isType() {
		return true;
	}


	@Override
	public String toString() {
		return "type " + getName();
	}
}
