/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.values;

import org.rascalmpl.value.IConstructor;

public abstract class BooleanValue extends Value {
	public static final BooleanValue TRUE = new TrueValue();
	public static final BooleanValue FALSE = new FalseValue();


	private BooleanValue() {
	}


	public Value and(Value other) {
		if(isTrue()) {
			return other;
		}
		else {
			return FALSE;
		}
	}


	@Override
	public String getKind() {
		return "predicate";
	}


	public Value iff(Value other) {
		if(this == other) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}


	public Value implies(Value other) {
		if(isTrue()) {
			return other;
		}
		else {
			return TRUE;
		}
	}


	@Override
	public boolean isBoolean() {
		return true;
	}


	@Override
	public boolean isValue() {
		return true;
	}


	public abstract Value not();


	public Value or(Value other) {
		if(isTrue()) {
			return TRUE;
		}
		else {
			return other;
		}
	}


	public static BooleanValue create(boolean b) {
		if(b) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}


	private static class FalseValue extends BooleanValue {
		@Override
		public boolean isFalse() {
			return true;
		}


		@Override
		public boolean isTrue() {
			return false;
		}


		@Override
		public BooleanValue not() {
			return TRUE;
		}


		@Override
		public String toString() {
			return "FALSE";
		}


	}


	private static class TrueValue extends BooleanValue {
		@Override
		public boolean isFalse() {
			return false;
		}


		@Override
		public boolean isTrue() {
			return true;
		}


		@Override
		public BooleanValue not() {
			return FALSE;
		}


		@Override
		public String toString() {
			return "TRUE";
		}
	}

}
