/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.values;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IMap;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.magnolialang.magnolia.evaluator.types.StructType;

public class StructValue extends Value implements IValueBuilder {
	private IMap fields;
	private final StructType type;


	public StructValue(StructType typeValue) {
		this.type = typeValue;
		this.fields = typeValue.getFields();
	}


	@Override
	public boolean equals(@Nullable Object obj) {
		if(this == obj) {
			return true;
		}
		if(obj == null) {
			return false;
		}
		if(getClass() != obj.getClass()) {
			return false;
		}
		StructValue other = (StructValue) obj;
		if(!fields.equals(other.fields)) {
			return false;
		}
		if(!type.equals(other.type)) {
			return false;
		}
		return true;
	}


	@SuppressWarnings("null")
	public Object get(IConstructor name) {
		return Value.objectOf((Value) fields.get(name));
	}


	@Override
	public String getKind() {
		return "struct";
	}


	@Override
	public Value getValue() {
		return this;
	}


	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + fields.hashCode();
		result = prime * result + type.hashCode();
		return result;
	}


	@Override
	public boolean isValue() {
		return true;
	}


	@SuppressWarnings("null")
	@Override
	public void put(IConstructor name, Object data) {
		if(type.hasField(name)) {
			fields = fields.put(name, Value.valueOf(data));
		}
		else {
			throw new EvaluationError("No such field: " + Magnolia.yieldName(name) + " in type " + type);
		}
	}


	@SuppressWarnings("null")
	@Override
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append(type.getName());
		buf.append("{");
		String sep = "";
		for(IValue k : fields) {
			buf.append(sep);
			buf.append(Magnolia.yieldName(k));
			buf.append("=");
			buf.append(fields.get(k));
			sep = ", ";
		}
		buf.append("}");
		return buf.toString();
	}
}
