/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.values;

import org.rascalmpl.value.IAnnotatable;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IExternalValue;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IWithKeywordParameters;
import org.rascalmpl.value.type.ExternalType;
import org.rascalmpl.value.type.Type;
import org.rascalmpl.value.type.TypeFactory;
import org.rascalmpl.value.visitors.IValueVisitor;
import org.eclipse.jdt.annotation.Nullable;
import org.nuthatchery.pica.terms.TermFactory;

public abstract class Value implements IExternalValue {

	public static final Value UNDEFINED = new Value() {
		@Override
		public String getKind() {
			return "value";
		}


		@Override
		public boolean isValue() {
			return true;
		}


		@Override
		public String toString() {
			return "?";
		}


	};
	public static final ExternalType VALUE_TYPE = new ExternalType() {
		@Override
		protected @Nullable Type glbWithExternal(@Nullable Type type) {
			if(type == this) {
				return this;
			}
			else {
				return TypeFactory.getInstance().voidType();
			}
		}


		@Override
		protected boolean isSubtypeOfExternal(@Nullable Type type) {
			return false;
		}


		@Override
		protected @Nullable Type lubWithExternal(@Nullable Type type) {
			if(type == this) {
				return this;
			}
			else {
				return TypeFactory.getInstance().valueType();
			}
		}
	};


	@Override
	@Nullable
	public <T, E extends Throwable> T accept(@Nullable IValueVisitor<T, E> arg0) throws E {
		return null;
	}


	@Override
	@Nullable
	public IAnnotatable<? extends IValue> asAnnotatable() {
		return null;
	}


	@Override
	@Nullable
	public IWithKeywordParameters<? extends IValue> asWithKeywordParameters() {
		return null;
	}


	public IConstructor getId() {
		return TermFactory.cons("Undefined");
	}


	public abstract String getKind();


	public String getName() {
		return "__undefined__";
	}


	@Override
	public Type getType() {
		return VALUE_TYPE;
	}


	@Override
	public boolean isAnnotatable() {
		return false;
	}


	/**
	 * @return True if this is a predicate value.
	 */
	public boolean isBoolean() {
		return false;
	}


	/**
	 * @return True if this value is callable.
	 */
	public boolean isCallable() {
		return false;
	}


	@Override
	public boolean isEqual(@Nullable IValue arg0) {
		if(arg0 instanceof Value) {
			return equals(arg0);
		}
		else {
			return false;
		}
	}


	/**
	 * @return True if the value represents 'false'.
	 */
	public boolean isFalse() {
		return false;
	}


	/**
	 * @return True if this value is a function. Implies {@link #isCallable()}
	 *         and {@link #isOperation()}.
	 */
	public boolean isFunction() {
		return false;
	}


	/**
	 * @return True if this value is an operation. Implies {@link #isCallable()}
	 *         .
	 */
	public boolean isOperation() {
		return false;
	}


	/**
	 * @return True if this value is a procedure. Implies {@link #isCallable()}
	 *         and {@link #isOperation()}.
	 */
	public boolean isProcedure() {
		return false;
	}


	/**
	 * @return True if this value represents 'true'.
	 */
	public boolean isTrue() {
		return false;
	}


	/**
	 * @return True if this value is the represenation of a type.
	 */
	public boolean isType() {
		return false;
	}


	/**
	 * @return True if this value is an actual value (as opposed to a type or an
	 *         operation).
	 */
	public boolean isValue() {
		return false;
	}


	/**
	 * @return True if this is a variable.
	 */
	public boolean isVar() {
		return false;
	}


	@Override
	public boolean mayHaveKeywordParameters() {
		return false;
	}


	@Override
	public String toString() {
		if(getName().equals("__undefined__")) {
			return getKind() + " " + super.toString();
		}
		else {
			return getKind() + " " + getName();
		}
	}


	/**
	 * Unwrap a value, if necessary.
	 *
	 * @param val
	 *            A value
	 * @return The value, as an object.
	 */
	public static Object objectOf(Value val) {
		if(val instanceof ExternalValue) {
			return ((ExternalValue) val).getData();
		}
		else {
			return val;
		}
	}


	/**
	 * Wrap a value, if necessary.
	 *
	 * Will turn any object into a Value; non-Values will be wrapped in
	 * ExternalValue.
	 *
	 * @param obj
	 *            The object
	 * @return A value
	 */
	public static Value valueOf(Object obj) {
		if(obj instanceof Value) {
			return (Value) obj;
		}
		else {
			return new ExternalValue(obj);
		}
	}


	@Override
	public IConstructor encodeAsConstructor() {
		// TODO Auto-generated method stub
		return null;
	}
}
