/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.values;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.evaluator.AlgoEvaluator;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.nuthatchery.pica.terms.TermFactory;

/**
 * Stores a variable in the Magnolia evaluator.
 *
 * @param <T>
 *            Java type of the variable.
 */
public class VarValue<T> extends Value {
	private static final IConstructor UPD = TermFactory.cons("Upd");

	@Nullable
	private T value;
	private IConstructor mode;


	public VarValue() {
		this(null, UPD);
	}


	@SuppressWarnings("unchecked")
	public VarValue(@Nullable T object, IConstructor mode) {
		this.value = (object instanceof VarValue ? ((VarValue<T>) object).get() : object);
		this.mode = mode;
	}


	/**
	 * Get value of variable.
	 *
	 * Throws an EvaluationError if the variable is not readable.
	 *
	 * @return The current value of the variable (null if undefined)
	 */
	@Nullable
	public T get() {
		if(AlgoEvaluator.isIn(mode)) {
			return value;
		}
		else {
			throw new EvaluationError("Variable not readable");
		}

	}


	@Override
	public String getKind() {
		return "var";
	}


	@Override
	public boolean isValue() {
		return true;
	}


	@Override
	public boolean isVar() {
		return true;
	}


	/**
	 * Set the current value of the variable
	 *
	 * Throws an EvaluationError if the variable is not writable.
	 *
	 * @param val
	 *            The new value
	 */
	public void set(T val) {
		if(AlgoEvaluator.isOut(mode)) {
			this.value = val;
			this.mode = UPD; // an out variable will be upd once written to
		}
		else {
			throw new EvaluationError("Variable not writable");
		}
	}


	/**
	 * Set the parameter mode of the variable.
	 *
	 * @param mode
	 *            New mode.
	 */
	public void setMode(IValue mode) {
		this.mode = (IConstructor) mode;
	}


}
