/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 Tero Hasu
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
// Parts taken from rascal::syntax::RascalRascal
module org::magnolialang::magnolia::grammar::Magnolia

// PROGRAMS

// start syntax Program = lex LAYOUT* RealProgram LAYOUT*;

start syntax Program
	= MagnoliaTree: PackageHead TopDecl*
	| TopExprTree: TopExpr+
	> MagnoliaTreeSugar: TopDecl*
	;

syntax PackageHead
	= @deprecated="Use \'package ...;\'" ModuleHead: "module" QualifiedNameIdentifier PackageClause* ";"
	| @Foldable PackageHead: "package" QualifiedNameIdentifier PackageClause* ";"
	| @Foldable PackageHeadSugar: PackageClause+ ";"
	;
	
syntax PackageClause
	= Imports: "imports" {ImportClause ","}*
	;
 
syntax ImportClause
	= ImportAll: QualifiedNameIdentifier
	;
	
// TOP-LEVEL

syntax TopDecl 
	= @Foldable TopDefineSugar:	Modifier* TopExprDeclarative SubClause* "=" TopExpr ";"
	| @Foldable TopDefineSugar:	Modifier* TopRenDeclarative SubClause* "=" RenameExpr ";"
	| @Foldable TopDefineSugar:	Modifier* TopSatDeclarative SubClause* "=" SatisfactionExpr ";"
	| @Foldable MetaDef:	"meta" MetaKind "{" MetaDecl* "}" ";"
	| Decl
	;

syntax TopExprDeclarative
	= ConceptClause: "concept" Name
	| ImplClause:    "implementation" Name
	| ProgramClause: "program" Name
	| ConceptClause: "signature" Name
	;
	 
syntax TopRenDeclarative
	= RenClause:	 "renaming" Name
	;

syntax TopSatDeclarative
	= SatClause:	 "satisfaction" Name
	;
keyword MetaKind = "top" | "algo";

syntax MetaDecl
	= @Foldable DefineSugar: 		Modifier* MetaDeclarative SubClause* ExprDefBody ";"
	;

syntax PartialTopExpr = TopExpr
	;
	
syntax TopExpr
	= ModuleName:	QualifiedNameIdentifier
	| DeclList: "{" Decl* "}"
 	| bracket  	"(" TopExpr ")"
	| Renamed:	TopExpr "[" {RenameExpr ","}* "]"
	| TopApp:	QualifiedNameIdentifier "(" {TopExpr ","}* ")"
	| TopAppSugar:	QualifiedNameIdentifier "{" Decl* "}"
	| TopMap:	"*" QualifiedNameIdentifier "(" TopExpr ")"
	| Expect:	"\\expect" "(" StringConstant "," TopExpr ")"
	> External:	"external" LanguageIdentifier QualifiedNameIdentifier TopExpr
	> left TopBinOpSugar:		TopExpr TOPBINOP TopExpr
	;
/*	| DeclList: "{" Decl* "}"
 	| bracket  		"(" TopExpr ")"
	> Renamed:		TopExpr "[" {Renaming ","}* "]"
	> TopPreOp:		TOPPREOP TopExpr
 	// Concepts
 	| left TopBinOp:	TopExpr MORPHISM TopExpr
 	// Libraries and Implementations
 	| @deprecated="Use \'{ require ...; ... }\'" OnDefines:		"on" TopExpr "defines" TopExpr
 	| @deprecated="Use \'{ ... }\'" Defines:		"defines" TopExpr
	| @deprecated="Use \'{ extend ...; ... }\'" ExtendsDefines:	"extends" TopExpr "defines" TopExpr	
	| @deprecated="Use \'{ extend ...; require ...; ... }\'" ExtendsOnDefines:	"extends" TopExpr "on" TopExpr "defines" TopExpr	
 	| @deprecated="Use \'{ require ...; ... }\'" External: 		"external" ExternalExpr
 	> left TopBinOp:	TopExpr TOPMULOP TopExpr
 	> left TopBinOp:	TopExpr TOPATOP TopExpr
	> left TopBinOp: 	TopExpr TOPADDOP TopExpr
	| left OnFilter:		TopExpr "on" TOPFILTEROP AlgDeclOrName
	| left DeclaredFilter: 	TopExpr "declared" TOPFILTEROP AlgDeclOrName
 	| left Times: 			TopExpr "times" TopExpr
	| left TopBinOp:		TopExpr TOPQUOTIENT AlgDeclOrName
	| left Homomorphism:	TopExpr "homomorphisms" TopExpr "on" TopExpr "with" TopExpr
*/



syntax RenameExpr
	= Rename:		OperationName "=\>" OperationName
	| ModuleName:		QualifiedNameIdentifier
	| RenameApp:	QualifiedNameIdentifier "(" {RenameExpr ","}* ")"
	| RenameList: 	"[" {RenameExpr ","}* "]"
	| RenameCompose: RenameExpr "[" {RenameExpr ","}* "]"
	;
	
keyword MORPHISM = "morphism";

keyword TOPPREOP 
	= "signature"
	| "full"
	| "on"
	| "declared"
	| "extend"
	;

keyword TOPBINOP 
	= "subalgebra"
	| "retain"
	| "remove"
	| "protect"
	| "datainvariant"
	| "congruence"
	| "quotient"
	;
	
	
syntax SatisfactionExpr
	= ModuleName: QualifiedNameIdentifier
	| Models: TopExpr "models" TopExpr
	| WithModels: TopExpr "with" TopExpr "models" TopExpr
	| Models: TopExpr "approximates" TopExpr
	| WithModels: TopExpr "with" TopExpr "approximates" TopExpr
	| Expect:	"\\expect" "(" StringConstant "," SatisfactionExpr ")"
	| DeclList: "{" Decl* "}"
	;

// DECLARATIONS
syntax Decl
	= @Foldable DefineSugar: 		Modifier* StatDeclarative SubClause* StatDefBody
	| @Foldable DefineSugar: 		Modifier* ExprDeclarative SubClause* ExprDefBody ";"
	| @Foldable DefineSugar: 		Modifier* TypeDeclarative SubClause* TypeDefBody ";"
	| @Foldable DefineSugar: 		Modifier* AlertDeclarative SubClause* AlertDefBody ";"	
	| RequireSugar:		"require" TopExpr ";"
	| ExtendSugar:		"extend" TopExpr ";"
	| UseSugar:			"use" TopExpr ";"
	| ProtectSugar:		"protect" TopExpr ";"
	| TopBinOpAsUnarySugar:	TOPBINOP TopExpr ";"
	| External:		"external" LanguageIdentifier QualifiedNameIdentifier TopExpr ";"
	| @deprecated="Use \'require ...;\'" Requires:		"requires" {TopExpr ","}+ ";"
	| PreserveSugar:	"preserve" TopExpr ";"
	| PreserveOn: 	"preserve" TopExpr "on" TopExpr ";"
	| GenerateBy:	"generate" {Type ","}+ "by" TopExpr ";"
	| FreeBy:		"free" {Type ","}+ "by" TopExpr ";"
	| PartitionBy: 	"partition" {Type ","}+ "by" TopExpr ";"
	| HomomorphismOnWith:"homomorphism" TopExpr "on" TopExpr "with" TopExpr ";"
	| Expect:	"\\expect" "(" StringConstant "," Decl ")"
	| Nop: ";"
	;

syntax StatDefBody
	= Body: "=" Stat
	| Body: "=" Proc
	| Body: BlockStat	
	| EmptyBody: ";"
	;

syntax StatDefBodyNS
	= Body: "=" Stat
	| Body: BlockStat	
	| EmptyBody:
	;
	
syntax ExprDefBody
	= Body: "=" Expr
	| EmptyBody:
	;
	
syntax TypeDefBody
	= Body: "=" Type
	| EmptyBody:
	;

syntax AlertDefBody
	= AlertBody:
	;
	
syntax DeclBody
	= DeclList: "{" Decl* "}"
	;

syntax AlgDecl
	= DeclBody
	| Decl: 		Modifier* StatDeclarative SubClause* StatDefBodyNS
	| Decl: 		Modifier* ExprDeclarative SubClause* ExprDefBody
	| Decl: 		Modifier* TypeDeclarative SubClause* TypeDefBody
	;

syntax Modifier
	= Protect: "protect"    // this decl's guard will extend
	| Retain: "retain"      // this declaration should be retained in the current list
	| Remove: "remove"      // this declaration should be removed from the current list
	| Require: "require"    // body of this decl must be provided elsewhere
	| Extend:  "extend"     // ? 
	| Abstract: "abstract"  // ? 
	| Default: "default"    // ? 
	| Export: "export"      // ? 
	| Congruence: "congruence"	// ?
	| Quotient: "quotient"	// this predicate defines equality for a type in the surrounding context
	| Datainvariant: "datainvariant"	// this predicate is preserved on all values of the type in the surrounding context
	| Derived: "derived"    // this decl was generated from another one
	| Builtin: "builtin"    // this is a builtin operation
	| Use: "use"            // this decl has been imported
	| External: "external"  // body of this decl will be provided externally
	| Preserve: "preserve"  // this decl preserves data invariant
	| Assume: "assume"      // we're assuming (in a satisfaction) that this decl will have a body somewhere
	| ImplyMod: "imply"     // we're imlpying (in a satisfaction) that this decl will be satisfied
	| NoBuild: "@nobuild"   // don't generate c++ code for this
	| Broken: "@broken"     // don't generate c++ code for this
	| AxiomModifier: "\\axiom"
	| ModifierSugar: "@" QualifiedName
	| Modifier: "@" QualifiedName "(" {Expr ","}* ")"
	| ModifierText: "@" QualifiedName "{" BraceBalancedString "}" 
	;
	
syntax SubClause
	= Guard: "guard" Expr
	| Alert: "alert" QualifiedName AlertSpec
	| AlertPreSugar: "precond" Expr 
	;

syntax AlertSpec
	= AlertPre:     "if" "pre" Expr
	| AlertPreNegSugar:  "unless" "pre" Expr
	| AlertThrows:  "if" "throwing" Name
	| AlertPost:    "if" "post" Expr
	| AlertPostNegSugar: "unless" "post" Expr
	;
	
syntax Attribute
	= Attr: QualifiedName "(" {Expr ","}* ")"
	| Attr: QualifiedName
	;
	
	
syntax StatDeclarative
	= ProcClause: "procedure" OperationName ParamList
	| AxiomClauseSugar: "axiom" OperationName ParamList
	| AxiomClauseSugar: "theorem" OperationName ParamList
	| AxiomClauseSugar: "proof" OperationName ParamList
	| TestClauseSugar: "test" OperationName ParamList
	;

syntax ExprDeclarative
	= FunClause: "function" OperationName ParamList ":" Type
	| PredClauseSugar: "predicate" OperationName ParamList
	| VarClause: "var" Name ":" Type
	;

syntax MetaDeclarative
	= MetaClause: "def" OperationName ParamList ":" Type
	;
	
syntax TypeDeclarative
	= TypeClause: "type" Name
	| LiteralClause: "literal" Name "(" {Expr ","}* ")"
	;

syntax AlertDeclarative
	= AlertClause:      "alert" Name "\<:" {QualifiedName ","}+
	| AlertClauseSugar: "alert" Name
	;

syntax ParamList
	= ParamList: "(" {Param ","}* ")"
	;

syntax Param
	= Param: ParamMode Name ":" Type
	| ParamSugar: Name ":" Type
	| ParamSugar: Type
	| AnonParam
	;

syntax AnonParamList
	= ParamList: "(" {AnonParam ","}* ")"
	;

syntax AnonParam
	= AnonParam: ParamMode Type
	;
	
syntax ParamMode
	= Obs: "obs"
    | Upd: "upd"
    | Out: "out"
    | Exp: "exp"
    | Giv: "giv"
    | Del: "del"
    | Nrm: "nrm"
	;


// TYPES

syntax Type
	= Type: QualifiedName
	| Tuple: "(" {Type ","}* ")"
	| Struct: "struct" "{" TypeField* "}"
	| Unknown: "?"
	| SpecialType: SpecialType
	| Expect:	"\\expect" "(" StringConstant "," Type ")"
	| InstantiatedType: QualifiedName "[" {RenameExpr ","}* "]"
	;
	
syntax SpecialType
	= Predicate: "\\predicate"
	| Axiom: "\\axiom"
	| Test: "\\test"
	| Expr: "\\expr" "[" Type "]"
	| Decl: "\\decl"
	| DeclList: "\\decllist"
	| Renaming: "\\renaming" 
	| Literal: "\\literal"
	;

syntax TypeField
	= TypeField: LetClause ";"
	;
	
// EXPRESSIONS


syntax Expr
	= @selectable Var: OperationName
	| @selectable Literal: IntegerLiteral
	| @selectable Literal: RealLiteral
	| @selectable Literal: StringLiteral
	| @selectable Literal: IntegerLiteral ":" Type
	| @selectable Literal: RealLiteral ":" Type
	| @selectable Literal: StringLiteral ":" Type
	| @selectable Tuple: "(" {Expr ","}* ")"
	| @selectable TypeExpr: ":" Type
	| @selectable Struct: Type "{" {InitSpec ","}* "}"
	| @selectable Apply: Fun "(" {Expr ","}* ")"
	| Expect:	"\\expect" "(" StringConstant "," Expr ")"
	| @selectable IfThenElseExpr: "if" Expr "then" Expr "else" Expr "end"
	| @selectable Guard: "guard" Expr "by" Expr "end"
	| @selectable BlockExprSugar1: "{" Stat* "}"
	| @selectable BlockExprSugar2: "begin" "{" Stat* Expr "}"
	> @selectable DotOp: Expr "." Name
	| @selectable DotOp: Expr "." DecimalIntegerLiteral
	| @selectable IndexSugar: Expr "[" {Expr ","}* "]"
	> @selectable ListCons: "[" {Expr ","}* "]"
	> @selectable PreOpSugar: NEGOP Expr
	| @selectable PreOpSugar: BTNOT Expr
	| @selectable PreOpSugar: LGNOT Expr
	> @selectable left BinOpSugar: Expr MULOP Expr
	> @selectable left BinOpSugar: Expr ADDOP Expr
	> @selectable left BinOpSugar: Expr SHFOP Expr
	> @selectable left BinOpSugar: Expr RNGOP Expr
	> left (
		@selectable BinOpSugar: Expr CMPOP Expr
	  | @selectable InSugar: Expr INOP Expr
	  | @selectable NotInSugar: Expr NOTINOP Expr
	  | @selectable BinOpSugar: Expr EQUOP Expr
	)
	> @selectable left BinOpSugar: Expr SUBST Expr
	> @selectable left BinOpSugar: Expr BTAND Expr
	> @selectable left BinOpSugar: Expr BTAND Expr
	> @selectable left BinOpSugar: Expr BTXOR Expr
	> @selectable left BinOpSugar: Expr BITOR Expr
	> @selectable left BinOpSugar: Expr LGAND Expr
	> @selectable left BinOpSugar: Expr LOGOR Expr
	> @selectable left BinOpSugar: Expr LGIMP Expr
	> @selectable left Recover: Expr "recover" Expr
	> @selectable Undefined: UNDERSCORE
	;
syntax Fun
	= Fun: QualifiedOperationName
	| FunOf: "\\f" "(" Fun "," DecimalIntegerLiteral ")"
	| InstantiatedFun: QualifiedOperationName "[" {RenameExpr ","}* "]"
	;
	
syntax InitSpec
	= Field: Name "=" Expr
	;

lexical INOP = "in";
syntax NOTINOP = "not" "in";
lexical DOTOP = ".";
lexical MULOP = [*/%] | "**";
lexical ADDOP = [\-+] |[\-+][\-+];
lexical SHFOP = [\<][\<] | [\>][\>];
lexical RNGOP = [.][.];
lexical CMPOP = [\<] | [\>] | [\<][=] | [\>][=];
lexical SUBST = [=][=] | [!][=] | [\-][\>];
lexical EQUOP = [=][=][=] | [!][=][=];
lexical BTAND = [&];
lexical BTXOR = [^];
lexical BITOR = [|];
lexical LGAND = [&][&];
lexical LOGOR = [|][|];
lexical LGIMP = [=][\>] | [\<][=][\>];
lexical LGNOT = [!];
lexical NEGOP = [+] | [\-];
lexical BTNOT = [~];
lexical UNOP = NEGOP |  LGNOT |  BTNOT;
lexical BINOP 
	= DOTOP | MULOP | ADDOP | SHFOP | RNGOP | CMPOP
	| EQUOP | BTAND | BTXOR | BITOR | LGAND
	| LOGOR | LGIMP | SUBST
	;
	
lexical UNDERSCORE = [+\-*/%\<\>.=!&^|~] !<< [_] !>> [+\-*/%\<\>.=!&^|~];
// STATEMENTS


syntax Stat
	= BlockStat
	| Nop:		";"
	| Nop:		"skip" ";"
	| If:		"if" Expr "then" Stat* "else" Stat* "end"
	| IfSugar:	"if" Expr "then" Stat* "end"
	| While:	"while" Expr "do" Stat* "end"
	| For:		"for" Identifier "in" Expr "do" Stat* "end"
	| Guard:	"guard" Stat* "by" Expr "end"
	| Call:		"call" Proc "(" {Expr ","}* ")" ";"
	| Yield:	"yield" Expr ";"
	| Break:	"break" ";"
	| Return:	"return" ";"
	| Return:	"return" Expr ";"
	| Let:		"let" LetClause "in" Stat* "end"
	| Assert:	"assert" Expr AssertClause* ";"
	| Assume:	"assume" Expr AssertClause* ";"
	| LetSugar:	LetClause ";"
	| Expect:	"\\expect" "(" StringConstant "," Stat ")"
	> Assign:   Expr "=" Expr ";"
	;
	
syntax BlockStat
	= Block:    "{" Stat* "}"
	;
	
syntax Proc
	= Proc: 		QualifiedOperationName
	| ProcOf:		"\\p" "(" Proc ")"
	| InstantiatedProc: 	QualifiedOperationName "[" {RenameExpr ","}* "]"
	;
	
	
syntax LetClause 
	= VarDeclSugar: "var" NameIdentifier ":" Type "=" Expr
	| VarDeclSugarNoType: "var" Name "=" Expr
	| VarDeclSugarNoExpr: "var" Name ":" Type
	| ValDeclSugar: "val" NameIdentifier ":" Type "=" Expr
	| ValDeclSugarNoType: "val" Name "=" Expr
	| VarDecl: ParamMode NameIdentifier ":" Type "=" Expr
	;
	
syntax AssertClause
	= By:   "by" Expr
	| By:   "by" "simplify" Expr
	| QED:	"qed"
	| Src:	"src" "{" BraceBalancedString "}"
	| Doc:	"doc" "{" BraceBalancedString "}"
	;
	
// NAMES

	
syntax Name
	= NameIdentifier
	;

syntax QualifiedName
	= QualifiedNameIdentifier
	| FunName: "function" QualifiedNameIdentifier ParamList ":" Type
	| ProcName: "procedure" QualifiedNameIdentifier ParamList
	| TypeName: "type" QualifiedNameIdentifier
	| AlertName: "alert" QualifiedNameIdentifier
	| VarName: "var" QualifiedNameIdentifier ":" Type
	;
	
syntax OperationName
	= NameOrOperatorIdentifier
	;
syntax NameOrOperatorIdentifier
	= OperatorIdentifier
	| NameIdentifier
	;
	
syntax QualifiedNameIdentifier
	= QName: {NameIdentifier "."}+ !>> "." ;

syntax QualifiedOperationName
	= OperationName
	> QNameSugar: {Name "."}+ "." OperationName
	!>> "." ;

syntax NameIdentifier
	= Name: NameIdentifierLex
	| IdName: NameIdentifierLex "«" UniqueIdentifierLex "»"
	;

syntax OperatorIdentifier
	= Name: OperatorIdentifierLex
	| IdName: OperatorIdentifierLex "«" UniqueIdentifierLex "»"
	;
	
lexical UniqueIdentifierLex
	= 
	[0-9 A-F a-f \-] !<<
	[0-9 A-F a-f \-]+
	!>> [0-9 A-F a-f \-]
	;
		
lexical NameIdentifierLex
	= 
	([0-9 A-Z _ a-z $#] !<<
	[A-Z _ a-z] [0-9 A-Z _ a-z]* [$#]? 
	!>> [0-9 A-Z _ a-z $#])
	\ MagnoliaReservedKeywords
	; 


syntax LanguageIdentifier
	= Lang: LanguageIdentifierLex
	;
	
lexical LanguageIdentifierLex
	= 
	([0-9 A-Z _ a-z $#] !<<
	[A-Z _ a-z] ![ \t \n \r \  ,]* 
	!>> ![ \t \n \r \  ,])
	; 

lexical OperatorIdentifierLex
	= "_" BINOP "_"
	| UNOP "_"
	| "_[_]"
	| "_[_]=_"
	| "_(_)"
	;

// LITERALS
	
lexical OctalIntegerLiteral
	=  [0-9] !<< [0] [0-7]+ !>> [0-9 A-Z _ a-z] ;

lexical HexIntegerLiteral
	=  [0-9] !<< [0] [X x] [0-9 A-F a-f]+ !>> [0-9 A-Z _ a-z] ;

lexical DecimalIntegerLiteral
	= [0] !>> [0-9 A-Z _ a-z]
	| [1-9] [0-9]* !>> [0-9 A-Z _ a-z] ;

lexical IntegerLiteral
	= DecimalIntegerLiteral: DecimalIntegerLiteral decimal 
	| HexIntegerLiteral: HexIntegerLiteral hex 
	| OctalIntegerLiteral: OctalIntegerLiteral octal
	;

lexical RealLiteral
	= [0-9]+ [D F d f] 
	| [0-9]+ [E e] [+ \-]? [0-9]+ [D F d f]?
	| [0-9]+ "." [0-9]* [D F d f]? 
	| [0-9]+ "." [0-9]* [E e] [+ \-]? [0-9]+ [D F d f]? 
	| [.] !<< "." [0-9]+ [D F d f]? 
	| [.] !<< "." [0-9]+ [E e] [+ \-]? [0-9]+ [D F d f]? 
	;
syntax StringLiteral
	= StringLiteral: StringConstant string;
	
lexical StringConstant
	= @category="Constant" "\"" StringCharacter* "\"" ;

lexical StringCharacter
	= "\\" [\" \' \< \> \\ b f n r t] 
	| UnicodeEscape 
	| OctalEscapeSequence 
	| ![\" \' \< \> \\] ;

lexical OctalEscapeSequence
	= "\\" [0-7] [0-7] !<< [0-7]
	| "\\" [0-3] [0-7] [0-7] 
	|  "\\" [0-7] !<< [0-7] ;

lexical UnicodeEscape
	= "\\" [u]+ [0-9 A-F a-f] [0-9 A-F a-f] [0-9 A-F a-f] [0-9 A-F a-f] ;

lexical BraceBalancedString
	= ([{] LAYOUTLIST) << ![ { } ]* >> [}]
	;
// LAYOUT

layout LAYOUTLIST
	= LAYOUT* !>> [\t-\n \r \ ] !>> "//" !>> "/*";

lexical LAYOUT
	= NestedComment
	> Comment 
	| whitespace: [\t-\n \r \ ] ;


lexical Comment
	= @category="Comment" "/*" !>> [\<] (![*] | [*] !>> [/])* "*/" 
	| @category="Comment" "//" ![\n]* [\n]
	;

lexical NestedComment
	= @category="Comment" "/*\<" (![/\>] | [\>] !>> "*/" | [/] !>> "*\<" | NestedComment)* "\>*/"
	;
	
keyword MagnoliaReservedKeywords
	= "var" | "type" | "function" | "procedure" | "_";

//anno str Tree@deprecated;

