/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::lib::CollectAxioms
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import Prelude;

data Axiom 
	= Axiom(str name, list[AST] args, AST expr)
	;	

data AST
	= True()
	| False()
	| Not(AST rhs)
	| Or(AST lhs, AST rhs)
	| And(AST lhs, AST rhs)
	| Implies(AST lhs, AST rhs)
	| Iff(AST lhs, AST rhs)
	| Equal(AST lhs, AST rhs)
	;

public Result[tuple[list[AST], list[Axiom]]] collectAxioms(AST tree) {
	list[AST] defs = [];
	list[Axiom] axioms = [];
	env = minimalEnv();
	
	tree = visit(tree) {
		case Equal(Fun(Name(leaf("_==_"))), x, y) => Equal(x, y)
		case Equal(Fun(Name(leaf("_!=_"))), x, y) => Not(Equal(x, y))
	}
	for(/d:Define(cls, attrs, body) <- tree) {
		if(hasModifier(d, Axiom()) && FunClause(name, params, _) := cls) {
			args = paramsToArgs(params);
			nameStr = shortNameToStr(name);
			set[Axiom] asserts = {};  
			if(Body(BlockExpr(seq(ss), rt)) := body) {
				for(s <- ss) {
					asserts = asserts + decodeAxiom(nameStr, args, True(), s);
				}
			}
			else {
				termPrintln("Don\'t know how to deal with axiom body: ", body);
			}
			axioms = axioms + toList(asserts);
		}
		else if(!hasModifier(d, External()) && !hasModifier(d, Derived()) && getDefBody(d) != EmptyBody() && !(TypeClause(_) := cls)) {
			defs = defs + d;
		}
	}
	return makeResult(<defs, axioms>, env);
}


default set[Axiom] decodeAxiom(str name, list[AST] args, AST cond, AST ax) {
	termPrintln("Unknown statement in axiom: ", ax);
	return {};
}

set[Axiom] decodeAxiom(str name, list[AST] args, AST cond, Assert(expr, clauses)) {
	if(cond == True()) {
		return {Axiom(name, args, decodeAssert(expr))};
			}
	else {
		return {Axiom(name, args, Implies(cond, decodeAssert(expr)))};
	}
}

AST decodeAssert(Equal(Fun(Name(leaf("_==_"))), x, y)) {
	if(getExprType(x) == SpecialType(Predicate()) && getExprType(y) == SpecialType(Predicate()))
		return Iff(decodeAssert(x), decodeAssert(y));
	else
		return Equal(x, y);
}

AST decodeAssert(Equal(Fun(Name(leaf("_!=_"))), x, y)) {
	if(getExprType(x) == SpecialType(Predicate()) && getExprType(y) == SpecialType(Predicate()))
		return Not(Iff(decodeAssert(x), decodeAssert(y)));
	else
		return Not(Equal(x, y));
}

AST decodeAssert(Apply(Builtin(_, leaf("_||_(\\predicate,\\predicate)")), seq([x, y]))) {
	return Or(decodeAssert(x), decodeAssert(y));
}

AST decodeAssert(Apply(Builtin(_, leaf("_&&_(\\predicate,\\predicate)")), seq([x, y]))) {
	return And(decodeAssert(x), decodeAssert(y));
}

AST decodeAssert(Apply(Builtin(_, leaf("_=\>_(\\predicate,\\predicate)")), seq([x, y]))) {
	return Implies(decodeAssert(x), decodeAssert(y));
}

AST decodeAssert(Apply(Builtin(_, leaf("_\<=\>_(\\predicate,\\predicate)")), seq([x, y]))) {
	return Iff(decodeAssert(x), decodeAssert(y));
}

AST decodeAssert(Apply(Builtin(_, leaf("!_(\\predicate)")), seq([x]))) {
	return Not(decodeAssert(x));
}

AST decodeAssert(Apply(Builtin(_, leaf("TRUE()")), seq([]))) {
	return True();
}

AST decodeAssert(Apply(Builtin(_, leaf("FALSE()")), seq([]))) {
	return False();
}

default AST decodeAssert(AST tree) {
	return tree;
}

set[Axiom] decodeAxiom(str name, list[AST] args, AST cond, stmt:If(c, seq(ts), seq(es))) {
	set[Axiom] result = {};
	if(cond != True()) {
		termPrintln("Nested if in axioms not supported yet", stmt);
	}
	
	for(s <- ts) {
		result = result + decodeAxiom(name, args, cond, s);
	}
	for(s <- es) {
		termPrintln("else clause in axiom conditional not supported yet: ", s);
		// result = result + decodeAxiom(name, args, cond, s);
	}
	return result;
}

set[Axiom] decodeAxiom(str name, list[AST] args, AST cond, Return(_)) {
	return {};
}

