/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::lib::CxxAPI
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::backend::OperatorInfo;
import org::magnolialang::magnolia::backend::CxxPP;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import Node;

public str makeCxxName(str qual, AST def) {
	name = makeCxxName(def);
	if(name notin reservedKeywords)
		return "<qual>::<name>";
	else
		return name;
}

public str makeJava(str qual, AST def) {
	name = makeJavaName(def);
	if(name notin reservedKeywords)
		return "<qual>.<name>";
	else
		return name;
}

public str makeCxxName(AST def) {
	defName = defNameOf(def);

	if(defName == SpecialType(Predicate()))
		return "bool";
	
	uname = shortNameToStr(unIdName(defName));
	if(uname in opMapCxx) {
		info = opMapCxx[uname];
		if(/:/ := info.opName)
			return info.opName;
		else
			defName = transformName(defName, str(str s) { return "op_<info.opName>";});
	}

	name = shortNameToStr(defName);
	name = replaceChars(name);
	types = typesInName(unqualify(defName)) ? [];
		
	
	if(types != []) {
		for(t <- types) {
			name += "_<ppName(t)>";
		}
	}	

	name = "<name><declSuffix(def)>";

	return name;
}

public str makeJava(AST def) {
	defName = defNameOf(def);
	if(defName == SpecialType(Predicate()))
		return "boolean";
		
	name = shortNameToStr(defName);
	name = replaceChars(name);

	types = typesInName(unqualify(defName));
		
	if(name in opMapJava) {
		info = opMapJava[name];
		if(/:/ := info.opName)
			return info.opName;
		else
			name = "op_<info.opName>";
	}
	
	if(types != []) {
		for(t <- types) {
			name += "_<ppName(t)>";
		}
	}	

	name = "<name><declSuffix(def)>";

	return name;
}

public str makeCxxNonKeywordName(AST def, str suf) {
	defName = defNameOf(def);
	if(defName == SpecialType(Predicate()))
		return "bool";

	name = shortNameToStr(defName);
	name = replaceChars(name);

	if(name in reservedKeywords)
		name = "<name><suf>";
	return name;
}

public str makeCxxNonKeywordName(str name, str suf) {
	if(name == "\\predicate")
		return "bool";
		
	if(name in reservedKeywords)
		name = "<name><suf>";
	return name;
}

public str makeJavaNonKeywordName(AST def, str suf) {
	defName = defNameOf(def);
	if(defName == SpecialType(Predicate()))
		return "boolean";

	name = shortNameToStr(defName);
	name = replaceChars(name);

	if(name in reservedJavaKeywords)
		name = "<name><suf>";
	return name;
}

public str makeJavaNonKeywordName(str name, str suf) {
	if(name == "\\predicate")
		return "boolean";
		
	if(name in reservedJavaKeywords)
		name = "<name><suf>";
	return name;
}
data Env; // no idea why this this is needed, since we import Environment; Rascal bug?

public str funSuffix = "_f";
public str procSuffix = "_p";
public str typeSuffix = "_t";

public AST cxxLanguage = leaf("C++");
public AST javaLanguage = leaf("Java");

public bool isCxxLanguage(AST language) {
	return Lang(leaf(/^[CcGgHh](\+\+|xx|pp)[0-9]*/)) := language;
}

public bool isJavaLanguage(AST language) {
	return Lang(leaf(/^[Jj]ava[0-9]*/)) := language;
}

public str getLanguageName(Lang(leaf(l))) {
	return l;
}

public AST checkCxxLanguage(AST language, Env env) {
	if(Lang(leaf(/^<cc:[CcGgHh]><plus:\+\+|xx|pp><version:[0-9]*>/)) := language) {
		if(plus != "++" || /[Hh]/ := cc) {
			env = errorMark(env, "Style: use language \'C++<version>\' instead of \'<nameToStr(language)>\'", Info(), locOf(language));
			plus = "++";
			cc = "C";
		}
		if(version == "" || version == "2011") {
			version = "11";
		}
		if(version != "11") {
			env = errorMark(env, "Only C++11 supported", Warning(), locOf(language));
		}
		return preserveAnnos(Lang(leaf("<cc><plus><version>")), language);
	}
}

public str declSuffix(AST decl) {
	if(isFunctionDecl(decl))
		return funSuffix;
	else if(isProcedureDecl(decl))
		return procSuffix;
	else if(isTypeDecl(decl))
		return typeSuffix;
	else if(isAlertDecl(decl))
		return "_a";
	else
		return "";
}

public str declSuffix(str kind) {
	switch(kind) {
		case "function":
			return funSuffix;
		case "procedure":
			return procSuffix;
		case "type":
			return typeSuffix;
		default:
			return "";
	}
}

private list[AST] typesInName(AST name) {
	list[AST] types = [t | AnonParam(_,t) <- paramsOf(name)];
	if(isFunctionDecl(name)) {
		types = types + [returnTypeOf(name)];
	}
	return types;
}

public str replaceChars(str name) {
	return visit(name) {
	case /«/ => "__"
	case /»/ => "_"
	case /-/ => "_"
	case /\+/ => "plus"
	case /\*/ => "times"
	case /\// => "slash"
	case /=/  => "eq"
	case /!/  => "not"
	case /&/  => "and"
	case /\|/ => "or"
	case /\^/ => "hat"
	case /\[/ => "lbracket"
	case /\]/ => "rbracket"
	case /%/  => "percent"
	case /~/  => "tilde"
	case /\</ => "lt"
	case /\>/ => "gt"
	};
}
