/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 Tero Hasu
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
module org::magnolialang::magnolia::lib::Environment
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::magnolia::evaluator::Values;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import List;
import util::Benchmark;

public bool DEBUG_OVERLOAD = false;
public bool DEBUG_TYPECHECK = false;
public bool DEBUG_INSTANCES = false;

anno AST AST@\type;

data Env
	= NilEnv() 
	| Env(
		AST scName,
		AST modName,
		rel[AST,NameKind,AST] names,
		map[AST,DefInfo] defs,
		map[AST,int] bindings,
		list[Value] stack,
		int stackTop,
		set[tuple[AST,NameKind]] thisScopeDefs,
		set[ErrorMark] errors,
		set[ErrorMark] warnings,
		SourceInfo info,
		list[tuple[str,str,int]] workStack,
		list[ErrorLoc] errorLocStack,
		list[tuple[str,int,list[str]]] expectedErrors, // regexp, number of errors expected
		Env next
	)
	| MinimalEnv(
		set[ErrorMark] errors,
		set[ErrorMark] warnings,
		SourceInfo info,
		list[tuple[str,str,int]] workStack,
		list[ErrorLoc] errorLocStack,
		list[tuple[str,int,list[str]]] expectedErrors // regexp, number of errors expected
	)
	;

data ErrorLoc = ErrorLoc(loc location, str errorPrefix, str warningPrefix);

data SourceInfo
	= SourceInfo(
		str currentFile,
		XRefs tokenRefs)
	| SourceInfo()
	;

data NameKind = ExprKind() | StatKind() | TypeKind() | TopKind() | ValKind() | AlertKind();

public Env minimalEnv() {
	return MinimalEnv({}, {}, SourceInfo(), [], [], []);
}

public Env minimalEnv(loc srcFile) {
	return MinimalEnv({}, {}, SourceInfo(srcFile.uri, {}), [], [], []);
}


public Env minimalEnv(loc srcFile, Result[&T] result) {
	return MinimalEnv(result.errors, result.warnings, SourceInfo(srcFile.uri, {}), [], []);
}

public Env newEnv(e:MinimalEnv(errors, warnings, info), AST scName, AST modName) {
	env = Env(scName, modName, {}, (), (), [], 0, {}, errors, warnings, info, [], [], [], e);
	return env; 
}

public Env newEnv(e:MinimalEnv(errors, warnings, _), AST scName, AST modName, loc srcFile) {
	env = Env(scName, modName, {}, (), (), [], 0, {}, errors, warnings, SourceInfo(srcFile.uri, {}), [], [], [], e);
	return env; 
}

public Env newEnv(AST scName, AST modName) {
	env = Env(scName, modName, {}, (), (), [], 0, {}, {}, {}, SourceInfo(), [], [], [], NilEnv());
	return env; 
}

public Env newEnv(AST scName, AST modName, loc srcFile) {
	env = Env(scName, modName, {}, (), (), [], 0, {}, {}, {}, SourceInfo(srcFile.uri, {}), [], [], [], NilEnv());
	return env; 
}


public Env enterScope(AST scName, Env env) {
	return Env(scName, env.modName, env.names, env.defs, env.bindings, env.stack, env.stackTop, {}, env.errors, env.warnings, env.info, env.workStack, env.errorLocStack, env.expectedErrors, env);
}

public Env enterScope(AST scName, loc srcFile, Env env) {
	return Env(scName, env.modName, env.names, env.defs, env.bindings, env.stack, env.stackTop, {}, env.errors, env.warnings, SourceInfo(srcFile.uri, {}), env.workStack, env.errorLocStack, env.expectedErrors, env);
}

public Env enterScope(Env env) {
	return Env(env.scName, env.modName, env.names, env.defs, env.bindings, env.stack, env.stackTop, {}, env.errors, env.warnings, env.info, env.workStack, env.errorLocStack, env.expectedErrors, env);
}

public Env enterNewScope(Env env) {
	// TODO: figure out a way to make this work so that it'll drop the toplevel stuff, but keep the algo-level stuff
	return Env(env.scName, env.modName, {}, (), (), env.stack, env.stackTop, {}, env.errors, env.warnings, env.info, env.workStack, env.errorLocStack, env.expectedErrors, env);
}

public Env exitScope(Env env) {
	Env result;
	
	if(SourceInfo() := env.info)
		result = env.next[defs = env.defs][errors = env.errors][warnings = env.warnings][stack = env.stack][workStack = env.workStack][expectedErrors = env.expectedErrors];
	else
		result = env.next[defs = env.defs][info = env.info][errors = env.errors][warnings = env.warnings][stack = env.stack][workStack = env.workStack][expectedErrors = env.expectedErrors];
		
	return result;
}

public Env exitDefScope(Env env) {
	Env result;
	
	if(SourceInfo() := env.info)
		result = env.next[errors = env.errors][warnings = env.warnings][stack = env.stack][workStack = env.workStack][expectedErrors = env.expectedErrors];
	else
		result = env.next[info = env.info][errors = env.errors][warnings = env.warnings][stack = env.stack][workStack = env.workStack][expectedErrors = env.expectedErrors];
		
	return result;
}

public Env joinEnv(list[Env] envs) {
	env = head(envs);
	envs = tail(envs);
	for(e <- envs)
		env = joinEnv(env, e);
}

public Env joinEnv(Env e1, Env e2) {
	if(e1.scName != e2.scName 
	   || e1.modName != e2.modName || e1.next != e2.next
	   || e1.stack != e2.stack || e1.stackTop != e2.stackTop  || e1.workStack != e2.workStack
	   || e1.errorLocStack != e2.errorLocStack
	   || e1.expectedErrors != e2.expectedErrors)
	     throw "Internal Compiler Error: trying to join two incompatible environments";

	info = e1.info;
	if(SourceInfo() := e1.info)
		info = e2.info;
	else if(!(SourceInfo() := e2.info)) {
		if(e1.info.currentFile != e2.info.currentFile)
	     throw "Internal Compiler Error: trying to join two incompatible environments";
		info[tokenRefs = info.tokenRefs + e2.info.tokenRefs];
	}
	return Env(e1.scName, e1.modName,
	           e1.names + e2.names,
	           e1.defs  + e2.defs,
	           e1.bindings + e2.bindings,
	           e1.stack, e1.stackTop,
	           e1.thisScopeDefs + e2.thisScopeDefs,
	           e1.errors + e2.errors,
	           e1.warnings + e2.warnings,
	           info,
	           e1.workStack,
	           e1.errorLocStack,
	           e1.expectedErrors,
	           e1.next);
}

public Env addName(AST name, AST qName, NameKind kind, Env env) {
	return addName(name, qName, kind, true, env);
}

public Env addName(AST name, AST qName, NameKind kind, bool checkForRedef, Env env) {
//	if(checkForRedef && <qName, kind> in env.thisScopeDefs) {
//		env = errorMark(env, "Redefinition of \'<dispNameOf(qName, env)>\'", locOf(qName));
//	}
//	else {
		entry = <basename(name),kind,qName>;
		if(entry in env.names)  {
			env.names -= {entry}; // so that we can override with the correct location anno
		}
		env.names += {entry};
		env.thisScopeDefs[qName] = kind;
//	}
	return env;
}

public Env addDef(AST qName, DefInfo def, Env env) {
	if(qName in env.defs) {
		termPrintln("Existing def: ", env.defs[qName].attrs);
		termPrintln("New def:      ", def.attrs);
		env = errorMark(env, "Symbol \'<dispNameOf(qName, env)>\' already defined", locOf(qName));
		return env;
	}
	else {
		env.defs[qName] = def;
		return env;
	}
}

public Env replaceDef(AST qName, DefInfo def, Env env) {
	env.defs[qName] = def;
	return env;
}

public set[AST] lookupName(AST name, NameKind kind, Env env) {
	if(name in env.defs)
		return {name};
		
	result = env.names[basename(name)][kind];

	return {preserveAnnos(r, name) | r <- result};
}

public set[AST] lookupNameNoError(AST name, NameKind kind, Env env) {
	if(name in env.defs)
		return {name};
	result = env.names[basename(name)][kind];
	return {preserveAnnos(r, name) | r <- result};
}

str nameKind(NameKind kind) {
	switch(kind) {
		case ExprKind():
			return "expression";
		case StatKind():
			return "statement";
		case TypeKind():
			return "type";
		case TopKind():
			return "top";
	}
}
public DefInfo lookupDef(AST qName, Env env) {
	return env.defs[qName];
}

public bool hasDef(AST qName, Env env) {
	return qName in env.defs;
}

public AST qualify(AST name, Env env) {
	return qualify(env.scName, name);
}

public str dispNameOf(AST tree, Env env) {
	return dispNameOf(unprefix(defNameOf(tree), env.modName));
}

public str defNameOf(AST tree, Env env) {
	return nameToStr(unprefix(defNameOf(tree), env.modName));
}

public Env addTokenRef(loc location, Ref ref, Env env) {
	info = env.info;
	if(!(SourceInfo() := info) && location.uri == info.currentFile) {
		if(location.offset >= 0 && location.length > 0) {
			info.tokenRefs[location] = ref;
			return env[info = info];
		}
	}
	
	return env;
}

public Env addTokenRef(AST tree, Ref ref, Env env) {
	if(!(SourceInfo() := env.info) && (tree@\loc)?)
		return addTokenRef(tree@\loc, ref, env);
	else
		return env;
}

public Env addTokenRef(AST tree, AST ref, Env env) {
	if(!(SourceInfo() := env.info) && (tree@\loc)?) {
		if(ref in env.defs) {
			info = env.defs[ref];
			if(Builtin() in info.attrs)
				return addTokenRef(tree@\loc, LocalRef(info), env);
			else
				return addTokenRef(tree@\loc, LocalRef(info), env);
				// TODO: return addTokenRef(tree@\loc, Ref(ref), env);
		}
		else {
			return addTokenRef(tree@\loc, LocalRef(definfo()[qName = ref][defName = ref][decl = ref][parent = env.scName][locInfo = LocInfo(locOf(ref))]), env);
		}
	}
	else
		return env;
}

public DefInfo makeDefInfo(AST name, Env env) {
	//nameNoAnno = delAnnotationsRec(name);
	nameNoAnno = locOnly(name);
	return definfo()[qName = nameNoAnno][defName = nameNoAnno][decl = nameNoAnno][parent = locOnly(env.scName)][locInfo = LocInfo(locOf(name))][srcLoc = locOf(name)];
}

public bool hasRefs(Env env) {
	return !(SourceInfo() := env.info);
}


public Env errorMark(Env env, str msg, loc location) {
	return errorMark(env, msg, defaultSeverity, location, true);
} 

public Env errorMark(Env env, str msg, Severity severity, loc location) {
	return errorMark(env, msg, severity, location, true);
}

public Env errorMark(Env env, str msg, Severity severity, loc location, bool checkExpected) {
	println("Error mark: <severity>, <msg>, <location>");
	if(env.errorLocStack != []) {
		el = env.errorLocStack[0]; // we can go through the stack if we want a full trace
		if(severity == Warning() || severity == Info())
			msg = el.warningPrefix + msg;
		else
			msg = el.errorPrefix + msg;
		msg += "\n(at <locInfo(location, el.location)>)";
		location = el.location;
	}
	if(checkExpected) {
		for(i <- [size(env.expectedErrors)-1 .. -1]) {
			<ex,nEx,msgs> = env.expectedErrors[i];
			if(size(msgs) < nEx && /<ex>/ := msg) {
				env.expectedErrors[i] = <ex, nEx, msgs + [msg]>;
				termPrintln("Found expected error: pattern: ", ex, " message: ", msg);
				return env;
			}
		}
	}
	if(severity == Warning() || severity == Info())
		env.warnings += {ErrorMark(msg, severity, location)};
	else
		env.errors += {ErrorMark(msg, severity, location)};
	return env;
}

public Env errorMarkInDecl(Env env, str msg, Severity severity, AST def) {
	defLoc = locOf(def);
	<via, ognLoc> = getLatestOrigin(def, defLoc);
	if(ognLoc != defLoc) {
		msg += " (original at <locInfo(defLoc, ognLoc)><via == "" ? "" : ", via <via>">)";
	}

	return errorMark(env, msg, severity, ognLoc, true);
}


public bool hasErrors(Env env) {
	return env.errors != {};
}

public bool hasMarks(Env env) {
	return env.errors != {} || env.warnings != {};
}

public set[ErrorMark] getMarks(Env env) {
	return env.errors + env.warnings;
}

public set[ErrorMark] getErrors(Env env) {
	return env.errors;
}

public Result[&T] makeResult(&T val, Env env) {
	return Result(val, getRefs(env), env.errors, env. warnings);
}

public Result[&T] makeResult(&T val, XRefs refs, Env env) {
	return Result(val, refs + getRefs(env), env.errors, env. warnings);
}

public Result[void] makeNoResult(str message, Env env) {
	return NoResult(message, env.errors, env. warnings);
}

public Env addVar(AST name, Env env) {
	if(size(env.stack) == env.stackTop)
		env.stack += [VoidValue(), VoidValue(), VoidValue(), VoidValue(), VoidValue(), VoidValue(), VoidValue(), VoidValue()];
	env.bindings[name] = env.stackTop;
	env.stackTop = env.stackTop + 1;
	return env; 
}

public Env setVar(AST name, Value val, Env env) {
	l = env.bindings[name];
	env.stack[l] = val;
	return env;
}

public Value getVar(AST name, Env env) {
	l = env.bindings[name];
	return env.stack[l];
}

public void dumpEnvDefs(Env env) {
	termPrintln("Environment definitions: {");
	for(qName <- env.defs) {
		termPrintln("    ", qName, ": ", env.defs[qName].body);
	}
	termPrintln("}");
}

public XRefs getRefs(Env env) {
	if(hasRefs(env))
		return env.info.tokenRefs;
	else
		return {};
}

///////////// WORK

public Env startWork(str message, str workUnit, Env env, bool enabled) {
	if(!enabled)
		message = "";
	env.workStack = [<message, workUnit, getMilliTime()>, *env.workStack];
	
	return env;
}

public Env startWork(str message, str workUnit, Env env) {
	return startWork(message, workUnit, env, true);
}

public Env endWork(Env env) {
	return endWork("FINISHED", env);
}

public Env endWork(str msg, Env env) {
	<message, unit, time> = head(env.workStack);
	env.workStack = tail(env.workStack);
	if(message != "")
		println("<message>: <msg> at <getMilliTime()-time> ms <unit != "" ? "(<unit>)" : "">");
	return env;
}

public Env markWork(str msg, Env env) {
	<message, unit, time> = head(env.workStack);
	if(message != "")
		println("<message>: <msg> at <getMilliTime()-time> ms <unit != "" ? "(<unit>)" : "">");
	return env;
}

////////////// ERROR LOCATION OVERRIDE
public Env pushErrorLoc(loc newLoc, str errorMsg, str warnMsg, Env env) {
	env.errorLocStack += [ErrorLoc(newLoc, errorMsg, warnMsg)];
	
	return env;
}

public Env popErrorLoc(Env env) {
	env.errorLocStack = delete(env.errorLocStack, size(env.errorLocStack)-1);
	
	return env;
}


/////////////// ERROR EXPECTS
public Env expectError(str regExp, int nExpected, Env env) {
	env.expectedErrors += [<regExp, nExpected, []>];
	
	return env;
}

public Env checkExpected(str regExp, loc location, Env env) {
	<ex, nExpected, msgList> = env.expectedErrors[size(env.expectedErrors)-1];
	env.expectedErrors = delete(env.expectedErrors, size(env.expectedErrors)-1);
	
	if(ex != regExp) {
		env = errorMark(env, "INTERNAL ERROR in checkExpected(): expected \"<regExp>\", was \"<ex>\"", location);
	}
	else {
		if(size(msgList) > 0) {
			plural = size(msgList) != 1 ? "s" : "";
			msgs = strJoin(["\"<m>\"" | m <- msgList], ", \n"); 
			env = errorMark(env, "Found <size(msgList)> expected error<plural>/warning<plural>:\<br\>\n<msgs>", Info(), location, false);
		}
		
		if(size(msgList) != nExpected) {
			plural = nExpected != 1 ? "s" : "";
			if(size(msgList) == 0)
				env = errorMark(env, "Expected <nExpected> error<plural>/warning<plural> matching \"<ex>\", got none", Error(), location, false);
			else
				env = errorMark(env, "Expected <nExpected> error<plural>/warning<plural> matching \"<ex>\", only got <size(msgList)>", Error(), location, false);
		}
	}
	
	return env;
}