/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
// run with rascal-src-with-magnolia

module org::magnolialang::magnolia::lib::Inliner

/*

  E.g.,

  (rimport "IO")
  (rimport "org::magnolialang::testutil::RascalPp")
  (rimport "org::magnolialang::magnolia::lib::Inliner")

  (run-inliner-tests)

  (rreload "org::magnolialang::magnolia::lib::Inliner")
  (rreload "org::magnolialang::magnolia::lib::Pretty")

 */

import org::magnolialang::compiler::Util;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Pretty;
import org::magnolialang::magnolia::typechecker::algo::Types;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::testutil::RascalExt;
import org::magnolialang::testutil::RascalPp;

import IO;
import List;
import Map;
import Node;
import Set;
import String;

data Exception
	= NoDefFound(str message, AST decl, loc location)
	| NoInlineableFound(int offset)
	| NoInlineableFound(int line, int column)
	;

// --------------------------------------------------
// 'inline', etc.
// --------------------------------------------------

// xxx These would not seem to be currently used anywhere, but 'inline' appears in TopLevelOps, in comments. Is this a handy API for compiler integration that should be implemented?

public &T inline(Env env, &T tree, set[AST] decls) {
	rel[AST, str, list[AST], AST] inlineDefs = {};
	for(d <- decls)
		inlineDefs += getInlineDefs(d);
	return applyInlining(env, tree, inlineDefs, ());
}

public rel[AST, str, list[AST], AST] getInlineDefs(AST decl) {
	switch(decl) {
		case Define(dcl,_,body0): {
			if(Body(body) := body0) {
				switch(dcl) {
					case FunClause(n,ParamList(seq(as)),t):
						return {<n, "fun", as, body>};
					case ProcClause(n,ParamList(seq(as))):
						return {<n, "proc", as, body>};
					case TypeClause(n):
						return {<n, "type", [], body>};
				}
			}
		}
		case DeclBody(seq(ds)): {
			return ({} | it + getInlineDefs(d) | d <- decl);
		}
	}
	return {};
}

data InlineInfo = inlineInfo(
		rel[AST, str, list[AST], AST] inlineDefs,
		map[AST, AST] renaming,
		set[AST] usedRenamings,
		set[AST] usedInlines,
		set[AST] foundDecls);
		
public &T applyInlining(Env env, &T tree, rel[AST, str, list[AST], AST] inlineDefs, map[AST, AST] renaming) {
	<result, info> = applyInliningInternal(env, tree, inlineInfo(inlineDefs, renaming, {}, {}, {})); 

	for(n <- info.inlineDefs<0>) {
		if(n notin info.foundDecls)
			env = errorMark(env, "Inlined operation <n> not declared in body\n", Warning(), locOf(n));
		if(n notin info.usedInlines)
			env = errorMark(env, "Inlined operation <n> not used in body\n", Warning(), locOf(n));
	}

	for(n <- domain(info.renaming)) {
		if(n notin info.usedRenamings)
			env = errorMark(env, "Renamed name <n> not used in body\n", Warning(), locOf(n));
	}
	
	return result;
}

tuple[&T, InlineInfo] applyInliningInternal(Env env, &T tree, InlineInfo info) {
	result = top-down-break visit(tree) {
		case app: Apply(Fun(funName), seq(args)): {
			<_, params, body> = overload(funName, "fun", info.inlineDefs, args);
			if(body != Nop()) {
				map[AST, AST] subst = ();
				<args0, info> = applyInliningInternal(env, seq(args), info);
				if(seq(inlinedArgs) := args0) {
					for(<Param(paramName, paramType), arg> <- [<params[i], inlinedArgs[i]> | i <- index(params)]) {
						termPrintln("applyInliningInternal: subst: ", paramName, " =\> ", arg);
						subst[paramName] = arg;
					}
					body = applyInlining(env, body, {}, subst);
					// println("<app> =\> <body>");
					info.usedInlines += {funName};
					insert preserveAnnos(body, app);
				}
					
			}
			else
				fail;
		}
		case call: Call(Proc(procName), seq(args)): {
			println("call <unparse(procName)>");
			<_, params, body> = overload(procName, "proc", info.inlineDefs, args);
			println("call <unparse(procName)>, <unparse(body)>");
			if(body != Nop()) {
				map[AST, AST] subst = ();
				<args0, info> = applyInliningInternal(env, seq(args), info);
				if(seq(inlinedArgs) := args0) {
					for(<Param(paramMode, paramName, paramType), arg> <- [<params[i], inlinedArgs[i]> | i <- index(params)]) {
						subst[paramName] = arg;
					}
					body = applyInlining(env, body, {}, subst);
					// println("<app> =\> <body>");
					info.usedInlines += {procName};
					insert preserveAnnos(body, call);
				}
					
			}
			else
				fail;
		}
		case typ: Type(typeName): {
			<n, _, body> = overload(typeName, "type", info.inlineDefs, []);
			if(body != Nop()) {
				// termPrintln(typ, " =\> ", body);
				info.usedInlines += {typeName};
				insert preserveAnnos(body, typ);
			}
			else
				fail;
		}
		case Define(FunClause(funName,ParamList(seq(args)),_),_,_): {
			<n, _, _> = overload(funName, "fun", info.inlineDefs, args);
			if(n != Nop()) {
				info.foundDecls += {funName};
				insert Nop();
			}
			else
				fail;
		}
		case Define(TypeClause(typeName),_,_): {
					<n, _, _> = overload(typeName, "type", info.inlineDefs, []);
			if(n != Nop()) {
				info.foundDecls += {typeName};
				insert Nop(); 
			}
			else
				fail;
		}
		case n: Name(_): {
			if(n in info.renaming) {
				if(!isName(info.renaming[n]))
					throw "applyInliningInternal: renaming: not a name <yieldTerm(info.renaming[n])>";
				info.usedRenamings += {n};
				insert info.renaming[n];
			}
			else
				fail;
		}
		case v: Var(n): {
			if(n in info.renaming) {
				info.usedRenamings += {n};
				insert info.renaming[n];
			}
			else
				fail;
		}
		case v: UserSyntax2(n): {
			if(n in info.renaming) {
				info.usedRenamings += {n};
				insert info.renaming[n];
			}
			else
				fail;
		}
	}
	return <result, info>;
}

public tuple[AST,list[AST],AST] overload(AST name, str kind, rel[AST,str,list[AST],AST] defs, list[AST] args) {
	result = <Nop(),[],Nop()>;
	cands = defs[name,kind];
	termPrintln("<name>:\n\t", cands);
	if(size(cands) > 0) {
		done = false;
		for(<params, body> <- cands, size(params) == size(args)) {
			if(done)
				println("Whoops: ambiguity in function call -- needs overload resolution:\n\t<unparse(funName)>(<unparse(args, ", ")>)");
			else
				result = <name, params, body>;	
		}
	}
	return result;
}

// --------------------------------------------------
// gensym
// --------------------------------------------------

str genIdSymK(str n, int k) = "<n> <k>";

str genIdSym0(str n) = genIdSymK(n, 0);

tuple[map[str, int], str] genIdSym(map[str, int] m, str n) {
  int k = 1;
  if (n in m)
    k = m[n] + 1;
  return <m + (n: k), genIdSymK(n, k)>;
}

tuple[str, int] getBaseNameWithNum(str basename) {
  int i = 0;
  if (/^<b:.*>_<n:\d+>$/ := basename) {
    basename = b;
    i = toInt(n);
  }
  return <basename, i>;
}

str makeBaseNameWithNum(str basename, int i) {
  return "<basename>_<i>";
}

tuple[set[str], str] genNameSym(set[str] takenNames, str basename) {
  if (basename notin takenNames)
    return <takenNames, basename>;

  int i;
  <basename, i> = getBaseNameWithNum(basename);
  i += 1;

  str n;
  while (true) {
    n = makeBaseNameWithNum(basename, i);
    if (n notin takenNames) {
      takenNames += {n};
      return <takenNames, n>;
    }
    i += 1;
  }
}

// --------------------------------------------------
// simple environment
// --------------------------------------------------

data Aenv = AenvFrame(map[str, AST] table, Aenv parent) | NoAenv();

Aenv emptyAenv = AenvFrame((), NoAenv());

Aenv newAenv() {
  return emptyAenv;
}

bool isEmpty(Aenv env) {
  while (env != NoAenv()) {
    if (!isEmpty(env.table))
      return false;
    env = env.parent;
  }
  return true;
}

set[str] allNames(Aenv env) {
  set[str] r = {};
  while (env != NoAenv()) {
    r += domain(env.table);
    env = env.parent;
  }
  return r;
}

str genUnusedName(Aenv env, str basename) {
  <_, n> = genNameSym(allNames(env), basename);
  return n;
}

bool defined(Aenv env, str name) {
  //println("<prettyPdbValue(env)> <name>");
  if (name in env.table)
    return true;
  if (env.parent != NoAenv())
    return defined(env.parent, name);
  return false;
}

/** Returns Nop() if no such name. */
AST lookup(Aenv env, str name) {
  if (name in env.table)
    return env.table[name];
  if (env.parent != NoAenv())
    return lookup(env.parent, name);
  return Nop();
}

AST mustLookup(Aenv env, str name) {
  AST decl = lookup(env, name);
  if (decl == Nop())
    throw "reference to undefined variable \'<n>\'";
  return decl;
}

Aenv enter(Aenv env) {
  return AenvFrame((), env);
}

Aenv exit(AenvFrame(_, p)) = p;

Aenv add(Aenv env, str name, AST decl) {
  table = env.table;
  table[name] = decl;
  return AenvFrame(table, env.parent);
}

// --------------------------------------------------
// general AST manipulation
// --------------------------------------------------

/** Gets a Name(...) node. */
AST getNameNode(FunName(n, _, _)) = n;
AST getNameNode(ProcName(n, _)) = n;

/** Gets the unique name of a definition. */
AST getDefName(Define(ProcClause(n,_),_,_)) = n;
AST getDefName(Define(FunClause(n,_,_),_,_)) = n;
AST getDefName(Define(PredClause(n,_),_,_)) = n;
AST getDefName(Define(TypeClause(n),_,_)) = n;
AST getDefName(Call(Proc(n), _)) = n;
AST getDefName(Apply(Fun(n), _)) = n;
AST getDefName(Apply(n : FunOf(_, _), _)) = n;
AST getDefName(Apply(Builtin(n, _), _)) = n;

default AST getDefName(AST x) {
  throw "getDefName undefined for <prettyPdbValue(x)>";
}

AST setDefName(ast : Define(c, a, b), AST n) {
  return preserveAnnos(Define(setDefName(c, n), a, b), ast);
}

AST setDefName(ast : ProcClause(_, p), AST n) {
  return preserveAnnos(ProcClause(n, p), ast);
}

AST setDefName(ast : FunClause(_, p, rt), AST n) {
  return preserveAnnos(FunClause(n, p, rt), ast);
}

default AST setDefName(AST x, AST n) {
  throw "setDefName undefined for <prettyPdbValue(x)>";
}

bool isDerivedDef(Define(_, seq(xs), _)) = (Derived() in xs);
default bool isDerivedDef(_) = false;

AST getBodyOfDef(Define(_,_,Body(b))) = b;

AST setBodyOfDef(def : Define(c,a,_), AST b) {
  return preserveAnnos(Define(c,a,Body(b)), def);
}

list[AST] getParamsOfName(ProcName(_, seq(parLst))) = parLst;
list[AST] getParamsOfName(FunName(_, seq(parLst), _)) = parLst;

AST setParamsOfName(ast : ProcName(n, _), list[AST] parLst) {
  return preserveAnnos(ProcName(n, seq(parLst)), ast);
}
AST setParamsOfName(ast : FunName(n, _, rt), list[AST] parLst) {
  return preserveAnnos(FunName(n, seq(parLst), rt), ast);
}

list[AST] getParamsOfDef(Define(FunClause(_,ParamList(seq(as)),_), _, _)) = as;
list[AST] getParamsOfDef(Define(ProcClause(_,ParamList(seq(as))), _, _)) = as;

AST setParamsOfDef(def : Define(FunClause(n, _, t), a, b),
                   list[AST] params) {
  return preserveAnnos(Define(FunClause(n, ParamList(seq(params)), t), 
                              a, b), def);
}

AST setParamsOfDef(def : Define(ProcClause(n, _), a, b),
                   list[AST] params) {
  return preserveAnnos(Define(ProcClause(n, ParamList(seq(params))), 
                              a, b), def);
}

str getNameStr(ProcName(Name(leaf(s)), _)) = s;
str getNameStr(FunName(Name(leaf(s)), _, _)) = s;
str getNameStr(TypeName(Name(leaf(s)))) = s;
//str getNameStr(FunOf(p, _)) = getNameStr(p);
str getNameStr(Name(leaf(s))) = s;

default str getNameStr(AST x) {
  throw "getNameStr undefined for <prettyPdbValue(x)>";
}

AST setNameStr(x : ProcName(Name(leaf(_)), p), str n) {
  return preserveAnnos(ProcName(Name(leaf(n)), p), x);
}

AST setNameStr(x : FunName(Name(leaf(_)), p, rt), str n) {
  return preserveAnnos(ProcName(Name(leaf(n)), p, rt), x);
}

default AST setNameStr(AST x, str n) {
  throw "setNameStr undefined for <prettyPdbValue(x)>";
}

str getVarNameStr(Assign(x, _)) = getVarNameStr(x);
str getVarNameStr(Param(_, Name(leaf(n)), _)) = n;
str getVarNameStr(Var(VarName(Name(leaf(n)), _))) = n;
str getVarNameStr(VarDecl(_, Name(leaf(n)), _, _)) = n;

default str getVarNameStr(AST x) {
  throw "getVarNameStr undefined for <prettyPdbValue(x)>";
}

AST setVarNameStr(ast : Param(m, Name(leaf(_)), t), str n) = 
  preserveAnnos(Param(m, Name(leaf(n)), t), ast);
AST setVarNameStr(ast : Var(VarName(Name(leaf(_)), t)), str n) = 
  preserveAnnos(Var(VarName(Name(leaf(n)), t)), ast);
AST setVarNameStr(ast : VarDecl(m, Name(leaf(_)), t, e), str n) = 
  preserveAnnos(VarDecl(m, Name(leaf(n)), t, e), ast);

default AST setVarNameStr(AST x, str n) {
  throw "setVarNameStr undefined for <prettyPdbValue(x)>";
}

AST getVarType(Param(_, _, t)) = t;
AST getVarType(VarDecl(_, _, t, _)) = t;
AST getVarType(Var(VarName(_, t))) = t;

AST getExprType(Var(VarName(_, t))) = t;
AST getExprType(x : DotOp(_, _, _)) = x@\type;
AST getExprType(BlockExpr(_, t)) = t;

// --------------------------------------------------
// evaluation (for testing)
// --------------------------------------------------

data MgValue = MgIntValue(int i) | 
  MgBoolValue(bool b) |
  MgStructValue(map[str, MgValue] lst);

/*
AST intType = TypeName(QName(seq([Name(leaf("int"))])));
AST boolType = SpecialType(Predicate());
*/

alias MgAddr = int;
alias MgStore = map[int, MgValue];
alias MgEnv = map[str, MgAddr];

data MgSt = MgSt(MgAddr lastAddr,
                 MgStore store,
                 MgEnv env,
                 map[AST, AST] globals);

/** Collects definitions for globals from the passed AST. */
MgSt newMgSt(AST tree) {
  map[AST, AST] globals = ();
  top-down-break visit (tree)
    {
    case ast : Define(_, _, _):
      {
        globals[getDefName(ast)] = ast;
      }
    };

  return MgSt(0, (), (), globals);
}

tuple[MgSt, MgAddr] storeAdd(MgSt st, MgValue val) {
  st.lastAddr += 1;
  MgAddr addr = st.lastAddr;
  st.store[addr] = val;
  return <st, addr>;
}

MgSt storeEnvAdd(MgSt st, str name, MgValue val) {
  <st, addr> = storeAdd(st, val);
  st.env[name] = addr;
  return st;
}

MgValue mgIntAdd(MgIntValue(x), MgIntValue(y)) = MgIntValue(x + y);
MgValue mgIntSub(MgIntValue(x), MgIntValue(y)) = MgIntValue(x - y);

map[str, MgValue (MgValue, MgValue)] mgIntBinOps =
  ("_+_": mgIntAdd, "_-_": mgIntSub);

tuple[MgSt, list[MgValue]] mgEvalLst(MgSt st, list[AST] expLst) {
  list[MgValue] valLst = [];
  for (e <- expLst) {
    <st, v> = mgEval(st, e);
    valLst += [v];
  }
  return <st, valLst>;
}

tuple[MgSt, MgValue] mgEval(MgSt st, Apply(f, seq(argLst))) {
  if (Builtin(bf, _) := f) {
    switch (bf)
      {
      case FunName(QName(seq([Name(leaf(n))])), _, 
                   SpecialType(Predicate())):
        {
          if (n == "TRUE")
            return <st, MgBoolValue(true)>;
          if (n == "FALSE")
            return <st, MgBoolValue(false)>;
        }
      };
    throw "do not know to apply built-in <prettyPdbValue(f)>";
  }

  if (Fun(x) := f)
    f = x;
  else
    throw "do not know to apply object <prettyPdbValue(f)>";
  
  // Literals are currently not supported. This is a workaround.
  if (FunName(Name(leaf(n)),
              seq([]),
              TypeName(Name(leaf("int")))) := f) {
    switch (n) {
    case "zero": return <st, MgIntValue(0)>;
    case "one": return <st, MgIntValue(1)>;
    case "two": return <st, MgIntValue(2)>;
    case "three": return <st, MgIntValue(3)>;
    case "four": return <st, MgIntValue(4)>;
    case "five": return <st, MgIntValue(5)>;
    case "six": return <st, MgIntValue(6)>;
    case "seven": return <st, MgIntValue(7)>;
    case "eight": return <st, MgIntValue(8)>;
    case "nine": return <st, MgIntValue(9)>;
    case "ten": return <st, MgIntValue(10)>;
    case "eleven": return <st, MgIntValue(11)>;
    case "twelve": return <st, MgIntValue(12)>;
    };
  }

  // 'int' arithmetic. Not really built-ins, but we shall implement
  // some operations as such.
  if (FunName(Name(leaf(n)),
              seq([AnonParam(Obs(),
                             TypeName(Name(leaf("int")))),
                   AnonParam(Obs(),
                             TypeName(Name(leaf("int"))))]),
              TypeName(Name(leaf("int")))) := f) {
    if (n in mgIntBinOps) {
      <st, valLst> = mgEvalLst(st, argLst);
      op_f = mgIntBinOps[n];
      v = op_f(valLst[0], valLst[1]);
      return <st, v>;
    }
  }

  // Built-in predicate operations.
  /// go here if we require them

  // User-defined function.
  if (FunName(_, _, _) := f) {
    decl = st.globals[f]; // gets a Define
    AST body = getBodyOfDef(decl); // body expression
    list[AST] params = getParamsOfDef(decl);
    MgEnv env_k = st.env; // store
    st.env = (); // fresh
    for (<par, arg> <- [<params[i], argLst[i]> | i <- index(params)]) {
      str p_n = getVarNameStr(par);
      if (m == Obs()) {
        <st, v> = mgEval(st, arg);
        st = storeEnvAdd(st, p_n, v); // pass-by-value
      } else {
        if (Var(_) !:= arg)
          throw "expected L-value: <prettyPdbValue(arg)>";
        str v_n = getVarNameStr(arg);
        MgAddr addr = env_k[v_n];
        st.env[p_n] = addr;
      }
    }
    <st, result> = mgEval(st, body);
    st.env = env_k; // restore
    return <st, result>;
  }

  throw "do not know to apply function <prettyPdbValue(f)>";
}

default tuple[MgSt, MgValue] mgEval(MgSt st, AST e) {
  throw "do not know how to eval <prettyPdbValue(e)>";
}

// Calls the specified function 'funName' in the specified program
// 'prog' with no arguments, returning the result.
public MgValue mgCompute(AST prog, str funName) {
  MgSt st = newMgSt(prog);
  AST f = Nop();
  for (gn <- domain(st.globals)) {
    if ((FunName(_, seq([]), _) := gn) &&
        (getNameStr(gn) == funName))
      f = gn;
  }
  if (f == Nop()) {
    throw "no function <funName> found in program";
  }
  AST e = Apply(Fun(f), seq([]));
  <st, v> = mgEval(st, e);
  return v;
}

// --------------------------------------------------
// splicing
// --------------------------------------------------

data AST = Splice(list[AST] lst) | Unsplice(AST ast);

bool hasSplice(list[AST] lst) {
  for (x <- lst) {
    if ((Splice(_) := x) || (Unsplice(_) := x))
      return true;
  }
  return false;
}

/** Note that Rascal 'headTail' is currently buggy.
    Does not throw 'EmptyList', but some internal error.
*/
AST doSeqSplice(list[AST] t) {
  list[AST] nss = [];
  while (!isEmpty(t)) {
    <h, t> = headTail(t);
    if (Splice(lst) := h) {
      nss += lst;
    } else if (Unsplice(x : Let(cs, seq(ss))) := h) {
      nss += [preserveAnnos(Let(cs, seq(ss + t)), x)];
      break;
    } else {
      nss += [h];
    }
  }
  return seqSplice(seq(nss));
}

AST seqSplice(AST tree) {
  tree = top-down-break visit (tree)
    {
    case ast : seq(ss): 
      {
        if (hasSplice(ss))
          insert doSeqSplice(ss);
        else
          fail;
      }
    };
  return tree;
}

// --------------------------------------------------
// locations
// --------------------------------------------------

/** Type checker does not preserve Apply annotations, for example, so
    using this is not appropriate. 'locOf' tries harder to get a
    location. */
loc getLoc(AST x) {
  if ((x@\loc)?)
    return x@\loc;
  else
    throw "no loc anno in <prettyPdbValueAnnos(x)>";
}

/** Compares file, line, and column to location. */
bool isPointInLocLC(loc oloc, str fil, int lin, int col) {
  return (oloc.uri == fil) && 
    ((oloc.begin.line < lin) ||
     ((oloc.begin.line == lin) && (oloc.begin.column <= col))) &&
    ((oloc.end.line > lin) ||
     ((oloc.end.line == lin) && (col < oloc.end.column)));
}

/** Compares file and offset to location. */
bool isPointInLocO(loc oloc, str fil, int off) {
  return (oloc.uri == fil) && 
    ((oloc.offset <= off) && 
     ((oloc.offset + oloc.length) > off));
}

public bool (loc) makePointPred(str uri, int off) {
  isPointWithin = bool (loc l) { return isPointInLocO(l, uri, off); };
  return isPointWithin;
}

public bool (loc) makePointPred(str uri, int lin, int col) {
  isPointWithin = bool (loc l) { return isPointInLocLC(l, uri, lin, col); };
  return isPointWithin;
}

public str getAstFileUri(AST ast) {
  return locOf(ast).uri;
}

// --------------------------------------------------
// const-ness
// --------------------------------------------------

/** Whether a const variable, i.e. whether an assignable variable.
 */
public default bool isConstVar(AST ast) = false;
public bool isConstVar(VarDecl(Obs(), _, _, _)) = true;

/** Whether a constant expression. We could check for Literal(_) and
    Obs() variables and all compositions thereof, but really, any
    expression without Upd() variables should have a fixed value. */
public bool isConstExpr(Env env, AST ast) {
  top-down-break visit(ast) {
  case Var(VarName(n, _)):
    {
      set[AST] defs = lookupName(n, ExprKind(), env);
      if (isEmpty(defs))
        return false; // unknown variable
      AST def = getOneFrom(defs); // should be no overloads
      if (VarDecl(m, n, t, e) := def) {
        // Note that an Out() variable should never appear as an R-value.
        assert m != Out();
        if (m == Upd())
          return false;
      } else {
        assert false : "expected VarDecl";
        return false; // was supposed to be a variable reference
      }
    }
  case BlockExpr(_,_):
    {
      // Might declare new variables, thus requiring 'env' tracking.
      // Might also have side effecting statements. We assume
      // non-const rather than analyze, for now.
      return false;
    }
  }
  return true;
}

// --------------------------------------------------
// statement lifting (from expressions)
// --------------------------------------------------

/** The fields of 'LiftStatExpr' contain (1) the type of the variable
    to receive the value of the lifted expression, and (2) the
    statement (which may be a block) computing and assigning the
    value. The statement computing the value of the expression must
    refer to its L-value as 'LiftVar'. The variable will be given a
    fresh name upon lifting.

    Any 'LiftStatExpr' in an expression are lifted around the
    statement containing the expression. Lifting is in left-to-right
    order of execution, which can matter if the lifted statements
    modify outer variables.
 */

AST LiftStatExpr(AST varType, AST stat) = LiftStatExpr(varType, stat, "lifted");

bool hasLiftStatExpr(AST tree) {
  visit (tree)
    {
    case LiftStatExpr(_, _, _):
      {
        return true;
      }
    };
  return false;
}

AST substLiftVar(AST tree, AST varRef) {
  return visit (tree)
    {
    case LiftVar():
      {
        insert varRef;
      }
    };
}

AST liftFromStat(AST stat) {
  set[str] takenNames = findAllVarNamesIn(stat);
  list[tuple[AST, AST]] varInfo = []; // <decl, ref>
  stat = top-down-break visit (stat) 
    {
      //case x: { termPrintln("liftFromStat", x); fail; } // debug
    case LiftStatExpr(t, s, bn):
      {
        str n; // name unique within statement 'stat'
        <takenNames, n> = genNameSym(takenNames, bn);
        AST r = Var(VarName(Name(leaf(n)), t));
        s = substLiftVar(s, r);
        AST d = VarDecl(Upd(), Name(leaf(n)), t, Undefined());
        varInfo += [<d, doLifting(s)>];
        insert r;
      }
    };
  for (<varDecl, varStat> <- reverse(varInfo)) {
    stat = Let(varDecl, seq([varStat, stat]));
  }
  return stat;
}

/** We traverse all possible (immediately) expression containing
    statements looking for expressions to lift. If any such statements
    are missing from here then we might end up lifting too high, which
    might lead to incorrect results. */
AST doLifting(AST tree) {
  tree = bottom-up visit (tree) 
    {
      //case x: { termPrintln("doLifting", x); fail; } // debug
    case stat : If(_, _, _):
      if (hasLiftStatExpr(stat))
        insert liftFromStat(stat);
    case stat : Guard(_, _):
      if (hasLiftStatExpr(stat))
        insert liftFromStat(stat);
    case stat : Call(_, _):
      if (hasLiftStatExpr(stat))
        insert liftFromStat(stat);
    case stat : Return(_):
      if (hasLiftStatExpr(stat))
        insert liftFromStat(stat);
    case stat : Assign(_, _):
      if (hasLiftStatExpr(stat))
        insert liftFromStat(stat);
    case stat : Let(_, _):
      if (hasLiftStatExpr(stat))
        insert liftFromStat(stat);
    case stat : Assert(_, _):
      if (hasLiftStatExpr(stat))
        insert liftFromStat(stat);
    case stat : Assume(_, _):
      if (hasLiftStatExpr(stat))
        insert liftFromStat(stat);
    };
  return tree;
}

// --------------------------------------------------
// selection of inlineables
// --------------------------------------------------

AST getDeclBody(AST ast) {
  top-down-break visit(ast) {
  case d : DeclBody(_): return d;
  }
  return Nop();
}

set[AST] getDeclBodySet(AST ast) {
  AST r = getDeclBody(ast);
  if (Nop() == r) return {};
  return {r};
}

/** Used to identify code to inline, currently for fun-like and
    variable inlining. */
anno int AST@inlId;

/** Used to identify the concept within which to inline, currently for
    fun-like and type inlining. */
anno bool AST@inlThis;

/** Collect inlineable definitions and references. Collects
    information for only a specific namespace (top-level declaration),
    identified by point.

    In the case of locals we are really only interested in individual
    functions (or rather their bodies), but upon calling this it is
    not known if the point identifies a local.

    The 'loc' keys are not necessarily unique, as there may be
    Derived() definitions with the exact same 'loc' as the
    derived-from definition. Also elements inside a derived definition
    may inherit the location from the derived-from definition, or its
    contents. We do not consider derived definitions or references
    thereof inlineable at this time.
 */
public tuple[AST, rel[int, loc, AST]]
findInlineablesInNsByPoint(AST tree,
                           bool (loc) isPointWithin) {
  rel[int, loc, AST] r = {};
  
  // Is there no way to request pruning of a subtree in a 'visit'?
  bool inNs = false;

  int nextId = 1; // 0 is reserved for a "none" value
  
  tree = top-down visit(tree) 
    {
    case x : DeclBody(b) : 
      {
        loc l = locOf(x);
        inNs = isPointWithin(l);
        if (inNs) {
          insert x[@inlThis = true];
        }
      }

    case x : Define(_,_,Body(b)): 
      if (inNs && !isDerivedDef(x)) {
        int inlId = nextId;
        nextId += 1;
        x = x[@inlId = inlId];
        r += {<inlId, locOf(x), x>};
        insert x;
      }

      // fun-like inlining objects

    case x : Call(_,_): 
      if (inNs) {
        int inlId = nextId;
        nextId += 1;
        x = x[@inlId = inlId];
        r += {<inlId, locOf(x), x>};
        insert x;
      }
    case x : Apply(y,_): 
      if (inNs && (FunOf(_, _) !:= y)) {
        int inlId = nextId;
        nextId += 1;
        x = x[@inlId = inlId];
        r += {<inlId, locOf(x), x>};
        insert x;
      }

      // type inlining objects

    case x : TypeName(_): 
      if (inNs) {
        int inlId = nextId;
        nextId += 1;
        x = x[@inlId = inlId];
        r += {<inlId, locOf(x), x>};
        insert x;
      }

      // variable inlining objects

    case x : Assign(_,_): 
      if (inNs) {
        int inlId = nextId;
        nextId += 1;
        x = x[@inlId = inlId];
        r += {<inlId, locOf(x), x>};
        insert x;
      }
    case x : VarDecl(_, _, _, e): // including assignment
      if (inNs) {
        if (e != Undefined()) {
          int inlId = nextId;
          nextId += 1;
          x = x[@inlId = inlId];
          r += {<inlId, locOf(x), x>};
          insert x;
        }
      }

      /*
      // Does it make sense to inline a single variable reference?
    case x : Var(_): ;
      */

      /*
        // Might also support attempts to inline a variable entirely,
        // meaning that the variable declaration could be removed.
    case x : VarDecl(_, _, _, _): ;
      */

    };

  return <tree, r>;
}

str getInlCategory(Assign(_, _)) = "var";
str getInlCategory(VarDecl(_, _, _, _)) = "var";

str getInlCategory(TypeName(_)) = "type";
str getInlCategory(Define(TypeClause(_),_,_)) = "type";

str getInlCategory(Call(_,_)) = "fun";
str getInlCategory(Apply(_,_)) = "fun";
str getInlCategory(Define(ProcClause(_,_),_,_)) = "fun";
str getInlCategory(Define(FunClause(_,_,_),_,_)) = "fun";

default str getInlCategory(AST x) {
  throw "cannot determine category for inlineable <x>";
}

rel[int, loc, AST] nsDropUnrelated(rel[int, loc, AST] ns, str c) {
  return { <i, l, x> | <i, l, x> <- ns, c == getInlCategory(x) };
}

/** Use to find out if an object is a definition or a reference. When
    a definition is inlined, we must inline all references, too. When
    a reference is inlined, we must check we have its definition (with
    a body). See 'getRefsForDef' and 'getDefForRef'.
    */
bool isDecl(Define(_,_,_)) = true;
default bool isDecl(AST x) = false;

rel[int, loc, AST] getRefsForDef(AST def, rel[int, loc, AST] ns) {
  rel[int, loc, AST] r = {};
  AST name = getDefName(def);
  for (x : <i, l, a> <- ns) {
    if (!isDecl(a)) {
      if (name == getDefName(a))
        r += {x};
    }
  }
  return r;
}

tuple[int, loc, AST] getDefForRef(AST ref, rel[int, loc, AST] ns) {
  AST name = getDefName(ref);
  for (x : <i, l, a> <- ns) {
    if (isDecl(a)) {
      if (name == getDefName(a))
        return x;
    }
  }
  return <0, unknownLoc, Nop()>;
}

/** Finds an inlineable object based on line and column position, or
    returns Nop() as the AST return value.

    When specifying the position by line and column, note that Tab
    characters count as a single character, and hence your text editor
    may not agree on column numbers.

    ns:: candidate objects to consider, along with inlIds and locations
    fil:: file URI
    lin:: line number
    col:: column number
    Returns:: information about the best match
*/
public tuple[int, loc, AST] 
findBestInlObjLC(rel[int, loc, AST] ns, str fil, int lin, int col) {
  best = <0, unknownLoc, Nop()>;
  for (v : <int id, loc oloc, AST oobj> <- ns) {
    //println(prettyPdbValue(v));
    if (isPointInLocLC(oloc, fil, lin, col)) {
      loc bestLoc = best[1];
      if ((bestLoc == unknownLoc) ||
          // whether bestLoc fully encloses oloc
          (bestLoc >= oloc))
        best = v;
    }
  }
  return best;
}

/** Finds an inlineable object based on file offset position, or
    returns Nop() as the AST return value.

    ns:: candidate objects to consider, along with inlIds and locations
    fil:: file URI
    off:: character offset
    Returns:: information about the best match
*/
public tuple[int, loc, AST] 
findBestInlObjO(rel[int, loc, AST] ns, str fil, int off) {
  best = <0, unknownLoc, Nop()>;
  for (v : <int id, loc oloc, AST oobj> <- ns) {
    if (isPointInLocO(oloc, fil, off)) {
      loc bestLoc = best[1];
      if ((bestLoc == unknownLoc) || (bestLoc >= oloc))
        best = v;
    }
  }
  return best;
}

/** 'id' is inlId. */
data InlObj = InlDef(int id)
  | InlRef(int id, AST decl)
  | InlNone();

InlObj getInlineObjMatch(set[InlObj] inlineObjs, int id) {
  for (x <- inlineObjs)
    if (x.id == id)
      return x;
  return InlNone();
}

bool isInInlineObjs(set[InlObj] inlineObjs, int id) {
  return getInlineObjMatch(inlineObjs, id) != InlNone();
}

default bool isProc(AST ast) = false;
bool isProc(Define(ProcClause(_,_),_,_)) = true; 

default bool isFunLike(AST ast) = false;
bool isFunLike(Define(ProcClause(_,_),_,_)) = true; 
bool isFunLike(Define(FunClause(_,_,_),_,_)) = true; 

tuple[int, loc, AST] findContainingFunLike(loc oloc, AST obj,
                                           rel[int, loc, AST] ns) {
  rel[int, loc, AST] contFun = {};
  for (cf : <id, floc, fobj> <- ns) {
    if (isFunLike(fobj) && !isDerivedDef(fobj) && (oloc <= floc))
      contFun += {cf};
  }
  assert size(contFun) == 1 : "expected exactly one containing function";
  return getOneFrom(contFun);
}

public tuple[int, loc, AST] findBestInlineable(AST ast, int lin, int col) {
  str uri = getAstFileUri(ast);
  rel[int, loc, AST] ns;
  isPointWithin = makePointPred(uri, lin, col);
  <ast, ns> = findInlineablesInNsByPoint(ast, isPointWithin);
  tuple[int, loc, AST] best = findBestInlObjLC(ns, uri, lin, col);
  return best;
}

public tuple[int, loc, AST] findBestInlineable(AST ast, int off) {
  str uri = getAstFileUri(ast);
  rel[int, loc, AST] ns;
  isPointWithin = makePointPred(uri, off);
  <ast, ns> = findInlineablesInNsByPoint(ast, isPointWithin);
  tuple[int, loc, AST] best = findBestInlObjO(ns, uri, off);
  return best;
}

/** Inlines inlineable object at position (if any).
    The position is identified by character offset.
*/
public AST inlineByPoint(AST ast, int off) {
  str uri = getAstFileUri(ast);
  rel[int, loc, AST] ns;
  isPointWithin = makePointPred(uri, off);
  <ast, ns> = findInlineablesInNsByPoint(ast, isPointWithin);
  tuple[int, loc, AST] best = findBestInlObjO(ns, uri, off);
  <o_id, oloc, obj> = best;
  if (obj == Nop())
    throw NoInlineableFound(off);
  return inlineBest(ast, ns, best);
}

/** Inlines inlineable object at position (if any).
    The position is identified by line and column number.
*/
public AST inlineByPoint(AST ast, int lin, int col) {
  str uri = getAstFileUri(ast);
  rel[int, loc, AST] ns;
  isPointWithin = makePointPred(uri, lin, col);
  <ast, ns> = findInlineablesInNsByPoint(ast, isPointWithin);
  tuple[int, loc, AST] best = findBestInlObjLC(ns, uri, lin, col);
  <o_id, oloc, obj> = best;
  if (obj == Nop())
    throw NoInlineableFound(lin, col);
  return inlineBest(ast, ns, best);
}

AST inlineBest(AST ast, 
               rel[int, loc, AST] ns, 
               tuple[int, loc, AST] best) {
  <o_id, oloc, obj> = best;
  //println("object chosen for inlining <prettyPdbValue(obj)>");
  switch (getInlCategory(obj))
    {
    case "var":
      {
        // 'best' is the assignment object
        <_, floc, fobj> = findContainingFunLike(oloc, obj, ns);
        return inlineVar(ast, fobj@inlId, obj@inlId);
      }
    case "fun":
      {
        ns = nsDropUnrelated(ns, "fun");
        //println("relevant namespace <prettyPdbValue(ns)>");
        set[InlObj] inlSet;
        if (isDecl(obj)) {
          inlSet = {InlDef(o_id)};
          AST decl = obj;
          if (isProc(decl))
            decl = translateBareReturnsInBody(decl);
          for (<r_id, rloc, robj> <- getRefsForDef(obj, ns))
            inlSet += {InlRef(r_id, decl)};
        } else {
          <d_id, dloc, dobj> = getDefForRef(obj, ns);
          if (dobj == Nop())
            throw NoDefFound(shortDispNameOf(getDefName(obj)), obj, oloc);
          AST decl = dobj;
          if (isProc(decl))
            decl = translateBareReturnsInBody(decl);
          inlSet = {InlRef(o_id, decl)};
        }
        return inlineFunLike(ast, inlSet);
      }
    case "type":
      {
        return inlineType(ast, obj);
      }
    default:
      {
        assert false;
      }
    };
}

// --------------------------------------------------
// varId
// --------------------------------------------------

// Locally unique ID for a variable.
anno str AST@varId;

bool hasInitExpr(VarDecl(_, _, _, e)) {
  return e != Undefined();
}

bool hasInitExpr(Param(_, _, _)) {
  return false;
}

AST getVarRef(Param(_, n, t)) = Var(VarName(n, t));
AST getVarRef(VarDecl(_, n, t, _)) = Var(VarName(n, t));
AST getVarRef(r : Var(_)) = r;

/** Useful in choosing locally unused variable names. */
set[str] findAllVarNamesIn(AST tree) {
  set[str] r = {};
  top-down visit(tree) 
    {
    case ast : Var(_): 
      {
        r += {getVarNameStr(ast)};
      }
    case ast : VarDecl(_, _, _, _):
      {
        r += {getVarNameStr(ast)};
      }
    case ast : Param(_, _, _):
      {
        r += {getVarNameStr(ast)};
      }
    };
  return r;
}

set[str] findAllVarIdsIn(AST tree) {
  set[str] r = {};
  top-down visit(tree) 
    {
    case ast : Var(_): 
      {
        r += {ast@varId};
      }
    case ast : VarDecl(_, _, _, _):
      {
        r += {ast@varId};
      }
    case ast : Param(_, _, _):
      {
        r += {ast@varId};
      }
    };
  return r;
}

tuple[AST, set[str]] renameVarsById(AST tree, 
                                    set[str] varIds,
                                    set[str] takenNames) {
  map[str, str] m = (); // id-to-name map

  AST makeAst(AST ast, str varId) {
    str nn;
    if (varId in m) {
      nn = m[varId];
    } else {
      str n = getVarNameStr(ast);
      <takenNames, nn> = genNameSym(takenNames, n);
      m[varId] = nn;
    }
    AST nast = setVarNameStr(ast, nn);
    return nast;
  }

  tree = visit(tree)
    {
    case ast : Var(_): 
      {
        str varId = ast@varId;
        if (varId in varIds) {
          AST nast = makeAst(ast, varId);
          insert preserveAnnos(nast, ast);
        }
      }
    case ast : VarDecl(_, _, _, e):
      {
        str varId = ast@varId;
        if (varId in varIds) {
          AST nast = makeAst(ast, varId);
          insert preserveAnnos(nast, ast);
        }
      }
    case ast : Param(_, _, _):
      {
        str varId = ast@varId;
        if (varId in varIds) {
          AST nast = makeAst(ast, varId);
          insert preserveAnnos(nast, ast);
        }
      }
    };
  return <tree, takenNames>;
}

/**
   For annotating variable IDs. A single fun-like Define(...) should
   be passed as 'tree'.
   */
tuple[AST, map[str, int]]
assignVarIds(AST tree, Aenv env, map[str, int] varMap) {
  tree = top-down-break visit(tree) 
    {
    case ast : Param(_, Name(leaf(n)), _):
      {
        <varMap, varId> = genIdSym(varMap, n);
        ast = ast[@varId = varId];
        env = add(env, n, ast);
        insert(ast);
      }

    case ast : Let(c, ss): // for VarDecl
      {
        Aenv nenv = env;
        if (VarDecl(m, n, t, e) := c) {
          str nn = getVarNameStr(c);
          <varMap, varId> = genIdSym(varMap, nn);
          c = c[@varId = varId];
          if (e != Undefined()) {
            <e, varMap> = assignVarIds(e, nenv, varMap);
          }
          c = preserveAnnos(VarDecl(m, n, t, e), c);
          nenv = add(nenv, nn, c);
          nc = c;
        } else {
          throw "unexpected LetClause <yieldTerm(c)>";
        }
        <ss, varMap> = assignVarIds(ss, nenv, varMap);
        insert preserveAnnos(Let(nc, ss), ast);
      }

    case ast : Var(_):
      {
        str n = getVarNameStr(ast);
        AST decl = lookup(env, n);
        if (decl == Nop())
          throw "reference to unknown variable \'<n>\': <prettyMagnolia(ast)>";
        str varId = decl@varId;
        ast = ast[@varId = varId];
        insert(ast);
      }
    };

  return <tree, varMap>;
}

AST assignVarIds(AST tree) {
  <tree, varMap> = assignVarIds(tree, emptyAenv, ());
  return tree;
}

// --------------------------------------------------
// inlining of locals
// --------------------------------------------------

// Locally unique ID for an assigned value.
anno str AST@valId;

public AST keepVarValIdAnnots(AST tree) {
  tree = keepAnnotationsRec(tree, {"varId", "valId"});
  return tree;
}

/** Collects Param and VarDecl nodes for the specified variable IDs
    and turns them into VarDecl's. */
map[str, AST] findVarDeclFor(AST tree, set[str] varIds) {
  map[str, AST] r = ();
  top-down visit(tree) 
    {
    case ast : VarDecl(m, n, t, e):
      {
        str varId = ast@varId;
        if (varId in varIds) {
          r[varId] = ast;
        }
      }
    case ast : Param(m, n, t):
      {
        str varId = ast@varId;
        if (varId in varIds) {
          ast = preserveAnnos(VarDecl(m, n, t, Undefined()), ast);
          r[varId] = ast;
        }
      }
    };
  return r;
}

/**
   For annotating variable and value IDs. A single fun-like
   Define(...) should be passed as 'tree'. Note that here we only
   assign unique names to variables and values that are being
   introduced; we cannot, for instance, know which value a variable
   refers to without data flow analysis, and possibly not even then.
   */
tuple[AST, map[str, int], map[str, int]]
vaiPreAnno(AST tree, Aenv env,
           map[str, int] varMap,
           map[str, int] valMap) {
  tree = top-down-break visit(tree) {
    // Params are in scope throughout, and all non-'out' params always
    // have an initial value, although we don't have the expression.
  case ast : Param(mode, Name(leaf(n)), _):
    {
      <varMap, varId> = genIdSym(varMap, n);
      ast = ast[@varId = varId];
      if (mode != Out()) {
        <valMap, valId> = genIdSym(valMap, varId);
        ast = ast[@valId = valId];
      }
      env = add(env, n, ast);
      insert(ast);
    }

  case ast : Let(c, ss): // for VarDecl
    {
      Aenv nenv = env;
      if (VarDecl(m, n, t, e) := c) {
        str nn = getVarNameStr(c);
        <varMap, varId> = genIdSym(varMap, nn);
        c = c[@varId = varId];
        if (e != Undefined()) {
          <e, varMap, valMap> = vaiPreAnno(e, nenv, varMap, valMap);
          if (m != Out()) {
            <valMap, valId> = genIdSym(valMap, varId);
            c = c[@valId = valId];
          }
        }
        c = preserveAnnos(VarDecl(m, n, t, e), c);
        nenv = add(nenv, nn, c);
        nc = c;
      } else {
        throw "unexpected LetClause <yieldTerm(c)>";
      }
      <ss, varMap, valMap> = vaiPreAnno(ss, nenv, varMap, valMap);
      insert preserveAnnos(Let(nc, ss), ast);
    }

  case ast : Assign(lv, rv):
    {
      str n = getVarNameStr(ast);
      AST decl = lookup(env, n);
      if (decl == Nop())
        throw "assign to unknown variable \'<n>\': <prettyMagnolia(ast)>";
      str varId = decl@varId;
      lv = lv[@varId = varId];
      <rv, varMap, valMap> = vaiPreAnno(rv, env, varMap, valMap);
      ast = preserveAnnos(Assign(lv, rv), ast);
      <valMap, valId> = genIdSym(valMap, varId);
      ast = ast[@varId = varId][@valId = valId];
      insert(ast);
    }

  case ast : Var(_):
    {
      str n = getVarNameStr(ast);
      AST decl = lookup(env, n);
      if (decl == Nop())
        throw "reference to unknown variable \'<n>\': <prettyMagnolia(ast)>";
      str varId = decl@varId;
      ast = ast[@varId = varId];
      insert(ast);
    }
  };

  return <tree, varMap, valMap>;
}

AST vaiDoPreAnno(AST tree) {
  <tree, varMap, valMap> = vaiPreAnno(tree, emptyAenv, (), ());
  return tree;
}

data ValId = ValId(str valId) | Phi(set[str] valIds) | NoValId();

set[str] toSet(NoValId()) = {};
set[str] toSet(ValId(n)) = {n};
set[str] toSet(Phi(ns)) = ns;

ValId sum(ValId x, ValId y) {
  set[str] s = toSet(x) + toSet(y);
  if (size(s) == 0)
    return NoValId();
  if (size(s) == 1)
    return ValId(getOneFrom(s));
  return Phi(s);
}

data AST = OwVar(str varId);

anno bool AST@owMarker;

/** df:: variable (ID) to value (ID) associations
*/
data VaiSt = VaiSt(AST n, AST t, AST e, str varId, str valId,
                   // loc l, 
                   int a_id,
                   map[str, ValId] df,
                   bool assignSeen, bool inlining, 
                   rel[str, str] owSet);

VaiSt newVaiSt(int a_id) {
  return VaiSt(Nop(), Nop(), Nop(), "", "", a_id, (), false, false, {});
}

map[str, ValId] sum(map[str, ValId] x, map[str, ValId] y) {
  for (k <- domain(y)) {
    if (k in x)
      x[k] = sum(x[k], y[k]);
    else
      x[k] = y[k];
  }
  return x;
}

VaiSt sum(VaiSt x, VaiSt y) {
  assert x.inlining == y.inlining : "cannot sum";
  x.assignSeen = (x.assignSeen || y.assignSeen);
  x.df = sum(x.df, y.df);
  x.owSet += y.owSet;
  return x;
}

ValId getValId(VaiSt st, str n) {
  if (n notin st.df)
    return NoValId();
  return st.df[n];
}

tuple[AST, VaiSt] doOw(AST tree, VaiSt st) {
  tree = top-down-break visit(tree) 
    {
    case ast : Var(_): 
      {
        str varId = ast@varId;
        ValId reqVal = ValId(genIdSym0(varId));
        if (getValId(st, varId) != reqVal) {
          str n = getVarNameStr(ast);
          st.owSet += {<n, varId>};
          insert OwVar(varId);
        }
      }
    };
  return <tree, st>;
}

map[str, ValId] newDf(AST tree) {
  map[str, ValId] r = ();
  top-down-break visit(tree) 
    {
    case ast : Var(arg): 
      {
        str varId = ast@varId;
        r[varId] = ValId(genIdSym0(varId));
      }
    };
  return r;
}

AST rewriteOwVar(AST tree, 
                 list[AST] declLst,
                 map[str, AST] varMap) {
  bool seen = false; // @owMarker
  tree = top-down visit (tree)
    {
    case ast : OwVar(varId): 
      {
        if (varId in varMap) {
          insert varMap[varId];
        }
      }
    case ast : Nop():
      {
        if (!seen && ast@owMarker?) {
          seen = true;
          insert Unsplice(LetMulti(declLst, seq([])));
        }
      }
    case ast : VarDecl(_, _, _, _):
      {
        if (!seen && ast@owMarker?) {
          seen = true;
          insert Splice(declLst + [ast]);
        }
      }
    };
  return tree;
}

AST vaiFixOw(AST tree, VaiSt st, set[str] takenNames) {
  rel[str, str] owSet = st.owSet; // <name, varId>

  if (isEmpty(owSet))
    return tree;

  set[str] idSet = owSet<1>;
  map[str, AST] declMap = findVarDeclFor(tree, idSet); // <varId, VarDecl>

  list[tuple[str, str, str, AST]] owLst = [];
  for (<n, id> <- sort(toList(owSet))) { // predictable gensym order
    <takenNames, nn> = genNameSym(takenNames, n);
    decl = declMap[id];
    owLst += <nn, id, n, decl>;
  }
  owLst = sort(owLst); // predictable declaration order

  // for adding declarations
  list[AST] declLst = [VarDecl(Obs(), Name(leaf(nn)), t,
                               Var(VarName(n, t))[@varId = id]) |
                       <nn, id, _, VarDecl(_, n, t, _)> <- owLst];

  // for resolving refs
  map[str, AST] varMap = (id : Var(VarName(Name(leaf(nn)), t)) | 
                          <nn, id, n, VarDecl(_, _, t, _)> <- owLst);

  tree = rewriteOwVar(tree, declLst, varMap);

  tree = seqSplice(tree);
  return tree;
}

/**
  Here we must proceed in statement order, as we are doing forward
  data flow analysis. Need not care about scope-related ordering,
  however, as all variables are uniquely annotated.

  When we find the assigned expression, we store it. We then replace
  all references to the assigned variable with the expression, up to
  the point where it still has said value/expression.

  If the value of the assigned-to variable changes in one
  conditional branch but not in the other, we perform the original
  assignment in the other branch to leave things in a consistent
  state.

  We patch things later to deal with shadowing and overwritten
  variables in the assigned expression. Here we do not track
  environment, and hence don't know what names are shadowed.
  However, it is here that we know which variables are overwritten,
  and will use the placeholder 'OwVar' instead of those variables,
  to support patching.
*/
tuple[AST, VaiSt] inlineAssignInDef(AST tree, VaiSt st)
{
  tree = top-down-break visit(tree) 
    {
    case ast : Var(_): 
      {
        if (st.inlining) {
          str varId = ast@varId;
          if (varId == st.varId) {
            // substitute
            <e, st> = doOw(st.e, st);
            insert e;
          }
        }
      }

    // Currently do not handle L-value being something other than a
    // variable reference. But by this point any other L-values should
    // have been translated to variable references.
    case ast : Assign(lv : Var(VarName(n, t)), rv):
      {
        str varId = ast@varId;
        str valId = ast@valId;
        if ((!st.assignSeen) && ast@inlId? && (ast@inlId == st.a_id)) {
          st.assignSeen = true;
          st.varId = varId;
          st.valId = valId;
          st.n = n;
          st.t = t;
          st.e = rv;
          st.df = newDf(st.e);
          st.inlining = true;
          ast = Nop();
          ast@owMarker = true;
          insert ast;
        } else if (st.assignSeen) {
          <rv, st> = inlineAssignInDef(rv, st);
          ast = preserveAnnos(Assign(lv, rv), ast);
          if (varId == st.varId)
            st.inlining = false;
          else
            st.df[varId] = ValId(valId);
          insert ast;
        }
      }

    case ast : VarDecl(m, n, t, rv):
      {
        if (ast@valId?) {
          str varId = ast@varId;
          str valId = ast@valId;
          if ((!st.assignSeen) && ast@inlId? && (ast@inlId == st.a_id)) {
            st.assignSeen = true;
            st.varId = varId;
            st.valId = valId;
            st.n = n;
            st.t = t;
            st.e = rv;
            st.df = newDf(st.e);
            st.inlining = true;
            ast = preserveAnnos(VarDecl(m, n, t, Undefined()), ast);
            ast@owMarker = true;
            insert ast;
          } else if (st.assignSeen) {
            <rv, st> = inlineAssignInDef(rv, st);
            ast = preserveAnnos(VarDecl(m, n, t, rv), ast);
            st.df[varId] = ValId(valId);
            insert ast;
          }
        }
      }

    case ast : If(ce, tb, eb):
      {
        if (st.inlining) {
          <ce, st> = inlineAssignInDef(ce, st);
          <tb, t_st> = inlineAssignInDef(tb, st);
          <eb, e_st> = inlineAssignInDef(eb, st);
          if (t_st.inlining != e_st.inlining) {
            str varId = st.varId;
            str valId = st.valId;
            AST lv = Var(VarName(st.n, st.t))[@varId = varId];
            AST asg = Assign(lv, st.e)[@varId = varId][@valId = valId];
            if (t_st.inlining) {
              // insert assignment to THEN branch
              t_st.inlining = false;
              <asg, t_st> = inlineAssignInDef(asg, t_st);
              tb = seq(tb.args + [asg]);
            } else if (e_st.inlining) {
              // insert assignment to ELSE branch
              e_st.inlining = false;
              <asg, e_st> = inlineAssignInDef(asg, e_st);
              eb = seq(eb.args + [asg]);
            }
          }
          st = sum(t_st, e_st);
          ast = preserveAnnos(If(ce, tb, eb), ast);
          insert ast;
        }
      }
    };

  return <tree, st>;
}

/** tree:: syntax tree to rewrite
    f_id:: inlId of the enclosing function definition
    a_id:: inlId of the assignment to inline (Assign or VarDecl with init expr)
*/
AST inlineVar(AST tree, int f_id, int a_id) 
{
  tree = top-down-break visit(tree) 
    {
    case ast : Define(c, _, _): 
      {
        if ((ProcClause(_, _) := c) ||
            (FunClause(_, _, _) := c)) {
          if (ast@inlId? && (ast@inlId == f_id)) {
            ast = translateReturnsInBody(ast);
            ast = vaiDoPreAnno(ast);
            <ast, st> = inlineAssignInDef(ast, newVaiSt(a_id));
            set[str] takenNames;
            set[str] e_ids = findAllVarIdsIn(st.e);
            //println("all expr var IDs <yieldTerm(e_ids)>");
            <ast, takenNames> = vaiFixScoping(ast, e_ids);
            ast = vaiFixOw(ast, st, takenNames);
            insert ast;
          }
        }
      }
    };
  return tree;
}

// --------------------------------------------------
// scope fixing
// --------------------------------------------------

/** Junction ID. Never 0 for a valid value. */
anno int AST@juncId;

tuple[int, int] getJuncId(int nid) {
  return <nid + 1, nid>;
}

/** Annotates all points in the program around (or within) which we
    could insert a new variable declaration, and where the subtree
    might contain variable references. Any statements with
    expressions, basically, and any composite statements, and
    'BlockExpr' (which is like a composite statement). */
tuple[AST, int] addJuncIds(AST tree) {
  int nid = 1;
  tree = bottom-up visit (tree) 
    {
    case ast : Block(_):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : If(_, _, _):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : Guard(_, _):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : Call(_, _):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : Return(_):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : Assign(_, _):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : Let(_, _):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : Assert(_, _):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : Assume(_, _):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    case ast : BlockExpr(_,_):
      { 
        int juncId; <nid, juncId> = getJuncId(nid);
        insert ast[@juncId = juncId];
      }
    };
  return <tree, nid>;
}

int countVarRefs(AST tree, str varId) {
  int i = 0;
  visit (tree) 
    {
    case ast : Var(_): 
      {
        if (ast@varId == varId)
          i += 1;
      }
    };
  return i;
}

/** For each variable reference of 'varId', finds out which junctions
    it is under. */
set[set[int]] juncsForVars(AST tree,
                           str varId,
                           bool skip,
                           set[int] juncSet,
                           set[set[int]] varSets) {
  top-down-break visit (tree)
    {
    case ast : Var(_): 
      {
        if (ast@varId == varId) {
          varSets += {juncSet};
        }
      }
      
    case seq(xs):
      {
        for (x <- xs) {
          set[set[int]] r = juncsForVars(x, varId, false, juncSet, {});
          varSets += r;
        }
      }

    case AST ast:
      {
        if (ast@juncId?) {
          if (!skip) {
            int juncId = ast@juncId;
            juncSet += {juncId};
            set[set[int]] r = juncsForVars(ast, varId, true, juncSet, {});
            varSets += r;
          } else {
            skip = false;
            fail;
          }
        } else
          fail;
      }
    };

  return varSets;
}

/** For the specified variable 'varId' in 'tree', finds a junction in
    which it should be declared for it to potentially be in scope
    everywhere where there is a 'Var' reference to it. (Shadowing may
    still occur.) Returns 0 if there is no suitable point to declare
    it. If the variable has no references, then an arbitrary junction
    is returned. */
int findJunction(AST tree, str varId) {
  set[set[int]] varSets = juncsForVars(tree, varId, false, {}, {});

  // Pick the first junction that "dominates" all the references.
  bottom-up visit (tree)
    {
    case AST ast:
      {
        if (ast@juncId?) {
          int juncId = ast@juncId;
          bool ok = true;
          for (juncSet <- varSets) {
            //termPrintln(juncSet);
            if (juncId notin juncSet) {
              ok = false;
              break;
            }
          }
          if (ok) {
            return juncId;
          }
        }
      }
    };

  return 0;
}

/** Finds all shadowed and out-of-scope 'Var' references in 'tree'.
    This relies on 'varId' annotations being in place. Returns two
    sets of 'varId': shadowed, and name-not-in-scope references. The
    sets may overlap. Formal parameters can never be in the latter
    set. (To fix shadowing, you can rename. To fix beyond-scope names,
    you can lift to some outer scope, which may introduce new
    shadowing. To fix both, you can do both.) */
tuple[set[str], set[str]]
findVarScopingMismatches(AST tree, Aenv env) {
  set[str] shad = {};
  set[str] beyo = {};
  top-down-break visit(tree) 
    {
    case ast : Var(_): 
      {
        str n = getVarNameStr(ast);
        AST decl = lookup(env, n);
        str thisVarId = ast@varId;
        if (decl == Nop()) {
          beyo += {thisVarId};
        } else {
          str declVarId = decl@varId;
          if (thisVarId != declVarId)
            shad += {thisVarId};
        }
      }
    case ast : Let(c, ss): // for VarDecl
      {
        Aenv nenv = env;
        if (VarDecl(m, n, t, e) := c) {
          str n = getVarNameStr(c);
          <r_shad, r_beyo> = findVarScopingMismatches(e, nenv);
          shad += r_shad;
          beyo += r_beyo;
          nenv = add(nenv, n, c);
        } else {
          throw "unexpected LetClause <yieldTerm(c)>";
        }
        
        <r_shad, r_beyo> = findVarScopingMismatches(ss, nenv);
        shad += r_shad;
        beyo += r_beyo;
      }
    case ast : Param(_, _, _):
      {
        str n = getVarNameStr(ast);
        env = add(env, n, ast);
      }
    };
  return <shad, beyo>;
}

/** Finds all shadowed 'Var' references in 'tree'. This relies on
    'varId' annotations being in place. Returns a set of 'varId'. */
set[str] findShadowedVarsIn(AST tree, Aenv env) {
  <r_shad, _> = findVarScopingMismatches(tree, env);
  return r_shad;
}

set[str] findOutOfScopeVarsIn(AST tree, Aenv env) {
  <_, r_beyo> = findVarScopingMismatches(tree, env);
  return r_beyo;
}

AST renameShadowedVars(AST tree) {
  set[str] takenNames = findAllVarNamesIn(tree);
  set[str] sh_ids = findShadowedVarsIn(tree, emptyAenv);
  <tree, takenNames> = renameVarsById(tree, sh_ids, takenNames);
  return tree;
}

bool declaresVarId(list[AST] cs, str varId) {
  for (c <- cs) {
    if (c@varId == varId)
      return true;
  }
  return false;
}

tuple[list[AST], list[AST], AST] splitLet(list[AST] cs, str varId) {
  list[AST] ncs1 = [];
  list[AST] ncs2 = [];
  AST decl = Nop();
  bool seen = false;
  for (c <- cs) {
    if (c@varId == varId) {
      decl = c;
      seen = true;
    } else {
      if (seen)
        ncs2 += [c];
      else
        ncs1 += [c];
    }
  }
  return <ncs1, ncs2, decl>;
}

/** Removes a declaration for the specified variable, but retains any
    (initial) assignment. Any 'varId' annotation is preserved. Returns
    the removed declaration as the second tuple member, or Nop() if
    not found. Any 'VarDecl' node is returned without initial
    assignment. */
tuple[AST, AST, int] removeVarDecl(AST tree, str varId, int nid) {
  AST decl = Nop();

  tree = top-down-break visit (tree)
    {
    case ast : Let(c, b : seq(bs)):
      {
        if (declaresVarId(c, varId)) {
          AST asg = Nop();
          if (VarDecl(m, n, t, e) := c) {
            if (e != Undefined()) {
              c = preserveAnnos(VarDecl(m, n, t, Undefined()), c);
              <nid, n_juncId> = getJuncId(nid);
              asg = Assign(Var(VarName(n, t))[@varId = varId], e)
                [@juncId = n_juncId];
            }
          } else {
            assert false : "expected VarDecl";
          }


          if (asg == Nop()) {
	     ast = preserveAnnos(Block(b), ast);
	  }
	  else {
              ast = preserveAnnos(Block(seq([asg] + bs)), ast);
          }
          insert ast;
        } else {
          fail;
        }
      }

    case ast : Param(_, _, _):
      {
        assert (ast@varId != varId) : "cannot remove formal parameter";
      }
    };
  return <tree, decl, nid>;
}

/** Adds a declaration, as given, with or without an initializer
    expression. Preserves and creates junction IDs. */
tuple[AST, int] addVarDecl(AST tree, int juncId, AST decl, int nid) {
  tree = top-down-break visit (tree)
    {
    case ast : Let(c, b):
      {
        if (ast@juncId == juncId) {
          insert preserveAnnos(Let(c, seq([Let(decl, b)])), ast);
        } else
          fail;
      }
    case ast : Block(b):
      {
        if (ast@juncId == juncId) {
          insert preserveAnnos(Let(decl, b), ast);
        } else
          fail;
      }
    case ast : BlockExpr(b,rt):
      {
        if (ast@juncId == juncId) {
          int n_juncId; // for new Let
          <nid, n_juncId> = getJuncId(nid);
          AST n_let = Let(decl, b)[@juncId = n_juncId];
          insert preserveAnnos(BlockExpr(seq([n_let]), rt), ast);
        } else
          fail;
      }
    case AST ast:
      {
        if (ast@juncId? && (ast@juncId == juncId)) {
          int n_juncId; // for new Let
          <nid, n_juncId> = getJuncId(nid);
          AST n_let = Let(decl, seq([ast]))[@juncId = n_juncId];
          insert n_let;
        } else
          fail;
      }
    };

  return <tree, nid>;
}

/** Fixes scoping issues, namely shadowed and moved-beyond-their-scope
    variables, but generally only for the specified variables. This
    function should not be used in an expression context. */
tuple[AST, set[str]] vaiFixScoping(AST tree, set[str] e_ids) {
  //println(prettyMagnolia(tree));
  set[str] takenNames = findAllVarNamesIn(tree);
  //println("taken names <yieldTerm(takenNames)>");
  set[str] bad_ids = findOutOfScopeVarsIn(tree, emptyAenv);
  //println("all out-of-scope var IDs <yieldTerm(bad_ids)>");
  bad_ids = bad_ids & e_ids;
  //println("specific out-of-scope var IDs <yieldTerm(bad_ids)>");
  if (!isEmpty(bad_ids)) {
    int nid;
    <tree, nid> = addJuncIds(tree);
    for (varId <- bad_ids) {
      // This may introduce a new variable reference in the form of an
      // assignment.
      <tree, decl, nid> = removeVarDecl(tree, varId, nid);
      assert decl != Nop() : "expected declaration";
      int juncId = findJunction(tree, varId);
      // This might happen in some expression context, in which case
      // we would have to insert a 'BlockExpr'. But in our current use
      // cases this should not happen.
      assert juncId != 0 : "unsupported vaiFixScoping context";
      //println("found junction <juncId>");
      <tree, nid> = addVarDecl(tree, juncId, decl, nid);
    }
  }
  set[str] sh_ids = findShadowedVarsIn(tree, emptyAenv);
  //println("all shadowed var IDs <yieldTerm(sh_ids)>");
  if (isEmpty(bad_ids)) {
    // If we messed with scoping, cannot afford to restrict to
    // expression variables, as we may have introduced new shadowings.
    sh_ids = sh_ids & e_ids;
    //println("specific shadowed var IDs <yieldTerm(sh_ids)>");
  }
  <tree, takenNames> = renameVarsById(tree, sh_ids, takenNames);
  return <tree, takenNames>;
}

// --------------------------------------------------
// inlining of function-like objects
// --------------------------------------------------

/** The set of inlinings to do (which calls, which definitions) is
    specified explicitly as 'inlSet'. This should guarantee high
    enough inlining granularity for just about any use case. 

    The 'amInlining' flag indicates whether we are processing the body
    of a function inlining. That is the only time when we do
    substitutions. Can be used for sanity checking and such.

    The 'env' is used to track scoping, only to notice shadowing due
    to Let nodes in inlined bodies. We do not really need the
    declarations associated with names.

    The 'subst' map is used both for renamings (to avoid shadowing)
    and to substitute argument expressions for formal parameters.
*/
data FuiSt = FuiSt(set[InlObj] inlSet,
                   bool amInlining,
                   Aenv env,
                   map[str, AST] subst);

FuiSt newFuiSt(set[InlObj] inlSet) {
  return FuiSt(inlSet, false, emptyAenv, ());
}

/** If the variable name declared by 'decl' is in 'env', then produces
    a new non-clashing name and adds a substitution for its
    references. Adds to the environment as well. Returns the
    declaration, renamed as necessary.
 */
tuple[AST, FuiSt] renameClash(AST decl, FuiSt st) {
  str oname = getVarNameStr(decl);
  str nname = oname;

  if (defined(st.env, oname)) {
    nname = genUnusedName(st.env, oname);
    decl = setVarNameStr(decl, nname);
    vref = getVarRef(decl);
    st.subst[oname] = vref;
    //println("added subst for <oname> to <prettyPdbValue(vref)>");
  }

  st.env = add(st.env, nname, decl);

  return <decl, st>;
}

// Ordered 'domain' for lists.
list[int] ixDomain(list[&T] lst) {
  return upTill(size(lst));
}

/**
   Here 'procName' should be a type resolved name. If it matches an
   entry in 'inlSet', then the call shall be inlined. We trust
   that type checking has ensured that the call can be made.
   
   Out() and Upd() params require L-values (always variable
   references at this stage), whereas for Obs() R-values are
   also okay. For L-values and constant R-values we can
   certainly apply substitution in the procedure body. In
   Magnolia this is generally okay even for Obs() variables,
   since an Obs() parameter must be independent of any mutable
   parameter, and hence the called procedure cannot do anything
   to change the value of the Obs() expression.
   
   In this case, however, an Obs() argument expression might be
   a 'BlockExpr', which might mutate variables from its
   surrounding scope, including variables passed as L-value
   parameters to the inlined procedure. This makes it necessary
   for us to compute and store the value of Obs() arguments
   before substitution. This also avoids duplication of
   computation involved when substituting a complex expression
   multiple times within a procedure body.
   
   We use 'Let' to declare a variable for each Obs() argument
   (with the same mode), initializing with the argument
   expression. This is only appropriate for Obs() parameters.

   A function body is always an expression (see
   'ExprDeclarative'), and an inlined function application
   remains as an expression. All function parameters effectively
   have an Obs() mode, and what applies to such parameters in
   the case of procedures applies also here. It is just harder
   to declare the variables due to the expression context.
*/
tuple[AST, FuiSt] fuiInlineCall(AST call, AST decl, FuiSt st) {
  str callType;
  AST args;
  if (Apply(Fun(_), a) := call) {
    args = a;
    callType = "fun";
  } else if (Call(Proc(_), a) := call) {
    args = a;
    callType = "proc";
  } else {
    assert false;
  }

  AST body = getBodyOfDef(decl); // body expression
  list[AST] params = getParamsOfDef(decl);

  <nargs, st> = fuiInline(args, st);
  FuiSt b_st = st;
  // Note that the .args accessor gets the list out of a 'seq'.
  nargLst = nargs.args;
  list[AST] tempVarDecls = []; // for Let
  for (<par : Param(paramMode, paramName, paramType), narg> <- 
         [<params[i], nargLst[i]> | i <- ixDomain(params)]) {
    if (paramMode == Obs()) {
      // Declare a local variable for the parameter, and add a
      // substitution from formal parameter name to that variable if
      // necessary.
      AST tv = VarDecl(paramMode, paramName, paramType, narg);
      <tv, b_st> = renameClash(tv, b_st);
      tempVarDecls += [tv];
    } else {
      assert (callType != "fun");
      // Substitute expression in place of param reference.
      b_st.subst[getVarNameStr(par)] = narg;
    }
  }

  //println("body is <prettyPdbValue(body)>");
  b_st.amInlining = true;
  <body, _> = fuiInline(body, b_st);
  if (!isEmpty(tempVarDecls)) {
    if (callType == "fun") {
      AST let_s = MultiLet(tempVarDecls, seq([Return(body)]));
      body = BlockExpr(seq([let_s]), Unknown());
    } else if (callType == "proc") {
      body = MultiLet(tempVarDecls, body.arg0);
    } else {
      assert false;
    }
  }

  return <body, st>;
}

tuple[&T, FuiSt] fuiInline(&T tree, FuiSt st) {
  //println("substitutions: <prettyPdbValue(st.subst)> <st.amInlining>");
  //println("defs to inline: <prettyPdbValue(st.inlSet)>");

  tree = top-down-break visit(tree) 
    {

      // We do some variable renaming for hygiene, and may also
      // substitute expression in place of a (parameter) variable.
      // Care must be taken that substitutions are only in effect when
      // the relevant variable is in scope.
      //
      // Var(VarName(Name(leaf("y")), TypeName(Name(leaf("Int"))
    case ast : Var(VarName(Name(leaf(n)), t)): 
      {
        if (st.amInlining && (n in st.subst)) {
          AST nast = st.subst[n];
          insert preserveAnnos(nast, ast);
        }
      }

    case ast : Param(_, _, _):
      {
        assert !st.amInlining;
        st.env = add(st.env, getVarNameStr(ast), ast);
      }

      // Here we just keep track of variable names introduced by the
      // 'let'. We may also rename to avoid harmful shadowing due to
      // inlining.
      //
      // Don't know if semantics are as for 'let' or 'let*', but
      // assuming the latter.
    case ast : Let(c, bs): 
      {
        FuiSt l_st = st; // for Let
        if (VarDecl(m, n, t, e) := c) {
          if (e != Undefined()) {
            <c.arg3, l_st> = fuiInline(e, l_st);
          }
          str name = getVarNameStr(c);
          // Any existing substitution by this name will no longer
          // be valid within the Let.
          l_st.subst -= (name: Nop());
          if (l_st.amInlining) {
            // This is required. See for example the InlineProc7.mg example.
            <c, l_st> = renameClash(c, l_st);
          } else {
            l_st.env = add(l_st.env, name, c);
          }
        } else {
          throw "unexpected LetClause <yieldTerm(c)>";
        }
        // process entire seq(_) at once
        <bs, l_st> = fuiInline(bs, l_st);
        insert preserveAnnos(Let(c, bs), ast);
      }

      // Here we just want to know if this function-like is among
      // those to be inlined. If so, we merely remove it. The name
      // already includes formal parameters (to be unique across
      // overloads), and it should have been collected from the AST,
      // so it should match exactly.
      //
      // Removing a procedure from the AST presumably causes a problem
      // if we have a functionalized version of it, e.g. \f(procedure
      // inc upd type Int, 0). The procedure should be inlined into the
      // functionalized variant first.
      //
      // Any 'FunName'/'ProcName' already includes formal parameters, and also any
      // return type. Any return type then gets matched here as well.
      //
      // Rascal note: Calling "fail" in a THEN branch of an "if" seems
      // to transfer execution to the ELSE branch?? Hopefully from the
      // ELSE it transfers to the next "visit" clause.
    case ast : Define(c, a, b): 
      {
        assert !st.amInlining && isEmpty(st.env) && isEmpty(st.subst);
        if ((ProcClause(_, _) := c) ||
            (FunClause(_, _, _) := c)) {
          if (ast@inlId? &&
              isInInlineObjs(st.inlSet, ast@inlId)) {
            insert Nop();
          } else {
            <c, d_st> = fuiInline(c, st);
            <b, _> = fuiInline(b, d_st);
            insert(preserveAnnos(Define(c, a, b), ast));
          }
        } else fail;
      }

    case call : Call(Proc(_), _):
      {
        InlObj cand = InlNone();
        if (call@inlId?)
          cand = getInlineObjMatch(st.inlSet, call@inlId);
        if (InlRef(_, decl) := cand) {
          <call, st> = fuiInlineCall(call, decl, st);
          insert(call);
        } else
          fail;
      }

    case call : Apply(Fun(_), _):
      {
        InlObj cand = InlNone();
        if (call@inlId?)
          cand = getInlineObjMatch(st.inlSet, call@inlId);
        if (InlRef(_, decl) := cand) {
          <call, st> = fuiInlineCall(call, decl, st);
          insert(call);
        } else
          fail;
      }

    }; // end 'visit'

  return <tree, st>;
}

&T inlineFunLikeInConcept(&T tree, set[InlObj] inlSet) {
  st = newFuiSt(inlSet);
  <tree, _> = fuiInline(tree, st);
  return tree;
}

/** Applies function inlinings to marked declaration body. */
&T inlineFunLike(&T tree, set[InlObj] inlSet) {
  return top-down-break visit (tree)
    {
    case subTree : DeclBody(_):
      {
        if (subTree@inlThis?) {
          subTree = inlineFunLikeInConcept(subTree, inlSet);
          //subTree = delAnnotation(subTree, "inlThis");
          insert(subTree);
        }
      }
    };
}

// --------------------------------------------------
// 'return' translation
// --------------------------------------------------

anno int AST@retId;

tuple[int, int] getRetId(int nid) {
  return <nid + 1, nid>;
}

data DidRet = NoRet() | YesRet() | MaybeRet() | 
  BrRet(DidRet l, DidRet r);

/** Computes a "sequential" sum of return statuses. If a return may
    previously have happened in 'p', there is no way current statement
    status 'c' will change that fact. Chances of returning can only
    increase, not decrease.
 */
DidRet seqSum(DidRet p, DidRet c) {
  if (p == NoRet() && c == NoRet())
    return p;
  if (p == YesRet() || c == YesRet())
    return YesRet();
  if (p == MaybeRet() || c == MaybeRet())
    return MaybeRet();
  if (BrRet(_, _) := p || BrRet(_, _) := c)
    return MaybeRet();
  assert false : "unexpected";
}

/** Computes a "branch" sum of return statuses.
 */
DidRet brSum(DidRet l, DidRet r) {
  if (l == NoRet() && r == NoRet())
    return l;
  if (l == YesRet() && r == YesRet())
    return l;
  if (l == MaybeRet() || r == MaybeRet())
    return MaybeRet();
  if (BrRet(_, _) := l || BrRet(_, _) := r)
    return MaybeRet();
  return BrRet(l, r);
}

/** Inserts 'UnlessReturned' blocks and (to an extent) restructures
    code to avoid having to insert such blocks. Also eliminates dead
    code after 'Return' statements. Assumes that the input 't' does
    not contain any return blocks; i.e. all appearing 'Return'
    statements must be associated with a single (outer and enclosing)
    return block. */
tuple[list[AST], DidRet, int] rtranStats(list[AST] t, 
                                         DidRet st,
                                         int nid) {
  list[AST] nss = [];
  while (!isEmpty(t)) {
    <h, t> = headTail(t);
    //println("head is <prettyPdbValue(h)>");
    if (Return() := h) {
      <nid, retId> = getRetId(nid);
      h = h[@retId = retId];
      nss += [h];
      st = YesRet();
      break; // drop dead code
    } else if (Nop() := h) {
      ; // drop
    } else if (Block(seq(bss0)) := h) {
      list[AST] bss;
      <bss, bst, nid> = rtranStats(bss0, st, nid);
      nb = preserveAnnos(Block(seq(bss)), h);
      nss += [nb];
      st = seqSum(st, bst);
      if (st == YesRet())
        break; // drop dead code
      if ((st == MaybeRet()) && !isEmpty(t)) {
        <css, cst, nid> = rtranStats(t, NoRet(), nid); // continuation
        st = seqSum(st, cst);
        nc = UnlessReturned(seq(css));
        nss += [nc];
        break; // 't' fully consumed
      }
    } else if (Let(c, seq(bss0)) := h) {
      // Rascal feature warning: Without this declaration the Rascal
      // type checker complains with a type error if the sequence
      // happens to be empty. Rascal infers list[void] type.
      list[AST] bss;
      <bss, bst, nid> = rtranStats(bss0, st, nid);
      nb = preserveAnnos(Let(c, seq(bss)), h);
      nss += [nb];
      st = seqSum(st, bst);
      if (st == YesRet())
        break; // drop dead code
      if ((st == MaybeRet()) && !isEmpty(t)) {
        <css, cst, nid> = rtranStats(t, NoRet(), nid); // continuation
        st = seqSum(st, cst);
        nc = UnlessReturned(seq(css));
        nss += [nc];
        break; // 't' fully consumed
      }
    } else if (If(c, seq(ts), seq(es)) := h) {
      <t_ss, t_st, nid> = rtranStats(ts, st, nid);
      <e_ss, e_st, nid> = rtranStats(es, st, nid);
      bst = brSum(t_st, e_st);
      AST nb;
      if (BrRet(NoRet(), YesRet()) == bst) {
        <n_t_ss, t_st, nid> = rtranStats(t, NoRet(), nid);
        t_ss += n_t_ss;
        bst = brSum(t_st, e_st);
        nb = preserveAnnos(If(c, seq(t_ss), seq(e_ss)), h);
        t = [];
      } else if (BrRet(YesRet(), NoRet()) == bst) {
        <n_e_ss, e_st, nid> = rtranStats(t, NoRet(), nid);
        e_ss += n_e_ss;
        bst = brSum(t_st, e_st);
        nb = preserveAnnos(If(c, seq(t_ss), seq(e_ss)), h);
        t = [];
      } else {
        nb = preserveAnnos(If(c, seq(t_ss), seq(e_ss)), h);
      }
      nss += [nb];
      st = seqSum(st, bst);
      if (st == YesRet())
        break; // drop dead code
      if ((st == MaybeRet()) && !isEmpty(t)) {
        <css, cst, nid> = rtranStats(t, NoRet(), nid); // continuation
        st = seqSum(st, cst);
        nc = UnlessReturned(seq(css));
        nss += [nc];
        break; // 't' fully consumed
      }
    } else {
      nss += [h];
    }
  }
  return <nss, st, nid>;
}

/** Traverse in "return-flow" order to find "reaching returns" for
    each 'UnlessReturned'. Non-reaching returns can simply be removed.
    The data flow analysis is done in the same way as for variable
    assignments, because that's what we will use. With the exception
    of 'UnlessReturned', within which outer assignments are not of
    interest, as the runtime value will have been false. */
tuple[set[int], set[int]]
rtranReaching(AST tree,
              set[int] retSet, // of 'retId'
              set[int] reachSet) // of 'retId'
{
  top-down-break visit (tree) 
    {
    case ast : Return():
      {
        retSet += {ast@retId};
      }
    case ast : UnlessReturned(b):
      {
        <x, y> = rtranReaching(b, {}, {});
        reachSet += (retSet + y);
        retSet = x;
      }
    };
  return <retSet, reachSet>;
}

AST rtranRwReturns(AST tree, set[int] reachSet,
                   AST flagRef, AST newRet) {
  return top-down visit (tree) 
    {
    case ast : Return():
      {
        int retId = ast@retId;
        if (retId in reachSet) {
          insert newRet[@retId = retId];
        } else {
          insert Nop()[@retId = retId];
        }
      }
    case ast : UnlessReturned(b):
      {
        insert If(flagRef, b, Nop());
      }
    };
}

/** Translates away Return() syntax. Returns a single statement,
    whether a Block or something else. */
AST rtranReturnsInStats(list[AST] t) {
  <ss, _, _> = rtranStats(t, NoRet(), 0);
  AST tree = seq(ss);
  <_, reachSet> = rtranReaching(tree, {}, {});
  str flagName = "not_done"; // flag variable name
  AST flagType = SpecialType(Predicate());
  AST flagRef = Nop();
  AST newRet = Nop();
  if (isEmpty(reachSet)) {
    tree = Block(tree);
  } else {
    set[str] takenNames = findAllVarNamesIn(tree);
    <_, flagName> = genNameSym(takenNames, flagName);
    AST flagInit = Apply(Builtin(FunName(QName(seq([Name(leaf("TRUE"))])), seq([]), flagType), leaf("TRUE()")), seq([])); // ??
    AST flagDecl = VarDecl(Upd(), Name(leaf(flagName)), flagType, flagInit);
    tree = Let(flagDecl, tree);
    flagRef = Var(VarName(Name(leaf(flagName)), flagType));
    newRet = Assign(flagRef, Apply(Builtin(FunName(QName(seq([Name(leaf("FALSE"))])), seq([]), flagType), leaf("FALSE()")), seq([]))); // ??
  }
  tree = rtranRwReturns(tree, reachSet, flagRef, newRet);
  return tree;
}

bool hasBareReturns(AST tree) {
  top-down-break visit (tree) {
  case Return(): return true;
  case BlockExpr(_,_): ; // prune
  };
  return false;
}

bool hasBlockExprs(AST tree) {
  visit (tree) {
  case BlockExpr(_,_): return true;
  };
  return false;
}

/** Ensures that a (procedure) body has no (expressionless) 'return'
    statements, such that appear directly, without any enclosing block
    return construct. (Note that this transformation is not applicable
    for functions, as they cannot have "bare" 'return'
    appearances.) */
AST translateBareReturnsInBody(AST tree) {
  if (hasBareReturns(tree)) {
    tree = bottom-up visit (tree) 
      {
        // A non-Block statement in a Body is not allowed.
      case ast : Body(Block(seq(ss))):
        {
          newStat = rtranReturnsInStats(ss);
          insert preserveAnnos(Body(newStat), ast);
        }
      case ast : Body(s):
        {
          newStat = rtranReturnsInStats([s]);
          insert preserveAnnos(Body(newStat), ast);
        }
      };
  }
  return tree;
}

AST rtranReturnExp(AST tree) {
  tree = top-down-break visit (tree)
    {
    case Return(e):
      {
        AST asg = Assign(LiftVar(), e);
        AST ne = Splice([asg, Return()]);
        insert ne;
      }
    };
  tree = seqSplice(tree);
  return tree;
}

/** Returns a translation for the 'BlockExpr'. There may not be
    further return blocks contained within these statements.
    'LiftStatExpr', on the other hand, are allowed, to make it
    possible to first translate away all return blocks, and then
    translate away liftings. This function in fact returns a
    'LiftStatExpr'. */
AST rtranBlockExpr(tree : BlockExpr(sSeq, t)) {
  // Both Return(_) and BlockExpr(_) appear to receive a type
  // annotation during type checking.
  //termPrintln("block type ", t);
  sSeq = rtranReturnExp(sSeq);
  AST s = rtranReturnsInStats(sSeq.args);
  tree = LiftStatExpr(t, s);
  return tree;
}

/** Translates away any appearing 'BlockExpr' syntax. Note that an
    outermost 'LiftStatExpr' nodes might remain if there is no
    statement context to lift to. */
AST rtranBlockExprs(AST tree) {
  bool mustLift = false;
  tree = bottom-up visit (tree) 
    {
    case ast : BlockExpr(_,_):
      {
        mustLift = true;
        insert rtranBlockExpr(ast);
      }
    };
  if (mustLift)
    tree = doLifting(tree);
  return tree;
}

/** Translates away 'BlockExpr' syntax, but may leave an outermost
    one, as required. */
public AST translateBlockExprsInExpr(AST tree) {
  tree = rtranBlockExprs(tree);
  if (LiftStatExpr(t, s, _) := tree) {
    s = Return(tree)[@\type = t]; // give it statement context
    s = liftFromStat(s); // lift around the context
    tree = BlockExpr(seq([s]), t); // turn into an expression
  }
  return tree;
}

public AST translateBlockExprsInBody(AST tree) {
  if (hasBlockExprs(tree)) {
    tree = bottom-up visit (tree) 
      {
      case ast : Body(s : Block(_)): // statement context
        {
          s = rtranBlockExprs(s);
          insert preserveAnnos(Body(s), ast);
        }
      case ast : Body(e): // expression context
        {
          e = translateBlockExprsInExpr(e);
          insert preserveAnnos(Body(e), ast);
        }
      };
  }
  return tree;
}

/** Translates away all possible 'return' statements in the 'Body' of
    the argument. */
public AST translateReturnsInBody(AST tree) {
  tree = translateBlockExprsInBody(tree);
  return translateBareReturnsInBody(tree);
}

// --------------------------------------------------
// type inlining
// --------------------------------------------------

AST findTypeDefByNameNode(AST tree, AST name) {
  top-down-break visit (tree)
    {
    case def : Define(TypeClause(n), _, _):
      {
        if (n == name)
          return def;
      }
    };
  return Nop();
}

bool isStructDef(AST tree) {
  return Define(TypeClause(_), _, Body(Struct(_))) := tree;
}

bool isTypeNameNodeIn(AST tree, AST name) {
  top-down-break visit (tree)
    {
    case x : TypeName(_):
      {
        if (x == name)
          return true;
      }
    };
  return false;
}

bool isTypeNameNodeInAnnos(AST tree, AST name) {
  top-down-break visit (tree)
    {
    case x : TypeName(_):
      {
        if (x == name)
          return true;
      }
    case AST x:
      {
        if (x@\type? &&
            (x@\type == name))
          return true;
        else
          fail;
      }
    };
  return false;
}

// Returns a list of 'VarDecl'.
list[AST] getTypeComps(Define(TypeClause(_), _, Body(Struct(seq(typeFields))))) {
  list[AST] r = [];
  for (fld <- typeFields) {
    if (TypeField(varDecl) := fld)
      r += [varDecl];
    else
      throw "expected TypeField";
  }
  return r;
}

// 'typeName' is the type name, and 'typeDef' is its definition.
// 'typeComps' is the list of type components (from the definition),
// stored as a list of VarDecl. 'compMap' maps compound varIds to maps
// mapping field names to component varIds. 'nameMap' maps varIds to
// variable names; this is useful for keeping track of what names we
// come up with for component variables. 'varMap' keeps track of state
// for varId assignment (to ensure uniqueness); use <varMap, varId> =
// genIdSym(varMap, n) to get a new varId. 'gloNameMap' maps uninlined
// names to inlined names. 'oldGloMap' maps uninlined global names
// (nodes) to their corresponding definitions; the values may be
// annotated with additional information as required. 'newGloMap' is
// the same for type inlined names. 'funOfMap' is needed for adjusting
// FunOf nodes to reflect new signatures; maps uninlined FunOf "names"
// to uninlined global names. 'returnIx' states which type component
// should be returned upon encountering a Return of the inlined type.
data TyiSt = TyiSt(
                   AST typeName, 
                   AST typeDef,
                   list[AST] typeComps,
                   map[str, map[str, str]] compMap,
                   map[str, str] nameMap,
                   map[str, int] varMap,
                   map[AST, AST] gloNameMap,
                   map[AST, AST] oldGloMap,
                   map[AST, AST] newGloMap,
                   map[AST, AST] funOfMap,
                   int returnIx
                   );

TyiSt newTyiSt(AST typeName, AST typeDef) {
  typeComps = getTypeComps(typeDef);
  return TyiSt(typeName, typeDef, typeComps, (), (), (), (), (), (), (), -1);
}

/** Resets all function-body-local state. */
TyiSt localReset(TyiSt st) {
  st.returnIx = -1;
  return st;
}

list[tuple[AST, int]] typeCompsWithIx(TyiSt st) {
  return [<st.typeComps[i], i> | i <- ixDomain(st.typeComps)];
}

int getIxByName(TyiSt st, str n) {
  for (<tc, i> <- typeCompsWithIx(st)) {
    if (n == getVarNameStr(tc))
      return i;
  }
  throw "no type component named <n>";
}

tuple[AST, TyiSt] tyiExprComp(tree : Undefined(), 
                              int ix, TyiSt st) {
  return <tree, st>;
}

/** 'fldLst' is a list of 'Field' nodes. It is notable here that there
    are no recursive types in Magnolia, and hence the type components
    cannot be of the inlined type. 

Struct(TypeName(Name(leaf("S"))),
       seq([Field(Name(leaf("x")),
                  Var(VarName(Name(leaf("z")),
                              TypeName(Name(leaf("X")))))),
            Field(Name(leaf("y")),
                  Var(VarName(Name(leaf("z")),
                              TypeName(Name(leaf("X"))))))])
*/
tuple[AST, TyiSt] tyiExprComp(tree : Struct(_, seq(fldLst)), 
                              int ix, TyiSt st) {
  AST fld = fldLst[ix];
  if (Field(n, e) := fld)
    // Type inline within the expression.
    return tyiStat(e, st);
  else
    throw "unexpected field <prettyPdbValue(fld)>";
}

tuple[AST, TyiSt] tyiExprComp(tree : Var(VarName(nAst, t)),
                              int ix, TyiSt st) {
  str varId = tree@varId;
  map[str, str] cm = st.compMap[varId]; // field to component varId
  AST typeComp = st.typeComps[ix];
  str fldName = getVarNameStr(typeComp);
  str cVarId = cm[fldName];
  str cVarName = st.nameMap[cVarId];
  AST cType = getVarType(typeComp);
  tree = Var(VarName(Name(leaf(cVarName)), cType))[@varId = cVarId];
  return <tree, st>;
}

// Information stored by Struct type inlining, and required by DotOp
// type inlining: old field name, component index, component VarDecl.
anno rel[str, int, AST] AST@tyiFldRel;

// The 'DotOp' expression 'e' is of the inlined type since the field
// 'nAst' is of that type. The field access, then, needs splitting
// into components.
tuple[AST, TyiSt] tyiExprComp(tree : DotOp(e, nAst, typ),
                              int ix, TyiSt st) {
  AST e_typeName = getExprType(e); // enclosing TypeName
  str orig_n = getNameStr(nAst); // original field name
  <e, st> = tyiStat(e, st); // inline within expression
  AST def = st.oldGloMap[e_typeName]; // old def for enclosing 'struct'
  AST vd = getOneFrom(def@tyiFldRel[orig_n, ix]);
  if (VarDecl(_, vnAst, vt, _) := vd) {
    nAst = vnAst;
  } else
    throw "expected VarDecl";
  tree = preserveAnnos(DotOp(e, nAst, typ), tree);
  return <tree, st>;
}

// A block expression of the inlined type. The type is actually
// determined by the 'return' statement(s) within, and hence it is at
// the 'return' points where the splitting must happen.
tuple[AST, TyiSt] tyiExprComp(tree : BlockExpr(_,_),
                              int ix, TyiSt st) {
  int oldIx = st.returnIx;
  st.returnIx = ix;
  <tree, st> = tyiStat(tree, st);
  st.returnIx = oldIx;
  return <tree, st>;
}

/** Intermediate language. 
 */
data AST = TyiExprOf(AST expr, int ix) |
  TyiNames(list[AST] names);

tuple[list[AST], TyiSt] tyiArgs(list[AST] parLst, list[AST] argLst, TyiSt st) {
  //println("tyiArgs " + prettyPdbValue(<parLst, argLst>));
  list[AST] nargLst = [];
  for (<par, arg> <- [<parLst[i], argLst[i]> | i <- ixDomain(parLst)]) {
    AST t = getVarType(par);
    if (t == st.typeName) {
      int ix;
      for (<tc, ix> <- typeCompsWithIx(st)) {
        AST narg;
        //println("component " + prettyPdbValue(tc));
        <narg, st> = tyiExprComp(arg, ix, st);
        nargLst += [narg];
      }
    } else {
      <arg, st> = tyiStat(arg, st);
      nargLst += [arg];
    }
  }
  return <nargLst, st>;
}

tuple[AST, TyiSt] tyiApplyFunOf(AST ast : 
                                Apply(funOfName : FunOf(procName, leaf(i_s)), 
                                      seq(argLst)),
                                int ix, TyiSt st) {
  int i = toInt(i_s);
  AST oFunName = st.funOfMap[funOfName];
  AST oFunDef = st.oldGloMap[oFunName];
  list[AST] funParLst = getParamsOfDef(oFunDef);          
  AST nProcName = st.gloNameMap[procName];
  AST oProcDef = st.oldGloMap[procName];
  // Procedure parameter list differs from the function parameter
  // list.
  list[AST] procParLst = getParamsOfDef(oProcDef);
  //println("procParLst <prettyPdbValue(procParLst)>");
  int j = tyiAdjustParamIndex(procParLst, i, ix, st);
  list[AST] nargLst;
  <nargLst, st> = tyiArgs(funParLst, argLst, st);
  str j_s = "<j>";
  ast = preserveAnnos(Apply(FunOf(nProcName, leaf(j_s)), 
                            seq(nargLst)), ast);
  return <ast, st>;
}

/** FunOf should only appear as Body(Apply(FunOf(_, _), _)), but in
    our case we may also have Body(TyiExprOf(Apply(FunOf(_, _),
    _), _)). The latter case is assumed here. */
tuple[AST, TyiSt] tyiExprComp(AST ast : 
                              Apply(funOfName : FunOf(_, _), _),
                              int ix, TyiSt st) {
  assert (funOfName in st.funOfMap);
  return tyiApplyFunOf(ast, ix, st);
}

/** Here we are dealing with applications of functions that yield the
    inlined type. We should already have type inlined functions
    corresponding to the specified one. */
tuple[AST, TyiSt] tyiExprComp(tree : Apply(Fun(funName), seq(argLst)),
                              int ix, TyiSt st) {
  AST tyiNames = st.gloNameMap[funName];
  AST nFunName;
  if (TyiNames(funLst) := tyiNames) {
    nFunName = funLst[ix];
  } else
    throw "expected TyiNames";
  AST oldDef = st.oldGloMap[funName]; // 'Define' with old params
  list[AST] parLst = getParamsOfDef(oldDef);
  list[AST] nargLst;
  <nargLst, st> = tyiArgs(parLst, argLst, st);
  tree = preserveAnnos(Apply(Fun(nFunName), seq(nargLst)), tree);
  return <tree, st>;
}

/** Produces the specified type component of an expression. 
    The 'tree' expression must be of the inlined type.
*/
default tuple[AST, TyiSt] tyiExprComp(AST tree, 
                                      int ix, TyiSt st) {
  throw "tyiExprComp undefined for <prettyPdbValue(tree)>";
}

// 'tree' is to be an expression of the inlined type, and therefore it
// will be split to multiple expressions, one for each component.
tuple[list[AST], TyiSt] tyiExpr(AST tree, TyiSt st) {
  list[AST] astLst = [];
  int ix = 0;
  for (tc <- st.typeComps) {
    <ast, st> = tyiExprComp(tree, ix, st);
    astLst += [ast];
    ix += 1;
  }
  return <astLst, st>;
}

int tyiAdjustParamIndex(list[AST] parLst, int n, int ix, TyiSt st) {
  int m = 0;
  int i = 0;
  int c = size(st.typeComps);
  for (Param(_, _, t) <- parLst) {
    if (t == st.typeName) {
      if (i == n)
        return m + ix;
      m += c;
    } else {
      if (i == n)
        return m;
      m += 1;
    }
    i += 1;
  }
  return m;
}

// Unlike the naming implies, 'tree' need not be a statement. It can
// be any AST subtree, really, that is not an expression of the
// inlined type. The semantics is that 'tree' is not immediately split
// into type components, but inlining is done within. For example, it
// is okay to pass a 'seq' of statements also as 'tree'. You may also
// pass a 'BlockExpr' long as it is not of the inlined type.
tuple[AST, TyiSt] tyiStat(AST tree, TyiSt st) {
  // 'Block' has no immediate expression, and hence we need not handle
  // it. 'If' has an expression, but it cannot be of a 'struct' type,
  // and hence we need not handle it. 'Assert' likewise has a boolean
  // expression.
  tree = top-down-break visit (tree)
    {
    case ast : Let(c, b):
      {
        if (VarDecl(m, nAst, t, e) := c) {
          if (t == st.typeName) {
            str n = getVarNameStr(c);
            map[str, str] cm = (); // field name to varId
            int ix = 0;
            for (vd : VarDecl(vm, vnAst, vt, vv) <- st.typeComps) {
              <ec, st> = tyiExprComp(e, ix, st);
              str vn = getVarNameStr(vd);
              str cn = "<n>_<vn>"; // may rename later for uniqueness
              <st.varMap, varId> = genIdSym(st.varMap, cn);
              c = VarDecl(m, Name(leaf(cn)), vt, ec)[@varId = varId];
              cm[vn] = varId;
              st.nameMap[varId] = cn;
              ix += 1;
            }
            st.compMap[c@varId] = cm;
          } else {
            <e, st> = tyiStat(e, st);
            c = preserveAnnos(VarDecl(m, nAst, t, e), c);
          }
        } else {
          throw "unexpected LetClause <yieldTerm(c)>";
        }
        <b, st> = tyiStat(b, st);
        ast = preserveAnnos(Let(c, b), ast);
        insert ast;
      }

      // Elsewhere in the inliner we may still be assuming that the
      // L-value of Assign can only be a 'Var', but in fact also
      // 'DotOp' may appear, at least.
      //
      // xxx to deal with variables appearing both on LHS and RHS
    case ast : Assign(lv, rv):
      {
        //println(prettyPdbValue(ast));
        AST t = getExprType(lv);
        //println(prettyPdbValue(t));
        if (t == st.typeName) {
          list[AST] asLst = []; // list of Assign
          for (<tc, ix> <- typeCompsWithIx(st)) {
            <lv_c, st> = tyiExprComp(lv, ix, st);
            <rv_c, st> = tyiExprComp(rv, ix, st);
            asLst += [Assign(lv_c, rv_c)];
          }
          insert Block(seq(asLst));
        } else
          fail;
      }

      // Note this this expression itself is not of the inlined type,
      // or it would be handled by tyiExprComp.
    case ast : DotOp(e, Name(leaf(n), _)):
      {
        //println(prettyPdbValue(ast));
        AST t = getExprType(e);
        //println(prettyPdbValue(t));
        if (t == st.typeName) {
          int ix = getIxByName(st, n);
          <e, st> = tyiExprComp(e, ix, st);
          insert e;
        } else
          fail;
      }

    case ast : Return(e):
      {
        AST t = ast@\type;
        if (t == st.typeName) {
          int returnIx = st.returnIx;
          assert returnIx != -1;
          st.returnIx = -1; // clear for better sanity checking
          <e, st> = tyiExprComp(e, returnIx, st);
          st.returnIx = returnIx;
          ast = Return(e)[@\type = t];
          insert ast;
        } else
          fail;
      }

    case ast : Call(Proc(procName), seq(argLst)):
      {
        //println("call " + prettyPdbValue(ast));
        if (isTypeNameNodeIn(procName, st.typeName)) {
          AST oldDef = st.oldGloMap[procName]; // 'Define' with old params
          list[AST] parLst = getParamsOfDef(oldDef);
          list[AST] nargLst;
          <nargLst, st> = tyiArgs(parLst, argLst, st);
          procName = st.gloNameMap[procName];
          ast = preserveAnnos(Call(Proc(procName), seq(nargLst)), ast);
          insert ast;
        } else
          fail;
      }

      // Even in the case where a function application expression is
      // not of the inlined type, we may have to rewrite the call to
      // reflect type inlined signatures. Only the function name and
      // the arguments are affected.
    case ast : Apply(Fun(funName), seq(argLst)):
      {
        if (funName in st.gloNameMap) {
          AST nFunName = st.gloNameMap[funName];
          AST oldDef = st.oldGloMap[funName]; // 'Define' with old params
          list[AST] parLst = getParamsOfDef(oldDef);
          list[AST] nargLst;
          <nargLst, st> = tyiArgs(parLst, argLst, st);
          ast = preserveAnnos(Apply(Fun(nFunName), seq(nargLst)), ast);
          insert ast;
        } else 
          fail;
      }

    case ast : Apply(funOfName : FunOf(_, _), _):
      {
        if (funOfName in st.funOfMap) {
          // The '0' argument should not matter here.
          <ast, st> = tyiApplyFunOf(ast, 0, st);
          insert ast;
        } else
          fail;
      }

    case ast : TyiExprOf(e, ix):
      {
        <e, st> = tyiExprComp(e, ix, st);
        insert e;
      }
    };
  return <tree, st>;
}

AST ensureUniqueDefName(AST nFunLikeName, TyiSt st) {
  <s, i> = getBaseNameWithNum(getNameStr(nFunLikeName));
  while (nFunLikeName in st.oldGloMap) {
    i += 1;
    nFunLikeName = setNameStr(nFunLikeName, makeBaseNameWithNum(s, i));
  }
  return nFunLikeName;
}

/** Does type inlining on the name of a function-like, e.g. on a
    ProcName or FunName node. (Such names appear e.g. in ProcClause,
    FunClause, and FunOf nodes.) Ensures that the inlined name is
    unique (it is possible for a clashing overload to already exist).
    Adds a mapping to gloNameMap. */
tuple[AST, TyiSt] tyiFunLikeName(AST funLikeName, TyiSt st) {
  list[AST] parLst = getParamsOfName(funLikeName);
  list[AST] nparLst = [];
  for (par : AnonParam(m, t) <- parLst) {
    if (t == st.typeName) {
      for (tc <- st.typeComps) {
        par = AnonParam(m, getVarType(tc));
        nparLst += [par];
      }
    } else
      nparLst += [par];
  }

  AST nFunLikeName = setParamsOfName(funLikeName, nparLst);
  nFunLikeName = ensureUniqueDefName(nFunLikeName, st);
  st.gloNameMap[funLikeName] = nFunLikeName;
  return <nFunLikeName, st>;
}

/** Splits compound parameters. 'tree' must be a function-like
    declaration. Parameters must already be varId annotated. The new
    parameters are set to 'tree', and varIds are also assigned.
    compMap, nameMap, and varMap are updated as appropriate.
 */
tuple[AST, TyiSt] tyiFunLikeParams(AST tree, TyiSt st) {
  list[AST] params = getParamsOfDef(tree);
  list[AST] nparams = [];
  //println(prettyPdbValue(params));
  //println(prettyPdbValue(st.typeComps));
  for (par : Param(m, nAst, t) <- params) {
    if (t == st.typeName) {
      str n = getVarNameStr(par);
      map[str, str] cm = (); // field name to varId
      for (vd : VarDecl(vm, vnAst, vt, vv) <- st.typeComps) {
        str vn = getVarNameStr(vd);
        str cn = "<n>_<vn>"; // may rename later for uniqueness
        <st.varMap, varId> = genIdSym(st.varMap, cn);
        AST cd = Param(m, Name(leaf(cn)), vt)[@varId = varId];
        cm[vn] = varId;
        st.nameMap[varId] = cn;
        nparams += [cd];
      }
      st.compMap[par@varId] = cm;
    } else {
      nparams += [par];
    }
  }
  //println(prettyPdbValue(nparams));
  <procName, st> = tyiFunLikeName(getDefName(tree), st);
  //println(prettyPdbValue(procName));
  AST ntree;
  ntree = setParamsOfDef(tree, nparams);
  ntree = setDefName(ntree, procName);
  st.newGloMap[procName] = ntree;
  return <ntree, st>;
}

// 'tree' is a 'Define' of a function-like.
tuple[AST, TyiSt] tyiFunLikeBody(AST tree, TyiSt st) {
  st = localReset(st);

  keepCompMap = st.compMap;
  keepNameMap = st.nameMap;

  AST body = getBodyOfDef(tree);
  // If 'isProc(tree)' then 'body' is a statement, otherwise it is an
  // expression of non-inlined type, in which case we treat it the
  // same as a statement, looking for something within to inline.
  <body, st> = tyiStat(body, st);
  tree = setBodyOfDef(tree, body);

  // Fix variable shadowing based on varId.
  tree = renameShadowedVars(tree);

  st.compMap = keepCompMap;
  st.nameMap = keepNameMap;

  return <tree, st>;
}

// xxx the following are for later use as a maintainability optimization -- the idea is to avoid splitting user implemented functions of inlined return type

/** Produces the canonical proceduralization of the given function
    definition. Given the same name as the function (but in a
    different namespace). Introduces a new 'out' parameter. */
AST proceduralize(tree : 
                  Define(FunClause(n, ParamList(seq(parLst)), rt), 
                         a, Body(b))) {
  set[str] takenNames = findAllVarNamesIn(tree);
  <takenNames, outName> = genNameSym(takenNames, "r");
  AST procName = ProcName(n, seq([AnonParam(Out(), rt)] + parLst));
  AST outParam = Param(Out(), Name(leaf(outName)), rt);
  list[AST] nparLst = [outParam] + parLst;
  AST procClause = ProcClause(procName, ParamList(seq(nparLst)));
  AST procStat = Assign(Var(VarName(Name(leaf(outName)), rt)), b);
  // Should any attributes be preserved? Or should this be marked as
  // Derived()?
  return Define(procClause, seq([]), Body(Block(seq([procStat]))));
}

/** Defines a function in terms of FunOf and the given canonical
    procedure. The function is given the same name as the
    procedure, and it is marked Derived(). */
AST makeFunOfCanonProc(tree : 
                       Define(ProcClause(procName, 
                                         ParamList(seq(parLst))), 
                              _, _)) {
  // Type of first procedure parameter, due to canonical form.
  AST rt = getVarType(parLst[0]);
  AST name = getNameNode(procName);
  // All except Out() procedure params, in the same order.
  list[AST] nparLst = [par | AST par : Param(m, n, t) <- parLst, m != Out()];
  // Variable references to nparLst parameters.
  list[AST] argLst = [Var(VarName(n, t)) | 
                      AST par : Param(m, n, t) <- nparLst];
  // nparLst contents, in a different format, among other things
  list[AST] aparLst = [AnonParam(m, t) | AST par : Param(m, n, t) <- nparLst];
  AST funName = FunName(name, seq(aparLst), rt);
  return Define(FunClause(funName,
                          ParamList(seq(nparLst)),
                          rt),
                seq([Derived()]),
                Body(Apply(FunOf(procName,
                                 leaf("0")),
                           seq(argLst))));
}
/* example result:
Define(FunClause(FunName(Name(leaf("q")),
                         seq([AnonParam(Obs(),
                                        TypeName(Name(leaf("S"))))]),
                         TypeName(Name(leaf("S")))),
                 ParamList(seq([Param(Obs(),
                                      Name(leaf("a")),
                                      TypeName(Name(leaf("S"))))])),
                 TypeName(Name(leaf("S")))),
       seq([Derived()]),
       Body(Apply(FunOf(ProcName(Name(leaf("q")),
                                 seq([AnonParam(Upd(),
                                                TypeName(Name(leaf("S"))))])),
                        leaf("0")),
                  seq([Var(VarName(Name(leaf("a")),
                                   TypeName(Name(leaf("S")))))]))))
*/

AST getFunOfNameFromBody(AST tree) {
  top-down-break visit (tree)
    {
    case Body(Apply(n : FunOf(_, _), _)):
      {
        return n;
      }
    };
  return Nop();
}

list[AST] tyiFunction(AST tree :
                      Define(FunClause(FunName(n, ap, _), np, _), a, Body(e)), 
                      TyiSt st) {
  list[AST] defLst = [];
  for (<tc, ix> <- typeCompsWithIx(st)) {
    AST thisRt = getVarType(tc);
    AST thisBody = Body(TyiExprOf(e, ix));
    str n_s = getNameStr(n);
    str f_s = getVarNameStr(tc);
    n_s = "<n_s>_<f_s>";
    AST thisName = FunName(Name(leaf(n_s)), ap, thisRt);
    thisName = ensureUniqueDefName(thisName, st);
    AST def = Define(FunClause(thisName, np, thisRt), a, thisBody);
    def = preserveAnnos(def, tree);
    defLst += [def];
  }
  return defLst;
}

/** Whether definition body requires type inlining. */
anno bool AST@tyiDoBody;

/** Records information about signatures/declarations, and on the
    inlinings of types in them, and also applies said inlinings. */
tuple[AST, TyiSt] tyiSignatures(AST tree, TyiSt st) {
  AST name = st.typeName;

  tree = top-down-break visit (tree)
    {
      // Handling of declarations of 'struct' types containing inlined
      // type. We "unpack" the inlined type; any field accesses into
      // values of the enclosing struct type will require the new
      // field naming information.
    case x : Define(TypeClause(typeName), _, _):
      {
        if ((typeName != name) && isTypeNameNodeIn(x, name)) {
          str tn = getNameStr(typeName); // enclosing 'struct' type name
          list[AST] flds0 = getTypeComps(x); // list of VarDecl
          set[str] takenNames = {getVarNameStr(vd) | AST vd <- flds0};
          list[AST] flds1 = [];
          rel[str, int, AST] tyiFldRel = {};
          // Could use 'append' here, with 'for' as an expression, but
          // having some syntactic problems. (Rascal bug?)
          for (fld : VarDecl(m, nAst, t, e) <- flds0) {
            str n = getVarNameStr(fld); // field name
            if (t == name) {
              for (<vd : VarDecl(vm, vnAst, vt, vv), ix> <- 
                     typeCompsWithIx(st)) {
                str vn = getVarNameStr(vd); // field name
                str cn = "<n>_<vn>";
                // Make the name unique in its 'struct' context.
                <takenNames, cn> = genNameSym(takenNames, cn);
                AST n_fld = VarDecl(m, Name(leaf(cn)), vt, e);
                flds1 += [n_fld];
                tyiFldRel += {<n, ix, n_fld>};
              }
            } else {
              flds1 += [fld];
            }
          }
          list[AST] t_flds = [TypeField(vd) | AST vd <- flds1];
          AST tyiDef = setBodyOfDef(x, Struct(seq(t_flds)));
          st.gloNameMap[typeName] = typeName;
          st.oldGloMap[typeName] = x[@tyiFldRel = tyiFldRel];
          st.newGloMap[typeName] = tyiDef;
          insert tyiDef;
        }
      }

    case x : Define(sig : ProcClause(n, _), _, _):
      {
        if (isTypeNameNodeInAnnos(x, name)) {
          // Assign unique identifiers to variables, so as not to have
          // to be concerned with scoping so much. As we split
          // variables, we'll have to assign new IDs for the
          // components, and track associations. We assign IDs before
          // new signatures are there, but then assign IDs also to the
          // variables in new signatures (those variables are not yet
          // referred to in the body at that point).
          <x, st.varMap> = assignVarIds(x, emptyAenv, st.varMap);
          if (isTypeNameNodeIn(sig, name)) {
            <x, st> = tyiFunLikeParams(x, st);
          }
          x = x[@tyiDoBody = true];
          insert x;
        }
      }

    case ast : Define(sig : FunClause(n, _, rt), _, _):
      {
        if (isTypeNameNodeInAnnos(ast, name)) {
          AST funOfName = getFunOfNameFromBody(ast);
          if (funOfName != Nop()) {
            // Needed to determine function corresponding to a FunOf.
            st.funOfMap[funOfName] = n;
          }
          AST x = ast[@tyiDoBody = true];
          if (rt == name) {
            list[AST] funLst = tyiFunction(x, st);
            list[AST] nFunLst = [];
            list[AST] funNames = [];
            for (f <- funLst) {
              <f, st.varMap> = assignVarIds(f, emptyAenv, st.varMap);
              <f, st> = tyiFunLikeParams(f, st);
              nFunLst += [f];
              AST funName = getDefName(f);
              funNames += [funName];
            }
            // Exceptionally, the mapping of old-to-new names is
            // 1-to-n.
            st.gloNameMap[getDefName(x)] = TyiNames(funNames);
            x = Splice(nFunLst);
          } else {
            <x, st.varMap> = assignVarIds(x, emptyAenv, st.varMap);
            if (isTypeNameNodeIn(sig, name)) {
              <x, st> = tyiFunLikeParams(x, st);
            }
          }
          insert x;
        }
      }
      
    };

  tree = seqSplice(tree);

  return <tree, st>;
}

/** Process the bodies of all functions and procedures marked as
    requiring type inlining, making the bodies match inlined
    signatures. */
tuple[AST, TyiSt] tyiBodies(AST tree, TyiSt st) {
  AST name = st.typeName;

  tree = top-down-break visit (tree)
    {
    case x : Define(ProcClause(_, _), _, _):
      {
        if (x@tyiDoBody? && x@tyiDoBody) {
          <x, st> = tyiFunLikeBody(x, st);
          insert x;
        }
      }
    case x : Define(FunClause(_, _, _), _, _):
      {
        if (x@tyiDoBody? && x@tyiDoBody) {
          <x, st> = tyiFunLikeBody(x, st);
          insert x;
        }
      }
    };

  return <tree, st>;
}

TyiSt tyiOldGloMap(AST tree, TyiSt st) {
  top-down-break visit (tree)
    {
    case x : Define(_, _, _):
      {
        st.oldGloMap[getDefName(x)] = x;
      }
    }
  return st;
}

AST inlineTypeInConcept(AST tree, AST obj) {
  AST decl;
  AST name;
  if (isDecl(obj)) {
    decl = obj;
    name = getDefName(decl);
  } else {
    name = obj;
    decl = findTypeDefByNameNode(tree, name);
  }

  if (decl == Nop())
    // Undefined, so cannot inline.
    return tree;
  if (!isStructDef(decl))
    // Not a 'struct', so cannot inline.
    return tree;

  // use prettyPdbValueAnnos to see annos
  //println("INLINING <prettyPdbValue(name)>\nDEFINED AS <prettyPdbValue(decl)>\nIN CONCEPT\n<prettyPdbValue(tree)>");

  TyiSt st = newTyiSt(name, decl);

  st = tyiOldGloMap(tree, st);

  // Do a pre-visit over signatures before processing fun-like bodies.
  // We must record both the old and the inlined signatures. Note:
  // This code would cause a Rascal parse error if not moved into a
  // separate function (a Rascal bug).
  <tree, st> = tyiSignatures(tree, st);

  <tree, st> = tyiBodies(tree, st);

  return tree;
}

/**
   The type specified by 'obj' must specify a 'struct' type, otherwise
   'tree' is returned unchanged.

   There must be an 'inlThis' in the 'DeclBody' (and only in that one)
   in which the type to inline is defined.

   tree:: syntax tree in which to inline
   obj:: identifies type to inline (either name or declaration)
 */
AST inlineType(AST tree, AST obj) {
  //println("asked to inline <prettyPdbValue(obj)>");

  return top-down-break visit (tree)
    {
    case subTree : DeclBody(_):
      {
        if (subTree@inlThis?) {
          subTree = inlineTypeInConcept(subTree, obj);
          //subTree = delAnnotation(subTree, "inlThis");
          //subTree = delAnnotationsRec(subTree, {"inlThis"});
          insert(subTree);
        }
      }
    };
}


// --------------------------------------------------
// find changes
// --------------------------------------------------
data Change
	= Replace(loc location, AST old, AST new)
	| Delete(loc location, AST old)
	| Insert(loc location, AST new)
	| Copy(loc location, loc from, AST old)
	| Nothing(loc location)
	;
	
public list[Change] treeEditDiff(AST oldTree, AST newTree) {
	return treeEditDiffHelper(oldTree, newTree, []);	
}

list[Change] treeEditDiffHelper(AST oldTree, AST newTree, list[Change] changes) {
	if(oldTree != newTree) {
		changeLoc = oldTree@\loc ? unknownLoc;
		switch(<oldTree, newTree>) {
			case <seq([]), seq(ns)>: {
				changes += Replace(changeLoc, oldTree, newTree);
			}
			case <seq(os), seq(ns)>: {
				if(size(os) == size(ns)) {
					for(i <- index(os)) {
						changes = treeEditDiffHelper(os[i], ns[i], changes);
					}
				}
				else {
					for(i <- index(os)) {
						if(i in index(ns))
							changes += replace(os[i], ns[i]);
						else
							changes += delete(os[i]);
					}
					for(i <- index(ns) - index(os)) {
						changes += insertAfter(changeLoc, last(os), ns[i]);
					}
				}
			}
			case <leaf(_), leaf(_)>:
				changes += Replace(changeLoc, oldTree, newTree);
			default: {
				if(consOf(oldTree) == consOf(newTree) && arity(oldTree) == arity(newTree)) {
					int i = 0;
					while(i < arity(oldTree)) {
						changes = treeEditDiffHelper(oldTree[i], newTree[i], changes);
						i = i + 1;
					}
				}
				else
					changes += Replace(changeLoc, oldTree, newTree);
			}
		}
	}

	return changes;	
}

Change delete(AST oldTree) {
	changeLoc = oldTree@\loc ? unknownLoc;
	if(changeLoc != unknownLoc) {
		return Delete(changeLoc, oldTree);
	}
	else
		termPrintln("No location for: ", oldTree);
	return Nothing(changeLoc);
}

Change replace(AST oldTree, AST newTree) {
	changeLoc = oldTree@\loc ? unknownLoc;
	if(changeLoc != unknownLoc) {
		return Replace(changeLoc, oldTree, newTree);
	}
	else
		termPrintln("No location for: ", oldTree);
	return Nothing(changeLoc);
}

Change insertAfter(loc changeLoc, AST oldTree, AST newTree) {
	if(changeLoc == unknownLoc)
		changeLoc = oldTree@\loc ? unknownLoc;
	if(changeLoc != unknownLoc) {
		changeLoc.offset = changeLoc.offset + changeLoc.length;
		changeLoc.length = 0;
		return Insert(changeLoc, newTree);
	}
	else
		termPrintln("No location for: ", oldTree);
	return Nothing(changeLoc);
}
