/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.lib;

import java.io.IOException;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicLong;

import nuthatch.rascal.adapter.ValuesCursor;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IValueFactory;
import org.rascalmpl.value.type.Type;
import org.magnolialang.infra.Infra;
import org.nuthatchery.pgf.trees.PrettyPrinter;
import org.nuthatchery.pica.errors.ImplementationError;
import org.rascalmpl.interpreter.IEvaluatorContext;
import org.rascalmpl.values.ValueFactoryFactory;

public class MagnoliaCompileLib {
	private final IValueFactory vf;
	private static final AtomicLong idCounter = new AtomicLong();
	private final PrettyPrinter<IValue, Type> printer = Pretty.makePrinter();


	public MagnoliaCompileLib() {
		this(ValueFactoryFactory.getValueFactory());
	}


	public MagnoliaCompileLib(IValueFactory vf) {
		this.vf = vf;
	}


	public IString getBundledResource(IString filePath, IEvaluatorContext ctx) {
		try {
			return vf.string(Infra.get().readFile(filePath.getValue()));
		}
		catch(IOException e) {
			throw new ImplementationError("Internal pkg not found: " + filePath, e);
		}
	}


	public IString makeId() {
		return makeSUID();
	}


	/**
	 * @return A sequential unique id (unique only within the JVM)
	 */
	public IString makeSUID() {
		return vf.string(String.valueOf(idCounter.getAndIncrement()));
	}


	/**
	 * @return A universally unique id (statistically unique in the entire
	 *         universe)
	 */
	public IString makeUUID() {
		return vf.string(UUID.randomUUID().toString());
	}


	public IString prettyMagnolia(IConstructor tree) {
		return vf.string(printer.toString(new ValuesCursor(tree)));
	}


	public static void refresh() {
		idCounter.set(0);
	}
}
