package org.magnolialang.magnolia.lib;

import static nuthatch.rascal.pattern.StaticValuesPatternFactory._;
import static org.magnolialang.magnolia.AbstractSyntax.*;
import static org.magnolialang.magnolia.grammar.MagnoliaPatterns.*;
import static org.nuthatchery.pgf.trees.PPUtil.*;

import java.util.Collection;

import nuthatch.library.ActionFactory;
import nuthatch.library.FactoryFactory;
import nuthatch.rascal.adapter.ValuesCursor;
import nuthatch.rascal.adapter.ValuesWalker;
import nuthatch.rascal.pattern.ValuesBuildContext;
import nuthatch.rascal.pattern.impl.ValuesPatternFactory;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.type.Type;
import org.eclipse.jdt.annotation.Nullable;
import org.nuthatchery.pgf.config.TokenizerConfig;
import org.nuthatchery.pgf.config.TokenizerConfigBase;
import org.nuthatchery.pgf.plumbing.ForwardStream;
import org.nuthatchery.pgf.tokens.CategoryStore;
import org.nuthatchery.pgf.tokens.DataToken;
import org.nuthatchery.pgf.tokens.Token;
import org.nuthatchery.pgf.trees.PPBuilder;
import org.nuthatchery.pgf.trees.PrettyPrinter;
import org.nuthatchery.pgf.trees.TreeToStream;
import org.nuthatchery.pica.terms.TermFactory;

public class Pretty {
	protected static final ActionFactory<IValue, Type, ValuesCursor, ValuesWalker> af = FactoryFactory.getActionFactory(IValue.class, Type.class, ValuesCursor.class, ValuesWalker.class);
	protected static final ValuesPatternFactory pf = ValuesPatternFactory.getInstance();
	protected static final ValuesBuildContext context = new ValuesBuildContext(TermFactory.vf, TermFactory.ts);

	static final TokenizerConfig config = new TokenizerConfigBase() {

		@Override
		protected @Nullable String cfgIdentifierRegex() {
			// TODO Auto-generated method stub
			return null;
		}


		@Override
		protected CatSet cfgInitialCategorySet() {
			return CatSet.CATSET_DEFAULT;
		}


		@Override
		protected @Nullable String cfgKeywordRegex() {
			// TODO Auto-generated method stub
			return null;
		}


		@Override
		protected @Nullable Collection<String> cfgKeywords() {
			// TODO Auto-generated method stub
			return null;
		}


		@Override
		protected boolean cfgUseDefaultLiteralClassification() {
			return true;
		}


		@Override
		protected void moreCategories(@Nullable CategoryStore store) {
			// TODO Auto-generated method stub

		}
	};


	public static PrettyPrinter<IValue, Type> makePrinter() {
		return makePrinter(PrinterMode.STANDARD);
	};


	public static PrettyPrinter<IValue, Type> makePrinter(PrinterMode mode) {
		PPBuilder<IValue, Type> builder = new PPBuilder<>(config, context);

		builder.addList(leaf(_), PPBuilder.custom(new TreeToStream<IValue>() {
			@Override
			public void printTree(IValue tree, ForwardStream<Token> output) {
				if(tree instanceof IString) {
					String s = ((IString) tree).getValue();
					output.put(new DataToken(s, config.getCatForLexical(s)));
				}
			}
		}, 1));

		builder.addTypedTmpl(RenameList(_), sort(""), // a list of renaming operations
				"[ <commaSep(1.1)> ]");
		builder.addTypedTmpl(RenameSeq(_, _), sort(""), // a sequence of renamings
				"<1> <2>");
		builder.addTypedTmpl(TopName(_, _), sort(""), // name, kind
				"<1>");
		builder.addTypedTmpl(Equal(_, _, _), sort(""), // lhs, op, rhs
				"<1> ( <2> , <3> )");
		builder.addTypedTmpl(Builtin(_, _), sort(""), // qname, idstring
				"<1>");
		builder.addTypedTmpl(Define(_, _, Instance(_, _)), sort(""), // dcl, attrs, body
				"<sepBy(2.1,' ')> \n <1>  =  <3> ; \n");
		builder.addTypedTmpl(Define(_, _, FunName(IdName(_, _), _, _)), sort(""), // dcl, attrs, body
				"<sepBy(2.1,' ')> \n <1>  =  <3> ; \n");
		builder.addTypedTmpl(Define(_, _, ProcName(IdName(_, _), _)), sort(""), // dcl, attrs, body
				"<sepBy(2.1,' ')> \n <1>  =  <3> ; \n");
		builder.addTypedTmpl(Define(_, _, TypeName(IdName(_, _))), sort(""), // dcl, attrs, body
				"<sepBy(2.1,' ')> \n <1>  =  <3> ; \n");
		builder.addTypedTmpl(Define(_, _, _), sort(""), // dcl, attrs, body
				"<sepBy(2.1,' ')> \n <1>  <3> ; \n");
		builder.addTypedTmpl(TopDefine(_, _, _), sort(""), // dcl, attrs, body
				"<sepBy(2.1,'\n')> \n <1>  <3>");
//		builder.addTypedTmpl(LiftStatExpr(_, _, _), sort(""), // sort("Expr")
//				"<1> <2> <3>");
//		builder.addTypedTmpl(LiftVar(), sort(""), //
//				"LiftVar()");
//		builder.addTypedTmpl(UnlessReturned(_), sort(""), //
//				"UnlessReturned()");
		builder.addTypedTmpl(ExternalType(_, _, _, _), sort(""), // language, extname, extparams, name
				"<2> . <4>");
		builder.addTypedTmpl(ExternalFun(_, _, _, _, _, _), sort(""), // language, extname, extparams, name, params, return
				"<2> . <4>");
		builder.addTypedTmpl(ExternalProc(_, _, _, _, _), sort(""), // language, extname, extparams, name, params
				"<2> . <4>");
		builder.addTypedTmpl(ExternalParams(_, _), sort(""), // types, operations (both are sorted lists of full names), sort("Modifier"),-- used in extparams above
				"");
		builder.addTypedTmpl(Instance(_, _), sort(""), // op/type, list of bindings (instantiation of a Fun, Proc, or a type)
				"<1> [ <commaSep(2.1)> ]");
		builder.addTypedTmpl(Binding(_, _), sort(""), // from-name, to-name (for use in Instance), sort("Modifier"),
				"<1> => <2>");
		if(mode == PrinterMode.FULL) {
			builder.addTypedTmpl(Concrete(_), sort(""), // paramlist (list of names). (potentially parameterised concrete implementation of something), sort("Modifier"),
					"@concrete ( <commaSep(1.1)> )");
		}
		if(mode == PrinterMode.FULL) {
			builder.addTypedTmpl(Api(), sort(""), // modifier for api definitions
					"@api");
		}
		if(mode == PrinterMode.FULL) {
			builder.addTypedTmpl(Impl(), sort(""), // modifier for implementation (non-api), sort("Modifier"),definitions
					"@impl");
		}
		builder.addTypedTmpl(DotOp(_, _, _), sort(""), // expr, name, type
				"<1> . <2>");
		builder.addTypedTmpl(DeclBody(_), sort(""), //
				"{ \n <sepBy(1.1,'')> \n }");
		builder.addTypedTmpl(Origins(_), sort(""), // attribute: a list of locations tracing the origin of a declaration
				"");
		builder.addTypedTmpl(Origin(_, _), sort(""), // attribute: a list of locations tracing the origin of a declaration
				"");
//		builder.addTypedTmpl(External(_, _), sort(""), // language, name
//				"");
		builder.addTypedTmpl(Undefined(), sort(""), //
				"_");
//		builder.addTypedTmpl(IntegerLiteralAttr(_, _, _, _, _, _), sort(""), // type, 0, 1, +, *, -
//				"");
		builder.addTypedTmpl(SignatureOf(_), sort(""), //
				"signatureOf ( <1> )");
		builder.addTypedTmpl(RequiresOf(_), sort(""), //
				"requiresOf ( <1> )");
		builder.addTypedTmpl(DefinesOf(_), sort(""), //
				"definesOf ( <1> )");
		builder.addTypedTmpl(Modifier(_), sort(""), //
				"modifier ( <1> )");
		builder.addTypedTmpl(Close(_), sort(""), //
				"close ( <1> )");

		builder.addTypedTmpl(Abstract(), sort("Modifier"),//
				"abstract");

		builder.addTypedTmpl(Alert(_, _), sort("SubClause"),//
				"@alert { <1>  <2> }");

		builder.addTypedTmpl(AlertBody(), sort("AlertDefBody"),//
				"");

		builder.addTypedTmpl(AlertClause(_, _), sort("AlertDeclarative"),//
				"alert  <1>  <:  <commaSep(2.1)>");

		builder.addTypedTmpl(AlertName(_), sort("QualifiedName"),//
				"<1>");

		builder.addTypedTmpl(AlertPost(_), sort("AlertSpec"),//
				"if  post  <1>");


		builder.addTypedTmpl(AlertPre(_), sort("AlertSpec"),//
				"if  pre  <1>");

		builder.addTypedTmpl(AlertThrows(_), sort("AlertSpec"),//
				"if  throwing  <1>");

		builder.addTypedTmpl(AnonParam(_, _), sort("AnonParam"),//
				"<1>  <2>");

		builder.addTypedTmpl(Apply(_, _), sort("Expr"),//
				"<1> ( <commaSep(2.1)> )");

		builder.addTypedTmpl(Assert(_, _), sort("Stat"),//
				"assert  <1>  <sepBy(2.1,' ')>  ;");

		builder.addTypedTmpl(Assign(_, _), sort("Stat"),//
				"<1>  =  <2>  ;");

		builder.addTypedTmpl(Assume(), sort("Modifier"),//
				"assume");

		builder.addTypedTmpl(Assume(_, _), sort("Stat"),//
				"assume  <1>  <2>  ;");

		builder.addTypedTmpl(Attr(_), sort("Attribute"),//
				"<1>");

		builder.addTypedTmpl(Attr(_, _), sort("Attribute"),//
				"<1> ( <commaSep(2.1)> )");

		builder.addTypedTmpl(SpecialType(Axiom()), sort("SpecialType"),//
				"\\axiom");


		builder.addTypedTmpl(Block(_), sort("BlockStat"),//
				"{ \n <sepBy(1.1,'\n')> \n }");

		builder.addTypedTmpl(BlockExpr(_, _), sort("Expr"),//
				"{ \n <sepBy(1.1,'\n')> \n }");

		builder.addTypedTmpl(Body(_), sort("ExprDefBody"),//
				"=  <1>");

		builder.addTypedTmpl(Body(_), sort("StatDefBody"),//
				"<1>");

		builder.addTypedTmpl(Body(_), sort("StatDefBody"),//
				"=  <1>");

		builder.addTypedTmpl(Body(_), sort("StatDefBodyNS"),//
				"<1>");

		builder.addTypedTmpl(Body(_), sort("StatDefBodyNS"),//
				"=  <1>");

		builder.addTypedTmpl(Body(_), sort("TypeDefBody"),//
				"=  <1>");

		builder.addTypedTmpl(Break(), sort("Stat"),//
				"break;");

		builder.addTypedTmpl(Broken(), sort("Modifier"),//
				"@broken");

		if(mode == PrinterMode.FULL) {
			builder.addTypedTmpl(Builtin(), sort("Modifier"),//
					"builtin");
		}

		builder.addTypedTmpl(By(_), sort("AssertClause"),//
				"by  <1>");

		builder.addTypedTmpl(By(_), sort("AssertClause"),//
				"by  simplify  <1>");

		builder.addTypedTmpl(Call(_, _), sort("Stat"),//
				"call  <1> ( <commaSep(2.1)> ) ;");

		builder.addTypedTmpl(ConceptClause(_), sort("TopExprDeclarative"),//
				"concept  <1>");

		builder.addTypedTmpl(Congruence(), sort("Modifier"),//
				"congruence");

		builder.addTypedTmpl(Datainvariant(), sort("Modifier"),//
				"datainvariant");

		builder.addTypedTmpl(DecimalIntegerLiteral(_), lex("IntegerLiteral"),//
				"<1>");

		builder.addTypedTmpl(SpecialType(Decl()), sort("SpecialType"),//
				"\\decl");

		builder.addTypedTmpl(Decl(_, _, _, _), sort("AlgDecl"),//
				"<1>  <2>  <3>  <4>");

		builder.addTypedTmpl(SpecialType(DeclList()), sort("SpecialType"),//
				"\\decllist");

		builder.addTypedTmpl(DeclList(_), sort("DeclBody"),//
				"{ \n <sepBy(1.1,'')> \n }");

		builder.addTypedTmpl(DeclList(_), sort("TopExpr"),//
				"{ \n <sepBy(1.1,'')> \n }");

		builder.addTypedTmpl(Default(), sort("Modifier"),//
				"default");

		builder.addTypedTmpl(Del(), sort("ParamMode"),//
				"del");

		if(mode == PrinterMode.FULL) {
			builder.addTypedTmpl(Derived(), sort("Modifier"),//
					"derived");
		}

		builder.addTypedTmpl(Doc(_), sort("AssertClause"),//
				"doc  {  <1>  }");

		builder.addTypedTmpl(DotOp(_, _), sort("Expr"),//
				"<1>  .  <2>");

		builder.addTypedTmpl(EmptyBody(), sort("ExprDefBody"),//
				"");

		builder.addTypedTmpl(EmptyBody(), sort("StatDefBody"),//
				";");

		builder.addTypedTmpl(EmptyBody(), sort("StatDefBodyNS"),//
				"");

		builder.addTypedTmpl(EmptyBody(), sort("TypeDefBody"),//
				"");

		builder.addTypedTmpl(Exp(), sort("ParamMode"),//
				"exp");

		builder.addTypedTmpl(Expect(_, _), sort("Decl"),//
				"\\expect ( <1> ,  <2> )");

		builder.addTypedTmpl(Expect(_, _), sort("Expr"),//
				"\\expect ( <1> ,  <2> )");

		builder.addTypedTmpl(Expect(_, _), sort("SatisfactionExpr"),//
				"\\expect ( <1> ,  <2> )");

		builder.addTypedTmpl(Expect(_, _), sort("Stat"),//
				"\\expect ( <1> ,  <2> )");

		builder.addTypedTmpl(Expect(_, _), sort("TopExpr"),//
				"\\expect ( <1> ,  <2> )");

		builder.addTypedTmpl(Expect(_, _), sort("Type"),//
				"\\expect ( <1> ,  <2> )");

		builder.addTypedTmpl(Export(), sort("Modifier"),//
				"export");

		builder.addTypedTmpl(SpecialType(Expr(_)), sort("SpecialType"),//
				"\\expr  [  <1>  ]");

		builder.addTypedTmpl(Extend(), sort("Modifier"),//
				"extend");

		if(mode == PrinterMode.FULL) {
			builder.addTypedTmpl(External(), sort("Modifier"),//
					"external");
		}

		builder.addTypedTmpl(External(_, _, _), sort("Decl"),//
				"external  <1>  <2>  <3>  ;");

		builder.addTypedTmpl(External(_, _, _), sort("TopExpr"),//
				"external  <1>  <2>  <3>");

		builder.addTypedTmpl(Field(_, _), sort("InitSpec"),//
				"<1>  =  <2>");

		builder.addTypedTmpl(For(_, _, _), sort("Stat"),//
				"for  <1>  in  <2>  do  <sepBy(3.1,'\n')>  end");

		builder.addTypedTmpl(FreeBy(_, _), sort("Decl"),//
				"free  <commaSep(1.1)>  by  <2>  ;");

		builder.addTypedTmpl(Fun(_), sort("Fun"),//
				"<1>");

		builder.addTypedTmpl(FunClause(_, _, _), sort("ExprDeclarative"),//
				"function  <1>  <2>  :  <3>");

		builder.addTypedTmpl(FunName(_, _, _), sort("QualifiedName"),//
				"<1>");

		builder.addTypedTmpl(FunOf(_, _), sort("Fun"),//
				"\\f ( <1> ,  <2> )");

		builder.addTypedTmpl(GenerateBy(_, _), sort("Decl"),//
				"generate  <commaSep(1.1)>  by  <2>  ;");

		builder.addTypedTmpl(Giv(), sort("ParamMode"),//
				"giv");

		builder.addTypedTmpl(Guard(_), sort("SubClause"),//
				"@guard ( <1> )");

		builder.addTypedTmpl(Guard(_, _), sort("Expr"),//
				"guard  <1>  by  <2>  end");

		builder.addTypedTmpl(Guard(_, _), sort("Stat"),//
				"guard  <1>  by  <2>  end");

		builder.addTypedTmpl(HexIntegerLiteral(_), lex("IntegerLiteral"),//
				"<1>");

		builder.addTypedTmpl(HomomorphismOnWith(_, _, _), sort("Decl"),//
				"homomorphism  <1>  on  <2>  with  <3>  ;");

		builder.addTypedTmpl(IdName(_, _), sort("NameIdentifier"),//
				"<1> « <2> »");

		builder.addTypedTmpl(If(_, _, _), sort("Stat"),//
				"if  <1>  then  <sepBy(2.1,'\n')>  else  <sepBy(3.1,'\n')>  end");


		builder.addTypedTmpl(IfThenElseExpr(_, _, _), sort("Expr"),//
				"if  <1>  then  <2>  else  <3>  end");

		builder.addTypedTmpl(ImplClause(_), sort("TopExprDeclarative"),//
				"implementation  <1>");

		builder.addTypedTmpl(ImplyMod(), sort("Modifier"),//
				"imply");


		builder.addTypedTmpl(ImportAll(_), sort("ImportClause"),//
				"<1>");

		builder.addTypedTmpl(Imports(_), sort("PackageClause"),//
				"imports  <commaSep(1.1)>");


		builder.addTypedTmpl(Lang(_), sort("LanguageIdentifier"),//
				"<1>");

		builder.addTypedTmpl(Let(_, _), sort("Stat"),//
				"let  <1>  in \n <sepBy(2.1,'\n')> \n end");

		builder.addTypedTmpl(ListCons(_), sort("Expr"),//
				"[  <commaSep(1.1)>  ]");

		builder.addTypedTmpl(SpecialType(Literal()), sort("SpecialType"),//
				"\\literal");

		builder.addTypedTmpl(Literal(_), sort("Expr"),//
				"<1>");

		builder.addTypedTmpl(Literal(_, _), sort("Expr"),//
				"<1>  :  <2>");

		builder.addTypedTmpl(LiteralClause(_, _), sort("TypeDeclarative"),//
				"literal  <1> ( <commaSep(2.1)> )");

		builder.addTypedTmpl(MagnoliaTree(_, _), sort("Program"),//
				"<1>  <2>");


		builder.addTypedTmpl(MetaClause(_, _, _), sort("MetaDeclarative"),//
				"def  <1>  <2>  :  <3>");

		builder.addTypedTmpl(MetaDef(_, _), sort("TopDecl"),//
				"meta  <1>  { \n <2> \n }  ;");

		builder.addTypedTmpl(Models(_, _), sort("SatisfactionExpr"),//
				"<1>  approximates  <2>");

		builder.addTypedTmpl(Models(_, _), sort("SatisfactionExpr"),//
				"<1>  models  <2>");

		builder.addTypedTmpl(ModuleHead(_, _), sort("PackageHead"),//
				"module  <1>  <2>  ;");

		builder.addTypedTmpl(ModuleName(_), sort("RenameExpr"),//
				"<1>");

		builder.addTypedTmpl(ModuleName(_), sort("SatisfactionExpr"),//
				"<1>");

		builder.addTypedTmpl(ModuleName(_), sort("TopExpr"),//
				"<1>");

		builder.addTypedTmpl(Name(_), sort("NameIdentifier"),//
				"<1>");

		builder.addTypedTmpl(Name(_), sort("OperationName"),//
				"<1>");

		builder.addTypedTmpl(NoBuild(), sort("Modifier"),//
				"@nobuild");

		builder.addTypedTmpl(Nop(), sort("Decl"),//
				";");

		builder.addTypedTmpl(Nop(), sort("Stat"),//
				";");

		builder.addTypedTmpl(Nop(), sort("Stat"),//
				"skip  ;");

		builder.addTypedTmpl(Nrm(), sort("ParamMode"),//
				"nrm");

		builder.addTypedTmpl(Obs(), sort("ParamMode"),//
				"obs");

		builder.addTypedTmpl(OctalIntegerLiteral(_), lex("IntegerLiteral"),//
				"<1>");

		builder.addTypedTmpl(Out(), sort("ParamMode"),//
				"out");

		builder.addTypedTmpl(PackageHead(_, _), sort("PackageHead"),//
				"package  <1>  <2> ;");

		builder.addTypedTmpl(Param(_, _, _), sort("ProcedureParam"),//
				"<1>  <2>  :  <3>");

		builder.addTypedTmpl(ParamList(_), sort("AnonParamList"),//
				"( <commaSep(1.1)> )");

		builder.addTypedTmpl(ParamList(_), sort("FunctionParamList"),//
				"( <commaSep(1.1)> )");

		builder.addTypedTmpl(ParamList(_), sort("ProcedureParamList"),//
				"( <commaSep(1.1)> )");

		builder.addTypedTmpl(PartitionBy(_, _), sort("Decl"),//
				"partition  <commaSep(1.1)>  by  <2>  ;");


		builder.addTypedTmpl(SpecialType(Predicate()), sort("SpecialType"),//
				"\\predicate");

		builder.addTypedTmpl(Preserve(), sort("Modifier"),//
				"preserve");

		builder.addTypedTmpl(PreserveOn(_, _), sort("Decl"),//
				"preserve  <1>  on  <2>  ;");

		builder.addTypedTmpl(Proc(_), sort("Proc"),//
				"<1>");

		builder.addTypedTmpl(ProcClause(_, _), sort("StatDeclarative"),//
				"procedure  <1>  <2>");

		builder.addTypedTmpl(ProcName(_, _), sort("QualifiedName"),//
				"<1>");

		builder.addTypedTmpl(ProcOf(_), sort("Proc"),//
				"\\p ( <1> )");

		builder.addTypedTmpl(ProgramClause(_), sort("TopExprDeclarative"),//
				"program  <1>");

		builder.addTypedTmpl(Protect(), sort("Modifier"),//
				"protect");


		builder.addTypedTmpl(QED(), sort("AssertClause"),//
				"qed");

		builder.addTypedTmpl(QName(_), sort("QualifiedNameIdentifier"),//
				"<sepBy(1.1,'.')>");

		builder.addTypedTmpl(Quotient(), sort("Modifier"),//
				"quotient");

		builder.addTypedTmpl(Recover(_, _), sort("Expr"),//
				"<1>  recover  <2>");

		builder.addTypedTmpl(Remove(), sort("Modifier"),//
				"remove");

		builder.addTypedTmpl(RenClause(_), sort("TopRenDeclarative"),//
				"renaming  <1>");

		builder.addTypedTmpl(Rename(_, _), sort("RenameExpr"),//
				"<1>  =>  <2>");

		builder.addTypedTmpl(RenameApp(_, _), sort("RenameExpr"),//
				"<1> ( <commaSep(2.1)> )");

		builder.addTypedTmpl(RenameCompose(_, _), sort("RenameExpr"),//
				"<1>  [  <commaSep(2.1)>  ]");

		builder.addTypedTmpl(RenameList(_), sort("RenameExpr"),//
				"[  <commaSep(1.1)>  ]");

		builder.addTypedTmpl(Renamed(_, _), sort("TopExpr"),//
				"<1>  [  <commaSep(2.1)>  ]");

		builder.addTypedTmpl(SpecialType(Renaming()), sort("SpecialType"),//
				"\\renaming");

		builder.addTypedTmpl(Require(), sort("Modifier"),//
				"require");

		builder.addTypedTmpl(Requires(_), sort("Decl"),//
				"requires  <commaSep(1.1)>  ;");

		builder.addTypedTmpl(Retain(), sort("Modifier"),//
				"retain");

		builder.addTypedTmpl(Return(), sort("Stat"),//
				"return  ;");

		builder.addTypedTmpl(Return(_), sort("Stat"),//
				"return  <1>  ;");

		builder.addTypedTmpl(SatClause(_), sort("TopSatDeclarative"),//
				"satisfaction  <1>");

		//builder.addTypedTmpl(SpecialType(_), sort("Type"),//
		//		"<1>");

		builder.addTypedTmpl(Src(_), sort("AssertClause"),//
				"");

		builder.addTypedTmpl(StringLiteral(_), sort("StringLiteral"),//
				"<1>");

		builder.addTypedTmpl(Struct(_), sort("Type"),//
				"struct  { \n <sepBy(1.1,'\n')> \n }");

		builder.addTypedTmpl(Struct(_, _), sort("Expr"),//
				"<1>  {  <commaSep(2.1)>  }");

		builder.addTypedTmpl(SpecialType(Test()), sort("SpecialType"),//
				"\\test");

		builder.addTypedTmpl(TopApp(_, _), sort("TopExpr"),//
				"<1> ( <commaSep(2.1)> )");


		builder.addTypedTmpl(TopExprTree(_), sort("Program"),//
				"<1>");

		builder.addTypedTmpl(TopMap(_, _), sort("TopExpr"),//
				"*  <1> ( <2> )");

		builder.addTypedTmpl(Tuple(_), sort("Expr"),//
				"( <commaSep(1.1)> )");

		builder.addTypedTmpl(Tuple(_), sort("Type"),//
				"( <commaSep(1.1)> )");

		builder.addTypedTmpl(Type(_), sort("Type"),//
				"<1>");

		builder.addTypedTmpl(TypeClause(_), sort("TypeDeclarative"),//
				"type  <1>");

		builder.addTypedTmpl(TypeExpr(_), sort("Expr"),//
				":  <1>");

		builder.addTypedTmpl(TypeField(_), sort("TypeField"),//
				"<1>  ;");

		builder.addTypedTmpl(TypeName(_), sort("QualifiedName"),//
				"<1>");

		builder.addTypedTmpl(Undefined(_), sort("Expr"),//
				"<1>");

		builder.addTypedTmpl(Unknown(), sort("Type"),//
				"?");

		builder.addTypedTmpl(Upd(), sort("ParamMode"),//
				"upd");

		if(mode == PrinterMode.FULL) {
			builder.addTypedTmpl(Use(), sort("Modifier"),//
					"use");
		}


		builder.addTypedTmpl(Var(_), sort("Expr"),//
				"<1>");

		builder.addTypedTmpl(VarClause(_, _), sort("ExprDeclarative"),//
				"var  <1>  :  <2>");

		builder.addTypedTmpl(VarDecl(_, _, _, _), sort("LetClause"),//
				"<1>  <2>  :  <3>  =  <4>");


		builder.addTypedTmpl(VarName(_, _), sort("QualifiedName"),//
				"<1>");

		builder.addTypedTmpl(While(_, _), sort("Stat"),//
				"while  <1>  do  <2>  end");

		builder.addTypedTmpl(WithModels(_, _, _), sort("SatisfactionExpr"),//
				"<1>  with  <2>  approximates  <3>");

		builder.addTypedTmpl(WithModels(_, _, _), sort("SatisfactionExpr"),//
				"<1>  with  <2>  models  <3>");

		builder.addTypedTmpl(Yield(_), sort("Stat"),//
				"yield  <1>  ;");

		builder.addTypedTmpl(whitespace(_), lex("LAYOUT"),//
				"<1>");

		if(mode == PrinterMode.FULL) {
			builder.addList(_, PPBuilder.custom(new TreeToStream<IValue>() {
				@Override
				public void printTree(IValue tree, ForwardStream<Token> output) {
					if(tree instanceof IConstructor) {
						output.put(new DataToken(((IConstructor) tree).getName(), config.cfgCategories().category("TXT")));
					}
				}
			}, 0), "(", PPBuilder.sepBy(0, ","), ")");
		}
		else {
			builder.addList(_, "");
		}

		return builder.compile();
	}


	enum PrinterMode {
		STANDARD, FULL
	}
}
