/**************************************************************************
 * Copyright (c) 2013 Anya Helene Bagge
 * Copyright (c) 2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::lib::Renaming

import org::magnolialang::compiler::Util;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;

import IO;
import List;
import Map;
import Node;
import Set;
import String;

// --------------------------------------------------
// simpleRename
// --------------------------------------------------

public tuple[map[AST, AST], Env] makeParamRenaming(FunClause(n1, ParamList(seq(ps1)), t1), FunClause(n2, ParamList(seq(ps2)), t2), Env env) {
	return makeParamRenaming2(ps1, ps2, env);
}

public tuple[map[AST, AST], Env] makeParamRenaming(ProcClause(n1, ParamList(seq(ps1))), ProcClause(n2, ParamList(seq(ps2))), Env env) {
	return makeParamRenaming2(ps1, ps2, env);
}

public default tuple[map[AST, AST], Env] makeParamRenaming(AST decl1, AST decl2, Env env) {
	return <(), env>;
}

private tuple[map[AST, AST], Env] makeParamRenaming2(list[AST] ps1, list[AST] ps2, Env env) {
	map[AST,AST] renaming = ();
	
	for(<Param(_, pn1, _), Param(_, pn2, _)> <- zip(ps1, ps2)) {
		if(pn1 != pn2) {
			if(!(Name(_) := pn1))
				termPrintln("Warning: makeParamRenaming: not a Name(_): ", pn1);
			renaming[pn1] = pn2;
		}
	}
	
	return <renaming, env>;
}
public tuple[&T, Env] simpleRename(Env env, &T tree, list[AST] renaming) {
	map[AST,AST] ren = ();
	for(Rename(a, b) <- renaming)
		ren = ren + (a : b);
	return simpleRename(env, tree, ren, true);
}

public tuple[&T, Env] simpleRename(Env env, &T tree, map[AST,AST] renaming) {
	return simpleRename(env, tree, renaming, true);
}

public tuple[&T, Env] simpleRenameNoWarn(Env env, &T tree, map[AST,AST] renaming) {
	return simpleRename(env, tree, renaming, false);
}

private tuple[&T, Env] simpleRename(Env env, &T tree, map[AST,AST] renaming, bool warn) {
	if(renaming == ())
		return <tree, env>;
		
	unused = domain(renaming);
	<tree, unused> = simpleRename(env, tree, renaming, unused);
	
	if(warn) {
		for(n <- unused) {
			env = errorMark(env, "Unused renaming \'<nameToStr(n)>\' =\> \'<nameToStr(renaming[n])>\'", Warning(), locOf(n));
		}
	}
	return <tree, env>;
}

private tuple[&T,set[AST]] simpleRename(Env env, &T tree, map[AST,AST] renaming, set[AST] unused) {
	tree = top-down-break visit(tree) {
		case x:Binding(a,n): {
			<m, unused> = applyRename(env, n, renaming, unused);
			insert x[arg1=m];
		}
		case x:Instance(n,bs): {
			//termPrintln("renamed instance: ", x);
			<bs, unused> = simpleRename(env, bs, renaming, unused);
			//termPrintln("               =\> ", x);
			insert preserveAnnos(Instance(n, bs), x);
		}
		case x:Define(dcl,attrs,body): {
				if(!isConcrete(x)) {
					<dcl, unused> = applyRename(env, dcl, renaming, unused);
					<attrs, unused> = applyRename(env, attrs, renaming, unused);
					<body, unused> = simpleRename(env, body, renaming, unused);
					//termPrintln("Rename: ", x);
					//termPrintln("     =\> ", Define(dcl, attrs, body));
					insert preserveAnnos(Define(dcl, attrs, body), x);
				}
				else {
					;//termPrintln("Not renaming impl: ", x);
				}
			}
		case x:ExternalType(lang, m, onPart, n): {
			<onPart, unused> = applyRename(env, onPart, renaming, unused);
			insert preserveAnnos(ExternalType(lang, m, onPart, n), x);
		}
		case x:ExternalFun(lang, m, onPart, n, ps, t): {
			<ps, unused> = applyRename(env, ps, renaming, unused);
			<t, unused> = applyRename(env, t, renaming, unused);
			<onPart, unused> = applyRename(env, onPart, renaming, unused);
			insert preserveAnnos(ExternalFun(lang, m, onPart, n, ps, t), x);
		}
		case x:ExternalProc(lang, m, onPart, n, ps): {
			<ps, unused> = applyRename(env, ps, renaming, unused);
			<onPart, unused> = applyRename(env, onPart, renaming, unused);
			insert preserveAnnos(ExternalProc(lang, m, onPart, n, ps), x);
		}
	}
	return <tree, unused>;
}

private tuple[&T,set[AST]] applyRename(Env env, &T tree, map[AST,AST] renaming, set[AST] unused) {
	tree = top-down-break visit(tree) {
		case n:Name(_):
			if(n in renaming) {
				unused -= {n};
				insert trackOrigin((renaming[n])[@\loc=locOf(n)], renaming[n]);
			}
//		case Concrete(_):
//			;
		case x:Instance(n,bs): {
			//termPrintln("renamed instance: ", x);
			<bs, unused> = simpleRename(env, bs, renaming, unused);
			//termPrintln("               =\> ", x);
			insert preserveAnnos(Instance(n, bs), x);
		}
	}
	return <tree, unused>;
}
