/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::random::RandomPrograms
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
//import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import List;
import util::Math;
import Exception;

data GenKind = Expression() | Statement() | Type();
data Generator = Gen(AST (GenKind, Generator) gen);

data RandomEnv = RandomEnv(
	list[AST] decls,
	set[AST] types,
	set[AST] funcs,
	set[AST] procs,
	set[AST] vars,
	int funcCtr,
	int typeCtr,
	int procCtr);

public RandomEnv RandomEnv() {
	return RandomEnv([], {}, {}, {}, {}, 0, 0, 0);
}

list[str] typeNames = ["apple", "orange", "pear", "kiwi", "melon",
	"grape", "date", "lemon", "lime", "fig", "banana", "pineapple",
	"papaya", "mango", "plum"];

@doc{Add a randomly generated type to the environment.}
public RandomEnv addType(RandomEnv env) {
	name = typeNames[env.typeCtr % size(typeNames)];
	c = env.typeCtr / size(typeNames);
	if(c > 0)
		name = "<name><c>";
	env.typeCtr = env.typeCtr + 1;
	
	return env[types = env.types + TypeName(Name(leaf(name)))]
	          [decls = env.decls + Define(TypeClause(Name(leaf(name))), seq([]), EmptyBody())];
}

list[str] funcNames = ["fee", "foo", "fie", "faz", "fez", "fib", "fab"];

public RandomEnv addFunc(RandomEnv env) {
	name = funcNames[env.funcCtr % size(funcNames)];
	c = env.funcCtr / size(funcNames);
	if(c > 0)
		name = "<name><c>";
	env.funcCtr = env.funcCtr + 1;
		
	if(size(env.types) == 0)
		env  = addType(env);
		
	retType = getOneFrom(env.types);
	nArgs = arbInt(6);
	if(nArgs > 3)
		nArgs = arbInt(6);
		
	params = [Param(Name(leaf("arg<i>")), basename(getOneFrom(env.types))) | i <- tail([0 .. nArgs])];
	aparams = [AnonParam(Obs(), TypeName(t)) | Param(_, t) <- params];
	return env[funcs = env.funcs + FunName(Name(leaf(name)), seq(aparams), retType)]
	          [decls = env.decls + Define(FunClause(Name(leaf(name)), ParamList(seq(params)),
	   		             basename(retType)), seq([]), EmptyBody())];
}

list[str] procNames = ["pee", "poo", "pie", "paz", "pez", "pib", "pab"];


public AST makeRandomProgram() {
	int nTypes = 10;
	int nFuncs = 10;
	
	env = RandomEnv();
	for(i <- [0 .. nTypes])
		env = addType(env);
	println("makeRandomProgram <env>");
	for(i <- [0 .. nFuncs])
		env = addFunc(env);
	
	decls = [addDef(d, env) | d <- env.decls];
	
	impl = ImplDef(seq([]), Name(leaf("MyImpl")), seq([]), DeclBody(seq(decls)));
	return MagnoliaTree(PackageHead(Name(leaf("RandomProg")), seq([])), seq([impl]));
}


public AST makeExpr(AST typ, RandomEnv env, int depth) {
	if(typ notin env.types)
		throw "Unknown type <typ>";
	funcs = [f | f:FunName(_,_,typ) <- env.funcs];
	vars = [v | v:VarName(_, typ) <- env.vars]; 
	names = funcs + vars;
		
	if(size(names) == 0 || depth > 10) {
		return Struct(basename(typ), seq([]));
	}
	n = getOneFrom(names);
	
	if(VarName(name, _) := n)
		return Var(name);
	else if(FunName(name, seq(ps), _) := n)
		return Apply(Fun(name), seq([makeExpr(t, env, depth + 1) | AnonParam(_, t) <- ps]));
}

public AST addDef(def:Define(FunClause(_, ParamList(seq(params)), retType),
                             _, EmptyBody()),
                  RandomEnv env) {

	for(Param(n, t) <- params)
		env.vars += {VarName(n, TypeName(t))};
	
	return def[arg2 = Body(makeExpr(TypeName(retType), env, 0))]; 	
}

public default AST addDef(AST def, RandomEnv env) {
	return def;
}
