/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Andreas Hjortland
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.repr;

import static nuthatch.pattern.StaticPatternFactory.parent;
import static nuthatch.rascal.pattern.StaticValuesPatternFactory.*;
import static org.magnolialang.magnolia.AbstractSyntax.*;
import static org.magnolialang.magnolia.evaluator.AlgoEvaluator.*;
import static org.magnolialang.magnolia.grammar.MagnoliaPatterns.*;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import nuthatch.library.Action;
import nuthatch.library.ActionFactory;
import nuthatch.library.BaseWalk;
import nuthatch.library.FactoryFactory;
import nuthatch.library.JoinPoints;
import nuthatch.library.MatchAction;
import nuthatch.library.MatchBuilder;
import nuthatch.library.Walk;
import nuthatch.pattern.Environment;
import nuthatch.pattern.EnvironmentFactory;
import nuthatch.pattern.Pattern;
import nuthatch.rascal.adapter.ValuesCursor;
import nuthatch.rascal.adapter.ValuesMatchAction;
import nuthatch.rascal.adapter.ValuesWalker;
import nuthatch.rascal.pattern.ValuesBuildContext;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.impl.reference.ValueFactory;
import org.rascalmpl.value.type.Type;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.grammar.MagnoliaPatterns;
import org.magnolialang.util.PdbEnvironment;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.nuthatchery.pica.util.depgraph.IDepGraph;
import org.nuthatchery.pica.util.depgraph.IWritableDepGraph;
import org.nuthatchery.pica.util.depgraph.UnsyncedDepGraph;

public class ConvertAST {
	private final IWritableDepGraph<IConstructor> depGraph = new UnsyncedDepGraph<IConstructor>();

	private boolean generated = false;
	private static final ValuesBuildContext context = new ValuesBuildContext(TermFactory.vf, TermFactory.ts);


	public Walk<ValuesWalker> cfWalk() {
		ActionFactory<IValue, Type, ValuesCursor, ValuesWalker> af = FactoryFactory.getActionFactory(IValue.class, Type.class, ValuesCursor.class, ValuesWalker.class);

		MatchBuilder<IValue, Type, ValuesCursor, ValuesWalker> mb = af.matchBuilder(context);

		Pattern<IValue, Type> mods = var("mods");
		Pattern<IValue, Type> name = var("name");
		Pattern<IValue, Type> ps = var("ps");
		Pattern<IValue, Type> t = var("t");
		Pattern<IValue, Type> subcls = var("subcls");
		Pattern<IValue, Type> body = var("body");

		mb.add(Define(var("subcls", FunClause(name, ps, t)), mods, body), subWalkAction(bodyWalk(depGraph)));
		mb.add(Define(var("subcls", ProcClause(name, ps)), mods, body), subWalkAction(bodyWalk(depGraph)));
		mb.add(Define(var("subcls", TypeClause(name)), mods, body), new MatchAction<IValue, Type, ValuesCursor, ValuesWalker>() {

			@Override
			public void init(ValuesWalker walker) {
				// TODO Auto-generated method stub

			}


			@Override
			public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
				return PARENT;
			}
		});

		//mb.add(var("x", pat), action)
		//af.down(af.match(pat, action))
		// af.down(whatToDoWhenGoingDown).up(...)
		return af.walk(mb.one());

	}


	public IDepGraph<IConstructor> getDFGraph(IConstructor tree) {
		ValuesWalker walker = new ValuesWalker(tree, cfWalk());
		walker.start();
		generated = true;
		DebugUtil.createGraph(tree, "asts/ast.dot");
		DebugUtil.WriteToDisk("graph.dot", depGraph);
		return depGraph;
	}


	public MatchAction<IValue, Type, ValuesCursor, ValuesWalker> subWalkAction(final Walk<ValuesWalker> walk) {
		return new MatchAction<IValue, Type, ValuesCursor, ValuesWalker>() {

			@Override
			public void init(ValuesWalker walker) {
			}


			@Override
			public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
				ValuesWalker subWalker = new ValuesWalker(walker.getData(), walk);
				subWalker.start();

				return PARENT;
			}
		};
	}


	private Walk<ValuesWalker> bodyWalk(IWritableDepGraph<IConstructor> depGraph) {
		return new MyWalk(depGraph);
	}


	private static class DebugUtil {

		public static void createGraph(IConstructor start, String filename) {
			try (PrintWriter pw = new PrintWriter(filename)) {
				pw.println("digraph ast {");
				//pw.println("    graph [rankdir=LR]");
				dfsGraphCreator(start, pw);
				pw.println("}");
			}
			catch(FileNotFoundException e) {
				e.printStackTrace();
			}
		}


		public static void WriteToDisk(String path, IDepGraph<IConstructor> graph) {
			try (PrintWriter pw = new PrintWriter(path)) {
				StringBuilder sb = new StringBuilder();

				sb.append("digraph depgrah {\n");
				//sb.append("    graph [rankdir=LR]\n");
				for(IConstructor node : graph.getElements()) {
					Set<IConstructor> dependents = graph.getDependents(node);
					if(dependents.size() > 0) {
						for(IConstructor dependent : dependents) {
							sb.append("    " + getID(node) + " -> " + getID(dependent) + ";\n");
						}
					}
					else {
						sb.append("    " + getID(node) + ";\n");
					}
				}
				sb.append("}\n");
				pw.println(sb.toString());
			}
			catch(FileNotFoundException e) {
				e.printStackTrace();
			}
		}


		private static void dfsGraphCreator(IConstructor val, PrintWriter pw) {
			for(IValue c : val.getChildren()) {
				if(c instanceof IConstructor) {
					IConstructor child = (IConstructor) c;
					pw.printf("    %s -> %s\n", getID(val), getID(child));
					dfsGraphCreator(child, pw);
				}
				if(c instanceof IList) {
					IList list = (IList) c;
					for(IValue elt : list) {
						IConstructor child = (IConstructor) elt;
						pw.printf("    %s -> %s\n", getID(val), getID(child));
						dfsGraphCreator(child, pw);
					}
				}
			}
		}


		private static String getID(IConstructor cons) {
			StringBuilder returnValue = new StringBuilder();

			switch(cons.getName()) {
			case "TypeName":
			case "ProcName":
			case "FunName":
			case "TopName":
			case "VarName":
				returnValue.append(cons.getName() + " - " + Magnolia.yieldName(cons));
				break;
			case "leaf":
				returnValue.append("leaf:'" + TermAdapter.yieldTerm(cons) + "'");
				break;
			case "Proc":
			case "Fun":
				returnValue.append(Magnolia.yieldName(cons.get(0)));
				break;
			default:
				returnValue.append(TermAdapter.yieldTerm(cons));
				break;
			}
			returnValue.append("@" + Integer.toHexString(System.identityHashCode(cons)));
			return '"' + returnValue.toString().replace('"', '\'') + '"';
		}
	}


	static class MyWalk extends BaseWalk<ValuesWalker> {
		private static final IConstructor NULL_CONSTRUCTOR = TermFactory.cons("error");
		private static IConstructor RETURN_CONSTRUCTOR;
		private static final IString RESULT = ValueFactory.getInstance().string("result");
		private final IWritableDepGraph<IConstructor> graph;
		private Environment<IConstructor> globalEnv = new PdbEnvironment<>();
		private Environment<IConstructor> bindings = new PdbEnvironment<>();
		private Action<ValuesWalker> globalMatchBuilder;


		public MyWalk(IWritableDepGraph<IConstructor> depGraph) {
			globalEnv.put(RESULT, NULL_CONSTRUCTOR);
			this.graph = depGraph;
		}


		@Override
		public void init(ValuesWalker walker) {
			ActionFactory<IValue, Type, ValuesCursor, ValuesWalker> af = FactoryFactory.getActionFactory(IValue.class, Type.class, ValuesCursor.class, ValuesWalker.class);

			MatchBuilder<IValue, Type, ValuesCursor, ValuesWalker> mb = af.matchBuilder(context);

			final Pattern<IValue, Type> mode = var("mode");
			final Pattern<IValue, Type> procname = var("ProcName");
			final Pattern<IValue, Type> mods = var("mods");
			final Pattern<IValue, Type> op = var("op");
			final Pattern<IValue, Type> name = var("name");
			final Pattern<IValue, Type> innerName = var("innerName");
			final Pattern<IValue, Type> as = listvar("as");
			final Pattern<IValue, Type> ps = var("ps");
			final Pattern<IValue, Type> t = var("t");
			final Pattern<IValue, Type> type = var("type");
			final Pattern<IValue, Type> expr = var("expr");
			final Pattern<IValue, Type> subcls = var("subcls");
			final Pattern<IValue, Type> body = var("body");

			mb.add(Define(var("subcls", or(FunClause(name, ps, t), ProcClause(name, ps))), mods, body), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					if(JoinPoints.down(walker)) {
						RETURN_CONSTRUCTOR = TermFactory.cons("return");
						IConstructor b = (IConstructor) walker.getData();
						DebugUtil.createGraph(b, "asts/ast-" + Integer.toHexString(b.hashCode()) + ".dot");
						IConstructor paramList = (IConstructor) env.get(ps).getData();
						bindings = bindings.enterScope();
						if(TermAdapter.isCons(paramList, "ParamList", 1)) { // ParamList(seq([p1, p2, ...]))
							for(IValue c : TermAdapter.getChildren(TermAdapter.getArg(paramList, 0))) {
								IConstructor param = (IConstructor) c; // Param(m,n.t)
								IConstructor var = TermFactory.cons("VarName", param.get(1), param.get(2));
								IConstructor node = TermFactory.cons("InputNode", var);
								graph.add(var);
								bindings.put(var, node);
								// System.out.println("Bound: " + TermAdapter.yieldTerm(var) + " - " + TermAdapter.yieldTerm(node));
							}
						}
						return 3;
					}
					else if(JoinPoints.up(walker)) {
						bindings = bindings.exitScope();
					}
					return PARENT;
				}
			});

			mb.add(Let(_, _), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					if(JoinPoints.down(walker)) {
						bindings = bindings.enterScope();
					}
					else if(JoinPoints.up(walker)) {
						bindings = bindings.exitScope();
					}
					return PROCEED;
				}
			});
			mb.add(Struct(name, _), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					IConstructor cons = (IConstructor) walker.getData();
					graph.add(cons);
					globalEnv.put(RESULT, cons);
					return PARENT;
				}

			});

			mb.add(Undefined(), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					IConstructor cons = (IConstructor) walker.getData();
					graph.add(cons);
					globalEnv.put(RESULT, cons);
					return PARENT;
				}
			});

			mb.add(VarDecl(mode, name, type, expr), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					if(JoinPoints.down(walker)) {
						globalEnv.put(RESULT, NULL_CONSTRUCTOR);
						return 4;
					}
					else if(JoinPoints.up(walker)) {
						IConstructor var = TermFactory.cons("VarName", env.get(name).getData(), env.get(type).getData());
						IConstructor node = globalEnv.get(RESULT);
						if(node.equals(NULL_CONSTRUCTOR)) {
							graph.add(NULL_CONSTRUCTOR, (IConstructor) walker.getBranchCursor(4).getData());
						}
						bindings.put(var, node);
						// System.out.println("Bound: " + TermAdapter.yieldTerm(var) + " - " + TermAdapter.yieldTerm(node));
					}
					return PARENT;
				}
			});

			//TODO:
			//mb.add(Return(), new ValuesMatchAction() {
			//	@Override
			//	public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
			//		IConstructor node = globalEnv.get(RESULT);
			//		if(node.equals(NULL_CONSTRUCTOR)) {
			//			graph.add(NULL_CONSTRUCTOR, (IConstructor) walker.getBranchCursor(walker.from()).getData());
			//		}
			//		else {
			//			graph.add(node, RETURN_CONSTRUCTOR);
			//		}
			//		return PARENT;
			//	}
			//});

			mb.add(Return(expr), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					if(JoinPoints.down(walker)) {
						globalEnv.put(RESULT, NULL_CONSTRUCTOR);
						return NEXT;
					}
					else if(JoinPoints.up(walker)) {
						IConstructor node = globalEnv.get(RESULT);
						if(node.equals(NULL_CONSTRUCTOR)) {
							graph.add(NULL_CONSTRUCTOR, (IConstructor) walker.getBranchCursor(walker.from()).getData());
						}
						else {
							graph.add(RETURN_CONSTRUCTOR, node);
						}
					}
					return PARENT;
				}
			});

			mb.add(Var(name), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					ValuesCursor key = env.get(name);
					IConstructor node = bindings.get(env.get(name).getData());
					globalEnv.put(RESULT, node);
					return PARENT;
				}
			});
			mb.add(Assign(Var(name), expr), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					if(JoinPoints.down(walker)) {
						return 2;
					}
					else if(JoinPoints.up(walker)) {
						bindings.put(env.get(name).getData(), globalEnv.get(RESULT));
						// System.out.println("Bound: " + TermAdapter.yieldTerm(env.get(name).getData()) + " - " + TermAdapter.yieldTerm(globalEnv.get(RESULT)));
					}
					return PROCEED;
				}
			});
			mb.add(Assign(DotOp(expr, name, type), expr), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					// TODO Auto-generated method stub
					return 0;
				}
			});
			mb.add(parent(seq(listvar("_"))), new ValuesMatchAction() {
				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					if(JoinPoints.afterChild(walker)) {
						int index = walker.from();
						IConstructor result = globalEnv.get(RESULT);
						// System.out.println("Result: " + result);
						// System.out.println("Child:  " + TermAdapter.yieldTerm(walker.getBranchHandle(index).getData()));
						globalEnv = globalEnv.exitScope();
						globalEnv.put(ValueFactory.getInstance().integer(index), result);
						globalEnv.put(RESULT, result);
					}
					if(JoinPoints.beforeChild(walker)) {
						globalEnv = globalEnv.enterScope();
					}
					return PROCEED;
				}
			});
			mb.add(Call(op, seq(as)), new ValuesMatchAction() {
				//TODO ExternalProc

				@Override
				public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
					// first, process 'as', get node for each argument
					if(JoinPoints.down(walker)) {
						return 2;
					}
					if(JoinPoints.up(walker)) {
						IList args = (IList) env.get(as).getData();
						List<IValue> params = getParams((IConstructor) env.get(op).getData());
						IConstructor node = (IConstructor) env.get(op).getData();
						// add a node for the procedure itself
						graph.add(node); // TODO: say that this is a procedure call
						int i = 1;
						for(IValue v : params) {
							if(AnonParam(mode, _).match(new ValuesCursor(v), env)) {
								IConstructor m = (IConstructor) env.get(mode).getData();
								if(isIn(m)) {
									IConstructor arg = globalEnv.get(ValueFactory.getInstance().integer(i));
									IConstructor s = arg;
									// add dependencies from procedure to each input node
									graph.add(node, s);
								}
								if(isOut(m)) {
									IValue arg = args.get(i - 1);
									if(Var(name).match(new ValuesCursor(arg), env)) {
										// for all output variables, add a node
										IConstructor varName = (IConstructor) env.get(name).getData();

										IConstructor output = TermFactory.cons("out" + i, ValueFactory.getInstance().string(Magnolia.yieldName(varName)));
										graph.add(output, node);
										// bind all output variables to output nodes
										bindings.put(varName, output);
										// System.out.println("Bound: " + TermAdapter.yieldTerm(varName) + " - " + TermAdapter.yieldTerm(node));
									}
								}
								i += 1;
								env = EnvironmentFactory.env();
							}
						}
					}
					return PROCEED;
				}


				private List<IValue> getParams(IConstructor op) {
					Environment<ValuesCursor> env = EnvironmentFactory.env();
					List<IValue> result = new ArrayList<>();
					// Dette er et hakk tryggere pga. typesjekking (se over om TermAdapter. 
					// Må via environment. mode er en metavariabel, og AnonParam er en statisk pattern fra MagnoliaPatterns)
					if(MagnoliaPatterns.Proc(ProcName(_, seq(as))).match(new ValuesCursor(op), env)) {
						for(IValue i : (IList) env.get(as).getData()) {
							IConstructor value = (IConstructor) i;
							result.add(value);
						}
					}
					return result;
					// Evaluator kode (AlgoEvaluator.evalProc(), tilsv. evalFun())
					// Just extract from the "op" thingy Proc(ProcName(name, ParamList(seq([AnonParam(mode,type)]))))
					// TermAdapter.
					// TermAdapter.isCons(op) // Is constructor (i motsetning til en listenode)
					// TermAdapter.isCons(op, "name"); // Om det er en konstruktør med navn "name", (vi bruker "Proc" eller "ExternalProc" i dette tilfellet)
					// TermAdapter.isCons(op, "name", 2); // 2 er antall barn (aritet)
					//
					// TermAdapter.getArg(op, num) // hent argument nr. num
					// TermAdapter.getChildren(op) // hent iterable av alle barn
					// TerAdapter.arity(op) // antall barn (aritet)
				}
			});
			globalMatchBuilder = mb.one();
		}


		@Override
		public int step(ValuesWalker walker) {
			return globalMatchBuilder.step(walker);
		}

	}

}
