/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.resources;

import java.net.URI;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IExternalValue;
import org.rascalmpl.value.ISourceLocation;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.resources.regions.ICodeRegion;
import org.nuthatchery.pica.tasks.ITaskMonitor;

public interface IMagnoliaCodeUnit extends IManagedCodeUnit {
	/**
	 *
	 * @return True if the file-relative offset is within this code unit
	 */
	boolean containsOffset(int offset);


	@Override
	IConstructor getId();


	/**
	 * Get information about this code unit.
	 *
	 * The qName field of the DefInfo will be equal to getId()
	 *
	 * @param tm
	 *            A progress monitor
	 * @return A DefInfo value
	 * @see org.magnolialang.magnolia.MagnoliaFacts#Type_DefInfo
	 */
	IConstructor getDefInfo(ITaskMonitor tm);


	/**
	 * @return Location of declaration's name
	 */
	ISourceLocation getNameLoc();


	/**
	 * @return Location of declaration's name
	 */
	ICodeRegion<URI> getNameRegion();


	/**
	 * Get the source location of this code unit.
	 *
	 * The location will span the entire code unit.
	 *
	 * @return Location of this code unit.
	 *
	 */
	ISourceLocation getSrcLoc();


	/**
	 * Get the source location of this code unit.
	 *
	 * The location will span the entire code unit.
	 *
	 * @return Location of this code unit.
	 *
	 */
	ICodeRegion<URI> getSrcRegion();


	/**
	 * Get the typechecked AST of a code unit.
	 *
	 * @param tm
	 *            A progress monitor
	 * @return A typechecked AST.
	 */
	IConstructor getTypechecked(ITaskMonitor tm);


	/**
	 * Get the raw AST of a code unit.
	 *
	 * The returned AST will not have been annotated with types, etc.
	 *
	 * @param tm
	 *            A progress monitor
	 * @return An AST.
	 */
	IConstructor getAST(ITaskMonitor tm);
}
