/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.resources;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.ISourceLocation;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.services.IMgXRefInfo;
import org.nuthatchery.pica.resources.IXRefInfo;
import org.nuthatchery.pica.tasks.NullTaskMonitor;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.interpreter.NullRascalMonitor;

public class XRefInfo {

	public static IXRefInfo makeFromDefInfo(IConstructor info, @Nullable IList bindings) {
		return new DefXRefInfo(info, bindings);
	}


	public static IXRefInfo makeFromHoverInfo(IConstructor info) {
		return new DefXRefInfo(info, null);
	}


	public static IXRefInfo makeFromName(IConstructor name) {
		return new NameXRefInfo(name);
	}


	public static IXRefInfo makeFromResource(IMagnoliaCodeUnit res) {
		return new ResXRefInfo(res);
	}


	static class DefXRefInfo implements IMgXRefInfo {
		private final IConstructor info;
		private final IList bindings;


		public DefXRefInfo(IConstructor info, @Nullable IList bindings2) {
			this.info = info;
			this.bindings = bindings2 != null ? bindings2 : TermFactory.vf.list(TermFactory.Type_AST);
		}


		@Override
		@Nullable
		public IList getBindings() {
			return bindings;
		}


		@Override
		@Nullable
		public IConstructor getDefInfo() {
			return info;
		}


		@Override
		public IConstructor getId() {
			return (IConstructor) info.get("qName");
		}


		@Override
		public ISourceLocation getSrcLoc() {
			return (ISourceLocation) info.get("srcLoc");
		}


		@Override
		public boolean hasBindings() {
			return !bindings.isEmpty();
		}


		@Override
		public boolean hasDefInfo() {
			return true;
		}

	}


	static class NameXRefInfo implements IMgXRefInfo {
		private final IConstructor name;


		public NameXRefInfo(IConstructor name) {
			this.name = name;
		}


		@Override
		public IList getBindings() {
			return TermFactory.vf.list(TermFactory.Type_AST);
		}


		@Override
		@Nullable
		public IConstructor getDefInfo() {
			return null;
		}


		@Override
		public IConstructor getId() {
			return name;
		}


		@Override
		@Nullable
		public ISourceLocation getSrcLoc() {
			return null;
		}


		@Override
		public boolean hasBindings() {
			return false;
		}


		@Override
		public boolean hasDefInfo() {
			return false;
		}
	}


	static class ResXRefInfo implements IMgXRefInfo {
		private final IMagnoliaCodeUnit res;


		public ResXRefInfo(IMagnoliaCodeUnit res) {
			this.res = res;
		}


		@Override
		public IList getBindings() {
			return TermFactory.vf.list(TermFactory.Type_AST);
		}


		@Override
		@Nullable
		public IConstructor getDefInfo() {
			if(res != null) {
				return res.getDefInfo(new NullTaskMonitor()).set("body", res.getTypechecked(new NullTaskMonitor()));
			}
			else {
				return null;
			}
		}


		@Override
		public IConstructor getId() {
			return res.getId();
		}


		@Override
		@Nullable
		public ISourceLocation getSrcLoc() {
			return (ISourceLocation) res.getDefInfo(new NullTaskMonitor()).get("srcLoc");
		}


		@Override
		public boolean hasBindings() {
			return false;
		}


		@Override
		public boolean hasDefInfo() {
			return true;
		}

	}

}
