/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.resources.internal;

import java.net.URI;
import java.util.Collection;
import java.util.List;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.resources.AbstractMagnoliaManagedResource;
import org.magnolialang.magnolia.resources.IDefinition;
import org.magnolialang.magnolia.resources.IMagnoliaModule;
import org.nuthatchery.pica.rascal.RascalUtil;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.IXRefInfo;
import org.nuthatchery.pica.resources.handles.IFileHandle;
import org.nuthatchery.pica.resources.internal.AbstractManagedResource;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.resources.managed.IManagedContainer;
import org.nuthatchery.pica.resources.managed.IManagedResource;
import org.nuthatchery.pica.resources.regions.ICodeRegion;
import org.nuthatchery.pica.resources.regions.IOffsetLength;
import org.nuthatchery.pica.resources.storage.IStorage;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.util.ISignature;
import org.nuthatchery.pica.util.Pair;

public class AlgoDefinition extends AbstractMagnoliaManagedResource implements IDefinition {

	private DefinitionData data;


	public AlgoDefinition(IFileHandle resource, DefinitionData data) {
		super(data.getURI(), resource);
		this.data = data;
	}


	@Override
	public boolean containsOffset(int offset) {
		return data.containsOffset(offset);
	}


	@Override
	public IConstructor getAST(ITaskMonitor tm) {
		return data.getAST();
	}


	@Override
	public IConstructor getAttribute(String name) {
		return data.getAttribute(name);
	}


	@Override
	public IConstructor getBaseName() {
		return data.getBaseName();
	}


	@Override
	public IConstructor getBody() {
		return data.getBody();
	}


	@Override
	public IConstructor getDefClause() {
		return data.getDefClause();
	}


	@Override
	public IConstructor getDefInfo(ITaskMonitor tm) {
		return data.getDefInfo();
	}


	@Override
	public Collection<? extends IManagedCodeUnit> getDepends(ITaskMonitor tm) {
		throw new UnsupportedOperationException();
	}


	@Override
	public IConstructor getId() {
		return data.getId();
	}


	@Override
	public String getKind(ITaskMonitor tm) {
		return data.getKind();
	}


	@Override
	public ILanguage getLanguage() {
		return data.getParent().getLanguage();
	}


	@Override
	public long getLength() {
		return data.getLength();
	}


	@Override
	public String getName() {
		return data.getFullName();
	}


	@Override
	public ISourceLocation getNameLoc() {
		return data.getNameLoc();
	}


	@Override
	public int getNumParams() {
		return data.getNumParams();
	}


	@Override
	public long getOffset() throws UnsupportedOperationException {
		return data.getOffset();
	}


	@Override
	public IConstructor getParamMode(int i) {
		return data.getParamMode(i);
	}


	@Override
	public IConstructor getParamName(int i) {
		return data.getParamName(i);
	}


	@Override
	public List<IConstructor> getParams() {
		return data.getParams();
	}


	@Override
	public IConstructor getParamType(int i) {
		return data.getParamType(i);
	}


	@Override
	@Nullable
	public IMagnoliaModule getParent() {
		return (IMagnoliaModule) data.getParent();
	}


	@Override
	public ISignature getSourceSignature(ITaskMonitor tm) {
		return data.getParent().getSourceSignature(tm);
	}


	@Override
	public ISourceLocation getSrcLoc() {
		return data.getSrcLoc();
	}


	@Override
	public ICodeRegion<URI> getSrcRegion() {
		return RascalUtil.toCodeRegion(data.getSrcLoc());
	}


	@Override
	@Nullable
	public IStorage getStorage() {
		return data.getParent().getStorage();
	}


	@Override
	public Collection<? extends IManagedCodeUnit> getTransitiveDepends(ITaskMonitor tm) {
		throw new UnsupportedOperationException();
	}


	@Override
	public IConstructor getTypechecked(ITaskMonitor tm) {
		return data.getAST();
	}


	@Override
	public IConstructor getUnqualifiedName() {
		return data.getUnqualifiedName();
	}


	@Override
	public Collection<Pair<IOffsetLength, Object>> getXRefs(ITaskMonitor tm) {
		throw new UnsupportedOperationException();
	}


	@Override
	@Nullable
	public IXRefInfo getXRefs(IOffsetLength loc, ITaskMonitor tm) {
		throw new UnsupportedOperationException();
	}


	@Override
	public boolean hasAttribute(String name) {
		return data.hasAttribute(name);
	}


	@Override
	public boolean hasIncompleteDepends(ITaskMonitor tm) {
		return false;
	}


	@Override
	public boolean hasModifier(String name) {
		return data.hasModifier(name);
	}


	@Override
	public boolean hasParams() {
		return data.hasParams();
	}


	@Override
	public boolean isCodeUnit() {
		return true;
	}


	@Override
	public boolean isFragment() {
		return true;
	}


	@Override
	public boolean isRoot() {
		return false;
	}


	@Override
	public void onDependencyChanged() {
	}


	@Override
	public void onResourceChanged() {
	}


	@Override
	public String toIdString() {
		return data.toIdString();
	}


	@Override
	public String toString() {
		return data.toString();
	}


	@Override
	public ISignature getFullSignature(ITaskMonitor tm) {
		return getSourceSignature(tm);
	}


	@Override
	public ICodeRegion<URI> getNameRegion() {
		return RascalUtil.toCodeRegion(getNameLoc());
	}


}
