/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::services::XRefs

import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::lib::Pretty;
import org::magnolialang::compiler::Util;

import org::nuthatchery::pica::terms::Terms;
import IO;
import String;
import Exception;

public str formatXRef(DefInfo info, list[AST] bindings) {
	name = info.qName;
	doc = formatDoc(info.doc);
	
	synopsis = quoteXML(shortDispNameOf(info));
	contents = formatContents(info, name);
	bindings = [Binding(x, y) | Binding(x, y) <- bindings, x != y];
	if(bindings != []) {
		synopsis = "<synopsis>[<strJoin(["<quoteXML(shortNameToStr(x))> =\> <quoteXML(shortNameToStr(y))>" | Binding(x, y) <- bindings, x != y], ", ")>]";
	}
	
	line = "\<p\>";
	for(i <- [0 .. 50])
		line += "–";
	line += "\</p\>";
	
	flag = "";
	if(Builtin() in info.attrs)
		flag = "builtin";
//	else if(!isQualifiedName(name))
//		flag = "local";
	
	if(flag == "")
		return "\<b\><synopsis>\</b\>\n<line><doc><contents>";
	else
		return "\<b\>\<i\><flag>\</i\> <synopsis>\</b\>\n<line><doc><contents>";
}

public str formatXRef(AST name, list[AST] bindings) {
	dispName = shortDispNameOf(name);
	synopsis = quoteXML(dispName);

	return "No info available for \<b\><synopsis>\</b\>\n";
}

public str formatDoc(str doc) {
	if(/^\</ := doc)
		return doc;
	else if(doc == "")
		return doc;
	else
		return "\<div class=\"doc\"\>\<p\><doc>\</p\>\</div\>";
}

public str formatContents(DefInfo info, value name) {
	str result = "";

	if(DeclBody(seq(ds)) := info.body) {
		reqs = "";
		exts = "";
		defs = "";
		dcls = "";
		axms = "";
		impls = "";
		for(def:Define(_,seq(attrs),body) <- sortDefs(ds)) {
			list[str] mods = [];
			// termPrintln("XRefs: def: ", def);
			entry = quoteXML(shortDispNameOf(def));
			//if(body != EmptyBody())
			//	entry = "<entry> = ...";
			if(Body(Apply(Builtin(_,leaf(b)), seq([]))) := body) {
				entry = "<entry> = <b>";
			}

			if(Guard(g) <- attrs) {
				entry += " \<b\>guard\</b\> <prettyMagnolia(g)>";
			}
			if(Alert(_,_) <- attrs) {
				entry += " \<i\>(alerted)\</i\>";
			}
			
			if(IntegerLiteralAttr(typeName, zeroName, oneName, plusName, mulName, negName) <- attrs) {
				entry += "\<br\>\n";
				entry += "            literal integer(";
				entry += "0 =\> <quoteXML(shortDispNameOf(zeroName))>";
				entry += ", 1 =\> <quoteXML(shortDispNameOf(oneName))>";
				entry += ", _+_ =\> <quoteXML(shortDispNameOf(plusName))>";
				entry += ", _*_ =\> <quoteXML(shortDispNameOf(mulName))>";
				if(negName != Nop())
					entry += ", -_ =\> <quoteXML(shortDispNameOf(negName))>";
				entry += ")";
			}
			if(hasModifier(def, Derived()))
				mods += "derived";
			if(hasModifier(def, External()))
				mods += "external";
			if(hasModifier(def, Default()))
				mods += "default";
			if(hasModifier(def, Predicate()))
				mods += "predicate";
			//if(hasModifier(def, Require()))
			//	mods += "require";
			//if(hasModifier(def, Extend()))
			//	mods += "extend";
			if(hasModifier(def, Use()))
				mods += "use";
			if(hasModifier(def, Protect()))
				mods += "protect";
			if(hasModifier(def, Broken()))
				mods += "@broken";
			if(hasModifier(def, Assume()))
				mods += "assume";
			if(hasModifier(def, ImplyMod()))
				mods += "imply";
			if(hasModifier(def, NoBuild()))
				mods += "@nobuild";

			if(mods != [])
				entry = "<entry> \<i\>(<strJoin(mods, ", ")>)\</i\>";

			if(Instance(i, seq(bs)) := body) {
				entry = "<entry> = <quoteXML(shortNameToStr(i))>";
				bindings = strJoin(["<quoteXML(shortNameToStr(x))> =\> <quoteXML(shortNameToStr(y))>" | Binding(x, y) <- bs, x != y], ", ");
				if(bindings != "") {
					entry = "<entry>[<bindings>]";
				}
			}

			if(Concrete(seq(ps)) <- attrs) {
				impls += "    <entry> [<strJoin([quoteXML(shortNameToStr(x)) | x <- ps], ", ")>]\<br\>\n";
			}
			else if(hasModifier(def, Require()))
				reqs += "    <entry>\<br\>\n";
			else if(hasModifier(def, Extend()))
				exts += "    <entry>\<br\>\n";
			else if(hasModifier(def, Axiom()))
				axms += "    <entry>\<br\>\n";
			else if(body == EmptyBody())
				dcls += "    <entry>\<br\>\n";
			else 
				defs += "    <entry>\<br\>\n";
		}
		
		if(dcls != "")
			result += "\<div\>\<b\>Declares API:\</b\>\<br\>\n<dcls>\</div\>\n";
		if(defs != "")
			result += "\<div\>\<b\>Defines API:\</b\>\<br\>\n<defs>\</div\>\n";
		if(reqs != "")
			result += "\<div\>\<b\>Requires API:\</b\>\<br\>\n<reqs>\</div\>\n";
		if(exts != "")
			result += "\<div\>\<b\>Extends API:\</b\>\<br\>\n<exts>\</div\>\n";
		if(axms != "")
			result += "\<div\>\<b\>Axioms:\</b\>\<br\>\n<axms>\</div\>\n";
		if(impls != "")
			result += "\<div\>\<b\>Implements:\</b\>\<br\>\n<impls>\</div\>\n";
	}
	else if(Body(Struct(seq(ds))) := info.body) {
		fields = "";
		
		for(TypeField(def:VarDecl(m,n,t,e)) <- ds) {
			fields += "    <quoteXML(shortDispNameOf(def))>\<br\>\n";
		}
		if(fields != "")
			result += "\<div\>\<b\>Fields:\</b\>\<br\>\n<fields>\</div\>\n";
	}
	else if(RenameList(seq(rns)) := info.body) {
		renames = "";
		for(Rename(a, b) <- rns) {
			renames += "    <quoteXML(shortDispNameOf(a))> =\> <quoteXML(shortDispNameOf(b))>\<br\>\n";
		}
		if(renames != "") {
			result += "\<div\>\<b\>Renamings:\</b\>\<br\>\n<renames>\</div\>\n";
		}
	}
	
	return visit(result) {
	case /\<\/?[ib]\>/ => ""
	case / / => " "
	};
}

public str quoteXML(str s) {
	return replaceAll(replaceAll(s, "&", "&amp;"), "\<", "&lt;");
}

