#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cstdlib>
#include <vector>
#include <limits>
/* INSERT INCLUDE FILES HERE */
/* INSERT IO LIBRARY HERE */
using std::string;
static string programName;


/* INSERT BODY HERE */

namespace Magnolia {
namespace MagicMain {

const bool FALLBACK_TO_STDOUT = true;
const int ARG_INFILE_GIVEN = 1;
const int ARG_OUTFILE_GIVEN = 2;
const int ARG_INSTRING_GIVEN = 4;
const int ARG_STDIN_GIVEN = 8;   // not used
const int ARG_STDOUT_GIVEN = 16; // not used
const int ARG_INPUT_GIVEN = ARG_INFILE_GIVEN | ARG_INSTRING_GIVEN | ARG_STDIN_GIVEN;
const int ARG_OUTPUT_GIVEN = ARG_OUTFILE_GIVEN | ARG_STDOUT_GIVEN;
const int ARG_FORMAT_AUTO = 0;
const int ARG_FORMAT_XML = 1;
const int ARG_FORMAT_TEXT = 2;
const int ARG_FORMAT_BINARY = 3;
const int PARAM_MODE_OBS = 1;
const int PARAM_MODE_OUT = 2;
const int PARAM_MODE_UPD = PARAM_MODE_OBS | PARAM_MODE_OUT;

struct arg {
	string data;
	string outFile;
	int tag;
	int out_format;
	
	arg(const char* _data, int _tag, int _format) : data(_data), outFile(), tag(_tag), out_format(_format) {}
	arg() : data(""), outFile(""), tag(0), out_format(ARG_FORMAT_AUTO) {}
};

static bool expectArg(const char* proc, const char* argName, int argNum, std::vector<arg>& args, int mode, bool inFallback, bool outFallback) {
	if(args.size() < argNum+1){
		args.resize(argNum+1, arg());
	}

	bool ok = true;

	if(mode & PARAM_MODE_OBS) {
		if(!(args[argNum].tag & ARG_INPUT_GIVEN)) {
			if(inFallback) {
				args[argNum].data = "-";
				args[argNum].tag |= ARG_INFILE_GIVEN;
			}
			else {
				std::cerr << "Missing input argument '" << argName << "' of " << proc << " (specify with -i" << argNum << ")\n";
				ok = false;
			}
		}
	}

	if(mode & PARAM_MODE_OUT) {
		if(!(args[argNum].tag & ARG_OUTPUT_GIVEN)) {
			if(outFallback) {
				args[argNum].outFile = "-";
				args[argNum].tag |= ARG_OUTFILE_GIVEN;
			}
			else {
				std::cerr << "Missing output argument '" << argName << "' of " << proc << " (specify with -o" << argNum << ")\n";
				ok = false;
			}
		}
	}

	return ok;
}

static void expectNoMoreArgs(const string& proc, int numArgs, const std::vector<arg>& args, bool ok) {
	for(int i = numArgs+1; i < args.size(); i++) {
		if(args[i].tag & ARG_INPUT_GIVEN) {
			std::cerr << "Extra input argument #" << i << " for '" << proc << "': '" << args[i].data << "'\n";
			ok = false;
		}
		if(args[i].tag & ARG_OUTPUT_GIVEN) {
			std::cerr << "Extra output argument #" << i << " for '" << proc << "': '" << args[i].outFile << "'\n";
			ok = false;
		}
	}
	if(!ok) {
		std::cerr << "(Try 'help " << proc << "' for command-specific help)\n";
		std::exit(2);
	}
}

// Patch the files given by '-i' / '-io' / '-o' into the normal argument vector
static void setIOArgs(std::vector<arg>& args, const arg& argIn, const arg& argOut, int firstIn, int firstOut) {
	if(firstIn < 1)
		firstIn = 1; // will cause error message later if there is no such argument
	if(firstOut < 1)
		firstOut = 1;

	if(args.size() < firstIn+1){
		args.resize(firstIn+1, arg());
	}
	if(args.size() < firstOut+1){
		args.resize(firstOut+1, arg());
	}

	if(argIn.tag & ARG_INPUT_GIVEN) {
		args[firstIn].tag |= argIn.tag;
		args[firstIn].data = argIn.data;
	}

	if(argOut.tag & ARG_OUTPUT_GIVEN) {
		args[firstOut].tag |= argOut.tag;
		args[firstOut].outFile = argOut.outFile;
		args[firstOut].out_format = argOut.out_format;
	}

}

static void help() {
	std::cout
	<<	"Usage: " << programName << " help [operation]\n"
	/* INSERT COMMAND USAGE HERE */
	<<	"Uppercase file suffix for xml data, lowercase for binary data.\n"
		"Use '-' for standard input or standard output where applicable.\n"
		"Standard in/out is default for single arguments.\n"
		"Data Format Options: \n"
		"         -b,  --binary   use binary data format for output *\n"
		"         -t,  --text     use text data format for output\n"
		"         -x,  --xml      use xml data format for output (default) \n"
		"\n"
		" (* unimplemented)\n"
		" Input format is auto-detected. Output defaults to XML, or the format\n"
		" of the input file, in case of -io arguments.\n";
}

static void fatal_unimplemented(const string& option) {
	std::cerr << "Unimplemented option '" << option << "'\n";
	std::cerr << "(Try -h or --help for help)\n";
	std::exit(2);
}

static void fatal_unknown_option(const string& option) {
	std::cerr << "Unknown option '" << option << "'\n";
	std::cerr << "(Try -h or --help for help)\n";
	std::exit(2);
}

static void fatal_missing_argument(const char* kind, const string& name) {
	std::cerr << "Missing " << kind << " argument '" << name << "'\n";
	std::cerr << "(Try -h or --help for help)\n";
	std::exit(2);
}

static void fatal_missing_option_argument(const char* name) {
	std::cerr << "Missing argument to option " << name << "\n";
	std::cerr << "(Try -h or --help for help)\n";
	std::exit(2);
}

static void fatal_error_file(const char* kind, const string& type,  string io) {
	std::cerr << "Error " << kind << " " << type << " from file '" << io << "'\n";
	std::exit(3);
}

static void fatal_error_stdio(const char* kind, const string& type,  string io) {
	std::cerr << "Error " << kind << " " << type << " from " << io << "\n";
	std::exit(3);
}

static void fatal_error_string(const char* kind, const string& type, string io) {
	std::cerr << "Error " << kind << " " << type << " from string '" << io << "'\n";
	std::exit(3);
}

static void fatal_unknown_command(const string& command) {
	std::cerr << programName << ": unknown command '" << command << "'\n"
		 << "(Try -h or --help for help)\n";
	std::exit(4);
}

static void fatal_no_command() {
	std::cerr << programName << ": no command given\n\n";
	help();
	std::exit(2);
}


template<typename T>
static int readValueFromStream(T& result, int format, std::istream* instream) {
	if(format == ARG_FORMAT_AUTO) {
		char c;
		*instream >> c;
		if(c == '<')
			format = ARG_FORMAT_XML;
		else if(c == '\0')
			format = ARG_FORMAT_BINARY;
		else
			format = ARG_FORMAT_TEXT;
		instream->putback(c);
	}

	if(format == ARG_FORMAT_XML) {
		Magnolia::IO::XmlIStream stream(*instream);
		result <<= stream;
	}
	else if(format == ARG_FORMAT_TEXT) {
		Magnolia::IO::TextIStream stream(*instream);
		result <<= stream;
	}
	else if(format == ARG_FORMAT_BINARY) {
		Magnolia::IO::BinaryIStream stream(*instream);
		result <<= stream;
	}
	return format;
}

template<typename T>
static T readValue(const string& type, arg& argument) {
	T result;
	int format;
	if(argument.tag & ARG_INFILE_GIVEN) {
		if(argument.data != "-") {
			std::ifstream file;
			file.open(argument.data, std::ifstream::in);
			format = readValueFromStream(result, ARG_FORMAT_AUTO, &file);
			file.close();
			if(!file)
				fatal_error_file("reading", type, argument.data);
		}
		else {
			format = readValueFromStream(result, ARG_FORMAT_AUTO, &std::cin);
			if(!stdin)
				fatal_error_stdio("reading", type, "standard input");
		}
	}
	else if(argument.tag & ARG_INSTRING_GIVEN) {
		std::istringstream strstream(argument.data);
		format = readValueFromStream(result, ARG_FORMAT_AUTO, &strstream);

		if(!strstream)
			fatal_error_string("reading", type, argument.data);
	}
	else {
		fatal_missing_argument("input", type);
	}
	if(argument.out_format == ARG_FORMAT_AUTO)
		argument.out_format = format;
	return result;
}

template<typename T>
static void writeValue(const string& type, const arg& argument, const T& value) {
	std::ostream* outstream;
	std::ofstream file;
	if(argument.tag & ARG_OUTFILE_GIVEN) {
		if(argument.outFile == "-")
			outstream = &std::cout;
		else {
			file.open(argument.outFile, std::ifstream::out);
			outstream = &file;
		}
	}
	else {
		fatal_missing_argument("output", type);
	}

	if(argument.out_format == ARG_FORMAT_XML || argument.out_format == ARG_FORMAT_AUTO) {
		Magnolia::IO::XmlOStream stream(*outstream);
		value >>= stream;
		stream.newline();
	}
	else if(argument.out_format == ARG_FORMAT_TEXT) {
		Magnolia::IO::TextOStream stream(*outstream);
		value >>= stream;
		stream.newline();
	}
	else if(argument.out_format == ARG_FORMAT_BINARY) {
		Magnolia::IO::BinaryOStream stream(*outstream);
		value >>= stream;
	}
	if(file.is_open())
		file.close();
	if(!*outstream)
		fatal_error_file("writing", type, argument.data);
}

static int getArgNum(string opt, int len) {
	int i = 0;
	if(opt.size() > len) {
		std::stringstream(opt.substr(len)) >> i;
	}
	return i;
}

static int magic_main(int argc, char **argv) {
	programName = argv[0];
	std::vector<arg> args;  // index 0 is command name
	
	if(argc <= 0) {
		Magnolia::MagicMain::help();
		return 0;
	}

	arg argIn = arg();
	arg argOut = arg();
	int argNum = 1;
	int format = ARG_FORMAT_AUTO;
	while(argNum < argc) {
		string opt = argv[argNum++];

		if(opt == "-t" || opt == "--text")
			format = ARG_FORMAT_TEXT;
		else if(opt == "-x" || opt == "--xml") {
			format = ARG_FORMAT_XML;
		}
		else if(opt == "-b" || opt == "--binary") {
			Magnolia::MagicMain::fatal_unimplemented(opt);
		}
		else if(opt == "-h" || opt == "--help") {
			Magnolia::MagicMain::help();
			return 0;
		}
		else if(opt.substr(0, 3) == "-io") {
			int i = Magnolia::MagicMain::getArgNum(opt, 3);
			if(argNum >= argc)
				Magnolia::MagicMain::fatal_missing_option_argument("-i");
			if(i == 0) {
				argIn.data = argv[argNum];
				argIn.tag |= ARG_INFILE_GIVEN;
				argOut.outFile = argv[argNum];
				argOut.tag |= ARG_OUTFILE_GIVEN;
				argOut.out_format = format;
			}
			else {
				if(args.size() <= i)
					args.resize(i+1, arg());
				args[i].data = argv[argNum];
				args[i].outFile = argv[argNum];
				args[i].tag |= ARG_INFILE_GIVEN | ARG_OUTFILE_GIVEN;
				args[i].out_format = format;
			}
			argNum++;
		} 
		else if(opt.substr(0, 2) == "-i") {
			int i = Magnolia::MagicMain::getArgNum(opt, 2);
			if(argNum >= argc)
				Magnolia::MagicMain::fatal_missing_option_argument("-i");
			if(i == 0) {
				argIn.data = argv[argNum++];
				argIn.tag |= ARG_INFILE_GIVEN;
			}
			else {
				if(args.size() <= i)
					args.resize(i+1, arg());
				args[i].data = argv[argNum++];
				args[i].tag |= ARG_INFILE_GIVEN;
			}
		}
		else if(opt.substr(0, 2) == "-o") {
			int i = Magnolia::MagicMain::getArgNum(opt, 2);
			if(argNum >= argc)
				Magnolia::MagicMain::fatal_missing_option_argument("-i");
			if(i == 0) {
				argOut.outFile = argv[argNum++];
				argOut.tag |= ARG_OUTFILE_GIVEN;
				argOut.out_format = format;
			}
			else {
				if(args.size() <= i)
					args.resize(i+1, arg());
				args[i].outFile = argv[argNum++];
				args[i].tag |= ARG_OUTFILE_GIVEN;
				args[i].out_format = format;
			}
		} 
		else if(opt == "--") {
			while(argNum < argc)
				args.push_back(arg(argv[argNum++], ARG_INSTRING_GIVEN, ARG_FORMAT_AUTO));
		}
		else if(opt[0] == '-' && opt.length() > 1) {
			Magnolia::MagicMain::fatal_unknown_option(opt);
		}
		else {
			args.push_back(arg(argv[argNum-1], ARG_INSTRING_GIVEN, ARG_FORMAT_AUTO));
		}
	}


	if(!args.empty()) {
		string cmd = args[0].data;

		if(cmd == "help") {
			if(args.size() == 2) {
				cmd = args[1].data;
			}
			if(args.size() != 2) {
				Magnolia::MagicMain::help();
			}
			/* INSERT COMMAND HELP HERE */
			else {
				Magnolia::MagicMain::fatal_unknown_command(cmd);
			}

			return 0;
		}
		else {
			if(false)
				;
			/* INSERT COMMAND CALLS HERE */
			else {
				Magnolia::MagicMain::fatal_unknown_command(cmd);
			}
		}
	}
	else {
		Magnolia::MagicMain::fatal_no_command();
	}

	return 0;
}

}
}

int main(int argc, char **argv) {
	try {
		return Magnolia::MagicMain::magic_main(argc, argv);
	} catch(const char* p) {
			std::cerr << "Error! " << p << std::endl;
	}
}
