/**************************************************************************
 * Copyright (c) 2013 Anya Helene Bagge
 * Copyright (c) 2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::algo::Alerts

import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::algo::Expressions;
import org::magnolialang::magnolia::typechecker::algo::Attributes;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import Set;
import Node;
import List;

/*
public default tuple[AST,Env] checkAlert(AST alert, Env env) {
	env = errorMark(env, "Internal Compiler Error: Alerts: no checkAlert rule for <consOf(alert)>", locOf(alert));
	return <alert, env>;
}
*/
public tuple[AST,Env] checkAlert(AST alert, Env env) {
	ts = lookupName(alert, AlertKind(), env);
	if(ts == {}) {
		env = errorMark(env, "Undefined alert \'<dispNameOf(alert, env)>\'", locOf(alert));
		if(AlertName(_) := alert)
			return <alert, env>;
		else
			return <preserveAnnos(AlertName(alert), alert), env>;
	}
	else if({t} := ts) {
		env = addTokenRef(alert, t, env);
		return <preserveAnnos(t, alert), env>;
	}
	else {
		env = errorMark(env, "Ambiguous name \'<dispNameOf(alert, env)>\'", locOf(alert));
		t = preserveAnnos(getOneFrom(ts), alert);
		for(tt <- ts)
			env = addTokenRef(alert, tt, env);
		if(AlertName(_) := t)
			return <t, env>;
		else
			return <preserveAnnos(Alert(t), alert), env>;
	}

}

public tuple[AST, Env] checkAlertSupers(AST supers, Env env) {
	if(seq(ss) := supers) {
		ss = for(s <- ss) {
			<s, env> = checkAlert(s, env);
			append s;
		}
		supers.args = ss;
	}
	else throw "Alerts::checkAlertSupers: invalid AST: <consOf(supers)> (expected seq)";
	
	return <supers, env>;
}

public tuple[AST, Env] checkAlertBody(AST body, Env env) {
	return <body, env>;
}
 
public tuple[AST, Env] mergeAlertDecls(AST decl1, AST decl2, Env env) {
	if(decl1 == decl2) {
		return mergeAttrs(decl1, decl2, env);
	}
	else {
		env = errorMark(env, "Merging of alert declarations not implemented yet", Warning(), locOf(decl2));
		return mergeAttrs(decl1, decl2, env);
	}
}

public tuple[AST, Env] mergeAlertSupers(AST decl1, AST decl2, AST alertName, Env env) {
	if(decl1.arg0 == decl2.arg0) {
		return <decl1.arg0, env>;
	}
	else {
		env = errorMark(env, "Merging of alert declarations not implemented yet", Warning(), locOf(decl2));
		return <decl1.arg0, env>;
	}
}
