/**************************************************************************
 * Copyright (c) 2011-2013 Anya Helene Bagge
 * Copyright (c) 2011-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::algo::DerivedOps
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::algo::Types;
import Set;
import Node;
import List;

public default tuple[list[tuple[AST,AST,NameKind,AST]], Env] deriveOps(AST defName, AST qName, NameKind kind, AST def, Env env) {
	return <[<defName, qName, kind, def>], env>;
} 

public tuple[list[tuple[AST,AST,NameKind,AST]], Env] deriveOps(AST defName, AST qName, NameKind kind, d:Define(def:ProcClause(procName,ParamList(seq(ps))), attrs, body), Env env) {
	list[tuple[AST,AST,NameKind,AST]] result = [<defName, qName, kind, d>];

	attrs.args += [Derived()];
			
	ins = [Param(Obs(), n, t) | Param(_, n, t) <- inParams(ps)];
	outs = outParams(ps);
	
	list[AST] names = [];
	
	if(size(outs) == 1)
		names = [basename(procName)];
	else
		names = [transformName(basename(procName), str(str x){ return "<x><i+1>"; }) | i <- index(outs)];
	
	if(DEBUG_TYPECHECK) {	
		termPrintln("Procedure ", procName, " args: ", ps);
		termPrintln("    Outs: ", outs);
	}
	for(<p:Param(m, n, t), name> <- zip(outs, names)) {
		funName = FunName(name, seq([AnonParam(Obs(), at) | Param(_,_,at) <- ins]), t);
		body = hasBody(d) ? Body(Apply(FunOf(procName, leaf("<indexOf(ps, p)>")), seq([Var(VarName(an,at)) | Param(_,an,at) <- ins]))) : EmptyBody();
		if(DEBUG_TYPECHECK)
			termPrintln("    Function ", unqualify(name), " : ", funName, " args: ", ins);
		result += <unqualify(name), funName, ExprKind(), d[arg0 = FunClause(funName, ParamList(seq(ins)), t)]
									 [arg1 = attrs]
								     [arg2 = body]>;
	}
	
	return <result, env>;
}

public tuple[list[tuple[AST,AST,NameKind,AST]], Env] deriveOps(AST defName, AST qName, NameKind kind, d:Define(def:TypeClause(typeName), attrs, body), Env env) {
	list[tuple[AST,AST,NameKind,AST]] result = [<defName, qName, kind, d>];

	attrs.args += [Derived()];
	
	if(DEBUG_TYPECHECK) {	
		termPrintln("Structure type: ", typeName);
		termPrintln("              = ", body);
	}
	if(Body(Struct(seq(ds))) := body) {
		for(f:TypeField(VarDecl(m, n, t, e)) <- ds) {
			<t, env> = makeType(t, env);
			if(DEBUG_TYPECHECK) {	
				termPrintln("  Field: ", n, " : ", t);
			}
			
			readName = transformName(basename(n), str(str x){ return "_get_<x>"; });
			writeName = transformName(basename(n), str(str x){ return "_set_<x>"; });
			
			funParams = [Param(Obs(), Name(leaf("o")), typeName)];
			funName = FunName(readName, seq([AnonParam(Obs(), typeName)]), t);
			funName = preserveAnnos(funName, n);
			funBody = Body(DotOp(Var(VarName(Name(leaf("o")), typeName)), n, t));
			if(DEBUG_TYPECHECK) {	
				termPrintln("    Access function:  ", funName);
				termPrintln("                   =  ", funBody);
			}
			
			result += <readName, funName, ExprKind(), //
				d[arg0=FunClause(funName, ParamList(seq(funParams)), t)] //
				 [arg1=attrs] //
				 [arg2=funBody]>;
				 
			procParams = [Param(Upd(), Name(leaf("o")), typeName), Param(Obs(), Name(leaf("v")), t)];
			procName = ProcName(writeName, seq([AnonParam(Upd(), typeName), AnonParam(Obs(), t)]));
			procName = preserveAnnos(procName, n);
			procBody = Body(Block(seq([Assign(DotOp(Var(VarName(Name(leaf("o")), typeName)), n, t), Var(VarName(Name(leaf("v")), t)))])));
			if(DEBUG_TYPECHECK) {	
				termPrintln("    Access procedure: ", procName);
				termPrintln("                    = ", procBody);
			}
			
			<procResults, env> = deriveOps(writeName, procName, StatKind(), //
				d[arg0=ProcClause(procName, ParamList(seq(procParams)))] //
				 [arg1=attrs] //
				 [arg2=procBody], env);
			result += procResults;
		}
	
	}
	return <result, env>;
}


public tuple[list[AST], AST, Env] deriveLiteralOps(lc:LiteralClause(qName, seq(as)), AST typ, Env env) {
	list[AST] result = [];
	AST attr = Nop();
	str kind = "";
	map[AST,AST] args = ();
	for(a <- as) {
		termPrintln("deriveLiteralOps: arg: ", a);
		if(Var(Name(leaf(s))) := a) {
			if(s in {"integer", "float", "string"}) {
				if(kind != "") {
					env = errorMark(env, "Overriding literal kind \'<kind>\' with \'<s>\'", Warning(), locOf(a));
				}
				kind = s;
			}
			else {
				env = errorMark(env, "Unknown literal clause argument \'<k>\'", Error(), locOf(x));
			}
		}
		else if(Apply(Fun(Name(leaf("_=\>_"))), seq([x, y])) := a) {
			if(Var(Name(k:leaf(s))) := x || Literal(DecimalIntegerLiteral(k:leaf(s))) := x) {
				if(k in args) {
					env = errorMark(env, "Overriding already specified argument \'<s>\'", Warning(), locOf(x));
				}
				if(Var(Name(v)) := y) {
					args[k] = v;
				}
				else {
					env = errorMark(env, "Argument should be a simple name", Error(), locOf(y));
				}
			}
			else {
				env = errorMark(env, "Left-hand side of keyword argument must be a name", locOf(x));
			}
		}
		else {
			env = errorMark(env, "Argument to literal type clause must be \'name\' or \'name =\> expr\'", locOf(a));
		}
	}

	bool checkArgs(set[AST] reqArgs, set[AST] optArgs) {
		for(k <- args) {
			if(k notin reqArgs && k notin optArgs) {
				env = errorMark(env, "Illegal argument \'<k.strVal>\' to declaration of literal <kind>", locOf(k));
			}
			else {
			
				reqArgs -= {k};
			}
		}
		for(k <- reqArgs) {
			env = errorMark(env, "Missing required argument \'<k.strVal>\' to declaration of literal <kind>", locOf(lc));
		}
		
		return size(reqArgs) == 0;
	}
	
	switch(kind) {
		case "integer": {
			if(checkArgs({leaf("0"), leaf("1"), leaf("_+_"), leaf("_*_")}, {leaf("-_")})) {
				name0 = setLocRec(FunName(Name(args[leaf("0")]), seq([]), typ), locOf(args[leaf("0")]));
				name1 = setLocRec(FunName(Name(args[leaf("1")]), seq([]), typ), locOf(args[leaf("1")]));
				namePlus = setLocRec(FunName(Name(args[leaf("_+_")]), seq([AnonParam(Obs(), typ), AnonParam(Obs(), typ)]), typ), locOf(args[leaf("_+_")]));
				nameMult = setLocRec(FunName(Name(args[leaf("_*_")]), seq([AnonParam(Obs(), typ), AnonParam(Obs(), typ)]), typ), locOf(args[leaf("_*_")]));
				nameNeg  = (args[leaf("-_")]?) ? setLocRec(FunName(Name(args[leaf("-_")], seq([AnonParam(Obs(), typ)]), typ)), locOf(args[leaf("-_")])) : Nop();
				binParams = ParamList(seq([Param(Obs(), Name(leaf("arg0")), Type(typ)), Param(Obs(), Name(leaf("arg1")), typ)]));
				
				result += setLocRec(Define(FunClause(name0, ParamList(seq([])), typ), seq([Require()]), EmptyBody()), locOf(args[leaf("0")]));
				result += setLocRec(Define(FunClause(name1, ParamList(seq([])), typ), seq([Require()]), EmptyBody()), locOf(args[leaf("1")]));
				result += setLocRec(Define(FunClause(namePlus, binParams, typ), seq([Require()]), EmptyBody()), locOf(args[leaf("_+_")]));
				result += setLocRec(Define(FunClause(nameMult, binParams, typ), seq([Require()]), EmptyBody()), locOf(args[leaf("_*_")]));
				if(args[leaf("-_")]?) {
					result += setLocRec(Define(FunClause(nameNeg, ParamList(seq([Param(Obs(), Name(leaf("arg0")), Type(typ))])), typ), seq([Require()]), EmptyBody()), locOf(args[leaf("_*_")]));
				}
				attr = IntegerLiteralAttr(typ, name0, name1, namePlus, nameMult, nameNeg);
			}
		}
		case "float": {
			if(checkArgs({}, {})) {
				;
			}
		}
		case "string": {
			if(checkArgs({}, {})) {
				;
			}
		}
	}
	return <result, attr, env>;
}