/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::algo::Expressions
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::algo::DerivedOps;
import org::magnolialang::magnolia::typechecker::algo::Statements;
import org::magnolialang::magnolia::typechecker::algo::Types;
import org::magnolialang::magnolia::typechecker::algo::Overload;
import org::magnolialang::magnolia::typechecker::algo::Attributes;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import org::nuthatchery::pica::memo::Memo;
import Set;
import Node;
import List;
import String;
import Exception;
 
public tuple[AST,AST,Env] checkExpr_impl(AST expr, Env env) {
	//termPrintln("checkExpr: ", expr);
	<e, t, env> = chkExpr(expr, env);
	if(!isType(t))
		env = errorMark(env, "Internal Compiler Error: Expressions::chkExpr(): assert failed; expression type is not a valid type: <yieldTerm(t)>", locOf(expr));
	return <preserveAnnos(e, expr)[@\type=t], t, env>;
}

public tuple[AST,AST,Env](AST,Env) checkExpr = memo(checkExpr_impl);

public tuple[AST,AST,Env] checkExprBody(b:Body(Block(body)), Env env) {
	<e, t, env> = checkExpr(BlockExpr(body, Unknown()), env);
	return <b[arg0 = e], t, env>;
}

public tuple[AST,AST,Env] checkExprBody(b:Body(body), Env env) {
	<e, t, env> = checkExpr(body, env);
	return <b[arg0 = e], t, env>;
}
public tuple[AST,AST,Env] checkExprBody(b:EmptyBody(), Env env) {
	return <b, Unknown(), env>;
}

public tuple[AST,AST,Env] checkExprBody(b:Instance(_, _), Env env) {
	return <b, getExprType(b), env>;
}

default tuple[AST,AST,Env] checkExprBody(AST tree, Env env) {
	println("CHECKEXPRBODY: no rule for <tree>");
	env = errorMark(env, "Internal Compiler Error: Expressions::checkExprBody: no checkExprBody rule for <consOf(tree)>", locOf(tree));
	return <tree, Unknown(), env>;
}


tuple[AST,AST,Env] chkExpr(expr:Expect(el:leaf(err), e), Env env) {
	env = expectError(err, 1, env);
	<e, t, env> = checkExpr(e, env);
	env = checkExpected(err, locOf(el), env);
	return <e, t, env>;
}

// Expression:   v
tuple[AST,AST,Env] chkExpr(expr:Var(name), Env env) {
	ns = lookupName(name, ExprKind(), env);
	declaredType = Unknown();
	if(VarName(_,t) := name) {
		<declaredType, _, env> = checkType(t, env);
		ns = {n | n <- ns, n == name};
	}
	if(ns == {}) {
		env = errorMark(env, "Undefined variable \'<dispNameOf(name, env)>\'", locOf(name));
		return <Var(VarName(name, Unknown())), declaredType, env>;
	}
	else if({v:TopName(n,k)} := ns) {
		env = addTokenRef(expr, v, env);
		t = SpecialType(DeclList());
		if(!isAssignable(declaredType, t, env))
			env = errorMark(env, "Use-site type of variable \'<dispNameOf(n, env)> : <dispNameOf(declaredType, env)>\' "
					+ " not compatible with definition type \'<dispNameOf(t, env)>\'", locOf(expr));
		return <Var(v), t, env>;
	}
	
	nsFiltered = {n | n:VarName(_,_) <- ns};
	if(nsFiltered == {})
		nsFiltered = {n | n:FunName(_,seq([]),_) <- ns};
	
	if({v:VarName(n,t)} := nsFiltered) {
		env = addTokenRef(expr, v, env);
		if(!isAssignable(declaredType, t, env))
			env = errorMark(env, "Use-site type of variable \'<dispNameOf(n, env)> : <dispNameOf(declaredType, env)>\' "
					+ " not compatible with definition type \'<dispNameOf(t, env)>\'", locOf(expr));
		return <Var(v), t, env>;
	}
	else if({v:FunName(n,_,t)} := nsFiltered) {
		env = addTokenRef(expr, v, env);
		if(!isAssignable(declaredType, t, env))
			env = errorMark(env, "Use-site type of function \'<dispNameOf(n, env)> : <dispNameOf(declaredType, env)>\' "
					+ " not compatible with definition type \'<dispNameOf(t, env)>\'", locOf(expr));
		f = Fun(v);
		if(hasDef(v, env) && Builtin(_,_) := lookupDef(v, env).body) {
			f = lookupDef(v, env).body;
		}
		return <Apply(f, seq([])), t, env>;
	}
	else if({} := nsFiltered) {
		cands = "";
		for(n <- ns) {
			env = addTokenRef(expr, n, env);
			cands += "  <dispNameOf(n)>\n";
		}
		env = errorMark(env, "\'<dispNameOf(name, env)>\' is not a variable:\n<cands>", locOf(expr));
		return <Var(VarName(name, Unknown())), Unknown(), env>;
	}
	else {
		cands = "";
		for(n <- nsFiltered) {
			env = addTokenRef(expr, n, env);
			cands += "  <dispNameOf(n)>\n";
		}
		termPrintln("Var error: ", name);
		println("Var error name: <name>");
		println("Var error expr: <expr>");
		env = errorMark(env, "Ambiguous variable name \'<dispNameOf(name, env)>\'. Candidates are:\n<cands>", locOf(expr));
		return <Var(VarName(name, Unknown())), Unknown(), env>;
	}

}

// TODO: Expression:   l
tuple[AST,AST,Env] chkExpr(expr:Literal(DecimalIntegerLiteral(leaf(l))), Env env) {
	return <expr, SpecialType(Literal()), env>;
}

tuple[AST,AST,Env] chkExpr(expr:Literal(HexIntegerLiteral(leaf(l))), Env env) {
	return <expr, SpecialType(Literal()), env>;
}

tuple[AST,AST,Env] chkExpr(expr:Literal(OctalIntegerLiteral(leaf(l))), Env env) {
	return <expr, SpecialType(Literal()), env>;
}

tuple[AST,AST,Env] chkExpr(expr:Literal(l, t), Env env) {
	<t, _, env> = checkType(t, env);
	if(isAssignable(t, SpecialType(Literal()), env)) {
		<expr, t, env> = checkLiteral(expr, t, t, env);
	}
	else {
		expr = Undefined();
		env = errorMark(env, "Type \'<shortDispNameOf(t)>\' does not support literals", locOf(expr));
	}
	
	return <expr, t, env>;
}

tuple[AST,AST,Env] chkExpr(expr:Literal(HexIntegerLiteral(leaf(l))), Env env) {
	return <expr, SpecialType(Literal()), env>;
}

tuple[AST,AST,Env] chkExpr(expr:Literal(OctalIntegerLiteral(leaf(l))), Env env) {
	return <expr, SpecialType(Literal()), env>;
}
tuple[AST,AST,Env] chkExpr(expr:Literal(l), Env env) {
	env = errorMark(env, "Literals not implemented yet", Warning(), locOf(expr));
	return <expr, Unknown(), env>;
}

tuple[AST,AST,Env] chkExpr(expr:ListCons(seq(as)), Env env) {
	list[AST] argTypes = [];
	list[AST] argValues = [];
	for(a <- as) {
		<av, at, env> = checkExpr(a, env);
		argTypes += at;
		argValues += av;
	}
	// TODO: check literals
	
	if(SpecialType(Renaming()) in argTypes || SpecialType(DeclList()) in argTypes) {
		return <RenameList(seq(argValues)), SpecialType(Renaming()), env>;
	}
	else {
		env = errorMark(env, "List literals not implemented yet", Warning(), locOf(expr));
		return <expr, Unknown(), env>;
	}
}

// TODO: Expression:   (es)
tuple[AST,AST,Env] chkExpr(expr:Tuple(seq(es)), Env env) {
	if([e] := es) {
		return checkExpr(e, env);
	}
	else {
		list[AST] checkedEs = [];
		list[AST] checkedTs = [];
		for(e <- es) {
			<e, t, env> = checkExpr(e, env);
			checkedEs += e;
			checkedTs += t;
		}
		env = errorMark(env, "Type checker: Tuples not implemented yet", Warning(), locOf(expr));
		return <Tuple(seq(checkedEs)), Tuple(seq(checkedTs)), env>;
	}
}

// TODO: Expression:   : typ
tuple[AST,AST,Env] chkExpr(expr:TypeExpr(typ), Env env) {
	<typ, _, env> = checkType(typ, env);
	env = errorMark(env, "Type checker: Type expressions not implemented yet", Warning(), locOf(expr));
	return <TypeExpr(typ), Unknown(), env>;
}

// TODO: Expression:   struct typ $ { inits }
tuple[AST,AST,Env] chkExpr(expr:Struct(typ, seq(inits)), Env env) {	
	<_, typ, env> = checkType(typ, env);
	if(typ != Unknown()) {
		try {
			info = lookupDef(typ, env);
			body = instantiate(info.body, env); 
			if(b:Body(Struct(seq(ds))) := body) {
				// check the initialisers
				list[AST] checkedInits = [];
				checkedInits = for(i <- inits) {
					<i, env> = chkInit(typ, b, i, env);
					append i;
				};
				// check that all fields are initialised
				loop:for(TypeField(VarDecl(m, fn, _, fe)) <- ds) {
					for(Field(n,_) <- inits) {
						if(fn == n)
							continue loop;
					}
					if(fe != Undefined())
						checkedInits += Field(fn, fe);
					else
						env = errorMark(env, "Uninitialized field \'<dispNameOf(fn)>\' for type \'<dispNameOf(typ, env)>\'", locOf(typ));
				} 
				return <Struct(typ, seq(checkedInits)), typ, env>;
			}
			else {
				env = errorMark(env, "Type \'<dispNameOf(typ, env)>\' does not have fields", locOf(typ));
				return <expr[arg0=typ], typ, env>;
			}
		}
		catch NoSuchKey(_): {
			env = errorMark(env, "Internal Compiler Error: Expressions: resolved name has no definition: <dispNameOf(typ, env)>", locOf(typ));
			return <expr, typ, env>;
		}
	}
	else
		return <expr, typ, env>;
	
}

tuple[AST,Env] chkInit(AST typ, AST body, f:Field(n, e), Env env) {
//termPrintln("chkInit: ", n);
	<ft, env> = getField(typ, body, n, env);
	<e, et, env> = checkExpr(e, env);

	<e, et, env> = checkLiteral(e, et, ft, env);

	if(!isAssignable(ft, et, env) && ft != Unknown() && et != Unknown()) {
		env = errorMark(env, "Type \'<dispNameOf(et, env)>\' not assignable to \'<dispNameOf(ft, env)>\'", locOf(f));
	}
	
	return <preserveAnnos(Field(n, e), f), env>;
}

// Expression:   func(as)
tuple[AST,AST,Env] chkExpr(expr:Apply(func, seq(as)), Env env) {
	origEnv = env;
	list[AST] argTypes = [];
	list[AST] argValues = [];
	for(a <- as) {
		<av, at, env> = checkExpr(a, env);
		argTypes += at;
		argValues += av;
	}
	
	if((Fun(Name(leaf("_==_"))) := func || Fun(Name(leaf("_!=_"))) := func)
			  && [x, y] := argValues && [t, u] := argTypes) {
		<x, t, env> = checkLiteral(x, t, u, env);
		<y, u, env> = checkLiteral(y, u, t, env);
		if((!isAssignable(t, u, env) || !isAssignable(u, t, env)) && t != Unknown() && u != Unknown()) {
			env = errorMark(env, "Incompatible types \'<dispNameOf(t, env)>\' and \'<dispNameOf(u, env)>\' in equality expression", Error(), locOf(func));
		}

		return <Equal(func, x, y), SpecialType(Predicate()), env>;
	} 
	else if(Fun(name) := func) {
		fs = lookupName(name, ExprKind(), env);
		//println("Name loc: <name@\loc ? "unknown">");
		//println("Candidates loc:");
		//for(f <- fs)
		//  println("  <f@\loc ? "unknown">");
		
		if(fs == {}) {
			env = errorMark(env, "Undefined function \'<dispNameOf(name, env)>\'", locOf(name));
			return <Apply(fakeFun(name, argTypes), seq(argValues)), Unknown(), env>;
		}
		else {
			ofs = overload(fs, argTypes, env);
			if(ofs == {}) {
				if(Fun(Name(leaf("_=\>_"))) := func && [Var(x), Var(y)] := as) {
					return <Rename(x, y), SpecialType(Renaming()), origEnv>;
				}
				else if(size(fs) == 1) {
					if(DEBUG_OVERLOAD) {
						termPrintln("Argument mismatch: formals: ", getOneFrom(fs).arg1.args);
						termPrintln("                   actuals: ", argTypes);
					}
					
					env = addTokenRef(name, Ref(getOneFrom(fs)), env);
					if(!(Unknown() in argTypes)) 
						env = errorMark(env, "Call to \'<dispNameOf(name, env)>(<dispArgList(argTypes, env)>)\' doesn\'t match parameters \'<dispArgList(getOneFrom(fs).arg1.args, env)>\'", locOf(name));
				}
				else {
					for(f <- fs)
						env = addTokenRef(name, f, env);
				
					if(!(Unknown() in argTypes))
						env = errorMark(env, "No matching overloaded function for \'<dispNameOf(name, env)>(<dispArgList(argTypes, env)>)\'  (out of <size(fs)>)", locOf(name));
				}
				return <Apply(fakeFun(name, argTypes), seq(argValues)), Unknown(), env>;
			}
			else if({f:FunName(fn,seq(fps),rt)} := ofs) {
				if(DEBUG_OVERLOAD) {
					termPrintln("Resolved <dispNameOf(name)>(<dispArgList(argTypes, env)>)");
					termPrintln("   =\> ", f);
				}
				
				<argValues, env> = checkLiteralArgs(argValues, fps, env);
				env = addTokenRef(name, f, env);
				
				if((Fun(Name(leaf("_==_"))) := func || Fun(Name(leaf("_!=_"))) := func)
				    && [x, y] := argValues) {
					return <Equal(preserveAnnos(Fun(f), name), x, y), rt, env>;
				}
				else if(hasDef(f, env)) {
					if(Builtin(_,_) := lookupDef(f, env).body) {
						return <Apply(preserveAnnos(lookupDef(f, env).body, name), seq(argValues)), rt, env>;
					}
					else {
						return <Apply(preserveAnnos(Fun(f), name), seq(argValues)), rt, env>;
					}
				}
				else {
					return <Apply(preserveAnnos(Fun(f), name), seq(argValues)), rt, env>;
				}
			}
			else if({f} := ofs) {
				env = addTokenRef(name, f, env);
				env = errorMark(env, "Internal Compiler Error: Expressions::chkExpr(): unknown function id <yieldTerm(f)>", locOf(func));
				return <Apply(func, seq(argValues)), Unknown(), env>;
			}
			else {
				cands = "";
				for(f <- ofs) {
					env = addTokenRef(name, f, env);
					cands += "  <dispNameOf(f)>\n";
				}
				// TODO: create literal terms if all arguments are literals
				env = errorMark(env, "Ambiguous name \'<dispNameOf(name, env)>\' in function application.  Candidates are:\n<cands>", locOf(name));
				return <Apply(fakeFun(name, argTypes), seq(argValues)), Unknown(), env>;
			}
		}
	}
	else if(ExternalFun(lang, ns, onPart, name, ParamList(seq(ps0)), rt) := func) {
		env = enterScope(env);
		<ps, env> = checkParams(ps0, env);
		<rt, _, env> = checkType(rt, env);
		env = exitScope(env);
		if(DEBUG_OVERLOAD) {
			termPrintln("Expressions: External function: ", lang, ns, name);
			termPrintln("  Param list: ", ps);
			termPrintln("  Return type: ", rt);
		}
		// TODO: add xref to C++ file
		return <Apply(ExternalFun(lang, ns, onPart, name, ParamList(seq(ps)), rt), seq(argValues)), rt, env>;
	}
	else if(Builtin(name, _) := func) {
		rt = getExprType(name);
		return <Apply(func, seq(argValues)), rt, env>;
	}
	else if(FunOf(p:ProcName(_,seq(ps)), leaf(i)) := func) {
		env = addTokenRef(func, Ref(p), env);
		if(AnonParam(_,rt) := ps[toInt(i)])
			return <Apply(func, seq(argValues)), rt, env>;
		else {
			env = errorMark(env, "Internal Compiler Error: Expressions::chkExpr(): unknown parameter <ps[toInt(i)]>", locOf(func));
			return <Apply(func, seq(argValues)), Unknown(), env>;
		}
	}
	else {
		env = errorMark(env, "Internal Compiler Error: Expressions::chkExpr(): unknown function kind <yieldTerm(func)>", locOf(func));
		return <Apply(func, seq(argValues)), Unknown(), env>;
	}
}

AST fakeFun(AST name, list[AST] argTypes) {
	return Fun(preserveAnnos(FunName(name, seq([preserveAnnos(AnonParam(Obs(), t), t) | t <- argTypes]), Unknown()), name));
}

tuple[AST,AST,Env] chkExpr(expr:Equal(f, e1, e2), Env env) {
	<e1, e1t, env> = checkExpr(e1, env);
	<e2, e2t, env> = checkExpr(e2, env);
		
	return <Equal(f, e1, e2), SpecialType(Predicate()), env>;
}


// Expression:   if e1 then e2 else e3 end
// Statement:   if e then s1 else s2
tuple[AST,AST,Env] chkExpr(expr:IfThenElseExpr(e1, e2, e3), Env env) {
	<e1, e1t, env> = checkExpr(e1, env);
	<e2, e2t, env> = checkExpr(e2, env);
	<e3, e3t, env> = checkExpr(e3, env);
	
	if(!isAssignable(SpecialType(Predicate()), e1t, env))
		env = errorMark(env, "If condition should be a predicate (type was <dispNameOf(e1t, env)>)", locOf(e1));

	<e2, e2t, env> = checkLiteral(e2, e2t, e3t, env);
	<e3, e3t, env> = checkLiteral(e3, e3t, e2t, env);

	t = lub(e2t, e3t);
	if(t == Unknown() && e2t != Unknown() && e3t != Unknown())
		env = errorMark(env, "If branch types \'<dispNameOf(e2t, env)>\' and \'<dispNameOf(e3t, env)>\' are not compatible", locOf(expr));
	
	
	return <IfThenElseExpr(e1, e2, e3), t, env>;
}

tuple[AST,AST,Env] chkExpr(expr:Recover(e1, e2), Env env) {
	<e1, e1t, env> = checkExpr(e1, env);
	<e2, e2t, env> = checkExpr(e2, env);

	<e1, e1t, env> = checkLiteral(e1, e1t, e2t, env);
	<e2, e2t, env> = checkLiteral(e2, e2t, e1t, env);

	t = lub(e1t, e2t);
	if(t == Unknown() && e1t != Unknown() && e2t != Unknown())
		env = errorMark(env, "If branch types \'<dispNameOf(e1t, env)>\' and \'<dispNameOf(e2t, env)>\' are not compatible", locOf(expr));
	
	
	return <Recover(e1, e2), t, env>;
}

tuple[AST,AST,Env] chkExpr(expr:Guard(e1, e2), Env env) {
	<e1, e1t, env> = checkExpr(e1, env);
	<e2, e2t, env> = checkExpr(e2, env);
	
	if(!isAssignable(SpecialType(Predicate()), e2t, env))
		env = errorMark(env, "Guard condition should be a predicate (type was <dispNameOf(e1t, env)>)", locOf(e1));

	return <Guard(e1, e2), e1t, env>;
}

// Expression:   { stats }
tuple[AST,AST,Env] chkExpr(expr:BlockExpr(ss, _), Env env) {
	env = enterScope(env);
	env = replaceDef(preserveAnnos(returnTypeName, expr), makeDefInfo(preserveAnnos(returnTypeName, expr), env)[body=Unknown()], env);
	<ss, env> = checkStat(ss, env);
	env = exitScope(env);
	set[AST] retTypes = {};
	top-down-break visit(ss) {
		case BlockExpr(_,_): ;
		case r:Return(_): retTypes += r@\type;
	}
	
	if({} := retTypes) {
		env = errorMark(env, "Block expression does not contain a return statement", Error(), locOf(expr));
		return <BlockExpr(ss, Unknown()), Unknown(), env>;
	}
	else if({t} := retTypes) {
		return <BlockExpr(ss, t), t, env>;
	}
	else if(Unknown() in retTypes) {
		return <BlockExpr(ss, Unknown()), Unknown(), env>;
	}
	else {
		retType = lub(retTypes);
		if(retType == Unknown()) 
			env = errorMark(env, "Block expression has incompatible return types", Error(), locOf(expr));
		return <BlockExpr(ss, retType), retType, env>;
	}
}

// TODO: Expression:   e . n
tuple[AST,AST,Env] chkExpr(expr:DotOp(e, n, t), Env env) {
	if(t == Unknown()) {
		<checked, ct, env> = checkExpr(Apply(Fun(Name(leaf("_get_<nameToStr(n)>"))), seq([e])), env);
		if(Apply(_, seq([ce])) := checked) {
			return <DotOp(ce, n, ct), ct, env>;
		}
		else { // shouldn't happen
			return <checked, ct, env>;
		}
	}
	else {
		return <expr, t, env>;
	}

	// OLD STUFF:
	/*
	<e, t, env> = checkExpr(e, env);
	if(t == SpecialType(DeclList())) {
		return <DotOp(e, n, SpecialType(Decl())), SpecialType(Decl()), env>;
	}
	else if(t != Unknown()) {
		try {
			info = lookupDef(t, env);
			body = instantiate(info.body, env); 
			// termPrintln("chkExpr DotOp: ", n);
			
			<ft, env> = getField(t, body, n, env);
			env = addTokenRef(n, ft, env);
			return <DotOp(e, n, ft), ft, env>;
		}
		catch NoSuchKey(_): {
			env = errorMark(env, "Internal Compiler Error: Expressions: resolved name has no definition: <dispNameOf(t, env)>", locOf(t));
		}
	}
	return <DotOp(e, n, Unknown()), Unknown(), env>;
	*/
}

tuple[AST,AST,Env] chkExpr(expr:Undefined(), Env env) {
	return <expr, Unknown(), env>;
}

default tuple[AST,AST,Env] chkExpr(AST tree, Env env) {
	println("CHECKEXPR: no rule for <tree>");
	env = errorMark(env, "Internal Compiler Error: Expressions::chkExpr: no checkExpr rule for <consOf(tree)>", locOf(tree));
	return <tree, Unknown(), env>;
}





public tuple[list[AST],Env] checkParams(list[AST] trees, Env env) {
	list[AST] result = [];
	for(a <- trees) {
		<b, env> = chkParam(a, env);
		result += preserveAnnos(b, a);
	}
	return <result, env>;
}

tuple[AST,Env] chkParam(p:Param(m, n, t), Env env) {
	if(n == Unknown())
		env = errorMark(env, "Parameter requires a name", Error(), locOf(p));
	dt = t;
	if(Type(_) := t)
		<dt, t, env> = checkType(t, env);
	env = addTokenRef(n, VarName(n, t), env);
	
	qName = preserveAnnos(VarName(n,t), n);
	info = makeDefInfo(qName, env);
	info.defName = n;
	info.srcLoc = locOf(p);
	info.decl = preserveAnnos(Param(m, n, dt), p);
	env = addName(n, qName[@hoverInfo = info], ExprKind(), env);
	
	return <Param(m, n[@hoverInfo = info], dt), env>;
}


tuple[AST,Env] getField(AST t, Body(Struct(seq(ds))), AST n, Env env) {
	if(leaf(kStr:/[0-9]+/) := n) {
		k = toInt(kStr);
		i = 1;
		for(TypeField(VarDecl(m, fn, ft, fi)) <- ds) {
			if(i == k)
				return <ft, addTokenRef(n, VarName(fn, ft), env)>;
			i = i + 1;
				
		}
		env = errorMark(env, "Unknown field #<k> in type \'<dispNameOf(t, env)>\'", locOf(n));
	}
	else {
		for(TypeField(VarDecl(m, fn, ft, fi)) <- ds) {
			if(fn == n)
				return <ft, addTokenRef(n, VarName(fn, ft), env)>;
		}
		env = errorMark(env, "Unknown field \'<nameToStr(n)>\' in type \'<dispNameOf(t, env)>\'", locOf(n));
	}
	return <Unknown(), env>;
}

default tuple[AST,Env] getField(AST t, AST body, AST n, Env env) {
	env = errorMark(env, "Type \'<dispNameOf(t, env)>\' does not have fields", locOf(n));
	return <Unknown(), env>;
}

str dispArgList(list[AST] args, Env env) {
	str dname(AnonParam(_, t)) { return dname(t); }
	str dname(AST n) { return nameToStr(unprefix(defNameOf(n), env.modName)); }
	return strJoin([dname(a) | a <- args], ", ");
}


// convert literal arguments to plain exprs, if any
public tuple[list[AST], Env] checkLiteralArgs(list[AST] actuals, list[AST] formals, Env env) {
	for(i <- index(actuals)) {
		t = getExprType(formals[i]);
		<actuals[i], _, env> = checkLiteral(actuals[i], t, t, env);
	}
	return <actuals, env>;
}

// check 'lit' and convert to plain expr, if it is a literal, otherwise do nothing
public tuple[AST, AST, Env] checkLiteral(AST expr, AST exprType, AST typ, Env env) {
	if(typ == SpecialType(Literal()) || typ == Unknown()) {
		if(Literal(l) := expr || Literal(l,_) := expr) {
			env = errorMark(env, "Literal must be converted to a proper type (context expects literal or unknown, should expect value of a known type)", locOf(expr));
		}
	}
	else if(Literal(l) := expr || Literal(l,_) := expr) {
		if(hasDef(typ, env)) { // if not, an error has already been reported
			info = lookupDef(typ, env);
			
			if(attr:IntegerLiteralAttr(_,_,_,_,_,_) := getAttr("IntegerLiteralAttr", info.attrs)) {
				termPrintln("literal integer: ", expr, " type: ", typ); 
				<l, env> = literalToExpr(l, attr, env);
				if(l != Nop()) {
					expr = l;
					exprType = typ;
				}
				termPrintln("  =\> ", expr);
			}
			else {
				// 
				// env = errorMark(env, "Type \'<shortDispNameOf(typ)>\' does not support literals", locOf(expr));
				termPrintln("literal other: ", expr, " type: ", typ, " type attrs: ", info.attrs); 
			}
		}
	}
	return <expr, exprType, env>;
}

tuple[AST, Env] literalToExpr(AST lit,	IntegerLiteralAttr(typeName, zeroName, oneName, plusName, mulName, negName), env) {
	AST base;
	str l;
	str sign = "";
	int radix;
	if(DecimalIntegerLiteral(leaf(/^<sgn:[\-+]?><s:.*>$/)) := lit) {
		base = makeDigit(10, zeroName, oneName, plusName);
		l = s;
		sign = sgn;
		radix = 10;
	}
	else if(OctalIntegerLiteral(leaf(/^0<s:.*>$/)) := lit) {
		base = makeDigit(8, zeroName, oneName, plusName);
		l = s;
		radix = 8;
	}
	else if(HexIntegerLiteral(leaf(/^0[xX]<s:.*>$/)) := lit) {
		base = makeDigit(16, zeroName, oneName, plusName);
		l = s;
		radix = 16;
	}
	else {
		env = errorMark(env, "Literal \'<unparse(lit)>\' not suitable as integer literal", locOf(lit));
		return <Nop(), env>;
	}
	
	AST e = Undefined();
	
	while(/^<head:[0-9a-fA-F]><tail:.*>$/ := l) {
		digit = makeDigit(toInt(head, radix), zeroName, oneName, plusName);
		if(e == Undefined()) {
			e = digit;
		}
		else {
			e = Apply(Fun(plusName), seq([Apply(Fun(mulName), seq([e, base])), digit]));  
		}
		l = tail;
	}
	if(e == "-") {
		if(negName != Nop()) {
			e = Apply(Fun(negName), seq([e]));
		}
		else {
			env = errorMark(env, "Type \'<shortDispNameOf(typeName)>\' doesn\' support negative literals", locOf(lit));
		}
	}
	e = setLocRec(e, locOf(lit));

	return <e, env>;
}

AST makeDigit(int d, AST zeroName, AST oneName, AST plusName) {
	if(d == 0) {
		return Apply(Fun(zeroName), seq([]));
	}
	else {
		oneExpr = Apply(Fun(oneName), seq([]));
		e = oneExpr;
		while(d > 1) {
			e = Apply(Fun(plusName), seq([e, oneExpr]));
			d = d - 1; 
		}
		return e;
	}
}
