/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::algo::Merge
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::lib::Renaming;
import org::magnolialang::magnolia::typechecker::algo::Expressions;
import org::magnolialang::magnolia::typechecker::algo::Attributes;
import org::magnolialang::magnolia::typechecker::algo::Statements;
import org::magnolialang::magnolia::typechecker::algo::DerivedOps;
import org::magnolialang::magnolia::typechecker::algo::Types;
import org::magnolialang::magnolia::typechecker::algo::Primitives;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import List;
import String;
import Map;
import util::Benchmark;
import util::Math;


public tuple[AST, Env] makeOverloadUnique(AST mdl, tuple[AST, Env](AST, AST, Env) combine, Env env) {
	if(DeclBody(seq(ds)) := mdl) {
		list[AST] result = [];
		map[AST, AST] defs = ();
		
		for(d <- ds) {
			if(Define(dcl, attrs, body) := d) {
				qName = defNameOf(dcl);
				if(qName in defs) {
					termPrintln("makeOverloadUnique: uniqifying ", qName);
					<defs[qName], env> = combine(d, defs[qName], env);
				}
				else {
					defs[qName] = d;
				}	
			}
			else {
				result += d;
			}
		}
		
		result += toList(range(defs));
		mdl.arg0.args = result;
	}
	else {
		env = errorMark(env, "Internal Compiler Error: Merge::makeOverloadUnique: expected DeclBody: <consOf(mdl)>", locOf(mdl));
	}
	return <mdl, env>;
}
public tuple[AST, Env] joinDef(AST thisDecl, AST prevDecl, Env env) {
	DEBUG_DECISIONS = false;
	// Inputs -- normal, derived, default, require, no body, axiom
	int Nrm = 0, Der = 1, Def = 2, Req = 3, Nbd = 4, Ax = 5, Pr = 6, RowSize = 7;
	// Decisions
	// I:choose thisDecl, O:choose prevDecl, E:error (unless equal), M:merge
	// MD: merge derived
	// IG:use thisDecl and merge guard, OG: use prevDecl and merge guard
	int I = 0, O = 1, E = 2, M = 3, IG = 4, OG = 5, MD = 6;
	//  -  Der Def Req Nop Ax   Pr: info
	matrix = [                      // prevDecl:
		E,	O,	O,	O,	O,	M,	OG,	// -
		I,	MD,	O,	O,	O,	M,	OG,	// Derived
		I,	I,	E,	O,	O,	M,	OG,	// Default
		I,	I,	I,	E,	O,	M,	OG,	// Require
		I,	I,	I,	I,	E,	I,	OG,	// Nop body
		M,	M,	M,	M,	O,	M,	E,	// Axiom
		IG,	IG,	IG,	IG,	IG,	E,	IG  // Protect
	];

	ij = <Nrm, Nrm>;
	for(<decl,i> <- [<thisDecl, 0>, <prevDecl, 1>]) {
		for(a <- getDeclAttrs(decl)) {
			if(a == Derived())
				ij[i] = max(ij[i], Der);
			else if(a == Default())
				ij[i] = max(ij[i], Def);
			else if(a == Require())
				ij[i] = max(ij[i], Req);
			else if(a == Axiom())
				ij[i] = max(ij[i], Ax);
			else if(a == Protect())
				ij[i] = max(ij[i], Pr);
		}
		if(getDefBody(decl) == EmptyBody())
			ij[i] = max(ij[i], Nbd);
	}
	
	if(DEBUG_DECISIONS) {
		termPrintln("Joining: ", defNameOf(thisDecl));
		println("  this is <ij[0]>, prev is <ij[1]>");
	}
	decision = matrix[ij[0] + RowSize * ij[1]];
	
	switch(decision) {
		case I: { // this
			if(DEBUG_DECISIONS) {
				println("  Decision: I");
				termPrintln("    thisDecl.body:  ", getDefBody(thisDecl));
				termPrintln("    otherDecl.body: ", getDefBody(prevDecl));
			}
			return mergeAttrs(thisDecl, prevDecl, env);
		}
		case O: { // other
			if(DEBUG_DECISIONS) {
				println("  Decision: O");
				termPrintln("    thisDecl.body: ", getDefBody(thisDecl));
				termPrintln("    prevDecl.body: ", getDefBody(prevDecl));
			}
			return mergeAttrs(prevDecl, thisDecl, env);
		}
		case IG: { // this + other guard
			if(DEBUG_DECISIONS) {
				println("  Decision: IG");
				termPrintln("    thisDecl.body: ", getDefBody(thisDecl));
				termPrintln("    prevDecl.body: ", getDefBody(prevDecl));
			}
			return mergeProtect(thisDecl, prevDecl, env);
		}
		case OG: { // other + this guard
			if(DEBUG_DECISIONS) {
				println("  Decision: OG");
				termPrintln("    thisDecl.body: ", getDefBody(thisDecl));
				termPrintln("    prevDecl.body: ", getDefBody(prevDecl));
			}
			return mergeProtect(prevDecl, thisDecl, env);
		}
		case E: { // error
			if(getDefBody(thisDecl) == getDefBody(prevDecl)) { // TODO: make more refined (only for concepts/require
				if(DEBUG_DECISIONS)
					println("  Decision: =");
			}
			else if(false && isTypeDecl(thisDecl)) { // not doing type merging anymore
				if(DEBUG_DECISIONS) {
					println("  Decision: MT");
					termPrintln("    thisDecl.body: ", getDefBody(thisDecl));
					termPrintln("    prevDecl.body: ", getDefBody(prevDecl));
				}
				<decl, env> = mergeTypeDecls(thisDecl, prevDecl, env);
				if(DEBUG_DECISIONS) {
					termPrintln("    result.body:   ", getDefBody(decl));
				}
				return <decl, env>;
			}
			else if(false && isAlertDecl(thisDecl)) { // not doing type merging anymore
				if(DEBUG_DECISIONS) {
					println("  Decision: MT");
					termPrintln("    thisDecl.body: ", getDefBody(thisDecl));
					termPrintln("    prevDecl.body: ", getDefBody(prevDecl));
				}
				<decl, env> = mergeAlertDecls(thisDecl, prevDecl, env);
				if(DEBUG_DECISIONS) {
					termPrintln("    result.body:   ", getDefBody(decl));
				}
				return <decl, env>;
			}
			else {
				if(DEBUG_DECISIONS) {
					println("  Decision: E");
				}
				env = warnConflict(prevDecl, thisDecl, env);							
			}
			return <prevDecl, env>;
		}
		case M: { // merge
			if(DEBUG_DECISIONS)
				println("  Decision: M");
			<prevDecl.arg2, env> = mergeAxiom(prevDecl, thisDecl, env);
			return mergeAttrs(prevDecl, thisDecl, env);
		}
		case MD: { // merge derived
			if(getDefBody(thisDecl) == getDefBody(prevDecl)) {
				if(DEBUG_DECISIONS)
					println("  Decision: MD=");
			}
			else if(<Body(Apply(FunOf(p,_),_)), Body(Apply(FunOf(p,_),_))> := <getDefBody(thisDecl), getDefBody(prevDecl)>) {
				if(DEBUG_DECISIONS)
					println("  Decision: MD");
				return mergeAttrs(prevDecl, thisDecl, env);
			}
			else {
				//if(DEBUG_DECISIONS) {
					println("  Decision: E");
				termPrintln("    thisDecl.body: ", getDefBody(thisDecl));
				termPrintln("    prevDecl.body: ", getDefBody(prevDecl));
				//}
				<via1, loc1> = getLatestOrigin(thisDecl, locOf(thisDecl));
				<via2, loc2> = getLatestOrigin(prevDecl, locOf(prevDecl));
				from1 = loc1 == locOf(thisDecl) ? "" : "from <locInfo(locOf(thisDecl), loc1)> ";
				from2 = loc2 == locOf(prevDecl) ? "" : "from <locInfo(locOf(prevDecl), loc1)> ";
				otherDef1 = "(other definition <from2><via2 == "" ? "" : "via <via2> ">at <locInfo(loc2, loc1)>)"; 
				otherDef2 = "(other definition <from1><via1 == "" ? "" : "via <via1> ">at <locInfo(loc1, loc1)>)"; 
				
				env = errorMark(env, "Conflicting definition of derived operation \'<dispNameOf(thisDecl, env)>\' <from1><otherDef1>", loc1);
				env = errorMark(env, "Conflicting definition of derived operation \'<dispNameOf(prevDecl, env)>\' <from2><otherDef2>", loc2);
			}
			return <prevDecl, env>;

		}
	}
}



public Env warnConflict(AST def1, AST def2, Env env) {
	msg = "Conflicting definition";
	level = Error();
	more = "";
	b1 = getDefBody(def1);
	b2 = getDefBody(def2);
	if(<Instance(iName1, bs1), Instance(iName2, bs2)> := <b1, b2>) {
		if(hasDef(iName1, env) && hasDef(iName2, env)) {
			implBody1 = lookupDef(iName1, env).body;
			implBody2 = lookupDef(iName2, env).body;
			
			if(bs1 == bs2 && implBody1 == implBody2) {
				msg = "Duplicate definition";
				level = Warning();
			}
			else if(iName1 == iName2) {
				msg = "Conflicting instantiation";
				termPrintln("b1: ", b1);
				termPrintln("b2: ", b2);
				more = ": <nameToStr(b1)> vs <nameToStr(b2)>";
			}
		}
	}
	else {
		<ren, env> = makeParamRenaming(getDeclarator(def2), getDeclarator(def1), env);
		<b2, env> = simpleRenameNoWarn(env, b2, ren);
	
		if(b1 == b2) {
			msg = "Duplicate definition";
			level = Warning();
		}
	}
	termPrintln("Attempting to merge: ", msg, "<more>");
	termPrintln("  1: ", def1);
	termPrintln("  2: ", def2);
	
	<via1, loc1> = getLatestOrigin(def1, locOf(def1));
	<via2, loc2> = getLatestOrigin(def2, locOf(def2));
	from1 = loc1 == locOf(def1) ? "" : "from <locInfo(locOf(def1), loc1)> ";
	from2 = loc2 == locOf(def2) ? "" : "from <locInfo(locOf(def2), loc1)> ";
	otherDef1 = "(other definition <from2><via2 == "" ? "" : "via <via2> ">at <locInfo(loc2, loc1)>)"; 
	otherDef2 = "(other definition <from1><via1 == "" ? "" : "via <via1> ">at <locInfo(loc1, loc1)>)"; 

	env = errorMark(env, "<msg> of \'<dispNameOf(def1, env)>\'<more> <from1><otherDef1>", level, loc1);
	env = errorMark(env, "<msg> of \'<dispNameOf(def2, env)>\'<more> <from2><otherDef2>", level, loc2);
	
	return env;
}
	
public tuple[AST,Env] mergeAxiom(AST def1, AST def2, Env env) {
	b1 = getDefBody(def1);
	b2 = getDefBody(def2);
	bInst1 = b1;
	bInst2 = b2;
	
	if(Instance(iName1, _) := getDefBody(def1) && hasDef(iName1, env)) {
		bInst1 = instantiate(b1, env);
		b1 = Body(Block(seq([Call(Proc(b1), seq([Var(VarName(an,at)) | Param(_,an,at) <- paramsOf(def1)]))])));
		//termPrintln("instantiated b1 = ", b1);
	}
	
	if(Instance(iName2, _) := getDefBody(def2) && hasDef(iName2, env)) {
		bInst2 = instantiate(b2, env);
		b2 = Body(Block(seq([Call(Proc(b2), seq([Var(VarName(an,at)) | Param(_,an,at) <- paramsOf(def1)]))])));
		//termPrintln("instantiated b2 = ", b2);
	}
	

	if(bInst1 == bInst2) { // instantiates to the same? use just one of them
		return <b1, env>;
	}
	
	<ren, env> = makeParamRenaming(getDeclarator(def2), getDeclarator(def1), env);
	<b2, env> = simpleRenameNoWarn(env, b2, ren);
	
	
	if(b1 == b2) { // same body modulo parameter names? use just one of them
		return <b1, env>;
	}
	else {
		if(<Body(Block(seq(ss1))), Body(Block(seq(ss2)))> := <b1, b2>) {
			return <Body(Block(seq([*ss1, *ss2]))), env>;
		}
		else {
			termPrintln("Failed!");
			termPrintln("  Body1: ", getDefBody(def1));
			termPrintln("       =\>", b1);
			termPrintln("  Body2: ", getDefBody(def2));
			termPrintln("       =\>", b2);
			<via1, loc1> = getLatestOrigin(def1, locOf(def1));
			<via2, loc2> = getLatestOrigin(def2, locOf(def2));
			otherDef1 = "(other definition <via2 == "" ? "" : "via <via2> ">at <locInfo(loc2, loc1)>)"; 
			otherDef2 = "(other definition <via1 == "" ? "" : "via <via1> ">at <locInfo(loc1, loc1)>)"; 
			env = errorMark(env, "Unable to merge axiom body of conflicting definition of \'<dispNameOf(def1, env)>\' <otherDef1>", Internal(), loc1);
			env = errorMark(env, "Unable to merge axiom body of conflicting definition of \'<dispNameOf(def2, env)>\' <otherDef2>", Internal(), loc2);
		}
	}
	
	return <b1, env>;
}
