/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps

import org::magnolialang::magnolia::typechecker::Flattener;
import org::magnolialang::magnolia::AbstractSyntax;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::typechecker::TypeChecker;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::lib::CxxAPI;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::evaluator::Values;
import org::magnolialang::magnolia::backend::OperatorInfo;
import IO;
import Set;
import List;
import Node;

public map[str,AST] modifierOps = (
	"protect" : Protect(),
	"remove" : Remove(),
	"retain" : Retain()
	);
	
public default tuple[AST, Env] applyOp(AST op, list[AST] as, loc location, Env env) {
	env = errorMark(env, "Unknown top-level operator <nameToStr(op)>", locOf(op));
	return <TopApp(op, seq(as)), env>;
}

public tuple[AST, Env] applyOp(op:Name(leaf("signature")), [rhs], loc location,  Env env) {
	<rhs, env> = flattenTopExpr(rhs, env);
	if(isFlat(rhs)) {
		rhs = DeclBody(seq(signatureOf(rhs)));
		env = addTokenRefForOp(op, rhs, env);
		return <rhs, env>;
	}
	else
		return <TopApp(op, seq([rhs])), env>;
}

public tuple[AST, Env] applyOp(op:Name(leaf("full")), [rhs], loc location,  Env env) {
	<rhs, env> = flattenTopExpr(rhs, env);
	if(isFlat(rhs)) {
		env = addTokenRefForOp(op, rhs, env);
		return <rhs, env>;
	}
	else
		return <TopApp(op, seq([rhs])), env>;
}

public tuple[AST, Env] applyOp(op:Name(leaf("on")), [rhs], loc location,  Env env) {
	<rhs, env> = flattenTopExpr(rhs, env);
	if(isFlat(rhs)) {
		rhs = DeclBody(seq(onOf(rhs)));
		env = addTokenRefForOp(op, rhs, env);
		return <rhs, env>;
	}
	else
		return <TopApp(op, seq([rhs])), env>;
}

public tuple[AST, Env] applyOp(op:Name(leaf("declared")), [rhs], loc location,  Env env) {
	<rhs, env> = flattenTopExpr(rhs, env);
	if(isFlat(rhs)) {
		rhs = DeclBody(seq(definesOf(rhs)));
		env = addTokenRefForOp(op, rhs, env);
		return <rhs, env>;
	}
	else
		return <TopApp(op, seq([rhs])), env>;
}

public tuple[AST, Env] applyOp(op:Name(leaf("extend")), [rhs], loc location,  Env env) {
	<rhs, env> = flattenTopExpr(rhs, env);
	if(isFlat(rhs)) {
		rhs = addModifier(rhs, Extend());
		env = addTokenRefForOp(op, rhs, env);		
	}
	return <rhs, env>;
}

public default tuple[AST, Env] applyDefOp(AST op, AST def, loc location, Env env) {
	env = errorMark(env, "Unknown top-level operator <nameToStr(op)>", locOf(op));
	return <def, env>;
}

// TODO
public tuple[AST, Env] applyDefOp(op:Name(leaf("datainvariant")), AST def, loc location,  Env env) {
	if(Define(_,_,_) := def) {
		env = addTokenRefForOp(op, def, env);		
		env = errorMark(env, "Not really implemented yet: <nameToStr(op)>",  Warning(), locOf(op));
	}
	else {
		env = errorMark(env, "Not applying <nameToStr(op)> to <consOf(def)>",  Warning(), locOf(op));
	}
	return <def, env>;
}

// TODO
public tuple[AST, Env] applyDefOp(op:Name(leaf("congruence")), AST def, loc location,  Env env) {
	if(Define(_,_,_) := def) {
		env = addTokenRefForOp(op, def, env);		
		env = errorMark(env, "Not really implemented yet: <nameToStr(op)>",  Warning(), locOf(op));
	}
	else {
		env = errorMark(env, "Not applying <nameToStr(op)> to <consOf(def)>",  Warning(), locOf(op));
	}
	return <def, env>;
}

public tuple[AST, Env] applyDefOp(op:Name(leaf("quotient")), AST def, loc location,  Env env) {
	if(Define(_,_,_) := def) {
		env = addTokenRefForOp(op, def, env);		
		env = errorMark(env, "Not really implemented yet: <nameToStr(op)>",  Warning(), locOf(op));
	}
	else {
		env = errorMark(env, "Not applying <nameToStr(op)> to <consOf(def)>",  Warning(), locOf(op));
	}
	return <def, env>;
}


public tuple[AST, Env] applyOp(op:Name(leaf("models")), [lhs, with, rhs], loc location,  Env env) {
	<combined, env> = flattenTopExpr(Close(DeclList(seq([TopMap(Modifier(preserveAnnos(Assume(), op)), lhs),
					TopMap(Modifier(preserveAnnos(Use(), op)), with),
					TopMap(Modifier(preserveAnnos(ImplyMod(), op)), TopMap(Modifier(preserveAnnos(Require(), op)), rhs))]))), env);
/*
	if(DeclBody(seq(ds)) := combined) {
		for(d <- ds) {
			if(hasModifier(d, Require())) {
				env = errorMarkInDecl(env, "Requirement \'<shortDispNameOf(d)>\' not satisfied", Error(), d);
			}
		}
*/
		env = addTokenRefForOp(op, combined, env);
		return <combined, env>;
/*	}
	else {
		return <TopApp(op, seq([lhs, with, rhs])), env>;
	}
	*/	
}

// TODO
public tuple[AST, Env] applyOp(op:Name(leaf("subalgebra")), [lhs, rhs], loc location,  Env env) {
	env = errorMark(env, "Not really implemented yet <nameToStr(op)>", Warning(),  locOf(op));
	<lhs, env> = flattenTopExpr(lhs, env);
	env = enterScope(env);
	<_, env> = loadNames(lhs, false, env);
	<rhs, env> = flattenTopExpr(rhs, env);
	env = exitScope(env);
	if(!isFlat(lhs) || !isFlat(rhs))
		return <TopApp(op, seq([lhs, rhs])), env>;
	termPrintln("dataInvariant: ", rhs);
	try {
		if(DeclBody(seq([def:Define(_,dcl,_,_)])) := rhs) { // && hasModifier(def, Predicate())) {
			bool found = false;
			list[AST] decls = [];
			for(d <- fullOf(lhs)) {
				if(canonical(def) == canonical(d)) {
					found = true;
				}
				decls += d;
			}
			if(!found)
				env = errorMark(env, "Data invariant predicate must also be defined at left-hand side", Warning(), locOf(dcl));
			lhs.arg0.args = decls;
			env = addTokenRefForOp(op, lhs, env);
			return <lhs, env>;
		}
		else {
			env = errorMark(env, "Quotient right-hand side should contain a single predicate", locOf(rhs));
		}
	}
	catch InternalError(msg, ast): {
		env = errorMark(env, msg, Internal(), locOf(ast));
	}
	return <TopApp(op, seq([lhs, rhs])), env>;
}

// TODO
public tuple[AST, Env] applyOp(op:Name(leaf("quotient")), [lhs, rhs], loc location,  Env env) {
	env = errorMark(env, "Not really implemented yet <nameToStr(op)>", Warning(), locOf(op));
	<lhs, env> = flattenTopExpr(lhs, env);
	env = enterScope(env);
	<_, env> = loadNames(lhs, false, env);
	<rhs, env> = flattenTopExpr(rhs, env);
	env = exitScope(env);
	if(!(isFlat(lhs) && isFlat(rhs)))
		return <TopApp(op, seq([lhs, rhs])), env>;
	try {
		if(DeclBody(seq([def:Define(_,dcl,_,_)])) := rhs) { // && hasModifier(def, Predicate())) {
			definedSet = {canonical(d) | Define(_,d,_,_) <- definesOf(lhs)};
			ps = paramsOf(dcl);
			if(canonical(dcl) notin definedSet)
				env = errorMark(env, "Predicate must also be defined at left-hand side", Warning(), locOf(dcl));
			eqName = FunName(Name(leaf("_==_")), seq([AnonParam(m,t) | Param(m, n, t) <- ps]), SpecialType(Predicate()));
			newDef = Define(seq([]), FunClause(eqName, ParamList(seq(ps)), SpecialType(Predicate())),
				seq([Predicate(), Derived()]),
				Body(Apply(Fun(defNameOf(dcl)), seq(paramsToArgs(ps)))));
			lhs.arg0.args += [newDef];
			env = addTokenRefForOp(op, lhs, env);
			return <lhs, env>;
		}
		else {
			env = errorMark(env, "Quotient right-hand side should contain a single predicate", locOf(rhs));
		}
	}
	catch InternalError(msg, ast): {
		env = errorMark(env, msg, Internal(), locOf(ast));
	}
	return <TopApp(op, seq([lhs, rhs])), env>;
}




public Env addTokenRefForOp(AST op, AST result, Env env) {
	return addTokenRefForOp(op, result, "", env);
}
public Env addTokenRefForOp(AST op, AST result, str doc, Env env) {
	info = definfo();
	info.locInfo = LocInfo();
	info.decl = FunName(Name(op), seq([AnonParam(Obs(), decllistType), AnonParam(Obs(), decllistType)]), decllistType); 
	info.defName = Name(op);
	info.qName = FunName(Name(op), seq([AnonParam(Obs(), decllistType), AnonParam(Obs(), decllistType)]), decllistType); 
	info.parent = QName(seq([]));
	info.attrs = [Builtin()];
	info.body = result;
	info.doc = doc;
	return addTokenRef(locOf(op), LocalRef(info), env);
}

public tuple[AST,Env] makeExternal(AST language, AST name, AST tree, Env env) {
	list[AST] onOps = [];
	list[AST] onTypes = [];
	
	for(Define(dcl,_,_) <- onOf(tree)) {
		if(isOperation(dcl)) {
			onOps += defNameOf(dcl);
			termPrintln("makeExternal: require op: ", defNameOf(dcl));
		}
		else if(isTypeDecl(dcl)) {
			onTypes += defNameOf(dcl);
			termPrintln("makeExternal: require type: ", defNameOf(dcl));
		}
	} 
	
	onOps = sort(onOps, bool(AST x, AST y) { return basename(x) < basename(y); });
	onTypes = sort(onTypes, bool(AST x, AST y) { return basename(x) < basename(y); });
	
	if(DeclBody(seq(ds)) := tree) {
		list[AST] names = [];
		for(i <- index(ds)) {
			switch(ds[i]) {
				case Define(def,attrs,body): {
					if(body == EmptyBody() && (size(toSet(attrs.args) & {Derived(), Require(), External()}) == 0)) {
						<def, n, env> = makeExternal(language, name, def, ds[i], ExternalParams(seq(onTypes), seq(onOps)), attrs, env);
						def = changeModifiers(def, {Use(), Extend()}, {External()});
						ds[i] = preserveAnnos(def, ds[i]);
						names += basename(n);
					}
				}
				default:
					env = errorMark(env, "TopLevelOps::makeExternal: Unknown declaration <consOf(ds[i])>", Error(), locOf(ds[i]));
			}
		}
		while(names != []) {
			<n, names> = <head(names), tail(names)>;
			if(n in names && !/#/ := nameToStr(n)) {
				env = errorMark(env, "Overloading (of name \'<nameToStr(n)>\') discouraged in external <getLanguageName(language)>", Warning(), locOf(n));
			}
		}
		return <DeclBody(seq(ds)), env>;
	}
	else
		throw InternalError("TopLevelOps::makeExternal: Top-level expression not fully flattened: <consOf(tree)>", tree);
}

private tuple[AST, Env] checkExternalName(AST language, AST def, AST decl, AST name, Env env) {
	if(isCxxLanguage(language)) {
		n = nameToStr(name);
	
		if(!/^[a-zA-Z][_a-zA-Z0-9]*$/ := n) {
			env = errorMarkInDecl(env, "Illegal external C++ name \'<n>\' (Use /[a-zA-Z][_a-zA-Z0-9]*/)", Error(),  decl);
			name = preserveAnnos(Name(leaf("__error__\n#error Illegal external C++ name \'<n>\'\n")), name);
		}
		else if(n in reservedKeywords) {
			env = errorMarkInDecl(env, "Illegal external C++ name: \'<n>\' is a reserved keyword", Error(), decl);
			name = preserveAnnos(Name(leaf("__error__\n#error Illegal external C++ name \'<n>\'\n")), name);
		}

		return <name, env>;
	}
	else if(isJavaLanguage(language)) {
		n = nameToStr(name);
	
		if(!/^[a-zA-Z][_a-zA-Z0-9]*$/ := n) {
			env = errorMarkInDecl(env, "Illegal external Java name \'<n>\' (Use /[a-zA-Z][_a-zA-Z0-9]*/)", Error(),  decl);
			name = preserveAnnos(Name(leaf("__error__")), name);
		}
		else if(n in reservedJavaKeywords) {
			env = errorMarkInDecl(env, "Illegal external Java name: \'<n>\' is a reserved keyword", Error(), decl);
			name = preserveAnnos(Name(leaf("__error__")), name);
		}

		return <name, env>;
	}
	else return <name, env>;
}

private tuple[AST,AST,Env] makeExternal(AST language, AST name, AST def, AST decl, AST onPart, AST attrs, Env env) {
	switch(def) {
		case TypeClause(n): {
			<n, env> = checkExternalName(language, def, decl, n, env);
			return <Define(def, attrs, Body(preserveAnnos(ExternalType(language, name, onPart, n), def))), n, env>;
		}
		case FunClause(n,ps,t): {
			<n, env> = checkExternalName(language, def, decl, n, env);
			return <Define(def, attrs, Body(Apply(preserveAnnos(ExternalFun(language, name, onPart, n, ps, t), def), seq(paramsToArgs(ps))))), n, env>;
		}
		case ProcClause(n,ps): {
			<n, env> = checkExternalName(language, def, decl, n, env);
			return <Define(def, attrs, Body(Block(seq([Call(preserveAnnos(ExternalProc(language, name, onPart, n, ps), def), seq(paramsToArgs(ps)))])))), n, env>;
		}
		default:
			throw InternalError("TopLevelOps::makeExternal: Unknown declaration clause <consOf(def)>", def);
	}
}


public list[AST] definesOf(AST expr) {
	return [d | d <- fullOf(expr), !hasModifier(d, {Require()})];
}

public list[AST] onOf(AST expr) {
	return [d | d <- fullOf(expr), hasModifier(d, Require())];
}

public list[AST] fullOf(AST expr) {
	if(DeclBody(seq(ds)) := expr)
		return ds;
	else if(Define(_,_,_) := expr)
		return [expr];
	else if(Nop() := expr)
		return [];
	else
		throw InternalError("TopLevelOps::fullOf: Top-level expression not fully flattened: <consOf(expr)>", expr);
}

@doc{Set of modifiers that should not appear in a signature}
public set[AST] noSigModifiers = {Assume(), Derived(), Broken(), Default(), Extend(), External(), NoBuild(), Protect(), Use(), Require()};
public list[AST] signatureOf(AST expr) {
	switch(expr) {
		case Define(_,seq(attrs),_):
			if(Axiom() in attrs || Impl() in attrs)
				return [];
			//else if(Derived() in attrs)
			//	return [changeModifiers(expr, noSigModifiers ,{})];   // TODO: is this really true?
			else
				return [changeModifiers(expr[arg2 = EmptyBody()], noSigModifiers, {})];
		case DeclBody(seq(ds)):
			return ([] | it + signatureOf(d) | d <- ds);
		case Nop():
			return [];
		default:
			throw InternalError("TopLevelOps::signatureOf: Don\'t know how to extract signature from <consOf(expr)>", expr);
	}
}

public AST canonical(AST decl) {
	result = decl;
	switch(decl) {
		case Define(dcl,_,_):
			result = decl[arg0 = canonical(decl.arg0)][arg1 = Nop()][arg2 = EmptyBody()];
		case Nop():
			result = decl;
		case Rename(_,_):
			result = decl;
		case TypeClause(_):
			result = decl;
		case FunClause(_,ParamList(seq(ps)),_):
			result = decl[arg1=ParamList(seq([anonParam(p) | p <- ps]))];
		case ProcClause(_,ParamList(seq(ps))):
			result = decl[arg1=ParamList(seq([anonParam(p) | p <- ps]))];
		default: {
			termPrintln("canonical: ", decl);
			throw InternalError("TopLevelOps::canonical: Don\'t know how to canonicalize <consOf(decl)>", decl);
		}
	}
	
	return preserveAnnos(result, decl);
}

public tuple[AST, AST] lhsAndRhsOf(AST expr) {
	switch(expr) {
		case Models(a, b):
			return <a, b>;
		case WithModels(a, w, b):
			return <a, b>;
		case Morphed(a, b):
			return <a, b>;
		case Protected(a, b):
			return <a, b>;
		case OnDefines(a, b):
			return <a, b>;
		case At(a, b):
			return <a, b>;
		case AtAt(a, b):
			return <a, b>;
		case Plus(a, b):
			return <a, b>;
		case PlusPlus(a, b):
			return <a, b>;
		case Times(a, b):
			return <a, b>;
		case Times(a, b):
			return <a, b>;
		default:
			throw InternalError("TopLevelOps::lhsAndRhsOf: Unknown top-level expr in lhsAndRhsOf(): <consOf(expr)>", expr);
	}
}

public AST lhsOf(AST expr) {
	return lhsAndRhsOf(expr)[0];
}

public AST rhsOf(AST expr) {
	return lhsAndRhsOf(expr)[1];
}

@doc{Set difference, with respect to the canonical representation of the
declarations in the sets.}
public list[AST] sigDiff(list[AST] lhs, list[AST] rhs) {
	rhsSet = {canonical(d) | d <- rhs};
	return [d | d <- lhs, canonical(d) notin rhsSet];
}

@doc{Set union, with respect to the canonical representation of the
declarations in the sets. Overlapping elements are taken from the lhs set.}
public list[AST] sigUnion(list[AST] lhs, list[AST] rhs) {
//	lhsSet = {canonical(d) | d <- lhs};
//	return lhs + [d | d <- rhs, canonical(d) notin lhsSet];
	lhsSet = toSet(lhs);
	return lhs + [d | d <- rhs, d notin lhsSet];
}

@doc{Set intersection, with respect to the canonical representation of the
declarations in the sets. Elements are taken from the lhs set.}
public list[AST] sigIsect(list[AST] lhs, list[AST] rhs) {
	rhsSet = {canonical(d) | d <- rhs};
	return [d | d <- lhs, canonical(d) in rhsSet];
}

public bool isIn(AST decl, AST declList) {
	set[AST] decls = {canonical(d) | d <- fullOf(declList)};
	return canonical(decl) in decls;
	
}
public AST decls(list[AST] ds) {
	return DeclBody(seq(ds));
}

public AST decls(list[AST] ds, AST from) {
	l = locOf(from);
	return DeclBody(seq([d[@from=l] | d <- ds]));
}
