/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::toplevel::TopPrimitives
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::algo::Primitives;
import org::magnolialang::magnolia::evaluator::Values;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import List;
import String;
import util::Benchmark;
import util::Math;

public Env addTopPrimitives(Env env) {
	for(t <- [Decl(), DeclList(), Renaming()]) {
		env = addName(SpecialType(t), SpecialType(t), TypeKind(), env);
		info = definfo();
		info.locInfo = LocInfo();
		info.decl = TypeClause(SpecialType(t));
		info.defName = SpecialType(t);
		info.qName = SpecialType(t);
		info.parent = QName(seq([]));
		info.attrs = [Builtin()];
		env = addDef(SpecialType(t), info, env);
	}

	env = addSpecialType("ast", "\<p class=\"dochead\"\>AST  – \<i\>basic top-level type.\</i\>\</p\>\<p\>\</p\>", env);
	env = addSpecialType("int", "\<p class=\"dochead\"\>Integer  – \<i\>basic top-level type.\</i\>\</p\>\<p\>\</p\>", env);
	env = addSpecialType("real", "\<p class=\"dochead\"\>Real  – \<i\>basic top-level type.\</i\>\</p\>\<p\>\</p\>", env);
	env = addSpecialType("str", "\<p class=\"dochead\"\>String  – \<i\>basic top-level type.\</i\>\</p\>\<p\>\</p\>", env);
	env = addSpecialType("void", "\<p class=\"dochead\"\>Void  – \<i\>basic top-level type.\</i\>\</p\>\<p\>\</p\>", env);
	env = addSpecialOp("_in_", [<"e1", declType>, <"e2", decllistType>], SpecialType(Predicate()),
		"\<p class=\"dochead\"\>Declaration membership  – \<i\>basic top-level predicate.\</i\>\</p\>\<p\>True iff declaration is a member of the given declaration list.\</p\>", env);
	env = addSpecialOp("_[_]", [<"e1", declType>, <"e2", renamingType>], declType,
		"\<p class=\"dochead\"\>Renaming  – \<i\>basic top-level operation.\</i\>\</p\>\<p\>Apply a renaming to the declaration.\</p\>", env);
	env = addSpecialOp("_[_]", [<"e1", decllistType>, <"e2", renamingType>], decllistType,
		"\<p class=\"dochead\"\>Renaming  – \<i\>basic top-level operation.\</i\>\</p\>\<p\>Apply a renaming to all declarations in the list.\</p\>", env);
	return env;
}

