/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::metaxa::BuildMatch

import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::metaxa::Env;
import org::magnolialang::compiler::Util;
import org::magnolialang::compiler::MetaXaAST;
import Node;
import List;
import Set;
import IO;
import String;

data cmd = jIf(exp cond, cmd consq, cmd alt) | jReturn(exp expr) | jDecl(str typ, exp var, exp expr, cmd) | jNop()
	| jExpr(exp expr) | jTryFinally(cmd tryPart, cmd finallyPart) | jFail()
	| jAssign(exp var, exp expr) | jSeq(cmd cmd1, cmd cmd2)
	| jDef(str name, list[tuple[str, exp]] args, cmd body);
data exp = jApply(exp obj, str meth, list[exp] args)
	| jApply(str meth, list[exp] args) 
	| jOp(exp e1, str op, exp o2)
	| jVar(str name)
	| jLiteral(str val)
	| q(AST ast); 
data AST = unquote(exp expr);

rule simplifySeq jSeq(jSeq(c1, c2), c3) => jSeq(c1, jSeq(c2, c3));

public cmd jBlock(list[cmd] cmds){
	if([] := cmds)
		return jNop();
	else if([c] := cmds)
		return c;
	else if([c1, c2] := cmds)
		return jSeq(c1, c2);
	else if([c, cs*] := cmds)
		return jSeq(c, jBlock(cs));
}

public str buildExpr(value term, Env env) {
	switch(term) {
		case leaf(s):
			return "leaf(\"<escape(s)>\")";
		case seq(s):
			return "seq([<strJoin([buildExpr(x, env) | x <- s], ", ")>])";
		case [ss*]:
			return "vf.list(<strJoin([buildExpr(x, env) | x <- ss], ", ")>)";
		case var(v):
			return "var(\"<v>\")";
		case unquote(e):
			return compileExpr(e, env);
		case node n:
			return "cons(\"<consOf(n)>\", <strJoin([buildExpr(x, env) | x <- getChildren(n)], ", ")>)"; 
	}
}

public str compileExpr(exp expr, Env env) {
	switch(expr) {
		case q(e):
			return buildExpr(e, env);
		case jApply(obj, meth, args):
			return "<compileExpr(obj, env)>.<meth>(<strJoin([compileExpr(a, env) | a <- args], ", ")>)";
		case jApply(meth, args):
			return "<meth>(<strJoin([compileExpr(a, env) | a <- args], ", ")>)";
		case jVar(v):
			return "<v>";
		case jLiteral(v):
			return "<v>";
		case jOp(e1, op, e2):
			return "(<compileExpr(e1, env)> <op> <compileExpr(e2, env)>)";
	}
}

public str compileCmd(cmd expr, Env env) {
	switch(expr) {
		case jIf(cond, consq, [jIf(cond2, cons2, alt2)]):
			return "if(<compileExpr(cond, env)>) {\n"
				+ "<indent(compileCmd(consq, env))>}\nelse "
				+ compileCmd(jIf(cond2, cons2, alt2), env);
		case jIf(cond, consq, alt):
			return "if(<compileExpr(cond, env)>) {\n"
				+ "<indent(compileCmd(consq, env))>}\nelse {\n"
				+ "<indent(compileCmd(alt, env))>}\n";
		case jReturn(e):
			return "return <compileExpr(e, env)>;\n";
		case jNop():
			return ";\n";
		case jFail():
			return "throw new Failure();\n";
		case jExpr(e):
			return "<compileExpr(e, env)>;\n";
		case jSeq(c1, c2):
			return "<compileCmd(c1, env)><compileCmd(c2, env)>";
		case jAssign(v, e):
			return "<compileExpr(v, env)> = <compileExpr(e, env)>;\n";
		case jTryFinally(tryPart, finallyPart):
			return "try {\n<indent(compileCmd(tryPart, env))>} "
				+ "finally {\n<indent(compileCmd(finallyPart, env))>}\n";
		case jDecl(t, jVar(v), e, cmds):
			return "<t> <v> = <compileExpr(e, env)>;\n<compileCmd(cmds, bind(v, env))>";
		case jDef(n, args, body):
			return "public IConstructor <n>(<strJoin(["<t> <compileExpr(a, env)>" | <t, a> <- args], ", ")>) {\n"
				+ "<indent(compileCmd(body, env))>}\n"; 
		default:
			println("Unknown command: <expr>");
	}
}

public exp jMatch(exp pat, exp expr) {
	return jApply("match", [pat, expr, jApply(jVar("env"), "enterScope", [])]);
}
public cmd mMatch(exp expr, list[tuple[AST, cmd]] conds) {
	potentials = {<consOf(cond),cond,consq> | <cond, consq> <- conds};
	list[tuple[exp, cmd]] compiledConds = [];
		
	for(<cond,consq> <- conds) {

		cons = consOf(cond);

		if(<cons,cond,consq> in potentials) {
			potentials -= {<cons,cond,consq>};

			if(var("_") := cond) {
				compiledConds += <jVar("true"), consq>;
			}
			else if(var(v) := cond) {
				if(isBound(v))
					compiledConds += <jApply(jVar(v), "isEqual", expr), consq>;
				else
					compiledConds += <jVar("true"), jDecl("IConstructor", jVar(v), expr, [consq])>;
			}
			else if(leaf(s) := cond) {
				compiledConds += <jApply(jLiteral("vf.string(\"<s>\")"), "isEqual", expr), consq>;
			}
			else {
				jcond = jOp(jOp(jApply("arity", [expr]), "==", jLiteral("<arity(cond)>")),
					 "&&",
					 jApply(jApply(expr, "consOf", []), "equals", [jLiteral("\"<consOf(cond)>\"")]));

				if(size(potentials[cons]) > 0) {
					vars = varsOf(cond);
					subConds = [<jMatch(q(cond), expr), jTryFinally([quoteVars(consq, vars)], [jExpr(jApply(jVar("env"), "exitScope", []))])>];

					for(<cond1, consq1> <- potentials[cons]) {
						vars = varsOf(cond);
						potentials -= {<cons, cond1, consq1>};
						subConds += <jMatch(q(cond1), expr), jTryFinally(quoteVars(consq1, vars), jExpr(jApply(jVar("env"), "exitScope", [])))>;
					}
					tmpVar = newName();
					compiledConds += <jcond, condToIf(subConds)>;
				}
				else {
					vars = varsOf(cond);
					jcond = jOp(jcond, "&&", jMatch(q(cond), expr));
					compiledConds += <jcond, jTryFinally(quoteVars(consq, vars), jExpr(jApply(jVar("env"), "exitScope", [])))>;
				}
			}
		}
	} 
	return condToIf(compiledConds);
}

private cmd condToIf(list[tuple[exp,cmd]] conds) {
	if(conds == [])
		return jFail();
	else
		return jIf(head(conds)[0], head(conds)[1], condToIf(tail(conds)));
}


public set[AST] varsOf(AST tree) {
	return {var(v) | /var(v) <- tree};
}

private cmd quoteVars(cmd c, set[AST] vars) {
	return top-down visit(c) {
		case jAssign(q(var(v)), e):
			if(var(v) in vars)
				insert jExpr(jApply(jLiteral("env"), "set", [jLiteral("\"<v>\""), e]));
			else
				fail;
		case var(v):
			if(var(v) in vars)
				insert unquote(jLiteral("env.get(\"<v>\")"));
			else
				fail;
	}
}

public str indent(str s) {
	return "\t" + replaceAll(replaceAll(s,"\n","\n\t"), "\t+$", "");
}

public str escape(str s) {
	return s;
}

private int newNameCounter = 0;

public str newName() {
	newNameCounter += 1;
	return "x<newNameCounter>";
}
public str newName(str name) {
	newNameCounter += 1;
	return "<name><newNameCounter>";
}

public str consOf(list[AST] foo) {
	if(size(foo) == 0)
		return "[]";
	else
		return "[<consOf(head(foo))>]";
}
