/**************************************************************************
 * Copyright (c) 2010 Anya Helene Bagge
 * Copyright (c) 2010 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::metaxa::GenerateSyntaxEmbedding
import IO;

public void generate(loc file, str moduleName, list[str] sorts, map[str,str] listSeps) {
	str prods = "";
	str lexprods = "";
	str lexrestr = "";
	str sqLeft = "[[", sqRight = "]]";
	str dqLeft = "{{", dqRight = "}}";
	
	for(sort <- sorts) {
		prods += "	\"<sqLeft>\" <sort>[[SQCtx]] \"<sqRight>\"				-\>	SQuoted[[Ctx]]	{cons(\"SQ<sort>\"),sortname(\"<sort>\"),quotes(\"single\")}\n";
		prods += "	\"<dqLeft>\" <sort>[[DQCtx]] \"<dqRight>\"				-\>	DQuoted[[Ctx]]	{cons(\"DQ<sort>\"),sortname(\"<sort>\"),quotes(\"double\")}\n";
		prods += "	\"<dqLeft>\" <sort>[[DQCtx]] \"<dqRight>\"				-\>	<sort>[[SQCtx]]	{cons(\"DQ<sort>\"),sortname(\"<sort>\"),quotes(\"double\")}\n";
		prods += "	\"<sort>\" \"<sqLeft>\" <sort>[[SQCtx]] \"<sqRight>\"			-\>	SQuoted[[Ctx]]	{cons(\"SQ<sort>\"),sortname(\"<sort>\"),quotes(\"single\")}\n";
		prods += "	\"<sort>\" \"<dqLeft>\" <sort>[[DQCtx]] \"<dqRight>\"			-\>	DQuoted[[Ctx]]	{cons(\"DQ<sort>\"),sortname(\"<sort>\"),quotes(\"double\")}\n";
		prods += "	\"<sort>\" \"<dqLeft>\" <sort>[[DQCtx]] \"<dqRight>\"			-\>	<sort>[[SQCtx]]	{cons(\"DQ<sort>\"),sortname(\"<sort>\"),quotes(\"double\")}\n";
		prods += "	\"<sort>\" \"*\" \"<sqLeft>\" <makelist(sort, "[[SQCtx]]","*",listSeps)> \"<sqRight>\"		-\>	SQuoted[[Ctx]]	{cons(\"SQ<sort>\"),sortname(\"<sort>\"),quotes(\"single\"),list}\n";
		prods += "	\"<sort>\" \"*\" \"<dqLeft>\" <makelist(sort, "[[DQCtx]]","*",listSeps)> \"<dqRight>\"		-\>	DQuoted[[Ctx]]	{cons(\"DQ<sort>\"),sortname(\"<sort>\"),quotes(\"double\"),list}\n";
		prods += "	\"<sort>\" \"*\" \"<dqLeft>\" <makelist(sort, "[[DQCtx]]","*",listSeps)> \"<dqRight>\"		-\>	<sort>[[SQCtx]]	{cons(\"DQ<sort>\"),sortname(\"<sort>\"),quotes(\"double\"),list}\n";
		prods += "\n";
		prods += "	<sort>METAID		-\>	<sort>MetaVar\n";
		prods += "	<sort>METAID \"*\"		-\> <sort>MetaVarList\n";
		prods += "	<sort>METAID \"+\"		-\> <sort>MetaVarList\n";
		prods += "	<sort>MetaVar		-\>	<sort>[[SQCtx]]	{cons(\"MetaVar\"), prefer, category(\"MetaVar\")}\n";
		prods += "	<sort>MetaVar		-\>	<sort>[[DQCtx]]	{cons(\"MetaVar\"), prefer, category(\"MetaVar\")}\n";
		prods += "	<sort>MetaVarList		-\>	<sort>[[SQCtx]]	{cons(\"MetaVar\"), prefer, category(\"MetaVar\"),list}\n";
		prods += "	<sort>MetaVarList		-\>	<sort>[[DQCtx]] {cons(\"MetaVar\"), prefer, category(\"MetaVar\"),list}\n";
		prods += "\n";
		prods += "	<sort>METAID		-\> Var {cons(\"<sort>\"), category(\"MetaVar\")}\n";
		prods += "	<sort>METAID		-\> VarToken {cons(\"<sort>\"), category(\"MetaVar\")}\n";
		prods += "	<sort>METAID REP	-\> ListVarToken	{cons(\"<sort>\"),prefer, category(\"MetaVar\")}\n";
		prods += "\n";
		
		lexprods += "	\"<sort>\"[s]*[0-9]*[\\\']*SUBID? -\> <sort>METAID\n";
		lexprods += "	<sort>METAID -\> ID[[SQCtx]] {reject}\n";
		lexprods += "	<sort>METAID -\> ID[[DQCtx]] {reject}\n";
		lexprods += "	<sort>METAID -\> VarID {reject}\n";
		lexrestr += "	<sort>METAID -/- [A-Za-z\\\'\\_0-9]\n";
	}
	prods += "	SQuoted[[Ctx]] -\> Quoted[[Ctx]]\n";
	prods += "	DQuoted[[Ctx]] -\> Quoted[[Ctx]]\n";
	
	str result = "module <moduleName>[Ctx SQCtx DQCtx]\n\nexports\n\ncontext-free syntax\n\n<prods>\nlexical syntax\n\n<lexprods>\nlexical restrictions\n\n<lexrestr>";

	writeFile(file, result);
}

str makelist(str sort, str param, str iter, map[str,str] listSeps) {
	if(sort in listSeps) 
		return "{<sort><param> \"<listSeps[sort]>\"}<iter>";
	else
		return "<sort><param><iter>";
}
