/**************************************************************************
 * Copyright (c) 2011 Anya Helene Bagge
 * Copyright (c) 2011 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::metaxa::MetaXa::MetaXa 
import ParseTree;

syntax Whitespace = lex [\ \t\n\f\r];

syntax Comment = lex "//" ![\n]* [\n];

layout Standard = WhitespaceOrComment*;

syntax WhitespaceOrComment = whitespace: Whitespace
                           | comment: Comment
                           ; 

syntax Program = lex WhitespaceOrComment* Module WhitespaceOrComment*;

syntax Module
	= Program: "module" Name Section*;


syntax Section
	= Construct: "construct" Name SortName ":" SubSection+ 
	| Operations: "operations" SortOpDecl* 
	| Sorts: "sorts" SortID+
	| Sort: "sort" Sort "\<:" Sort ":" SubSection+ 
	| Sort: "sort" SortName ":" SubSection+ ;


syntax QuotedTokens
	= QuotedTokens: "[+" SimpleToken* "+]" ;

syntax SimpleToken
	= BNFExpr
	| Keyword: Keyword 
	| Punct: Punct ;

syntax SubSection
	= Syntax: "syntax" SyntaxRules 
	| Doc: "doc" "{" DocString "}" ;

syntax SyntaxRules
	= SyntaxRules: SyntaxRule* 
	| SyntaxRules: SyntaxRule* "where" SyntaxRule* ;

syntax SyntaxRule
	= QuotedTokens 
//	| Quoted 
// | SugarDef 
	| BNFRule ;

syntax BNFRule
	= SyntaxProd: SortName "=" {BNFTopExpr "|"}* ";";

syntax BNFTopExpr
	= Cons: Cons ":" BNFExpr* 
	| AutoCons: "_" ":" BNFExpr* 
	| NoCons: BNFExpr* ;

syntax BNFExpr
	= Sort: SortName 
	| Token: Token 
	| "(" BNFExpr ")" 
	| Iter: BNFExpr REP 
	| IterSep: "{" BNFExpr Token "}" REP 
	| Opt: BNFExpr "?" ;

syntax Token
	= lex "\"" TOKENPART* "\"" ;

syntax TOKENPART =
	  lex [\\][\\] 
	| lex [\\][\"]
	| lex ![\\\"];
	
syntax SName
	= Name: LName 
	| Name: UName ;
	
syntax Name
	= SName
	| QName: Name "." SName;
	
syntax Sort 
	= Sort: SortName
	| Iter: SortName REP;

syntax Cons = Name;

syntax SortName 
	= UName 
	| QName: Name "." UName;


syntax UName = lex [A-Z][a-z A-Z 0-9 _]+
          # [a-z A-Z 0-9 _]
          ;

syntax LName = lex [a-z][a-z A-Z 0-9 _]+
          # [a-z A-Z 0-9 _]
          ;

syntax Keyword = lex [a-z][a-z A-Z 0-9 _]+
          # [a-z A-Z 0-9 _]
          ;

syntax SYMBOL = lex SYMBOLCHAR+;

syntax SYMBOLCHAR = 
	  lex [\\][\"]
	| lex [\\][\\]
	| lex ![\"\\ \t\n]
	# ![\"\\ \t\n]
	;
	
	
syntax REP = lex [*+];

syntax Punct = lex [\[\](){}\<\>|,;]+;


public Tree parseMetaXa(loc fil) {
  return parse(#Program, fil);
}
