/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::metaxa::MetaXaChecker


import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::metaxa::MetaXaCompiler;
import org::magnolialang::compiler::Util;
import org::magnolialang::compiler::MetaXaAST;
import org::magnolialang::programdb::ProgramDB;
import List;
import Set;
import IO;
import Node;
import String;
import ParseTree;

data Symbol = sort(AST);

public void checkSym(value db, value srcDb, AST sym) {
	set[AST] contents = {};
	def = getDef(srcDb, sym);
	attrmap = getSym(srcDb, sym);
	
	switch(attrmap[ATTR_KIND]) {
		case "module":
			putSym(db, sym, attrmap);
		case "construct":
			checkConstruct(db, sym, def, attrmap);
		case "sort":
			checkSort(db, sym, def, attrmap);
		case "language":
			checkLanguage(db, sym, def, attrmap);
		case "file":
			;
		default:
			throw "unexpected <attrmap[ATTR_KIND]>";
	}
}

void checkConstruct(value db, AST sym, AST def, map[str,value] attrmap) {
	if(seq(defs) := def) {
		attrmap = checkCSections(db, sym, getAs(attrmap, "sort", #AST), defs, attrmap);
		putSym(db, sym, attrmap);
	}
	else
		throw "unexpected <def>";
}

void checkSort(value db, AST sym, AST def, map[str,value] attrmap) {
	if(seq(defs) := def) {
		attrmap = checkCSections(db, sym, sym, defs, attrmap);
		putSym(db, sym, attrmap);
	}
	else
		throw "unexpected <def>";
}

void checkLanguage(value db, AST sym, AST def, map[str,value] attrmap) {
	if(seq(defs) := def) {
		attrmap = checkCSections(db, sym, sym, defs, attrmap);
		putSym(db, sym, attrmap);
	}
	else
		throw "unexpected <def>";
}

map[str,value] checkCSections(value db, AST sym, AST sortName, list[AST] defs, map[str, value] attrmap) {
	for(d <- defs) {
		switch(d) {
			case Syntax(syn):
				addSyntax(db, sym, extractSyntax(db, sym, sortName, syn));
			case Doc(s):
				attrmap[ATTR_DOC] = s;
			case Attributes(seq(as)):
				attrmap["attrs"] = as;
			case Alias(seq(als)):
				termPrintln("TODO: aliases: ", als);
			case Patterns(seq(pats)):
				attrmap["patterns"] = pats;
			case SyntaxStart(seq(syms)):
				attrmap["starts"] = {sort(makeSortName(s)) | s <- syms};
			default:
				throw "missed a case <d>";
		}
	;
	}
	return attrmap;
}



set[Production] extractSyntax(value db, AST name, AST sortName, AST rules) {
	rel[Symbol, Production] prods = {};
	list[AST] getRHS(AST ast) {
		switch(ast) {
			case NoCons(seq(x)):
				return x;
			case Cons(_, seq(x)):
				return x;
		}
	}
	nameStr = unparse(name);
	
	set[AST] mainRules = {};
	set[AST] auxRules = {};
	switch(rules) {
		case SyntaxRules(seq(rs)):
			mainRules = toSet(rs);
		case SyntaxRulesWhere(seq(rs), seq(ws)): {
			mainRules = toSet(rs);
			auxRules = toSet(ws);	
		}
	}

	map[AST,AST] renamings = ();	
	for(r <- auxRules) {
		switch(r) {
			case SyntaxProd(lhs, seq(rhss)): {
				renamings += (lhs : QName(name, lhs)[@concrete=[child(0),token("."),child(1)]]);
				println("<lhs> =\> <QName(name, lhs)>");
			}
		}
	}
	
	allRules = top-down-break visit(mainRules + auxRules) {
			case v:Var(_) => renamings[v] ? v
	};
	
	for(r <- allRules) {
		switch(r) {
			case SyntaxProd(lhs, seq(rhss)): {
				addSyntax(db, makeSortName(lhs), {prod([extractBNF(sym) | sym <- getRHS(rhs)], sort(lhs),
											makeConsAttrs(makeSortName(lhs), rhs, nameStr, sortName))
						  | rhs <- rhss});
			}
			case SyntaxSugar(NewSynRule(TokenList(seq(toks)), _, _)): {
				addSyntax(db, sortName, {prod([extractBNF(sym) | sym <- toks], sort(sortName),
											makeConsAttrs(sortName, NoCons(seq(toks)), nameStr, sortName))});
			}
			case SyntaxTokens(TokenList(seq(toks))): {
				addSyntax(db, sortName, {prod([extractBNF(sym) | sym <- toks], sort(sortName),
											makeConsAttrs(sortName, NoCons(seq(toks)), nameStr, sortName))});
			}
			case SyntaxExample(_): {
				;
			}
			default: {
				putMarker(db, r, SEVERITY_FATAL, "extractSyntax() missed a case: \'<yieldTerm(r)>\'");
			}
		}
	}
			
	return {};
}

void addSyntax(value db, AST sym, set[Production] prods) {
	// check syntax
	top-down-break visit(prods) {
		case sort(QName(Name(leaf("sort")), _)):
			;
		case sort(n):
			termPrintln("Ill-formed sort name: ", n);
		case n:QName(_,_):
			termPrintln("Misplaced name: ", n);
		case n:Name(_):
			termPrintln("Misplaced name: ", n);
	}
	
	if(peekSym(db, sym) == ())
		putAttr(db, sym, ATTR_KIND, "sort");
	set[Production] oldProds = {};
	oldProds = get(peekSym(db, sym), "syntax", oldProds);
	putAttr(db, sym, "syntax", oldProds + prods);
}

Symbol extractBNF(AST rhs) {
	switch(rhs) {
		case Sort(s):
			return \sort(makeSortName(s));
		case SortToken(s):
			return extractBNF(s);
		case Symbol(t):
			return \lit(unquote(unparse(t)));
		case String(t): // merge String and Symbol?
			return \lit(unquote(unparse(t)));
		case Keyword(s):		// TokenList
			return \lit(unquote(unparse(s)));
		case \seq(s1, s2):
			return \seq([extractBNF(s1), extractBNF(s2)]);
		case \alt(s1, s2):
			return \seq([extractBNF(s1), extractBNF(s2)]);
		case IterSep(e, t, leaf("+")):
			return \iter-seps(extractBNF(e), [extractBNF(t)]);
		case IterSep(e, t, leaf("*")):
			return \iter-star-seps(extractBNF(e), [extractBNF(t)]);
		case Iter(e, leaf("+")):
			return \iter(extractBNF(e));
		case Iter(e, leaf("*")):
			return \iter-star(extractBNF(e));
		case Opt(e):
			return \opt(extractBNF(e));
		default: {
			if(((rhs@category) ? "") == "MetaVar")
				return \sort(makeSortName(Name(leaf(consOf(rhs)))));
			throw "missed a case <rhs>";
		}
	}
}

map[AST, int] consNameGen = ();

Attributes makeConsAttrs(AST sortName, AST rhs, str constructName, AST constructSortName) {
	if(Cons(consName, _) := rhs)
		return attrs([term("cons"(undot(unparse(consName))))]);
	else if(NoCons(seq([Sort(_)])) := rhs)
		return \no-attrs();
	else if(sortName == constructSortName)
		return attrs([term("cons"(undot(constructName)))]);
	else {
		n = consNameGen[sortName] ? 0;
		consNameGen[sortName] = n + 1;
		return attrs([term("cons"(undot("<sortName><n>")))]);
	}
}

str undot(str name) {
	return replaceAll(name, "\\.", "_");
}
