/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::metaxa::MetaXaCompiler

import org::magnolialang::metaxa::MetaXaLoader;
import org::magnolialang::metaxa::MetaXaChecker;
import org::magnolialang::metaxa::MetaXaGenerator;
import org::magnolialang::metaxa::BuildMatch;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::metaxa::Env;
import org::magnolialang::compiler::Util;
import org::magnolialang::compiler::MetaXaAST;
import org::magnolialang::programdb::ProgramDB;
import lang::rascal::syntax::Grammar;
import lang::rascal::syntax::Grammar2Rascal;
import List;
import Set;
import ParseTree;
import IO;
import Node;
import String;


public void loader(value db, value srcDb, AST sym) {
	loadModule(db, srcDb, sym);
}

public void checker(value db, value srcDb, AST sym) {
	// trigger processing in previous stage
	def = getSym(srcDb, sym);

	for(s <- symbols(srcDb)) {
		attrmap = getSym(srcDb, s);
		if(attrmap[ATTR_KIND] == KIND_FILE)
			putSym(db, s, attrmap);
	}
	
	for(s <- symbols(srcDb))
		checkSym(db, srcDb, s);

}

public void generator(value db, value srcDb, AST sym) {
	// trigger processing in previous stage
	def = getSym(srcDb, sym);
	for(s <- symbols(srcDb)) {
		putSym(db, s, getSym(srcDb, s));
	}

	generateGrammar(db, sym);		
}

public AST makeSortName(AST name) {
	return QName(Name(leaf("sort")), name);
}

public AST makeConstructName(AST name) {
	return QName(Name(leaf("construct")), name);
}

public AST makeLanguageName(AST name) {
	return QName(Name(leaf("language")), name);
}
public AST makePlainName(AST name) {
	if(QName(Name(leaf(k)), n) := name && k in ["sort", "construct"])
		return n;
	else
		return name;
}

public void putMarker(value db, loc location, int severity, str message) {
	map[str,value] attrmap = getSym(db, leaf(location.uri)) ? ();
	
	
	putSym(db, leaf(location.uri), addMarker(attrmap, location, severity, message));

/*	else if(hasAttr(db, sym, ATTR_PARENT)) {
		putMarker(db, getAttr(db, sym, ATTR_PARENT), location, severity, message);
	}
	else if(hasAttr(db, sym, ATTR_FILE)) {
		putMarker(db, getAttr(db, sym, ATTR_FILE), location, severity, message);
	}*/
}

public loc findContainingUri(value db, value sym) {
	println("Find containter: <sym>");
	attrmap = getSym(db, sym);
	if(hasAttr(db, sym, ATTR_URI))
		return getAttrAs(db, sym, ATTR_URI, #loc);
	else if(hasAttr(db, sym, ATTR_PARENT))
		return findContainingUri(db, getAttr(db, sym, ATTR_PARENT));
	else if(hasAttr(db, sym, ATTR_FILE))
		return findContainingUri(db, getAttr(db, sym, ATTR_FILE));
	else
		throw "symbol has no uri or parent: <sym>";
}

public void putMarker(value db, AST subject, int severity, str message) {
	putMarker(db, subject@\loc, severity, message);
}
