/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::metaxa::MetaXaGenerator

import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::metaxa::Env;
import org::magnolialang::compiler::Util;
import org::magnolialang::compiler::MetaXaAST;
import org::magnolialang::programdb::ProgramDB;
import org::magnolialang::metaxa::MetaXaCompiler;
import lang::rascal::syntax::Grammar;
import lang::rascal::syntax::Grammar2Rascal;
import List;
import Set;
import ParseTree;
import IO;
import Node;
import String;
data Symbol = sort(AST);

public void generateGrammar(value db, AST sym) {
	sorts = {sort(makeSortName(s)) | s <- symbols(db), 
					hasAttr(db, s, ATTR_PARENT, sym), 
					hasAttr(db, s, ATTR_KIND, "sort")};

	println("Generate sorts: <sorts>");
	if(sorts != {}) {
		g = extractGrammar(db, sorts);
		putAttr(db, sym, "grammar", g);
		putAttr(db, sym, "text", grammar2rascal(g, unparse(sym)));
	} 	
}

public Grammar extractGrammar(value db, set[Symbol] starts) {
	set[Symbol] sorts = {};
	set[Production] prods = {};
	set[Production] emptyProds = {};
	for(sym <- symbols(db), hasAttr(db, sym, ATTR_KIND, "sort")) {
		if(hasAttr(db, sym, "super")) {
			prods += prod([sort(sym)], sort(getAttrAs(db, sym, "super", #AST)), attrs([term("inject"())]));
		}

		sorts += {sort(sym)};
		prods += getAttr(db, sym, "syntax", emptyProds);
	}

	<sorts, starts, prods> = visit(<sorts, starts, prods>) {
		case sort(n) => sort(unparse(n)) 
	}
	return grammar(starts, prods);
}
