/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::metaxa::MetaXaLoader


import org::magnolialang::metaxa::MetaXaCompiler;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::metaxa::Env;
import org::magnolialang::compiler::Util;
import org::magnolialang::compiler::MetaXaAST;
import org::magnolialang::programdb::ProgramDB;
import Set;
//import IO;
//import String;

public void loadModule(value db, value srcDb, AST modName) {
	set[AST] contents = {};
	AST tree = getDef(srcDb, modName);
	
	if(Program(modNameDef, seq(sections)) := tree) {
		for(sec <- sections) {
			map[str,value] attrs = ();
			AST symbol;
			switch(sec) {
				case Import(seq(mods)): {
					for(imp <- mods) {
						if(peekSym(db, imp) != ()) {
							termPrintln("import", imp, " (loaded)");
							if(getAttr(db, imp, ATTR_KIND) == KIND_PACKAGE)
								;
							else
								termPrintln("ERROR:", imp, "not a module");
						}
						else if(peekSym(srcDb, imp) != ()) {
							termPrintln("import", imp, " (parsed)");
							if(getAttr(db, imp, ATTR_KIND) == KIND_PACKAGE)
								loadModule(db, srcDb, imp);
							else
								termPrintln("ERROR:", imp, "not a module");
						}
						else {
							termPrintln("ERROR:", imp, "not found");
						}
					}
				}
				case SortDef(name, defs): {
					symbol = makeSortName(name);
					attrs[ATTR_DEF] = defs;
					attrs[ATTR_KIND] = "sort";
				}
				case SubSortDef(name, super, defs):  {
					symbol = makeSortName(name);
					attrs[ATTR_DEF] = defs;
					attrs["super"] = makeSortName(super);
					attrs[ATTR_KIND] = "sort";
				}
				case ConstructDef(name, sortName, defs): {
					symbol = makeConstructName(name);
					attrs[ATTR_DEF] = defs;
					attrs["sort"] = makeSortName(sortName);
					attrs[ATTR_KIND] = "construct";
				}
				case LanguageDef(name, defs): {
					symbol = makeLanguageName(name);
					attrs[ATTR_DEF] = defs;
					attrs[ATTR_KIND] = "language";
				}
			}
			if(attrs != ()) {
				attrs[ATTR_LOCS] = [sec@\loc] ? [];
				attrs[ATTR_PARENT] = modName;
				putSym(db, symbol, attrs);
				contents += {symbol};
			}
		}
	
		putSym(db, modName, (ATTR_DEF : seq(toList(contents)),
						  ATTR_KIND : KIND_PACKAGE,
						  ATTR_FILE : getAttr(srcDb, modName, ATTR_FILE)));
		putSym(db, getAttr(srcDb, modName, ATTR_FILE), getSym(srcDb, getAttr(srcDb, modName, ATTR_FILE)));
	}
	else
		throw "unexpected <tree>";
}
