/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::metaxa::MetaXaToRascalGrammar


import org::magnolialang::metaxa::BuildMatch;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::metaxa::Env;
import org::magnolialang::compiler::Util;
import org::magnolialang::compiler::MetaXaAST;
import lang::rascal::syntax::Grammar;
import lang::rascal::syntax::Grammar2Rascal;
import List;
import Set;
import ParseTree;
import IO;
import Node;
import String;

data ConstructDef = constructDef(
	set[Production] syntax,
	str doc,
	list[AST] aliases,
	list[AST] attrs,
	map[str, AST] etc
);
public ConstructDef constructDef() {
	return constructDef({}, "", [], [], ());
}

public Grammar extractGrammar(AST tree, loc outFile) {
	set[Symbol] sorts = {};
	set[Production] prods = {};
	
	if(Program(modName, seq(sections)) := tree) {
		for(sec <- sections) switch(sec) {
			case SortDef(name, defs): {
				println("sort <unparse(name)>");
				sorts += sort(unparse(name));
			}
			case SubSortDef(name, super, defs):  {
				println("sort <unparse(name)> \<: <unparse(super)>");
				sorts += sort(unparse(name));
				prods += prod([sort(unparse(name))], sort(unparse(super)), attrs([term("inject"())]));
			}
			case ConstructDef(name, sortName, defs): {
				println("construct <unparse(name)>");
				def = extractDef(unparse(name), unparse(sortName), defs);
				prods += def.syntax;
			}
		}
	
		rscName = replaceAll(unparse(modName), "\\.", "::");
		Grammar g =  grammar({sort("START")}, prods);
		writeFile(outFile, grammar2rascal(g, rscName));
		return g; 
	}
	else
		throw "unexpected <tree>";
	
}

ConstructDef extractDef(str name, str sortName, AST tree) {
	ConstructDef def = constructDef();
	
	if(seq(defs) := tree) {
		for(d <- defs) {
			switch(d) {
				case Syntax(syn):
					def.syntax = extractSyntax(name, sortName, syn);
				case Doc(s):
					def.doc = s;
				case Attributes(seq(attrs)):
					def.attrs = attrs;
				case Aliases(seq(als)):
					def.aliases = als;
				case Patterns(seq(pats)):
					def.etc["patterns"] = seq(pats);
				default:
					throw "missed a case <d>";
					
			}
		;
		}
		return def;
	}
	throw "unexpected <tree>";
}

set[Production] extractSyntax(str name, str sortName, AST rules) {
	rel[Symbol, Production] prods = {};
	list[AST] getRHS(AST ast) {
		switch(ast) {
			case NoCons(seq(x)):
				return x;
			case Cons(_, seq(x)):
				return x;
		}
	}
	
	set[AST] mainRules = {};
	set[AST] auxRules = {};
	switch(rules) {
		case SyntaxRules(seq(rs)):
			mainRules = toSet(rs);
		case SyntaxRulesWhere(seq(rs), seq(ws)): {
			mainRules = toSet(rs);
			auxRules = toSet(ws);	
		}
	}

	map[AST,AST] renamings = ();	
	for(r <- auxRules) {
		switch(r) {
			case SyntaxProd(lhs, seq(rhss)): {
				renamings += (lhs : QName(Name(leaf(name)), lhs)[@concrete=[child(0),token("."),child(1)]]);
				println("<lhs> =\> <QName(Name(leaf(name)), lhs)>");
			}
		}
	}
	
	for(r <- mainRules + auxRules) {
		r = top-down-break visit(r) {
			case v:Name(_) => renamings[v] ? v
		};
		switch(r) {
			case SyntaxProd(lhs, seq(rhss)): {
				prods += {<sort(replaceAll(unparse(lhs), "\\.", "_")),
							prod([extractBNF(sym) | sym <- getRHS(rhs)], sort(unparse(lhs)),
											makeConsAttrs(unparse(lhs), rhs, name, sortName))>
						  | rhs <- rhss};
			}
			case SyntaxSugar(NewSynRule(TokenList(seq(toks)), _, _)): {
				prods += {<sort(sortName), prod([extractBNF(sym) | sym <- toks], sort(sortName),
											makeConsAttrs(sortName, NoCons(seq(toks)), name, sortName))>};
			}
			case SyntaxTokens(TokenList(seq(toks))): {
				prods += {<sort(sortName), prod([extractBNF(sym) | sym <- toks], sort(sortName),
											makeConsAttrs(sortName, NoCons(seq(toks)), name, sortName))>};
			}
			case SyntaxExample(_): {
				;
			}
			default: {
				throw "missed a case <r>";
			}
		}
	}
			
	return {choice(sym, prods[sym]) | sym <- prods<0>};
}

Symbol extractBNF(AST rhs) {
	switch(rhs) {
		case Sort(s):
			return \sort(replaceAll(unparse(s), "\\.", "_"));
		case Symbol(t):
			return \lit(unquote(unparse(t)));
		case Keyword(s):		// TokenList
			return \lit(unquote(unparse(s)));
		case \seq(s1, s2):
			return \seq([extractBNF(s1), extractBNF(s2)]);
		case \alt(s1, s2):
			return \seq([extractBNF(s1), extractBNF(s2)]);
		case IterSep(e, t, leaf("+")):
			return \iter-seps(extractBNF(e), [extractBNF(t)]);
		case IterSep(e, t, leaf("*")):
			return \iter-star-seps(extractBNF(e), [extractBNF(t)]);
		case Iter(e, leaf("+")):
			return \iter(extractBNF(e));
		case Iter(e, leaf("*")):
			return \iter-star(extractBNF(e));
		case Opt(e):
			return \opt(extractBNF(e));
		default: {
			if(((rhs@category) ? "") == "MetaVar")
				return \sort(consOf(rhs));
			rawPrintln(getAnnotations(rhs));
			throw "missed a case <rhs>";
		}
	}
}

map[str, int] consNameGen = ();

Attributes makeConsAttrs(str sortName, AST rhs, str constructName, str constructSortName) {
	if(Cons(consName, _) := rhs)
		return attrs([term("cons"(undot(unparse(consName))))]);
	else if(NoCons(seq([Sort(_)])) := rhs)
		return \no-attrs();
	else if(sortName == constructSortName)
		return attrs([term("cons"(undot(constructName)))]);
	else {
		n = consNameGen[sortName] ? 0;
		consNameGen[sortName] = n + 1;
		return attrs([term("cons"(undot("<sortName><n>")))]);
	}
}

str undot(str name) {
	return replaceAll(name, "\\.", "_");
}
