/**************************************************************************
 * Copyright (c) 2009-2012 CWI, UiB
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 Tero Hasu
 * Copyright (c) 2012-2013 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * + Arnold Lankamp
 * + Jurgen J. Vinju
 * + Paul Klint
 * * Tero Hasu
 * + Tijs van der Storm
 *
 *************************************************************************/

package org.magnolialang.testutil;

import java.io.File;
import java.io.PrintWriter;
import java.net.URI;
import java.util.Collection;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IValueFactory;
import org.rascalmpl.value.type.Type;
import org.magnolialang.infra.ConsoleInfra;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.interpreter.Evaluator;
import org.rascalmpl.interpreter.env.GlobalEnvironment;
import org.rascalmpl.interpreter.env.ModuleEnvironment;
import org.rascalmpl.interpreter.result.Result;
import org.rascalmpl.values.ValueFactoryFactory;
import org.rascalmpl.values.uptr.RascalValueFactory;
import org.rascalmpl.values.uptr.TreeAdapter;

/*
 * Calling toString() on a VoidResult gives a NullPointerException,
 * which can cause trouble here and there.
 * This is a more polite implementation.
 *
 * public class VoidResult extends Result<VoidResult.Void> {
 * abstract class Void implements IValue {}
 *
 * public VoidResult(Type type, IEvaluatorContext ctx) {
 * super(type, null, ctx);
 * }
 *
 * public String toString() {
 * return getType().toString();
 * }
 *
 * public String toString(int length) {
 * return toString();
 * }
 * }
 */

public class RascalEval {
	private final static int LINE_LIMIT = 200;
	private static final String EVAL_MODULE = "$eval$";

	private final Evaluator evaluator;


	public RascalEval() {
		GlobalEnvironment heap = new GlobalEnvironment();
		ModuleEnvironment root = heap.addModule(new ModuleEnvironment(EVAL_MODULE, heap));
		PrintWriter stderr = new PrintWriter(System.err);
		PrintWriter stdout = new PrintWriter(System.out);
		IValueFactory vf = ValueFactoryFactory.getValueFactory();
		evaluator = new Evaluator(vf, stderr, stdout, root, heap);

		evaluator.addRascalSearchPath(TermFactory.vf.sourceLocation(new File(".", "src").toURI()));
		for(String s : ConsoleInfra.getRascalSearchPath()) {
			evaluator.addRascalSearchPath(TermFactory.vf.sourceLocation(new File(s).toURI()));
		}
	}


	public RascalEval(Evaluator evaluator) {
		this.evaluator = evaluator;
	}


	public IValue call(String name, Collection<IValue> args) {
		//System.err.println("call " + name + " with " + args);
		IValue[] arr = new IValue[] {};
		if(args != null) {
			arr = args.toArray(arr);
		}
		return evaluator.call(name, arr);
	}


	public void doImport(String modName) {
		eval("import " + modName + ";");
	}


	public Result<IValue> eval(String statement) {
		return evaluator.eval(null, statement, TermFactory.vf.sourceLocation(URI.create("prompt:///")));
	}


	/**
	 * @param statement
	 * @return result of evaluation
	 * @throws NullPointerException
	 *             if there's no actual result produced
	 */
	public Result<IValue> evalMore(String statement) {
		return evaluator.evalMore(null, statement, TermFactory.vf.sourceLocation(URI.create("prompt:///")));
	}


	/*
	public IValue call(String name, IValue[] args) {
		// Note about Java varargs: IValue... is sugar for IValue[], so this is okay.
		return evaluator.call(name, args);
	}*/

	public Evaluator getEvaluator() {
		return evaluator;
	}


	public IValue[] toValueArray(Collection<IValue> args) {
		IValue[] arr = new IValue[] {};
		if(args != null) {
			arr = args.toArray(arr);
		}
		return arr;
	}


	public static String resultToString(Result<IValue> value) {
		IValue v = value.getValue();
		Type type = value.getType();

		if(v == null || type == null) {
			return "ok";
		}

		// Calling Result<IValue>.toString should be safe if both getValue
		// and getType yield non-null.
		String s = value.toString(LINE_LIMIT);

		if(type.isAbstractData() && type.isSubtypeOf(RascalValueFactory.Tree)) {
			return "`" + TreeAdapter.yield((IConstructor) v) + "`\n" + s;
		}

		return s;
	}

}
