/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 Tero Hasu
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
package org.magnolialang.testutil;

import java.util.Collection;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IListWriter;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.ITuple;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IValueFactory;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.interpreter.utils.RuntimeExceptionFactory;

public class RascalExt {
	private final IValueFactory vf;


	public RascalExt(IValueFactory vf) {
		this.vf = vf;
	}


	public IString javaToString(IValue v) {
		return vf.string(v.toString());
	}


	/**
	 * It is possible to invoke Clojure APIs from Rascal, via Java. Like
	 * this. But this would introduce a dependency on pdb_values_pprint,
	 * which would have to be compiled, using "lein compile". Which is
	 * bad for "pure Eclipse" developers, even when the Rascal APIs are
	 * not used, since Java linking must still succeed.
	 * 
	 * public IString prettyPdbValue(IValue v) {
	 * return vf.string(pdb_values_pprint.prettyPdbValue(v));
	 * }
	 * 
	 * public IString prettyPdbValueAnnos(IValue v) {
	 * return vf.string(pdb_values_pprint.prettyPdbValueAnnos(v));
	 * }
	 */
	public IString prettyPdbValue(IValue v) {
		return TermFactory.vf.string(TermAdapter.yieldTerm(v, false));
	}


	public IString prettyPdbValueAnnos(IValue v) {
		return TermFactory.vf.string(TermAdapter.yieldTerm(v, true));
	}


	/**
	 * @param val
	 *            a tuple (do not know how to declare this generically on Rascal
	 *            side as any tuple, so using a value)
	 * @return a list of the values of the tuple fields, in order
	 */
	public IValue tupleToList(IValue val) {
		if(!val.getType().isTuple())
			throw RuntimeExceptionFactory.illegalArgument(val, null, null, "argument of type tuple is required");
		ITuple tp = (ITuple) val;

		IListWriter w = vf.listWriter();

		for(IValue v : tp) {
			w.append(v);
		}

		return w.done();
	}


	public static IValue ctorGetField(IValue o, String name) {
		return ((IConstructor) o).get(name);
	}


	public static IValue ctorGetVal(IValue o) {
		return ctorGetField(o, "val");
	}


	public static boolean isIValue(Object o) {
		return o instanceof IValue;
	}


	public static IValue[] toValueArray(Collection<IValue> args) {
		IValue[] arr = new IValue[] {};
		if(args != null) {
			arr = args.toArray(arr);
		}
		return arr;
	}

}
