/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
module org::magnolialang::testutil::RascalExt

/** This module does have dependencies on Magnolia compiler libraries,
    but may not have dependencies on a compiler instance. */

// import org::magnolialang::magnolia::MagnoliaFacts;
// import org::magnolialang::magnolia::lib::Names;

import IO;
import List;
import Map;
import Node;

@doc{Call Java toString to turn a Rascal value into a string.}
@javaClass{org.magnolialang.testutil.RascalExt}
public java str javaToString(value v);

@doc{Converts tuple field values to a list.}
@javaClass{org.magnolialang.testutil.RascalExt}
public java list[value] tupleToList(value tp);

@doc{Turns a Rascal value into a pretty string.}
@javaClass{org.magnolialang.testutil.RascalExt}
public java str prettyPdbValue(value v);

@doc{Turns a Rascal value (with annotations) into a pretty string.}
@javaClass{org.magnolialang.testutil.RascalExt}
public java str prettyPdbValueAnnos(value v);

// --------------------------------------------------
// annotations
// --------------------------------------------------

/** Deletes all but the named annotations. */
public &T keepAnnotationsRec(&T v, set[str] keep) {
  return visit(v) { 
  case m : node n:
    {
      map[str, value] amap = getAnnotations(m);
      map[str, value] namap = ();
      for (x <- domain(amap)) {
        if (x in keep)
          namap[x] = keepAnnotationsRec(amap[x], keep);
      }
      // 'setAnnotations' puts, does not delete
      m = delAnnotations(m);
      insert(setAnnotations(m, namap));
    }
  };
}

/** Deletes the named annotations. */
public &T delAnnotationsRec(&T v, set[str] dontKeep) {
  return visit(v) { 
  case m : node n:
    {
      map[str, value] amap = getAnnotations(m);
      map[str, value] namap = ();
      for (x <- domain(amap)) {
        if (x notin dontKeep)
          namap[x] = delAnnotationsRec(amap[x], dontKeep);
      }
      // 'setAnnotations' puts, does not delete
      m = delAnnotations(m);
      insert(setAnnotations(m, namap));
    }
  };
}
