/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 Tero Hasu
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
// run with rascal-src-plain-in-magnolia

/*
  The formatting is designed to accommodate large annotations,
  but one better have a large size console.

  (rimport "org::magnolialang::testutil::RascalPp")
  (rcall "runTestRascalPp")
  (rprint (get-fact-imploded-tree "basic.Basic") true)
  (rprint (get-fact-type-checked "override.Override") true)
  (rprint (get-fact-type-checked "inlining.InlineProc4")) ;; with XRefs
  (rprint (ctor-get-val (get-fact-type-checked "inlining.InlineProc4")) true)

  '@loc' is a common annotation, for 'loc' type information see
  http://tutor.rascal-mpl.org/Courses/Rascal/Expressions/Values/Location/Location.html
*/

module org::magnolialang::testutil::RascalPp

import org::magnolialang::testutil::RascalExt;

import org::nuthatchery::pgf::rascal::engines::FmtEngine;
import org::nuthatchery::pgf::rascal::FmtUtil;
import org::nuthatchery::pgf::rascal::Groupings;
import org::nuthatchery::pgf::rascal::Token;
import org::nuthatchery::pgf::rascal::Tseq;

import util::Benchmark;
import IO;
import List;
import Map;
import Node;
import Set;
import Type;

data PpStyle = 
  StyleA(bool withAnnos) |
  StyleC(bool withAnnos);

Token br_able = union(Text(""), br);

/* bad performance
Tseq ppBrackets(Tseq ts, str ob, str cb, 
                list[value] vs, PpStyle withStyle,
                Tseq (Tseq, value, PpStyle) ppv) {
  bool first = true;
  ts = tseq(ts, ob, align, group_);
  for (v <- vs) {
    if (first)
      first = false;
    else
      ts = tseq(ts, ",", br);
    ts = ppv(ts, v, withStyle);
  }
  return tseq(ts, _group, dedent, cb);
}
*/

// this also performs badly despite no groupings use;
// badly but better than with groupings
Tseq ppBrackets(Tseq ts, str ob, str cb, 
                list[value] vs, withStyle : StyleA(_),
                Tseq (Tseq, value, PpStyle) ppv) {
  Tseq pp(value v) {
    return ppv(tseqNull, v, withStyle);
  }
  return tseq(ts, ob, align, 
              group(sepBy(tseq(",", br),
                          mapper(vs, pp))),
              dedent, cb);
}

Tseq ppAdt(Tseq ts, str ob, str cb, 
           list[value] vs, withStyle : StyleA(_),
           Tseq (Tseq, value, PpStyle) ppv) {
  Tseq pp(value v) {
    return ppv(tseqNull, v, withStyle);
  }
  return tseq(ts, ob, indent(2), br,
              sepBy(tseq(",", br),
                    mapper(vs, pp)),
              dedent, br, cb);
}

/*
Tseq ppAdt(Tseq ts, str ob, str cb, 
           list[value] vs, PpStyle withStyle,
           Tseq (Tseq, value, PpStyle) ppv) {
  bool first = true;
  ts = tseq(ts, ob, indent(2), br);
  for (v <- vs) {
    if (first)
      first = false;
    else
      ts = tseq(ts, ",", br);
    ts = ppv(ts, v, withStyle);
  }
  return tseq(ts, dedent, br, cb);
}
*/

Tseq ppBrackets(Tseq ts, str ob, str cb, 
                list[value] vs, withStyle : StyleC(_),
                Tseq (Tseq, value, PpStyle) ppv) {
  Tseq pp(value v) {
    return ppv(tseqNull, v, withStyle);
  }
  return tseq(ts, ob, align, 
              group(sepBy(tseq(",", br),
                          mapper(vs, pp))),
              dedent, cb);
}

Tseq ppAdt(Tseq ts, str ob, str cb, 
           list[value] vs, withStyle : StyleC(_),
           Tseq (Tseq, value, PpStyle) ppv) {
  return ppBrackets(ts, ob, cb, vs, withStyle, ppv);
}

Tseq ppMapPair(Tseq ts, value pv, PpStyle withStyle) {
  if (<k, v> := pv) {
    Tseq pp(value x) {
      return ppValue(tseqNull, x, withStyle);
    }
    return tseq(ts, pp(k), ": ", pp(v));
  }
  return ppValue(ts, pv, withStyle);
}

Tseq ppAnnoPair(Tseq ts, value pv, PpStyle withStyle) {
  if (<k, v> := pv) {
    return tseq(ts, "@", k, " = ",
                ppValue(tseqNull, v, withStyle));
  }
  return ppValue(ts, pv, withStyle);
}

Tseq ppValue(Tseq ts, value v, PpStyle withStyle) {
  switch (v) {
  case node n: {
    ts = ppAdt(ts, getName(n) + "(", ")", getChildren(n), 
               withStyle, ppValue);
    map[str, value] as = getAnnotations(n);
    bool withAnnos = (StyleA(true) := withStyle) ||
      (StyleC(true) := withStyle);
    if (withAnnos && !isEmpty(as)) {
      ts = ppBrackets(ts, "[", "]", toList(as), withStyle, ppAnnoPair);
    }
    return ts;
  }
  case set[value] vs: {
    return ppBrackets(ts, "{", "}", toList(vs), withStyle, ppValue);
  }
  case list[value] vs: {
    return ppBrackets(ts, "[", "]", vs, withStyle, ppValue);
  }
  case map[value, value] vs: {
    return ppBrackets(ts, "(", ")", toList(vs), withStyle, ppMapPair);
  }}
  t = typeOf(v);
  switch (t) {
  case \tuple(_): {
    return ppBrackets(ts, "\<", "\>", tupleToList(v), withStyle, ppValue);
  }
  }
  return tseq(ts, javaToString(v));
}

public FmtSt ppRascalValue(FmtSt st, value v, PpStyle withStyle) {
  Tseq s = ppValue(tseqNull, v, withStyle);
  FmtSt st = write(st, s);
  return st;
}

public FmtSt ppRascalValue(value v, PpStyle withStyle) {
  return ppRascalValue(newFmtSt(200), v, withStyle);
}

public FmtSt ppRascalValue(value v) {
  return ppRascalValue(v, StyleC(false));
}

public str prettyRascalValue(value v, PpStyle withStyle) {
  return pretty(ppRascalValue(v, withStyle));
}

//import Benchmark;

int bm_t0 = 0;
int bm_t1 = 0;
int bm_t2 = 0;

public str prettyRascalValue(value v) {
  bm_t0 = realTime();
  st = ppRascalValue(v, StyleC(false));
  bm_t1 = realTime();
  s = pretty(st);
  bm_t2 = realTime();
  return s;
}

public void printPrettyRascalValue(value v) {
  bm_t0 = realTime();
  st = ppRascalValue(v, StyleC(false));
  bm_t1 = realTime();
  prettyPrintFlushLn(st);
  bm_t2 = realTime();
}

/*
  (use 'org.magnolialang.testutil.bankers-deque)
  (init-compiler)
  (rimport "org::magnolialang::testutil::RascalPp")
  (rcall "runTestRascalPp")
  (println (.getValue (rcall "prettyRascalValue" (get-fact-type-checked "basic.Basic"))))
  (rcall "lastBench")
  (rcall "printPrettyRascalValue" (get-fact-type-checked "basic.Basic"))
  (rcall "lastBench")
  (let [ast (get-fact-type-checked "basic.Basic")]
    (for [i (range 3 20)]
      (binding [*dq-factor* i]
        (rcall "prettyRascalValue" ast)
        [i (rcall "lastBench")])))
*/
public str lastBench() {
  return "<bm_t0>, <bm_t1>, <bm_t2>; build pp tree <bm_t1 - bm_t0> ms; format <bm_t2 - bm_t1> ms";
}

data A = A() | B(int x, int y);

public void runTestRascalPp() {
  list[value] vs = 
    [
     true, 1, "baz", "foo\nbar", A(), B(7, 8),
     {1, 2}, [1, 3], (1: 2, 3: 4), <1, 2, 3>
     ];
  for (v <- vs) {
    //FmtSt st = ppRascalValue(v, StyleA(false));
    FmtSt st = ppRascalValue(v, StyleC(false));
    prettyPrintFlushLn(st);
  }
}

public void main(list[str] args) {
  runTestRascalPp();
}
