(ns org.magnolialang.testutil.high-level
  (:use
   org.magnolialang.testutil.core
   org.magnolialang.testutil.pdb-values
   org.magnolialang.testutil.pdb-values-pprint
   clojure.pprint
   ))

;;; 
;;; inlining
;;; 

;; E.g.
;; (rimport "org::magnolialang::magnolia::lib::Inliner")
;; (rprint-f-no-annos (find-inlineable-obj-by-point (get-fact-type-checked "inlining.InlineVar1") 18 6))
(defn find-inlineable-obj-by-point
  ([tree line column]
     (rcall "findBestInlineable" tree (rinteger line) (rinteger column)))
  ([tree offset]
     (rcall "findBestInlineable" tree (rinteger offset))))

(defn- inline-in-module
  [mod-name & {:keys [line-column offset
                      pre-pt?
                      pre-mg? post-mg?
                      pre-ast? post-ast?
                      annos? keep del]
               :or {line-column nil
                    offset nil
                    pre-pt? false
                    pre-mg? true
                    post-mg? true
                    pre-ast? false
                    post-ast? false
                    annos? false
                    keep nil
                    del nil
                    }}]
  (when pre-pt?
    (rcall "println" (get-fact-contents-string mod-name)))
  (let [tree (get-fact-type-checked mod-name)
        pr-a (fn [x]
               (let [x (cond
                        keep (apply keep-annos x keep)
                        del (apply del-annos x del)
                        :else x)]
                 (rprint-p x :annos? annos?)))]
    (when pre-mg?
      (print-ast-as-mg tree))
    (when pre-ast?
      (pr-a tree))
    (let [new-tree
          (cond
           line-column (let [[line column] line-column]
                         (rcall "inlineByPoint" tree
                                (rinteger line) (rinteger column)))
           offset (rcall "inlineByPoint" tree (rinteger offset))
           :else (assert false "no point specified"))]
      (when post-mg?
        (print-ast-as-mg new-tree))
      (when post-ast?
        (pr-a new-tree))
      nil)))

(defn inline-in-module-by-point [mod-name line column & kw]
  (apply inline-in-module mod-name :line-column [line column] kw))

(defn inline-in-module-by-offset [mod-name offset & kw]
  (apply inline-in-module mod-name :offset offset kw))

(defn run-inliner-tests []
  (inline-in-module-by-point "inlining.InlineProc1" 9 5)
  (inline-in-module-by-point "inlining.InlineProc2" 8 15)
  (inline-in-module-by-point "inlining.InlineProc3" 8 15)
  (inline-in-module-by-point "inlining.InlineProc4" 8 8) ;; whole proc
  (inline-in-module-by-point "inlining.InlineProc4" 17 10) ;; one call
  (inline-in-module-by-point "inlining.InlineProc4" 18 10) ;; another call
  (inline-in-module-by-point "inlining.InlineProc5" 10 12)
  (inline-in-module-by-point "inlining.InlineProc5" 14 12)
  (inline-in-module-by-point "inlining.InlineProc6" 10 12)
  (inline-in-module-by-point "inlining.InlineProc7" 10 12)
  (inline-in-module-by-point "inlining.InlineProc8" 10 12) ;; 'f'
  (inline-in-module-by-point "inlining.InlineProc8" 14 12) ;; 'g'
  (inline-in-module-by-point "inlining.InlineProc9" 12 7) ; inline 'f'
  (inline-in-module-by-point "inlining.InlineProc9" 18 7) ; inline 'g'
  (inline-in-module-by-point "inlining.InlineProc9" 26 7) ; inline 'h'
  (inline-in-module-by-point "inlining.InlineProcA" 10 6) ; inline 'f'
  (inline-in-module-by-point "inlining.InlineProcB" 10 6) ;; 'f'
  (inline-in-module-by-point "inlining.InlineProcB" 16 6) ;; 'g'
  (inline-in-module-by-point "inlining.InlineProcB" 22 6) ;; 'h'
  (inline-in-module-by-point "inlining.InlineFun1" 9 7) ; inline 'one'
  (inline-in-module-by-point "inlining.InlineFun2" 10 5) ; inline 'two'
  (inline-in-module-by-point "inlining.InlineFun2" 11 5) ; inline 'three'
  (inline-in-module-by-point "inlining.InlineFun2" 12 5) ; inline 'four'
  (inline-in-module-by-point "inlining.InlineFun3" 10 5) ; inline 'inc'
  (inline-in-module-by-point "inlining.InlineFun3" 11 5) ; inline 3-param 'add'
  (inline-in-module-by-point "inlining.InlineFun3" 18 10) ; inline 'double' appl.
  (inline-in-module-by-point "inlining.InlineFun4" 10 6) ; inline 'f'
  (inline-in-module-by-point "inlining.InlineFun4" 12 6) ; inline 'g'
  (inline-in-module-by-point "inlining.InlineFun4" 14 6) ; inline 'h'
  (inline-in-module-by-point "inlining.InlineFun4" 23 8) ; inline first 'h' application
  (inline-in-module-by-point "inlining.InlineVar1" 18 6)
  (inline-in-module-by-offset "inlining.InlineVar1" 418)
  (inline-in-module-by-point "inlining.InlineVar2" 19 6)
  (inline-in-module-by-point "inlining.InlineVar3" 18 6)
  (inline-in-module-by-point "inlining.InlineVar4" 19 7)
  (inline-in-module-by-point "inlining.InlineVar5" 17 14)
  (inline-in-module-by-point "inlining.InlineVar6" 18 16)
  (inline-in-module-by-point "inlining.InlineVar7" 12 10)
  (inline-in-module-by-point "inlining.InlineVar7" 15 6)
  (inline-in-module-by-point "inlining.InlineVar8" 14 8)
  (inline-in-module-by-point "inlining.InlineVar9" 17 14)
  (inline-in-module-by-point "inlining.InlineType1" 7 12)
  (inline-in-module-by-point "inlining.InlineType1" 11 20)
  (inline-in-module-by-point "inlining.InlineType1" 12 25)
  (inline-in-module-by-point "inlining.InlineType2" 8 4)
  (inline-in-module-by-point "inlining.InlineType3" 7 4)
  (inline-in-module-by-point "inlining.InlineType4" 7 4)
  (inline-in-module-by-point "inlining.InlineType5" 8 4)
  (inline-in-module-by-point "inlining.InlineType6" 8 4)
  (inline-in-module-by-point "inlining.InlineType7" 8 4)
  (inline-in-module-by-point "inlining.InlineType8" 9 4)
  (inline-in-module-by-point "inlining.InlineType9" 8 4)
  (inline-in-module-by-point "inlining.InlineTypeA" 8 4)
  (inline-in-module-by-point "inlining.InlineTypeB" 8 4)
  (inline-in-module-by-point "inlining.InlineTypeC" 8 4)
  (inline-in-module-by-point "fixes.BooleanLib" 9 10)
  (inline-in-module-by-point "fixes.BooleanLib" 10 10)
  )

;;; 
;;; pretty printing
;;; 

(defn try-flatten []
  (rimport "org::nuthatchery::pgf::rascal::engines::FmtEngine")
  (rimport "org::nuthatchery::pgf::rascal::Token")
  (rimport "org::nuthatchery::pgf::rascal::Tseq")
  (rcall "pretty" (rinteger 80) (rcall "flatten" (rcall "tseq" (rstring "first") (reval "Line()") (rstring "second")))))
