(ns org.magnolialang.testutil.pdb-values-pprint
  (:use
   ;;org.magnolialang.testutil.core
   org.magnolialang.testutil.pdb-values
   clojure.pprint
   )
  (:gen-class
   ;; http://walkwithoutrhythm.net/blog/2012/03/26/how-to-call-clojure-1-dot-3-functions-from-java/
   :methods
   [
    #^{:static true}
    [prettyPdbValue [org.rascalmpl.value.IValue] String]
    #^{:static true}
    [prettyPdbValueAnnos [org.rascalmpl.value.IValue] String]
    ]
   ))

;; E.g.
;; (with-pprint-dispatch rascal-dispatch (pprint (c-r (list 1 2 3))))
(defmulti rascal-dispatch class)

(defn- rascal-method [dispatch-val func]
  ;; '.' special form explained at http://clojure.org/java_interop
  (. rascal-dispatch addMethod dispatch-val func))

(defn- rprint-integer [v]
  (write-out (r-c-integer v)))

(defn- rprint-real [v]
  (write-out (r-c-real v)))

(defn- rprint-string [v]
  (write-out (r-c-string v)))

(defn- rprint-bool [v]
  (write-out (r-c-bool v)))

;; These collection functions are just about the same as those in
;; dispatch.clj of pprint in Clojure.

(defn- direct-write-out [s]
  (.write ^java.io.Writer *out* s))

(defn- rprint-list [v]
  (pprint-logical-block
   :prefix "[" :suffix "]"
   ;; Yes, must use 'seq' here. Note that (if () 1 2) => 1 whereas (if
   ;; (seq ()) 1 2) => 2. This is because () != nil.
   (print-length-loop [aseq (seq v)]
                      (when aseq
                        (write-out (first aseq))
                        (when (next aseq)
                          (direct-write-out ", ")
                          (pprint-newline :linear)
                          (recur (next aseq)))))))

(defn- rprint-set [v]
  (pprint-logical-block
   :prefix "{" :suffix "}"
   (print-length-loop [aseq (seq v)]
                      (when aseq
                        (write-out (first aseq))
                        (when (next aseq)
                          (direct-write-out ", ")
                          (pprint-newline :linear)
                          (recur (next aseq)))))))

(defn- rprint-tuple [v]
  (pprint-logical-block
   :prefix "<" :suffix ">"
   (print-length-loop [aseq (seq v)]
                      (when aseq
                        (write-out (first aseq))
                        (when (next aseq)
                          (direct-write-out ", ")
                          (pprint-newline :linear)
                          (recur (next aseq)))))))

(defn- rprint-map [v]
  (pprint-logical-block
   :prefix "(" :suffix ")"
    (print-length-loop [it (iterator-seq (.entryIterator v))]
      (when it
        (let [h (first it)
              t (next it)
              k (.getKey h)
              v (.getValue h)]
          (pprint-logical-block 
           (write-out k)
           (direct-write-out ": ")
           (write-out v))
          (when t
            (direct-write-out ", ")
            (pprint-newline :linear)
            (recur t)))))))

;; A 'loc', looking something like |file:/...|(104,0,<5,8>,<5,8>).
;; See IValue visitSourceLocation(ISourceLocation o) in StandardTextWriter.
(defn- rprint-loc [v]
  (let [uri (.toString (.getURI v))
        n (count uri)
        uri (if (> n 30)
              (str (.substring uri 0 10)
                   "..."
                   (.substring uri (- n 20)))
              uri)]
    (direct-write-out
     (str 
      "|"
      uri
      "|"
      (if-not (.hasOffsetLength v)
        ""
        (str "("
             (.getOffset v) "," (.getLength v)
             (if-not (.hasLineColumn v)
               ""
               (str
                ","
                "<" (.getBeginLine v) "," (.getBeginColumn v) ">"
                ","
                "<" (.getEndLine v) "," (.getEndColumn v) ">"))
             ")"))))))

;; For the traditional format.
;; (defn- rprint-loc [v]
;;   (direct-write-out (.toString v)))

;; (def ^{:private true} rprint-loc
;;   (let [f (formatter-out "~<#{~;~@{~w~^ ~:_~}~;}~:>")]
;;     (fn [v]
;;       (cl-format true f xxx))))

(def ^:dynamic
 ^{:doc "Whether to print Rascal 'node' annotations."}
 *print-annos* true)

(defn- rprint-annos [amap]
  (pprint-logical-block
   :prefix "[" :suffix "]"
   (print-length-loop
    [it (iterator-seq (.entryIterator amap))]
    (when it
      (let [e (first it)
            k (.getKey e)
            v (.getValue e)
            t (next it)]
        (direct-write-out "@")
        (direct-write-out k)
        (direct-write-out " = ")
        (write-out v)
        (when t
          (direct-write-out ", ")
          (pprint-newline :linear)
          (recur t)))))))

;; IConstructor is-a INode
(defn- rprint-node [v]
  (pprint-logical-block
   :prefix (str (.getName v) "(") :suffix ")"
   (print-length-loop
    [it (seq (.getChildren v))]
    (when it
      (let [h (first it)
            t (next it)]
        (write-out h)
        (when t
          (direct-write-out ", ")
          (pprint-newline :linear)
          (recur t))))))
  (let [as (.getAnnotations v)]
    (when (and (not (empty? as)) *print-annos*)
      (pprint-newline :linear)
      (rprint-annos as))))
  
(rascal-method org.rascalmpl.value.IInteger rprint-integer)
(rascal-method org.rascalmpl.value.IReal rprint-real)
(rascal-method org.rascalmpl.value.IString rprint-string)
(rascal-method org.rascalmpl.value.IBool rprint-bool)
(rascal-method org.rascalmpl.value.IList rprint-list)
(rascal-method org.rascalmpl.value.ISet rprint-set)
(rascal-method org.rascalmpl.value.ITuple rprint-tuple)
(rascal-method org.rascalmpl.value.IMap rprint-map)
(rascal-method org.rascalmpl.value.INode rprint-node)
(rascal-method org.rascalmpl.value.ISourceLocation rprint-loc)
(rascal-method :default simple-dispatch)

(defmacro with-rascal-pprint
  ([& act]
     `(with-pprint-dispatch rascal-dispatch
        ~@act)))

(defn set-rascal-pprint []
  (set-pprint-dispatch rascal-dispatch))

(defn rprint-f [v]
  (with-rascal-pprint
    (pprint v)))

(defn rprint-f-no-annos [v]
  (binding [*print-annos* false]
    (rprint-f v)))

(defn rprint-p [v & {:keys [annos?] :or {annos? false}}]
  (binding [*print-annos* annos?]
    (rprint-f v)))

(defn pretty-rascal-value [v]
  (with-rascal-pprint
    (write v :stream nil)))

(defn -prettyPdbValue [v]
  (binding [*print-annos* false]
    (with-rascal-pprint
      (write v :stream nil))))

(defn -prettyPdbValueAnnos [v]
  (with-rascal-pprint
    (write v :stream nil)))

(comment
  (use 'clojure.pprint)
  (use 'org.magnolialang.testutil.pdb-values)
  (use 'org.magnolialang.testutil.pdb-values-pprint :reload-all)

  (defn g [f] (with-rascal-pprint (pprint (f))))
  (g (fn [] (c-r [1 true "foo" 4.5])))
  (g (fn [] (c-r [[1 2] [3 4]])))
  (g (fn [] (c-r '((1 2) (3 4)))))
  (g (fn [] (c-r {1 2 3 4})))
  (g (fn [] (c-r #{1 2 3 4})))
  (g (fn [] (c-r [{1 2} {3 4}])))
  (g (fn [] (c-r {{1 2} {3 4} {5 6} {7 8}})))
  (g (fn [] (r-empty-relation)))
  (g (fn [] (apply r-relation (map c-r [[1 2 "a"] [4 5 "b"]]))))

  (with-rascal-pprint (write (c-r {{1 2} {3 4} {5 6} {7 8}}) :stream nil))
  
  (init-compiler)
  (let [mod-name "basic.Basic" tree (ctor-get-val (get-fact-type-checked mod-name))] (with-rascal-pprint (pprint tree)))
  (let [mod-name "basic.Basic" tree (ctor-get-val (get-fact-type-checked mod-name))] (with-rascal-pprint (pprint (keep-annos tree "concrete"))))
  (let [mod-name "basic.Basic" tree (ctor-get-val (get-fact-type-checked mod-name))] (with-rascal-pprint (pprint (keep-annos tree "loc"))))
  (binding [*print-annos* false] (let [mod-name "basic.Basic" tree (ctor-get-val (get-fact-type-checked mod-name))] (with-rascal-pprint (pprint (keep-annos tree "loc")))))
  (let [mod-name "basic.Basic" tree (ctor-get-val (get-fact-type-checked mod-name))] (rprint-f tree))
  
  ) ;; end test code
