/**************************************************************************
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Andreas Hjortland
 * 
 *************************************************************************/
package org.magnolialang.util;

import java.util.Iterator;
import java.util.Stack;

import nuthatch.pattern.Environment;

import org.rascalmpl.value.IMap;
import org.rascalmpl.value.IValue;
import org.nuthatchery.pica.terms.TermFactory;

public class PdbEnvironment<T extends IValue> implements Environment<T> {
	private final PdbEnvironment<T> parentEnv;
	private final Stack<IMap> tmpScope;

	private IMap scope;


	public PdbEnvironment() {
		tmpScope = new Stack<>();
		scope = TermFactory.vf.mapWriter().done();
		parentEnv = null;
	}


	private PdbEnvironment(PdbEnvironment<T> parent) {
		tmpScope = new Stack<>();
		scope = parent.scope;
		parentEnv = parent;
	}


	@Override
	public void begin() {
		tmpScope.push(scope);
	}


	@Override
	public void commit() {
		if(tmpScope.isEmpty()) {
			throw new IllegalStateException("There is no started transaction");
		}
		tmpScope.pop();
	}


	@Override
	public Environment<T> enterScope() {
		return new PdbEnvironment<>(this);
	}


	@Override
	public Environment<T> exitScope() {
		return parentEnv;
	}


	@SuppressWarnings("unchecked")
	@Override
	public <K> T get(K var) {
		if(!(var instanceof IValue)) {
			throw new IllegalArgumentException("Key is not instance of IValue");
		}
		return (T) scope.get((IValue) var);
	}


	@Override
	public boolean isEmpty() {
		return isEmptyScope() && parentEnv.isEmpty();
	}


	@Override
	public boolean isEmptyScope() {
		return scope.isEmpty();
	}


	@Override
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public Iterator<Object> iterator() {
		return (Iterator) scope.iterator();
	}


	@Override
	public <K> void put(K var, T value) {
		if(!(var instanceof IValue)) {
			throw new IllegalArgumentException("Key is not instance of IValue");
		}
		scope = scope.put((IValue) var, value);

	}


	@Override
	public void rollback() {
		if(tmpScope.isEmpty()) {
			throw new IllegalStateException("There is no started transaction");
		}
		scope = tmpScope.pop();
	}
}
