(ns org.magnolialang.test.core
  (:use
   org.magnolialang.testutil.core
   org.magnolialang.testutil.high-level
   clojure.test))

(defn setup [f]
  (ensure-common)
  (f))

(use-fixtures :once setup)

(deftest test-add
  (is (= 2 (+ 1 1))))

(deftest test-hello
  (is (= (hello-str) "This is Magnolia.")))

(deftest test-ensuring
  (let [lst ["IO"
             "org::magnolialang::testutil::RascalExt"
             "org::magnolialang::testutil::RascalPp"
             "org::magnolialang::magnolia::lib::Pretty"]]
    (doseq [x lst]
      (ensure-rimport x)
      (is (rimported? x)))
    (apply ensure-rimport lst)
    (is (apply rimported? lst))))

(deftest test-tree-facts
  (get-fact-imploded-tree "basic.Basic")
  (get-fact-desugared-tree "basic.Basic")
  (ensugar-tree (get-fact-desugared-tree "basic.Basic")))

(deftest test-view-flattened-module
  (view-flattened-module "basic.Basic")
  (view-flattened-module "basic.Integer")
  (view-flattened-module "errors.ConstantOverloading")
  (view-flattened-module "mutification.Integer")
  (view-flattened-module "mutification.Mutification")
  (view-flattened-module "mutification.Mutification2")
  (view-flattened-module "override.Override")
  (view-flattened-module "scoping.Scoping")
  (view-flattened-module "toplevel.Operators")
  (view-flattened-module "toplevel.Parens")
  (view-flattened-module "toplevel.Predicates")
  (view-flattened-module "toplevel.Renaming")
  (view-flattened-module "typecheck.OnDefines")
  (view-flattened-module "typecheck.Requires")
  (view-flattened-module "typecheck.TypeDecl"))

(deftest test-rreload
  (ensure-rimport "org::nuthatchery::pgf::rascal::tests::TestFmtEngine")
  (rreload "org::nuthatchery::pgf::rascal::tests::TestFmtEngine"))

(deftest test-spacer
  (ensure-rimport "org::nuthatchery::pgf::rascal::tests::RunTestSpacerDirect")
  (reval-str "runRunTestSpacerDirect()"))

(deftest test-fmt-engine
  (ensure-rimport "org::nuthatchery::pgf::rascal::tests::TestFmtEngine")
  (reval-more-str "runTestFmtEngine();")
  )

(deftest test-token
  (ensure-rimport "org::nuthatchery::pgf::rascal::tests::TestToken")
  (reval-str "runTestToken()"))

(deftest test-rascal-pp
  (ensure-rimport "org::magnolialang::testutil::RascalPp")
  (rcall "runTestRascalPp")
  (rprint (get-fact-imploded-tree "basic.Basic"))
  (rprint (get-fact-imploded-tree "basic.Basic") false)
  (rprint (get-fact-imploded-tree "basic.Basic") true))

(deftest test-inliner
  (ensure-rimport "org::magnolialang::magnolia::lib::Inliner")
  (run-inliner-tests))
