#ifndef _MAGNOLIA_PRELUDE_IO_HXX
#define _MAGNOLIA_PRELUDE_IO_HXX
namespace Magnolia {
namespace IO {

struct Type {
	const char* name;
	const char* fields;

	constexpr Type(const char* n, const char* fs) : name(n), fields(fs) {}
};

template<typename Stream>
struct MgIOStream {
	Stream& stream;
	int nesting;
	bool bol;

	MgIOStream(Stream& s) : stream(s), nesting(0), bol(true) {}

	void nest() { ++nesting; }
	void unnest() { --nesting; }
};

struct MgIStream : MgIOStream<std::istream> {
	MgIStream(std::istream& s) : MgIOStream(s) {
		s.setf(std::ios_base::skipws);
	}

	void expect(char expected) {
		char c;
		stream >> c;
		if(expected != c)
			stream.setstate(std::ios_base::failbit);
	}

	void readId(std::string& outString) {
		char c;
		while(stream.good()) {
			stream.get(c);
			if(c == '/' || c == '<' || c == '>' || c == '[' || c == ']' || c == '(' || c == ')' || isspace(c)) {
				stream.unget();
				break;
			}
			outString += c;
		}
	}
};

struct MgOStream : MgIOStream<std::ostream> {
	MgOStream(std::ostream& s) : MgIOStream(s) {}

	void newline() {
		if(!bol) {
			stream << std::endl;
			bol = true;
		}
	}
	void indent() {
		if(bol) {
			for(int i = 0; i < nesting; i++)
				stream << "  ";
			bol = false;
		}
	}

	template<typename T>
	void write(const T& t) {
		stream << t;
		bol = false;
	}

	template<typename T>
	void writeBinary(const T& t) {
		stream << t;
		bol = false;
	}
};


struct startData {
	const Type& type;
	startData(const Type& t) : type(t) {}
};

struct sepData {
	sepData() {}
};

struct endData {
	const Type& type;
	endData(const Type& t) : type(t) {}
};

struct startSequence {
	const Type& type;
	const int length;
	startSequence(const Type& t, int l) : type(t), length(l) {}
};

struct sepSequence {
	sepSequence() {}
};

struct endSequence{
	const Type& type;
	endSequence(const Type& t) : type(t) {}
};

}
}

typedef int int_t;
#endif
