package Turtle;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;

import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.magnolialang.magnolia.evaluator.values.VarValue;

public class JavaTurtle {
	public final Class<Turtle> Turtle = Turtle.class;
	public final Class<Canvas> PictureWorld = Canvas.class;
	public final Class<Coord> Coord = Coord.class;
	/*
	procedure go(obs t : Turtle, upd w : PictureWorld);
	procedure penUp(obs t : Turtle, upd w : PictureWorld);
	procedure penDown(obs t : Turtle, upd w : PictureWorld);
	procedure turnLeft(obs t : Turtle, upd w : PictureWorld);
	procedure turnRight(obs t : Turtle, upd w : PictureWorld);
	procedure turnLeft90(obs t : Turtle, upd w : PictureWorld);
	procedure turnRight90(obs t : Turtle, upd w : PictureWorld);
	procedure newTurtle(out t : Turtle, upd w : PictureWorld);
	procedure newPicture(out w : PictureWorld);
*/
	
	public void display(VarValue<Canvas> w) {
		w.get().repaint();
	}
	
	public void getCoord(VarValue<Coord> c, Turtle t) {
		c.set(new Coord());
		c.get().x = t.x;
		c.get().y = t.y;
	}
	public void go(VarValue<Turtle> tv, VarValue<Canvas> w) {
		Turtle t = tv.get();
		if(!w.get().isDisplayable())
			throw new EvaluationError("Display no longer active");
		t.go(w.get());
		t.draw(w.get());
		w.get().repaint();
		try {
			Thread.sleep(125);
		} catch (InterruptedException e) {
		}
	}
	
	public void draw(Coord c, VarValue<Canvas> w) {
		w.get().setDot(Math.round(c.x), Math.round(c.y), Color.GREEN); 
		w.get().repaint();
	}

	public void penUp(VarValue<Turtle> tv, VarValue<Canvas> w) {
		Turtle t = tv.get();
		t.penUp();
	}
	
	public void penDown(VarValue<Turtle> tv, VarValue<Canvas> w) {
		Turtle t = tv.get();
		t.penDown();
		t.draw(w.get());
		w.get().repaint();
	}
	
	public void turnLeft(VarValue<Turtle> tv, VarValue<Canvas> w) {
		Turtle t = tv.get();
		t.turnLeft();
	}
	
	public void turnRight(VarValue<Turtle> tv, VarValue<Canvas> w) {
		Turtle t = tv.get();
		t.turnRight();
	}
	
	public void turnLeft90(VarValue<Turtle> tv, VarValue<Canvas> w) {
		Turtle t = tv.get();
		t.turnLeft90();
		
	}
	
	public void turnRight90(VarValue<Turtle> tv, VarValue<Canvas> w) {
		Turtle t = tv.get();
		t.turnRight90();
	}

	public void newTurtle(VarValue<Turtle> tv, VarValue<Canvas> w) {
		Turtle turtle = new Turtle();
		turtle.x = w.get().W/2;
		turtle.y = w.get().H/2;
		tv.set(turtle);
		w.get().turtles.add(turtle);
	}

	public void newPicture(VarValue<Canvas> w) {
		Frame frame = new Frame();		
		w.set(frame.canvas);
	}
	
	static class Frame extends JFrame {
		Canvas canvas;
		public Frame() {
			setTitle("TurtleWorld");
	        setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);

	        canvas = new Canvas();
	        add(canvas);
	        setSize(canvas.W*canvas.SCALE, canvas.H*canvas.SCALE);
	        setLocationRelativeTo(null);
	        setVisible(true);
		}
	}
	
	static class Canvas extends JComponent {
		int W = 200;
		int H = 200;
		int SCALE = 3;
		Color[] dots = new Color[W*H];
		List<Turtle> turtles = new ArrayList<>();
		
		@Override
	    public void paintComponent(Graphics g) {
	        super.paintComponent(g);
	        Graphics2D g2 = (Graphics2D) g;
	        for(int x = 0; x < W; x++) {
	        	for(int y = 0; y < W; y++) {
	        		Color dot = getDot(x, y);
	        		if(dot != null) {
	        			g2.setColor(dot);
	        			
	        			g2.drawOval(x*SCALE + SCALE/2, (H-y)*SCALE + SCALE/2, SCALE, SCALE);
	        		}
	        	}
	        }
	        
	        for(Turtle t : turtles) {
	        	g2.setColor(t.getColor());
	        	int xPos = Math.round(t.getX()*SCALE + SCALE/2.0f);
	        	int yPos = Math.round((H-t.getY())*SCALE + SCALE/2.0f);
	        	int xPoints[] = new int[]{xPos-2, xPos-4, xPos+10, xPos-4};
	        	int yPoints[] = new int[]{yPos, yPos+6, yPos, yPos-6};
	        	AffineTransform transform = g2.getTransform();
	        	g2.rotate(-t.direction, xPos, yPos);
	        	g2.drawPolygon(xPoints, yPoints, 4);
	        	g2.setTransform(transform);
	        }
	        
	    }
	    
	    public Color getDot(int x, int y) {
	    	if(x < 0 || x >= W || y < 0 || y >= H)
	    		throw new IllegalArgumentException();
	    	return dots[y*W+x];
	    }
	    
	    public void setDot(int x, int y, Color c) {
	    	if(x < 0 || x >= W || y < 0 || y >= H)
	    		throw new IllegalArgumentException();
	    	dots[y*W+x] = c;
	    }
	    
	}
	
	static class Coord {
		float x;
		float y;
	}
	
	static class Turtle {
		Color color = Color.RED;
		boolean penDown = true;
		float x;
		float y;
		double direction = Math.PI/2.0;
		
		public void go(Canvas canvas) {
			x += Math.cos(direction);
			y += Math.sin(direction);
			
			if(x < 0)
				x += canvas.W;
			if(y < 0)
				y += canvas.H;
			if(x >= canvas.W)
				x -= canvas.W;
			if(y >= canvas.H)
				y -= canvas.H;
		}
		
		public void turnLeft90() {
			direction += Math.PI/2.0;
		}
		
		public void turnRight90() {
			direction -= Math.PI/2.0;
		}
		
		public void turnLeft() {
			direction += 10.0*Math.PI/180.0;
		}
		
		public void turnRight() {
			direction -= 10.0*Math.PI/180.0;
		}
		
		public void penUp() {
			penDown = false;
		}
		public void penDown() {
			penDown = true;
		}
		
		public void draw(Canvas canvas) {
			if(penDown)
				canvas.setDot(Math.round(x), Math.round(y), color);
		}
		
		public float getY() {
			return y;
		}

		public float getX() {
			return x;
		}

		public Color getColor() {
			return color;
		}
		
	}
}
